#define BENCHMARK "Cross Modified OSU MPI Bandwidth Test"
/*
 * Copyright (C) 2002-2008 the Network-Based Computing Laboratory
 * (NBCL), The Ohio State University. 
 *
 * Contact: Dr. D. K. Panda (panda@cse.ohio-state.edu)
 *
 * This program is a modified version of osu_bw.c and is part of the
 * SST/Macroscale package at https://bitbucket.org/cljanss/sstmacro.
 * Questions should be directed to contact info found at that site
 * for this program.
 */

/*
This program is available under BSD licensing.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

(1) Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

(2) Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

(3) Neither the name of The Ohio State University nor the names of
their contributors may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*
 * 08/2009 evensky modified osu_bw.c to allow for 2 streams, an outer
 * measured stream, and an inner interfering stream. Router router bw
 * can be inferred by arranging for a shared link between the two
 * streams. (e.g. on a Cray XT5:
 *   aprun -n4 -N1 ./crossed_osu_bw -i0:2 -o1:3)
 */

#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#ifndef HOST_NAME_MAX
 #define HOST_NAME_MAX 255
#endif
#include "osu.h"
#include <math.h>
#include <assert.h>

#define MAX_REQ_NUM 1000

#define MAX_ALIGNMENT 65536
#define MAX_MSG_SIZE (1<<22)
#define MYBUFSIZE (MAX_MSG_SIZE + MAX_ALIGNMENT)

int loop = 100;
int window_size = 64;
int skip = 10;

int loop_large = 20;
int window_size_large = 64;
int skip_large = 2;

int large_message_size = 8192;

char s_buf1[MYBUFSIZE];
char r_buf1[MYBUFSIZE];

MPI_Request request[MAX_REQ_NUM];
MPI_Status  reqstat[MAX_REQ_NUM];

void usage(char *s) {
  printf("Usage: %s (-i<innertalker>:<innerlistner>) (-o<outertalker>:<outlistener>) {-h} {-v}\n",(s?s:"argv[0]"));
}

void assign_pair(const char *range, int delim, int *src, int *dest) {
  int i;
  char *s = index(range, delim);
  if (s) {
    *src = (int)strtof(range,0);
    *dest = (int)strtof(s+1,0);
  } else {
    usage(0);
    exit(1);
  }
}

void check_values(const int x,const char * const x_str,const int lower,const int upper) {
  if (x < lower || x > upper) {
    printf("ERROR: %s needs to be in [%d,%d]\n",x_str,lower,upper);
    exit(1);
  }
}

int main(int argc, char *argv[])
{
  int myid, numprocs, i, j;
  int size, align_size;
  char *s_buf, *r_buf;
  double t_start = 0.0, t_end = 0.0, t = 0.0;

  int outersrc = 0;
  int outerdest = 2;
  int innersrc = 1;
  int innerdest = 3;

  int verbose = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);


    while(1) {
      int c = getopt(argc,argv,"hvi:o:");
      if (c < 0) break;
      switch(c) {
      case 'h':
	usage(argv[0]);
	exit(0);
	break;
      case 'v':
	++verbose;
	break;
      case 'i':
	assign_pair(optarg,':',&innersrc,&innerdest);
	break;
      case 'o':
	assign_pair(optarg,':',&outersrc,&outerdest);
	break;
      default:
	usage(argv[0]);
	exit(0);
	break;
      }
    }

    check_values(innersrc,"innersrc",0,3);
    check_values(innerdest,"innerdest",0,3);
    check_values(outersrc,"outersrc",0,3);
    check_values(outerdest,"outerdest",0,3);

    if (myid == 0) {
      printf("i<%d-%d> o<%d-%d>\n",innersrc,innerdest,outersrc,outerdest);
    }

  char hname[HOST_NAME_MAX+1];
  gethostname(hname,HOST_NAME_MAX);
  printf("%s: rank %d of %d\n",hname,myid,numprocs);

  align_size = getpagesize();
  assert(align_size <= MAX_ALIGNMENT);

  s_buf =
    (char *) (((unsigned long) s_buf1 + (align_size - 1)) /
	      align_size * align_size);
  r_buf =
    (char *) (((unsigned long) r_buf1 + (align_size - 1)) /
	      align_size * align_size);

  if(numprocs != 4) {
    if(myid == 0) {
      fprintf(stderr, "This test requires exactly four processes\n");
    }
    
    MPI_Finalize();
    return EXIT_FAILURE;
  }

  if(myid == outersrc) {
    fprintf(stdout, "# %s %s\n", BENCHMARK, OMB_VERSION);
    fprintf(stdout, "%-*s%*s\n", 10, "# Size", FIELD_WIDTH,
	    "Bandwidth (MB/s)");
    fflush(stdout);
  }

  /* Bandwidth test */
  for(size = 1; size <= MAX_MSG_SIZE; size *= 2) {
    /* touch the data */
    for(i = 0; i < size; i++) {
      s_buf[i] = 'a';
      r_buf[i] = 'b';
    }

    if(size > large_message_size) {
      loop = loop_large;
      skip = skip_large;
      window_size = window_size_large;
    }

    MPI_Barrier(MPI_COMM_WORLD);

    if(myid == outersrc) {
      for(i = 0; i < loop + skip; i++) {
	if(i == skip) {
	  t_start = MPI_Wtime();
	}

	for(j = 0; j < window_size; j++) {
	  MPI_Isend(s_buf, size, MPI_CHAR, outerdest,
		    100, MPI_COMM_WORLD,
		    request + j);
	}

	MPI_Waitall(window_size, request, reqstat);
	MPI_Recv(r_buf, 4, MPI_CHAR, outerdest,
		 101, MPI_COMM_WORLD,
		 &reqstat[0]);
      }

      t_end = MPI_Wtime();
      t = t_end - t_start;
    }
    else if(myid == innersrc) {
      for(i = 0; i < loop + skip; i++) {
	if(i == skip) {
	  t_start = MPI_Wtime();
	}

	for(j = 0; j < window_size; j++) {
	  MPI_Isend(s_buf, size, MPI_CHAR, innerdest,
		    100, MPI_COMM_WORLD,
		    request + j);
	}

	MPI_Waitall(window_size, request, reqstat);
	MPI_Recv(r_buf, 4, MPI_CHAR, innerdest,
		 101, MPI_COMM_WORLD,
		 &reqstat[0]);
      }

      t_end = MPI_Wtime();
      t = t_end - t_start;
    }
    else if(myid == innerdest) {
      for(i = 0; i < loop + skip; i++) {
	for(j = 0; j < window_size; j++) {
	  MPI_Irecv(r_buf, size, MPI_CHAR, innersrc, 100, MPI_COMM_WORLD,
		    request + j);
	}

	MPI_Waitall(window_size, request, reqstat);
	MPI_Send(s_buf, 4, MPI_CHAR, innersrc, 101, MPI_COMM_WORLD);
      }
    }
    else if(myid == outerdest) {
      for(i = 0; i < loop + skip; i++) {
	for(j = 0; j < window_size; j++) {
	  MPI_Irecv(r_buf, size, MPI_CHAR, outersrc, 100, MPI_COMM_WORLD,
		    request + j);
	}

	MPI_Waitall(window_size, request, reqstat);
	MPI_Send(s_buf, 4, MPI_CHAR, outersrc, 101, MPI_COMM_WORLD);
      }
    }

    MPI_Barrier(MPI_COMM_WORLD);

    if(myid == outersrc) {
      double tmp = size / 1e6 * loop * window_size;

      fprintf(stdout, "%-*d%*.*f\n", 10, size, FIELD_WIDTH,
	      FLOAT_PRECISION, tmp / t);
      fflush(stdout);
    }
  }

  MPI_Finalize();

  return EXIT_SUCCESS;
}

/* vi:set sw=4 sts=4 tw=80: */
