/** \file
 * Executable which runs hellocomm::helloapp.
 * hello-comm-world is a very simple executable to run the hellocomm::helloapp
 * example skeleton mpi application within the SST/Macro simulator.
 */

#include "sstmac/process/mpilaunch.h"
#include "sstmac/process/scheduler.h"
#include "sstmac/simulator/manager.h"
#include "sstmac/simulator/ordered_flower.h"
#include "sstmac/simulator/sharedcircuit.h"
#include "sstmac/network/line.h"
#include "sstmac/network/network.h"
#include "sstmac/simulator/eventmap.h"
#include "sstmac/simulator/simplenode.h"
#include "sstmac/simulator/taskid.h"
#include "sstmac/processor/simpleprocessor.h"
#include "sstmac/util/debug.h"
#include <unistd.h> // getopt
#include "helloapp.h"

using namespace sstmac;
using namespace hellocomm;

int main () {

  // Uncommenting this produces copious debugging output
  //debug().to_stderr();

  // Specify a stack size for the user context stack threads
  const size_t stack(32768);

  // Hardcode for 2 nodes, 1 processor per node
  const int nodes(2);
  const int ppn(1);

  //--------------------------------------
  // Section 1. Set up simulation manager
  //--------------------------------------

  // Construct the simulation manager.
  boost::shared_ptr<manager> mgr = manager::construct(); 

  //--------------------------------------
  // Section 2. Set up interconnect model
  //--------------------------------------

  // Construct a network topology.
  // Use a line network (only 2 nodes after all) 
  // more interesting networks are available.
  boost::shared_ptr<network> net = line::construct(nodes);

  // Construct an interconnect model (sharedcircuit)
  // and give it a bandwidth allocation model (flower).
  const timestamp latency(1e-9); // latency in seconds
  const int bandwidth(125000000); // bandwidth in bytes/second
  boost::shared_ptr<flower> fl =
    ordered_flower::construct(net,bandwidth,bandwidth);
  boost::shared_ptr<interconnect> ic =
    sharedcircuit::construct(latency,bandwidth,net,fl);
 
  mgr->set_interconnect(ic);

  //----------------------
  // Section 3. Add nodes
  //----------------------

  // Add a node to the manager, manager will clone it "nodes" times
  // and assign those nodes to network slots.
  const double memory_bandwidth(1e10); // bandwidth in bytes/second
  boost::shared_ptr<simplenode> node =
    simplenode::construct(memory_bandwidth); // bandwidth in bytes/second
  mgr->add_nodes(node, nodes);

  //-------------------------------------
  // Section 4. Add application launcher 
  //-------------------------------------

  // Construct an application object
  boost::shared_ptr<mpiapp> application(helloapp::construct());

  // Construct a launcher
  boost::shared_ptr<mpilaunch> job = mpilaunch::construct(application);

  const timestamp tstart(0);
  mgr->add_launcher(tstart,job,nodes,ppn);

  //-------------------------------
  // Section 5. Run the simulation
  //-------------------------------

  mgr->run();

}
