/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/netsharing.h>
#include <sstmac/simulator/blockindexing.h>
#include <sstmac/simulator/allocationstrategy.h>
#include <sstmac/simulator/manager.h>
#include <sstmac/simulator/circuit.h>
#include <sstmac/simulator/original_flower.h>
#include <sstmac/simulator/sharedcircuit.h>
#include <sstmac/simulator/eventmap.h>
#include <sstmac/simulator/simplenode.h>
#include <sstmac/simulator/allocationstrategy.h>
#include <sstmac/simulator/randomindexing.h>
#include <sstmac/skeleton/osubenchmarks/osubw.h>
#include <sstmac/process/mpilaunch.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/network/torus.h>
#include <sstmac/network/crossbar.h>
#include <sstmac/network/gamma.h>
#include <sstmac/network/fattree.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <iterator>
#include <string>
#include <map>
#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <boost/lexical_cast.hpp>

using namespace sstmac;

using boost::shared_ptr;
using boost::lexical_cast;

enum nettype {
  RSQTORUS, TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3
};

std::ostream& operator<<(std::ostream &os, const nettype &nt) {
  switch(nt) {
  case TORUS:    os << "TORUS";    break;
  case CROSSBAR: os << "CROSSBAR"; break;
  case GAMMA:    os << "GAMMA";    break;
  case FATTREE:  os << "FATTREE";  break;
  case TORUS3:   os << "TORUS3";   break;
  default:       os << "GARBAGE";  break;
  }
  return os;
}

enum circuittype {
  EXCLUSIVE, SHARED
};

std::ostream& operator<<(std::ostream &os, const circuittype &ct) {
  switch(ct) {
  case EXCLUSIVE: os << "EXCLUSIVE"; break;
  case SHARED:    os << "SHARED";    break;
  default:        os << "GARBAGE";   break;
  }
  return os;
}

struct opts {
  int help;
  int verbose;
  double timescale;
  int ppn;
  std::vector<int> geometry;
  int64_t bandwidth;
  timestamp latency;
  int64_t cpubandwidth;
  timestamp cpulatency;
  int64_t router_router_bw;
  int64_t router_node_bw;
  netsharing netshare;
  int   n_rebalance_max;
  nettype network;
  circuittype circuit;
  int random_indexing;
  int random_allocation;
  std::string outfile;

  opts() :
    help(0), verbose(0), timescale(1), ppn(1), bandwidth(2500000000ll),
    latency(timestamp::exact_nsec(1300)), cpubandwidth(10000000000ll),
    cpulatency(timestamp::exact_nsec(1)), router_router_bw(250000000),
    router_node_bw(250000000), netshare(ORDERED), n_rebalance_max(-1),
    network(TORUS), circuit(SHARED), random_indexing(0), random_allocation(0)
  {}
};

/// This is what the various network creation routines spit back.
struct runsetup {
  shared_ptr<network> net;
  int nodes;
  int totnodes;
};

std::ostream& operator<<(std::ostream &os, const opts &oo) {
  os << "  --help=" << oo.help << "\\\n"
     << "  --verbose=" << oo.verbose << "\\\n"
     << "  --timescale=" << oo.timescale << "\\\n"
     << "  --ppn=" << oo.ppn << "\\\n"
     << "  --geometry=(";
  std::copy(oo.geometry.begin(), oo.geometry.end(),
            std::ostream_iterator<int>(os, ", "));
  os << ")\\\n";
  os << "  --bandwidth=" << oo.bandwidth << "\\\n"
     << "  --latency=" << oo.latency << "\\\n"
     << "  --cpubandwidth=" << oo.cpubandwidth << "\\\n"
     << "  --cpulatency=" << oo.cpulatency << "\\\n"
     << "  --router_router_bw=" << oo.router_router_bw << "\\\n"
     << "  --router_node_bw=" << oo.router_node_bw << "\\\n"
     << "  --netsharing=" << oo.netshare << "\\\n"
     << "  --n_rebalance_max=" << oo.n_rebalance_max << "\\\n"
     << "  --network=" << oo.network << "\\\n"
     << "  --circuit=" << oo.circuit << "\\\n"
     << "  --random_indexing=" << oo.random_indexing << "\\\n"
     << "  --random_allocation=" << oo.random_allocation << "\\\n"
     << "  --outfile=" << oo.outfile << "\n";
  return os;
}

int parse_opts(int argc, char **argv, opts &oo);
double get_positive_float(const char *value, bool &errorflag);
int get_int(const char *value, bool &errorflag);
int get_positive_int(const char *value, bool &errorflag);
std::vector<int> get_intvec(const char *value, bool &errorflag);
timestamp get_timestamp(const char *value, bool &errorflag);
int64_t get_bandwidth(const char *value, bool &errorflag);
netsharing get_netsharing(const char *value, bool &errorflag);
nettype get_network(const char *value, bool &errorflag);
circuittype get_circuit(const char *value, bool &errorflag);
void print_help(int argc, char **argv);
void run(const opts &oo);
void build_torus(runsetup &sup, std::vector<int> geometry);
void build_xbar(runsetup &sup, std::vector<int> geometry);
void build_gamma(runsetup &sup, std::vector<int> geometry);
void build_fattree(runsetup &sup, std::vector<int> geometry);
void build_torus3(runsetup &sup, std::vector<int> geometry);

int main(int argc, char **argv) {
  opts oo;
  if(! parse_opts(argc, argv, oo)) {
    if(oo.verbose)
      debug().to_stderr();
    // Now start grinding away.
    run(oo);
  }
  else {
    return 1;
  }
}

int parse_opts(int argc, char **argv, opts &oo) {
  option gopt[] = {
    {"help", no_argument, NULL, 'h'},
    {"verbose", no_argument, NULL, 'v'},
    {"timescale", required_argument, NULL, 't'},
    {"ppn", required_argument, NULL, 'p'},
    {"bandwidth", required_argument, NULL, 'b'},
    {"latency", required_argument, NULL, 'l'},
    {"cpubandwidth", required_argument, NULL, 'B'},
    {"cpulatency", required_argument, NULL, 'L'},
    {"router_router_bw", required_argument, NULL, 'R'},
    {"router_node_bw", required_argument, NULL, 'S'},
    {"netsharing", required_argument, NULL, 'T'},
    {"n_rebalance_max", required_argument, NULL, 'U'},
    {"network", required_argument, NULL, 'N'},
    {"geometry", required_argument, NULL, 'G'},
    {"circuit", required_argument, NULL, 'C'},
    {"random_indexing", required_argument, NULL, 'I'},
    {"random_allocation", required_argument, NULL, 'A'},
    {"outfile", required_argument, NULL, 'o'},
    {NULL, 0, NULL, '\0'}
  };
  int ch;
  bool errorflag = false;
  while((ch = getopt_long(argc, argv, "hvt:p:b:l:B:L:R:S:T:U:N:G:C:I:A:o:",
                          gopt, NULL)) != -1)
    {
      switch(ch) {
      case 'h': oo.help = 1; break;
      case 'v': oo.verbose = 1; break;
      case 't': oo.timescale = get_positive_float(optarg, errorflag); break;
      case 'p': oo.ppn = get_positive_int(optarg, errorflag); break;
      case 'b': oo.bandwidth = get_bandwidth(optarg, errorflag); break;
      case 'l': oo.latency = get_timestamp(optarg, errorflag); break;
      case 'B': oo.cpubandwidth = get_bandwidth(optarg, errorflag); break;
      case 'L': oo.cpulatency = get_timestamp(optarg, errorflag); break;
      case 'R': oo.router_router_bw = get_bandwidth(optarg, errorflag); break;
      case 'S': oo.router_node_bw = get_bandwidth(optarg, errorflag); break;
      case 'T': oo.netshare = get_netsharing(optarg, errorflag); break;
      case 'U': oo.n_rebalance_max = get_int(optarg, errorflag); break;
      case 'N': oo.network = get_network(optarg, errorflag); break;
      case 'G': oo.geometry = get_intvec(optarg, errorflag); break;
      case 'C': oo.circuit = get_circuit(optarg, errorflag); break;
      case 'I': oo.random_indexing = get_int(optarg, errorflag); break;
      case 'A': oo.random_allocation = get_int(optarg, errorflag); break;
      case 'o': oo.outfile = optarg; break;
      default: std::cerr << "Unhandled input flag.\n"; errorflag = true; break;
      }
      if(oo.help) {
        print_help(argc, argv);
        return 1;
      }
      if(errorflag) {
        std::cerr << "Error return\n";
        return 2;
      }
    }
  std::cerr << oo;
  return 0;
}

void run(const opts &oo) {
  runsetup sup;
  switch(oo.network) {
  case TORUS:    build_torus(sup, oo.geometry);    break;
  case CROSSBAR: build_xbar(sup, oo.geometry);     break;
  case GAMMA:    build_gamma(sup, oo.geometry);    break;
  case FATTREE:  build_fattree(sup, oo.geometry);  break;
  case TORUS3:   build_torus3(sup, oo.geometry);   break;
  default: throw valueerror("runosubw:  Invalid net type\n");
  }
  shared_ptr<manager> mgr = manager::construct();
  if(oo.circuit == EXCLUSIVE)
    mgr->set_interconnect(circuit::construct(oo.latency,oo.bandwidth,sup.net));
  else {
    shared_ptr<original_flower> flow = original_flower::construct(sup.net, oo.router_node_bw,
                                                oo.router_router_bw,
                                                oo.netshare);
    flow->n_rebalance_max = oo.n_rebalance_max;
    mgr->set_interconnect(sharedcircuit::construct(oo.latency, oo.bandwidth,
                                                   sup.net, flow));
  }
  mgr->set_eventmanager(eventmap::construct());
  mgr->add_nodes(simplenode::construct(oo.cpulatency, oo.cpubandwidth, oo.ppn),
                 sup.totnodes);
  shared_ptr<osubw> bwtest = osubw::construct(oo.outfile);
  bwtest->loop = bwtest->loop_large = 1;
  bwtest->skip = bwtest->skip_large = 0;
  shared_ptr<indexstrategy> idx;
  shared_ptr<allocationstrategy> all;
  if(oo.random_indexing)
    idx = randomindexing::construct(oo.random_indexing);
  if(oo.random_allocation)
    all = randomallocation::construct(oo.random_allocation);
  int ppn = (oo.ppn == 1 ? 1 : 2);
  int targetnodes = (oo.ppn == 1 ? 2 : 1);
  shared_ptr<mpilaunch> job =
    mpilaunch::construct(bwtest, false, 2);
  job->set_scheduler(scheduler::construct(81920, 10240000, true));
  std::cerr << "Adding launcher for " << sup.nodes << " peers.\n";
  mgr->add_launcher(timestamp(0), job, targetnodes, ppn, idx, all);
  timestamp runtime = mgr->run();
  fprintf(stderr, "# %5s %10s %10s %10s %10s %10s\n",
          "procs", "timescale", "bandwidth", "latency", "nettype", "runtime");
  fprintf(stderr, "%7d %10.4e %10lld %10.4e %10s %10.4f\n",
          sup.nodes, oo.timescale, (long long)(oo.bandwidth),
          oo.latency.sec(), lexical_cast<std::string>(oo.network).c_str(),
          runtime.sec());
}

double get_positive_float(const char *value, bool &errorflag) {
  char *endptr = NULL;
  double val = strtod(value, &endptr);
  // This is really pretty bad form
  errno = 0;
  if(*value == '\0' || *endptr != '\0') {
    errorflag = true;
    std::cerr << "Not a valid floating point value: " << value << "\n";
  }
  if(errno == ERANGE) {
    errorflag = true;
    std::cerr << "Value outside double precision range: " << value << "\n";
  }
  if(val <= 0) {
    errorflag = true;
    std::cerr << "Not a positive floating point value: " << value << "\n";
  }
  return val;
}

/// Multiply two 64 bit integer values and check for overflow.
/// This would be two lines if we had 128-bit integers.
int64_t multiply64(int64_t a, int64_t b, bool &errorflag) {
  // Use detection of 32-bit overflow to help us.
  //  let 
  //      signed    scaling = 1 << 32
  //      unsigned  mask = scaling >> 1
  //      signed    au = a / scaling       
  //      unsigned  al = a & mask
  //      signed    bu = b / scaling       
  //      unsigned  bl = b & mask
  //  then
  //      a * b = (au << 32 + al) * (bu << 32 + bl)
  //            = (au * bu << 64) + ((au * bl + bu * al) << 32) + (al * bl)
  // so we have an overflow if any of the following hold:
  //      au and bu are both non-zero
  //      au*bl + bu*al is greater than 0x7fffffff (now that au or bu is zero)
  //      al * bl is greater than 0x7fffffff
  //      the sum ((au*bl + bu*al) << 32) + (al*bl) overflows
  int64_t scaling = int64_t(1) << 32;
  int64_t mask = scaling -1;
  int64_t max32  = scaling>>1;
  uint64_t max64 = (~uint64_t(0)) >> 1;
  int64_t au = a/scaling, bu = b/scaling;
  int64_t al = a & mask, bl = b & mask;
  int64_t mul1 = au * bl + bu * al;
  uint64_t mul2 = uint64_t(al) * uint64_t(bl);
  uint64_t uval = uint64_t(mul1) * uint64_t(scaling) + mul2;
  if((au && bu) || (mul1 > max32) || (mul2 > max64) || (uval > max64))  {
    errorflag = true;
    std::cerr << "64-bit integer overflow in " << a << "*" << b << "\n";
  }
  return int64_t(uval);
}

void populate_timestamp_names(std::map<std::string, int64_t> &value) {
  // Accept ps, ns, us, ms, and s with various (lower-case) forms
  // Store time in picoseconds.
  value["ps"] = value["psec"] = value["pseconds"] = value["picoseconds"] = 1;
  value["ns"] = value["nsec"] = value["nseconds"] = value["nanoseconds"] = 1000;
  value["us"] = value["usec"] = value["useconds"] =
    value["microseconds"] = 1000*1000;
  value["ms"] = value["msec"] = value["mseconds"] =
    value["milliseconds"] = 1000LL*1000LL*1000LL;
  value["s"] = value["sec"] = value["seconds"] = 1000LL*1000LL*1000LL*1000LL;
}

/// Get a timestamp possiblly suffixed with any of the identifiers
/// psec, nsec, usec, msec, sec, ps, ns, us, ms, s
timestamp get_timestamp(const char *value, bool &errorflag) {
  static std::map<std::string, int64_t> mulmap;
  if(mulmap.empty())
    populate_timestamp_names(mulmap);

  char *endptr = NULL;
  int64_t multiplier = 1;
  int64_t val = strtoll(value, &endptr, 10);
  errno = 0;
  // Figure out whether we have a value.
  if(*value == '\0' || endptr == value) {
    errorflag = true;
    std::cerr << "Invalid time value: " << value << "\n";
  }
  if(errno == EINVAL) {
    errorflag = true;
    std::cerr << "Invalid characters in value: " << value << "\n";
  }
  if(errno == ERANGE) {
    errorflag = true;
    std::cerr << "Value outside 64-bit integer range: " << value << "\n";
  }
  // This is where we may have the units appended to the input value.
  if(! errorflag) {
    while(*endptr == ' ') ++endptr;
    if(*endptr != '\0') {
      std::string units(endptr);
      for(unsigned i = 0; i < units.size(); ++i) units[i] = tolower(units[i]);
      if(mulmap.find(units) != mulmap.end())
        multiplier = mulmap[units];
      else {
        errorflag = true;
        std::cerr << "Invalid time units: " << units << "\n";
      }
    }
    if(! errorflag) {
      // Now expand the value.
      val = multiply64(val, multiplier, errorflag);
    }
  }
  return timestamp::exact_psec(val);
}

int get_int(const char *value, bool &errorflag) {
  char *endptr;
  errno = 0;
  long val = strtol(value, &endptr, 10);
  if(*value == '\0' || *endptr != '\0') {
    errorflag = true;
    std::cerr << "Invalid integer value: " << value << "\n";
  }
  if(errno == ERANGE ||
     val > std::numeric_limits<int>::max() ||
     val < std::numeric_limits<int>::min())
    {
      errorflag = true;
      std::cerr << "Value outside 32-bit integer range: " << value << "\n";
    }
  if(errno == EINVAL) {
    errorflag = true;
    std::cerr << "Value contains invalid characters: " << value << "\n";
  }
  return val;
}

int get_positive_int(const char *value, bool &errorflag) {
  int v = get_int(value, errorflag);
  if((! errorflag) && v <= 0) {
    errorflag = true;
    std::cerr << "Not a positive non-zero integer value: " << value << "\n";
  }
  return v;
}

std::vector<int> get_intvec(const char *value, bool &errorflag) {
  std::vector<int> retval;
  char *brkt;
  char *val = strdup(value);
  const char *sep = " ,;";
  for(char *word = strtok_r(val, sep, &brkt);
      word != NULL;
      word = strtok_r(NULL, sep, &brkt))
    {
      retval.push_back(get_int(word, errorflag));
      if(errorflag) break;
    }
  free(val);
  return retval;
}

void populate_bandwidth_names(std::map<std::string, int64_t> &abbrname,
                              std::map<std::string, int64_t> &fullname)
{
  abbrname["bps"] = abbrname["b/s"] = abbrname["b/sec"] =
    fullname["bits/s"] = fullname["bits/sec"] = fullname["bits/second"] =
    fullname["bits per second"] = 1;
  // Bytes per second.
  abbrname["Bps"] = abbrname["B/s"] = abbrname["B/sec"] =
    fullname["bytes/s"] = fullname["bytes/sec"] = fullname["bytes/second"] =
    fullname["bytes per second"] = 8;
  // Kilobits per second.
  abbrname["kbps"] = abbrname["Kbps"] = abbrname["kb/s"] =
    abbrname["Kb/s"] = fullname["kbits/s"] = fullname["kbits/sec"] =
    fullname["kbits/second"] = fullname["kilobits/s"] =
    fullname["kilobits/sec"] = fullname["kilobits/second"] =
    fullname["kilobits per second"] = 1000;
  // Kibibits per second.
  abbrname["kibps"] = abbrname["Kibps"] = abbrname["kib/s"] =
    abbrname["Kib/s"] = fullname["kibits/s"] = fullname["kibits/sec"] =
    fullname["kibits/second"] = fullname["kibibits/s"] =
    fullname["kibibits/sec"] = fullname["kibibits/second"] =
    fullname["kibibits per second"] = 1024;
  // Kilobytes per second.
  abbrname["kBps"] = abbrname["KBps"] = abbrname["kB/s"] =
    abbrname["KB/s"] = fullname["kbytes/s"] = fullname["kbytes/sec"] =
    fullname["kbytes/second"] = fullname["kilobytes/s"] =
    fullname["kilobytes/sec"] = fullname["kilobytes/second"] =
    fullname["kilobytes per second"] = 1000 * 8;
  // Kibibytes per second.
  abbrname["kiBps"] = abbrname["KiBps"] = abbrname["kiB/s"] =
    abbrname["KiB/s"] = fullname["kibytes/s"] = fullname["kibytes/sec"] =
    fullname["kibytes/second"] = fullname["kibibytes/s"] =
    fullname["kibibytes/sec"] = fullname["kibibytes/second"] =
    fullname["kibibytes per second"] = 1024 * 8;
  // Megabits per second.
  abbrname["mbps"] = abbrname["Mbps"] = abbrname["mb/s"] =
    abbrname["Mb/s"] = fullname["mbits/s"] = fullname["mbits/sec"] =
    fullname["mbits/second"] = fullname["megabits/s"] =
    fullname["megabits/sec"] = fullname["megabits/second"] =
    fullname["megabits per second"] = 1000 * 1000;
  // Mebibits per second.
  abbrname["mibps"] = abbrname["Mibps"] = abbrname["mib/s"] =
    abbrname["Mib/s"] = fullname["mibits/s"] = fullname["mibits/sec"] =
    fullname["mibits/second"] = fullname["mebibits/s"] =
    fullname["mebibits/sec"] = fullname["mebibits/second"] =
    fullname["mebibits per second"] = 1024 * 1024;
  // Megabytes per second.
  abbrname["mBps"] = abbrname["MBps"] = abbrname["mB/s"] =
    abbrname["MB/s"] = fullname["mbytes/s"] = fullname["mbytes/sec"] =
    fullname["mbytes/second"] = fullname["megabytes/s"] =
    fullname["megabytes/sec"] = fullname["megabytes/second"] =
    fullname["megabytes per second"] = 1000 * 1000 * 8;
  // Mebibytes per second.
  abbrname["miBps"] = abbrname["MiBps"] = abbrname["miB/s"] =
    abbrname["MiB/s"] = fullname["mibytes/s"] = fullname["mibytes/sec"] =
    fullname["mibytes/second"] = fullname["mebibytes/s"] =
    fullname["mebibytes/sec"] = fullname["mebibytes/second"] =
    fullname["mebibytes per second"] = 1024 * 1024 * 8;
  // Gigabits per second.
  abbrname["gbps"] = abbrname["Gbps"] = abbrname["gb/s"] =
    abbrname["Gb/s"] = fullname["gbits/s"] = fullname["gbits/sec"] =
    fullname["gbits/second"] = fullname["gigabits/s"] =
    fullname["gigabits/sec"] = fullname["gigabits/second"] =
    fullname["gigabits per second"] = 1000 * 1000 * 1000;
  // Gibibits per second.
  abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
    abbrname["Gib/s"] = fullname["gibits/s"] = fullname["gibits/sec"] =
    fullname["gibits/second"] = fullname["gibibits/s"] =
    fullname["gibibits/sec"] = fullname["gibibits/second"] =
    fullname["gibibits per second"] = 1024 * 1024 * 1024;
  // Gigabytes per second.
  abbrname["gBps"] = abbrname["GBps"] = abbrname["gB/s"] =
    abbrname["GB/s"] = fullname["gbytes/s"] = fullname["gbytes/sec"] =
    fullname["gbytes/second"] = fullname["gigabytes/s"] =
    fullname["gigabytes/sec"] = fullname["gigabytes/second"] =
    fullname["gigabytes per second"] = 1000LL * 1000LL * 1000LL * 8LL;
  // Gibibytes per second.
  abbrname["giBps"] = abbrname["GiBps"] = abbrname["giB/s"] =
    abbrname["GiB/s"] = fullname["gibytes/s"] = fullname["gibytes/sec"] =
    fullname["gibytes/second"] = fullname["gibibytes/s"] =
    fullname["gibibytes/sec"] = fullname["gibibytes/second"] =
    fullname["gibibytes per second"] = 1024LL * 1024LL * 1024LL * 8;
  // Terabits per second.
  abbrname["tbps"] = abbrname["Tbps"] = abbrname["tb/s"] =
    abbrname["Tb/s"] = fullname["tbits/s"] = fullname["tbits/sec"] =
    fullname["tbits/second"] = fullname["terabits/s"] =
    fullname["terabits/sec"] = fullname["terabits/second"] =
    fullname["terabits per second"] = 1000LL * 1000LL * 1000LL * 1000LL;
  // Tebibits per second.
  abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
    abbrname["Gib/s"] = fullname["tebits/s"] = fullname["tebits/sec"] =
    fullname["tebits/second"] = fullname["tebibits/s"] =
    fullname["tebibits/sec"] = fullname["tebibits/second"] =
    fullname["tebibits per second"] = 1024LL * 1024LL * 1024LL * 1024LL;
  // Terabytes per second.
  abbrname["tBps"] = abbrname["TBps"] = abbrname["tB/s"] =
    abbrname["TB/s"] = fullname["tbytes/s"] = fullname["tbytes/sec"] =
    fullname["tbytes/second"] = fullname["terabytes/s"] =
    fullname["terabytes/sec"] = fullname["terabytes/second"] =
    fullname["terabytes per second"] = 1000LL*1000LL*1000LL*1000LL*8LL;
  // Tebibytes per second.
  abbrname["tiBps"] = abbrname["TiBps"] = abbrname["tiB/s"] =
    abbrname["TiB/s"] = fullname["tibytes/s"] = fullname["tibytes/sec"] =
    fullname["tibytes/second"] = fullname["tebibytes/s"] =
    fullname["tebibytes/sec"] = fullname["tebibytes/second"] =
    fullname["tebibytes per second"] = 1024LL * 1024LL * 1024LL * 8;
  // Petabits per second.
  abbrname["pbps"] = abbrname["Pbps"] = abbrname["pb/s"] =
    abbrname["Pb/s"] = fullname["pbits/s"] = fullname["pbits/sec"] =
    fullname["pbits/second"] = fullname["petabits/s"] =
    fullname["petabits/sec"] = fullname["petabits/second"] =
    fullname["petabits per second"] = 1000LL*1000LL*1000LL*1000LL*1000LL;
  // Pebibits per second.
  abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
    abbrname["Gib/s"] = fullname["pebits/s"] = fullname["pebits/sec"] =
    fullname["pebits/second"] = fullname["pebibits/s"] =
    fullname["pebibits/sec"] = fullname["pebibits/second"] =
    fullname["pebibits per second"] = 1024LL*1024LL*1024LL*1024LL*1024LL;
  // Petabytes per second.
  abbrname["tBps"] = abbrname["TBps"] = abbrname["tB/s"] =
    abbrname["TB/s"] = fullname["tbytes/s"] = fullname["tbytes/sec"] =
    fullname["tbytes/second"] = fullname["petabytes/s"] =
    fullname["petabytes/sec"] = fullname["petabytes/second"] =
    fullname["petabytes per second"] = 1000LL*1000LL*1000LL*1000LL*1000LL*8LL;
  // Pebibytes per second.
  abbrname["piBps"] = abbrname["PiBps"] = abbrname["piB/s"] =
    abbrname["PiB/s"] = fullname["pibytes/s"] = fullname["pibytes/sec"] =
    fullname["pibytes/second"] = fullname["pebibytes/s"] =
    fullname["pebibytes/sec"] = fullname["pebibytes/second"] =
    fullname["pebibytes per second"] = 1024LL * 1024LL * 1024LL * 8;
}

int64_t get_bandwidth(const char *value, bool &errorflag) {
  // This map does case-sensitive matching of abbreviated names
  static std::map<std::string, int64_t> abbrname;
  // and this map does case-insenstive matching of non-abbreviated names.
  static std::map<std::string, int64_t> fullname;
  // Populate the maps with the multipliers for all recognized units.
  // We store the multipliers in bits/second to make everything integers.
  // Bits per second.
  if(abbrname.empty() || fullname.empty())
    populate_bandwidth_names(abbrname, fullname);
  // Go.
  char *endptr = NULL;
  int64_t multiplier = 1;
  int64_t val = strtoll(value, &endptr, 10);
  errno = 0;
  // Figure out whether we have a value.
  if(*value == '\0' || endptr == value) {
    errorflag = true;
    std::cerr << "Invalid bandwidth value: " << value << "\n";
  }
  if(errno == EINVAL) {
    errorflag = true;
    std::cerr << "Invalid characters in value: " << value << "\n";
  }
  if(errno == ERANGE) {
    errorflag = true;
    std::cerr << "Value outside 64-bit integer range: " << value << "\n";
  }
  // This is where we may have the units appended to the input value.
  if(! errorflag) {
    while(*endptr == ' ') ++endptr;
    if(*endptr != '\0') {
      std::string units(endptr);
      std::string lcunits = units;
      for(unsigned i = 0; i < lcunits.size(); ++i)
        lcunits[i] = tolower(lcunits[i]);
      if(abbrname.find(units) != abbrname.end())
        multiplier = abbrname[units];
      else if(fullname.find(lcunits) != fullname.end())
        multiplier = fullname[units];
      else {
        errorflag = true;
        std::cerr << "No match for the bandwidth units " << units << "\n";
      }
    }
    if(! errorflag) {
      // Now expand the value.
      val = multiply64(val, multiplier, errorflag);
      val /= 8;  // because the simulator works in bytes/sec.
    }
  }
  return val;
}

netsharing get_netsharing(const char *value, bool &errorflag) {
  std::string arg = value;
  for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
  netsharing ns = ORDERED;
  if(arg == "overcommit") ns = OVERCOMMIT;
  else if(arg == "throttle_overcommits" || arg == "throttle-overcommits" ||
          arg == "throttle")
    ns = THROTTLE_OVERCOMMITS;
  else if(arg == "ordered")
    ns = ORDERED;
  else {
    errorflag = true;
    std::cerr << "Invalid netsharing type: " << value << "\n";
  }
  return ns;
}

nettype get_network(const char *value, bool &errorflag) {
  // TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3
  std::string arg = value;
  for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
  nettype nt = TORUS;
  if(arg == "torus") nt = TORUS;
  else if(arg == "crossbar" || arg == "xbar" || arg == "cross") nt = CROSSBAR;
  else if(arg == "gamma" || arg == "gammagraph" || arg == "gamma-graph")
    nt = GAMMA;
  else if(arg == "fattree" || arg == "fat-tree" || arg == "fat") nt = FATTREE;
  else if(arg == "torus3" || arg == "torus3d" || arg == "3dtorus" ||
          arg == "3d-torus")
    nt = TORUS3;
  else {
    errorflag = true;
    std::cerr << "Invalid network type: " << value;
  }
  return nt;
}

circuittype get_circuit(const char *value, bool &errorflag) {
  // EXCLUSIVE, SHARED
  std::string arg(value);
  for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
  circuittype ct = SHARED;
  if(arg == "exclusive") ct = EXCLUSIVE;
  else if(arg == "shared") ct = SHARED;
  else {
    errorflag = true;
    std::cerr << "Invalid circuit type: " << value << "\n";
  }
  return ct;
}

void build_torus(runsetup &sup, std::vector<int> geometry) {
  if(geometry.empty()) {
    geometry.push_back(32);
    geometry.push_back(32);
  }
  if(geometry.size() == 2) geometry.push_back(1);
  if(geometry.size() == 3) geometry.push_back(1);
  if(geometry.size() != 4 || geometry[0] <= 0 || geometry[1] <= 0 ||
     geometry[1] < 0 || geometry[2] <= 0)
    {
      std::cerr << "build_torus:  Torus geometry must have two to four entries: "
                << "width > 0, height > 0, nperswitch >= 0, and width > 0\n";
      std::copy(geometry.begin(), geometry.end(),
                std::ostream_iterator<int>(std::cerr, ", "));
      std::cerr << "\n";
      throw valueerror("build_torus:  Invalid torus geometry.");
    }
  sup.totnodes = geometry[0]*geometry[1];
  sup.net = torus::construct_2D(geometry[0],geometry[1],geometry[2],geometry[3]);
  sup.nodes = sup.totnodes;
}

void build_xbar(runsetup &sup, std::vector<int> geometry) {
  if(geometry.empty()) {
    geometry.push_back(1024);
  }
  if(geometry.size() != 1 || geometry[0] <= 0) {
    std::cerr << "build_xbar: Xbar geometry must have one entry: size > 0\n";
    std::copy(geometry.begin(), geometry.end(),
              std::ostream_iterator<int>(std::cerr, ", "));
    std::cerr << "\n";
    throw valueerror("build_xbar:  Invalid xbar geometry.");
  }
  sup.totnodes = geometry[0];
  sup.net = crossbar::construct(sup.totnodes);
  sup.nodes = sup.totnodes;
}

void build_gamma(runsetup &sup, std::vector<int> geometry) {
  std::cerr << "build_gamma:  Sorry, I don't have a clue how to compute "
            << "the size of a gamma graph\n";
  throw unimplementederror("undumpi.cc:  build_gamma");
  (void)sup; (void)geometry;
  /*
  sup.net = sstmac::gamma::construct(8,3,1);
  sup.nodes = meta->getnumprocs();
  */
}

void build_fattree(runsetup &sup, std::vector<int> geometry) {
  if(geometry.empty()) {
    geometry.push_back(3);
    geometry.push_back(24);
  }
  if(geometry.size() == 2) geometry.push_back(0);
  if(geometry.size() == 3) geometry.push_back(1);
  if(geometry.size() != 4 || geometry[0] <= 0 || geometry[1] <= 0 ||
     geometry[2] < 0 || geometry[3] <= 0)
    {
      std::cerr << "build_fattree:  Tree geometry must have two to four entries:"
                << " height > 0, half-links > 0, npswitch >= 0, and width > 0\n";
      std::copy(geometry.begin(), geometry.end(),
                std::ostream_iterator<int>(std::cerr, ", "));
      std::cerr << "\n";
      throw valueerror("build_fattree:  Invalid tree geometry.");
    }
  sup.net = fattree::construct(3,12);
  /// GH: complained that arguments were ints for me, casted to double.
  sup.totnodes = int(pow((double)geometry[1]/2, (double)geometry[0]));
  sup.nodes = sup.totnodes;
}

void build_torus3(runsetup &sup, std::vector<int> geometry) {
  if(geometry.empty()) {
    geometry.resize(3, 8);
  }
  if(geometry.size() == 3) geometry.push_back(0);
  if(geometry.size() == 4) geometry.push_back(1);
  if(geometry.size() != 5 || geometry[0] <= 0 || geometry[1] <= 0 ||
     geometry[2] <= 0 || geometry[3] < 0 || geometry[4] <= 0)
    {
     std::cerr << "build_torus3:  Torus geometry must have three to five "
               << "entries: {x,y,z}-size > 0, nperswitch >= 0, and width > 0\n";
     std::cerr << "  Current geometry: (";
     std::copy(geometry.begin(), geometry.end(),
               std::ostream_iterator<int>(std::cerr, ", "));
     std::cerr << "\n";
     throw valueerror("build_torus3:  Invalid torus geometry.");
    }
  sup.net = torus::construct_3D(8,8,8,1);
  sup.totnodes = geometry[0] * geometry[1] * geometry[2];
  sup.nodes = sup.totnodes;
}

/*
    {"help", no_argument, NULL, 'h'},
    {"verbose", no_argument, NULL, 'v'},
    {"timescale", required_argument, NULL, 't'},
    {"bandwidth", required_argument, NULL, 'b'},
    {"latency", required_argument, NULL, 'l'},
    {"cpubandwidth", required_argument, NULL, 'B'},
    {"cpulatency", required_argument, NULL, 'L'},
    {"router_router_bw", required_argument, NULL, 'R'},
    {"router_node_bw", required_argument, NULL, 'S'},
    {"netsharing", required_argument, NULL, 'T'},
    {"n_rebalance_max", required_argument, NULL, 'U'},
    {"network", required_argument, NULL, 'N'},
    {"circuit", required_argument, NULL, 'C'},
*/
void print_help(int, char **argv) {
  char *name = argv[0];
  char *tail = name + strlen(name);
  while(tail != name && *tail != '/') --tail;
  if(tail != name) ++tail;
  std::cerr << tail << " \\\n"
            << "    [(--help|h)]                           \\\n"
            << "    [(--verbose|-v)]                       \\\n"
            << "    [(--timescale|-t)         <value> ]    \\\n"
            << "    [(--ppn|-p)               <value> ]    \\\n"
            << "    [(--geometry|-g)      <v1 v2 ...> ]    \\\n"
            << "    [(--bandwidth|-b)         <value> ]    \\\n"
            << "    [(--latency|-l)           <value> ]    \\\n"
            << "    [(--cpubandwidth|-B)      <value> ]    \\\n"
            << "    [(--cpulatency|-L)        <value> ]    \\\n"
            << "    [(--router_router_bw|-R)  <value> ]    \\\n"
            << "    [(--router_node_bw|-S)    <value> ]    \\\n"
            << "    [(--netsharing|-T)        <string>]    \\\n"
            << "    [(--n_rebalance_max|-U)   <int>   ]    \\\n"
            << "    [(--network|-N)           <string>]    \\\n"
            << "    [(--random_indexing|-I)   <value> ]    \\\n"
            << "    [(--random_allocation|-A) <value> ]    \\\n"
            << "    [(--circuit|-C)]          <value> ]    \\\n"
            << "    [(--outfile|-o)           <string>]\n"
            << "Processors per node (ppn) defaults to 1 and must be > 0\n"
            << "Geometry has different meanings for different networks,\n"
            << "each network type has a default size of 1024 nodes\n"
            << "       - Torus takes 2 to 4 arguments: width, height,\n"
            << "         nodes-per-switch (def 0) and edge-width (def 1)\n"
            << "       - Torus3 takes 3 arguments (width, height, depth)\n"
            << "       - Fattree takes 2 to 4 arguments: levels, half-links, \n"
            << "         nodes-per-switch (def. 0) and edge-width (def. 1)\n"
            << "       - Gamma takes 2 to 4 arguments: degree, diameter,\n"
            << "         nodes-per-switch (def. 0) and edge-width (def. 1)\n"
            << "       - Crossbar takes one argument: number-of-nodes\n"
            << "Valid arguments for --network are\n"
            << "       \"torus\", \"fattree\", \"gamma\",\n"
            << "       \"torus3\" and \"crossbar\"\n"
            << "Valid arguments for --circuit are\n"
            << "       \"exclusive\" and \"shared\"\n"
            << "Valid arguments for --netsharing are\n"
            << "       \"ordered\", \"overcommit\",\n"
            << "       and \"throttle_overcommits\"\n"
            << "Bandwidth arguments take an integer value and optional units\n"
            << "       bits/sec,   bytes/sec,   kbits/sec, kibits/sec,\n"
            << "       kbytes/sec, kibytes/sec, mbits/sec, mibits/sec,\n"
            << "       mbytes/sec, mibytes/sec, gbits/sec, gibits/sec,\n"
            << "       gbytes/sec, gibytes/sec, tbits/sec, tibits/sec,\n"
            << "       tbytes/sec, tibytes/sec, pbits/sec, pibits/sec,\n"
            << "       pbytes/sec, or pibytes/sec\n"
            << "     abbreviated variants are also recognized, for example:\n"
            << "       bps, Bps, Kbps, Kibps, KBps, KiBps, .., PiBps, PBps\n"
            << "Latency arguments take an integer value and optional units\n"
            << "       psec, ps, nsec, ns, usec, us, msec, ms, sec, s\n"
            << "Random indexing and allocation take an integer seed value.\n";
}
