/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_MACROBRIDGE_MACROSENDER_H
#define SSTMAC_MACROBRIDGE_MACROSENDER_H

#include <sstmac/simulator/eventmanager.h>
#include <sstmac/process/context.h>
//#include <sst_config.h>
#include <sst/core/component.h>
#include <sst/core/eventFunctor.h>
#include <sst/core/link.h>

namespace sstmac {

  /**
   * A bridge component to instantiate a bridge between SST/macroscale
   * and SST/microscale.  The bridge will allow SST/macroscale to run
   * over the SST/microscale event manager and (eventually) transfer
   * events between the two scales.
   *
   * Since the SST/microscale implementation doesn't allow a component to
   * send events to itself, we provide two interfaces for the
   * event handler.  This is the sender (all event scheduling happens here).
   */
  class macrosender : public SST::Component {
    /// The port wiring this component to others.
    SST::Link *port_;

    /// Nested event manager type.
    class mgr;

    /// A static singleton of a manager.
    static boost::shared_ptr<mgr> mgr_;

    /// The stack size used for the python context.
    static const size_t python_stacksize_;

    /// The actual stack used for the python context.
    char *python_stack_;

    /// Static routine to run the python script in a separate context.
    static void python_routine(void *self);

    /// The main context (that was used to call Setup).
    threadcontext_t macro_context_;

    /// The context we will be using to run python.
    threadcontext_t python_context_;

    /// The script we plan to run.
    std::string script_;

  private:
    /// Copy constructor is blocked so we can identify possible copy issues.
    macrosender(const macrosender&) : 
      Component(*this)
    {}
  
    /// Assignment operator is blocked to identify possible copy issues.
    void operator=(const macrosender&) {}

    /// Private method to handle incoming events.
    /// This should never be called, but we can't do the wiring without it.
    void handle_events(SST::Event *evt);

    /// Schedule an event.
    void schedule(const timestamp &target,
                  const boost::shared_ptr<eventhandler> &handler);

    /// Private method to set up the wiring between components.
    void get_wired(const std::string &target);

  public:
    /// This constructor is called by the SST runtime framework.
    macrosender(SST::ComponentId_t id,
                SST::Component::Params_t &params);

    virtual ~macrosender();

    /// As far as I can tell, this method gets called by the runtime system
    /// after setup is complete (well that sure makes a lot of sense).
    int Setup();

    /// This should be what happens when we're all done
    /// (although you never can tell with some people).
    int Finish();

    /// This is needed so the scripting interface can get access
    /// to the event manager.
    /// This manager becomes available as soon as Setup is called.
    /// If needed, we may need to have a map of active managers to
    /// make sure that we always have an available one (this lack of
    /// memory management is killing me).
    /// \throw illformederror if Setup has not been called yet.
    static boost::shared_ptr<eventmanager> manager();
  };


} // end of namespace sstmac.

#endif // ! SSTMAC_MACROBRIDGE_MACROSENDER_H
