/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// crossbar.cc: Implementation of crossbar networks.
//
// Author: Curtis Janssen <cljanss@sandia.gov>

#include <sstream>
#include <sstmac/network/crossbar.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac {

  crossbar::crossbar(int nnode)
  {
    nodes_.clear();
    for (int i=0; i<nnode; i++) {
      std::vector<int> node(1);
      node[0] = i;
      nodes_.insert(node);
    }

    set<vertex> all_vertices(nodes_);
    std::vector<int> sw(1);
    sw[0] = -1;
    vertex swvertex(sw);
    all_vertices.insert(swvertex);

    set<edge> edges;
    for (set<vertex>::iterator i = nodes_.begin();
         i!=nodes_.end();
         i++) {
      edge e1(*i,swvertex);
      edge e2(swvertex,*i);
      edges.insert(e1);
      edges.insert(e2);
    }

    g_.init(all_vertices, edges);
    sw_index_ = g_.vertex_to_index(sw);
    init();
  }

  boost::shared_ptr<crossbar> crossbar::construct(int nnode) {
    return boost::shared_ptr<crossbar>(new crossbar(nnode));
  }

  crossbar::~crossbar() {
  }

  double
  crossbar::a() const
  {
    return 2.0;
  }

  int64_t
  crossbar::b() const
  {
    return n()/2;
  }

  void
  crossbar::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    vertex_index v1i = g().vertex_to_index(v1);
    vertex_index v2i = g().vertex_to_index(v2);
    if (v1i == v2i) {
      w.ev().clear();
      return;
    }
    w.ev().resize(2);
    w.ev()[0].init(v1i,sw_index_);
    w.ev()[1].init(sw_index_,v2i);
  }

  int64_t
  crossbar::diameter() const
  {
    return 2;
  }

  bool
  crossbar::vertex_symmetric() const
  {
    return true;
  }

  string
  crossbar::name() const
  {
    std::ostringstream ostr;
    ostr << "crossbar(" << n() << ")";
    return ostr.str();
  }

  bool
  crossbar::is_switch(vertex_index vi) const
  {
    return vi == sw_index_;
  }

}
