/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// edge.h: Interface for edges.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_NETWORK_EDGE_H
#define SSTMAC_NETWORK_EDGE_H

#include <sstmac/network/vertex.h>
#include <sstmac/util/cast.h>
#include <iosfwd>

namespace sstmac {

class graph;

/** Each edge in a graph is defined by a pair of vertices.  An edge is
    directional, leading from v1 to v2.
*/
class edge {
    vertex v1_, v2_;
    int id_; // need to distinguish edges between the same nodes
  public:
    edge(): id_(0) {}
    edge(const vertex &v1,const vertex &v2,
        int id = 0): v1_(v1), v2_(v2), id_(id) {}
    void init(const vertex &v1, const vertex &v2, int id=0) {
      v1_ = v1; v2_ = v2; id_ = id;
    }
    void init(const std::vector<int> &v1, const std::vector<int> &v2,
              int id=0) {
      v1_.init(v1); v2_.init(v2); id_ = id;
    }
    const vertex &from() const { return v1_; }
    const vertex &to() const { return v2_; }
    int id() const { return id_; }
    bool operator < (const edge &e) const;
    void print(std::ostream &o = std::cout) const;
};

/// Print out information about the given edge.
std::ostream& operator<<(std::ostream &o, const edge &e);

/** Each edge in a graph is defined by a pair of vertices.  An edge is
    directional, leading from v1 to v2. This is a lightweight edge,
    which uses vertex index objects rather that full indices.
*/
class lwedge {
    vertex_index v1_, v2_;
    int id_; // need to distinguish edges between the same nodes
  public:
    lwedge(): id_(0) {}
    lwedge(const vertex_index &v1,const vertex_index &v2,
            int id = 0): v1_(v1), v2_(v2), id_(id) {}
    lwedge(const graph &g,
            const vertex &v1,const vertex &v2,
            int id = 0);
    lwedge(const graph &g,
            const std::vector<int> &v1, const std::vector<int> &v2,
            int id = 0);
    void init(const vertex_index &v1, const vertex_index &v2, int id=0) {
      v1_ = v1; v2_ = v2; id_ = id;
    }
    void init(const graph &g, const vertex &v1, const vertex &v2, int id=0);
    void init(const graph &g,
              const std::vector<int> &v1, const std::vector<int> &v2,
              int id=0);
    const vertex_index &from() const { return v1_; }
    const vertex_index &to() const { return v2_; }
    int id() const { return id_; }

    bool operator < (const lwedge &e) const
    {
      if (v1_ < e.v1_) return true;
      else if (v1_ == e.v1_) {
        if (v2_ < e.v2_) return true;
        else if (v2_ == e.v2_) {
            if (id_ < e.id_) return true;
          }
      }
      return false;
    }

    bool operator == (const lwedge &e) const
    {
      if (v1_ == e.v1_)
        if (v2_ == e.v2_)
          if (id_ == e.id_) return true;
      return false;
    }

    void print(std::ostream &o = std::cout) const;

#ifndef SWIG // Do not parse nested classes with swig.
    class lwedge_hash {
    public:
      std::size_t operator() (const lwedge e) const {
        // use 16 bits from v1 and v2
        return ( (e.v1_.value()<<16) | (e.v2_.value() & 0xffff) ) & 0xffffffff;
      }
    };
#endif

};

/// Print out information about the given edge.
std::ostream& operator<<(std::ostream &o, const lwedge &e);

class edge_index {
public:
  typedef int edge_index_type;
private:
  edge_index_type ei_;
public:
  edge_index(): ei_(0) {}
  explicit edge_index(const int64_t&e):
  ei_(check_cast<edge_index_type>(e)) {}
  edge_index(const edge_index&e): ei_(e.ei_) {}
  //edge_index_type operator edge_index_type() const { return ei_; }
  edge_index operator = (const edge_index&e) {
    ei_ = e.ei_;
    return *this;
  }
  edge_index operator = (const edge_index_type&e) {
    ei_ = e;
    return *this;
  }
  bool operator != (const edge_index&e) const { return ei_ != e.ei_; }
  bool operator == (const edge_index&e) const { return ei_ == e.ei_; }
  bool operator < (const edge_index&e) const { return ei_ < e.ei_; }
  bool operator > (const edge_index&e) const { return ei_ > e.ei_; }
  bool operator <= (const edge_index&e) const { return ei_ <= e.ei_; }
  bool operator >= (const edge_index&e) const { return ei_ >= e.ei_; }
  edge_index_type value() const { return ei_; }
};

std::ostream& operator<<(std::ostream&s,const edge_index&e);

}

#endif
