/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// fattree.cc: Implementation of fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstream>
#include <sstmac/network/fattree.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac {

  static bool
  incr_vector(int l, int k, vector<int>& v)
  {
    for (int i=l-1; i>=0; i--) {
      v[i]++;
      if (v[i] == k) {
        v[i] = 0;
      }
      else return true;
    }
    return false;
  }

  static bool
  incr_vector2(int l, int ki, int kl, vector<int>& v)
  {
    int i=l-1;
    v[i]++;
    if (v[i] == kl) {
      v[i] = 0;
    }
    else return true;
    i--;
    for (; i>=0; i--) {
      v[i]++;
      if (v[i] == ki) {
        v[i] = 0;
      }
      else return true;
    }
    return false;
  }

  static void
  compute_nodes(int l, int k, set<vertex> &vertices, int nndpersw)
  {
    vertices.clear();

    // The node vertices are all possible l-digit numbers.
    // The first l-1 digits are base k.
    // The last digit is base nndpersw.
    vertex v(l);
    fill(v.v().begin(), v.v().end(), 0);
    do {
      vertices.insert(v);
    } while (incr_vector2(l,k,nndpersw,v.v()));
  }

  static void
  compute_switches(int l, int k, set<vertex> &vertices,
                   vector<set<vertex> > &tiers)
  {
    vertices.clear();

    // Switch vertices have an extra digit with the tier and
    // the last significant digit is zero.

    for (int i=1; i<=l; i++) {
      tiers[i].clear();
      vector<int> v(l-1);
      fill(v.begin(), v.end(), 0);
      do {
        vertex vert;
        vert.v().resize(l);
        // switches are distigushed from nodes by the have a negative
        // coordinate (the tier number)
        vert.v()[0] = -i;
        for (int j=1; j<l; j++) vert.v()[j] = v[j-1];
        // For the top tier, the next to last significant digit rounded
        // down to the nearest even number, since the up links are
        // used as down links.
        if (i == l) vert.v()[l-1] &= ~1;
        vertices.insert(vert);
        tiers[i].insert(vert);
      } while (incr_vector(l-1,k,v));
    }
  }

  static void
  compute_edges(int l, int k, set<edge> &edges,
                vector<set<vertex> > &tiers, int nndpersw)
  {
    edges.clear();

    // Compute the edges between the nodes and tier 1.
    vertex sv(l); // sv is a switch index
    sv[0] = -1;
    // loop through the nodes (tiers[0])
    for (set<vertex>::iterator i = tiers[0].begin();
         i != tiers[0].end(); i++) {
      // Copy the nodes indices into the switch index.  The last node
      // index is ignored: it enumerates the nodes on a single switch.
      for (int j=0; j<i->v().size()-1; j++) {
        sv.v()[j+1] = i->v()[j];
      }
      edges.insert(edge(sv, *i));
      edges.insert(edge(*i, sv));
    }

    // Compute the remaining edges.
    for (int il=1; il<l; il++) {
      for (set<vertex>::iterator iv = tiers[il].begin();
           iv != tiers[il].end(); iv++) {
        for (int ik=0; ik<k; ik++) {
          vertex up(*iv);
          up.v()[0]--;
          up.v()[l-il] = ik;
          int edge_id = 0;
          if (il+1==l) {
            // On the top tier, the next to last digit is even.
            // Edges that are folded into a different switch
            // are given an id of 1.
            if (up.v()[l-1] & 1) edge_id = 1;
            up.v()[l-1] &= ~1;
          }
          edges.insert(edge(*iv,up,edge_id));
          edges.insert(edge(up,*iv,edge_id));
        }
      }
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::fattree",
  //                               sstmac::xmlclass::ctor_template<fattree>);

  // fattree::fattree(TiXmlHandle h)
  // {
  //   int l = 2;
  //   int k = 2;
  //   h.ToElement()->QueryIntAttribute("l",&l);
  //   h.ToElement()->QueryIntAttribute("k",&k);
  //   init_fattree(l,k);
  // }

  fattree::fattree(int l, int k, int nndpersw, int widtharg)
  {
    init_fattree(l,k,nndpersw,widtharg);
  }

  void
  fattree::init_fattree(int l, int k, int nndpersw, int widtharg)
  {
    width = widtharg;
    l_ = l;
    k_ = k;

    if (l<2) throw std::runtime_error("fattree: must have l > 1");
    if (k&1) throw std::runtime_error("fattree: must have k even");

    vector<set<vertex> > tiers(l+1);

    if (nndpersw == 0) nndpersw = k;

    compute_nodes(l, k, nodes_, nndpersw);
    tiers[0] = nodes_;

    set<vertex> switches;
    compute_switches(l, k, switches, tiers);

    set<edge> edges;
    compute_edges(l, k, edges, tiers, nndpersw);

    set<vertex> all_vertices;
    all_vertices.insert(nodes_.begin(), nodes_.end());
    all_vertices.insert(switches.begin(), switches.end());
    g_.init(all_vertices, edges);
    init();
  }

  boost::shared_ptr<fattree> fattree::construct(int levels, int half_links,
                                                int nndpersw, int widtharg) {
    return boost::shared_ptr<fattree>(new fattree(levels, half_links,
                                                  nndpersw, widtharg));
  }

  fattree::~fattree() {
  }

  size_t
  fattree::size(int l, int k)
  {
    size_t r = 1;
    for (int i=0; i<l; i++) {
      r *= k;
    }
    return r;
  }

  double
  fattree::a() const
  {
    return l_*2.0;
  }

  int64_t
  fattree::b() const
  {
    return n()/2;
  }

  void
  fattree::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    throw std::runtime_error("fattree::shortest_path: not unique");
  }

  // This algorithm routes messages using the least significant bits as a
  // hash.  This will better distribute the load in the network.  In a fat
  // tree with three layers of switches, the route from (s2,s1,s0) to
  // (t2,t1,t0) is:
  //     (s2,s1,s0)
  //   (1,s2,s1, 0)
  //   (2,s2,t0, 0)
  //   (3,t1,t0, 0)  Pairs of these switches get folded into a single switch
  //   (2,t2,t0, 0)
  //   (1,t2,t1, 0)
  //     (t2,t1,t0)
  void
  fattree::route(const vertex &v1, const vertex &v2, walk &w) const
  {
    if (v1 == v2) {
      walk empty;
      w = empty;
      return;
    }

    vertex old_vertex(v1);
    vector<lwedge> edges;
    edges.reserve(2*l_);

    // find the level at which target and source switch differ
    int diff_level;
    for (diff_level = l_; diff_level > 0; diff_level--) {
      if (v1[l_-diff_level] != v2[l_-diff_level]) break;
    }

    // find the target switch
    vertex target_switch;
    target_switch.v().resize(v1.v().size());
    target_switch.v()[0] = -1; // tier 1 is the last switch hop
    for (int i=1; i<v2.v().size(); i++) target_switch.v()[i]=v2.v()[i-1];

    // take us to the first switch
    vertex new_vertex;
    new_vertex.v().resize(v1.v().size());
    new_vertex.v()[0] = -1; // tier 1 is the first hop
    for (int i=1; i<v1.v().size(); i++) new_vertex.v()[i]=v1.v()[i-1];
    edges.push_back(g().generate_lwedge(old_vertex,new_vertex));
    old_vertex = new_vertex;

    // move up the tree
    int edge_id = 0;
    for (int level=2; level<=diff_level; level++) {
      new_vertex[0] = -level;
      // This uses v2's indices to route up, having the effect
      // of load balancing routes. The mod k_ is needed because
      // there may be more nodes on a switch than k_.
      new_vertex[l_-level+1] = v2.v()[l_-level+1]%k_;
      if (level == l_) {
        // The top-level switches have twice as many downward links
        // available, and half as many switches are used. This
        // computes the correct switch address.
        edge_id = new_vertex[l_-1] & 1;
        new_vertex[l_-1] &= ~1;
      }
      edges.push_back(g().generate_lwedge(old_vertex,new_vertex,edge_id));
      old_vertex = new_vertex;
    }

    // move down the tree
    for (int level=diff_level-1; level>0; level--) {
      new_vertex[0] = -level;
      new_vertex[l_-1] += edge_id;
      new_vertex[l_-level] = target_switch[l_-level];
      edges.push_back(g().generate_lwedge(old_vertex,new_vertex,edge_id));
      edge_id = 0;
      old_vertex = new_vertex;
    }

    // take us to the target
    edges.push_back(g().generate_lwedge(target_switch,v2));
    walk r(edges);
    w = r;
  }

  int
  fattree::l() const
  {
    return l_;
  }

  int
  fattree::k() const
  {
    return k_;
  }

  int64_t
  fattree::diameter() const
  {
    return (l()+1)*2;
  }

  bool
  fattree::vertex_symmetric() const
  {
    return true;
  }

  string
  fattree::name() const
  {
    std::ostringstream ostr;
    ostr << "FatTree(" << l_ << "," << k_ << ")";
    return ostr.str();
  }

  int
  fattree::edge_width(edge_index e) const
  {
    const lwedge &lwe = g().index_to_edge(e);
    if (is_switch(lwe.from()) && is_switch(lwe.to())) return width;
    return 1;
  }

  bool
  fattree::is_switch(vertex_index vi) const
  {
    const vertex &v = g().index_to_vertex(vi);
    if (v[0] < 0) return true;
    return false;
  }

}
