/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// gamma.cc: Implementation of gamma graph networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

// Gamma_Delta(D)
// Delta = d = degree
// D     = k = diameter

// a vertex is a word of length k
// the alphabet has d + 1 letters: [0, 1, ..., d]

// the gamma graph vertices are all permutations of k distinct
// letters of this alphabet


#include <stdexcept>
#include <algorithm>
#include <sstream>
#include <sstmac/network/gamma.h>

#include <sstmac/network/walk.h>

using std::set;
using std::vector;
using std::string;
using std::find;

namespace sstmac {

  // c(x_0,...,x_(k-1)) = |{x in T_k | x != x_i and x > x_(k-1) in Tk}|
  static int
  c_Tk(int k, const vector<int> &Tk, const vertex &v)
  {
    // Note that Tk and v have an extra digit which gives the node
    // number (if 0, then this is a network element).  That is why
    // --v.v().end() and --Tk.end() are used below.
    int n = 0;
    for (vector<int>::const_iterator x = Tk.begin();
         x != --Tk.end();
         x++) {
      if (find(v.v().begin(), --v.v().end(), *x) != --v.v().end()) continue;
      if (find(x, --Tk.end(), v[k-1]) == --Tk.end()) n++;
    }
    return n;
  }

  static void
  compute_network_vertices(const set<int> &letters, int i, int k,
                   vertex &v, set<vertex> &vertices)
  {
    if (i == k) {
      vertices.insert(v);
      return;
    }
  
    for (set<int>::const_iterator l = letters.begin();
         l != letters.end(); l++) {
      set<int> next_letters(letters);
      v[i] = *l;
      next_letters.erase(*l);
      compute_network_vertices(next_letters, i+1, k, v, vertices);
    }
  }

  static void
  compute_network_vertices(int d, int k, set<vertex> &vertices)
  {
    // The first k elements are the switch address.
    // Element k is 0 for a switch and nonzero for non-direct nodes.
    vertex v(k+1); v[k] = 0;
    set<int> letters;
    for (int i=0; i<=d; i++) letters.insert(i);

    compute_network_vertices(letters, 0, k, v, vertices);
  }

  // Faber and Moore edges
  static void
  compute_edges_1(set<vertex> &vertices, int d, int k, set<edge> &edges)
  {
    for (set<vertex>::iterator v = vertices.begin();
         v != vertices.end();
         v++) {
      vertex from = *v;
      vertex to(k+1); to[k] = 0;
      for (int i=1; i<k; i++) {
        for (int j=1; j<=i; j++) to[j] = from[j-1];
        for (int j=i+1; j<k; j++) to[j] = from[j];
        to[0] = from[i];
        edges.insert(edge(from,to));
      }
      set<int> unused_letters;
      for (int i=0; i<=d; i++) unused_letters.insert(i);
      for (int i=0; i<k; i++) unused_letters.erase(from[i]);
      for (int i=0; i<k-1; i++) to[i+1] = from[i];
      for (set<int>::iterator l = unused_letters.begin();
           l != unused_letters.end();
           l++) {
        to[0] = *l;
        edges.insert(edge(from,to));
      }
    }
  }

  // Comellas and Fiol edges
  // These are compatible with the shortest_path routine.
  static void
  compute_edges_2(set<vertex> &vertices, int d, int k, set<edge> &edges)
  {
    for (set<vertex>::iterator v = vertices.begin();
         v != vertices.end();
         v++) {
      vertex from = *v;
      vertex to(k+1); to[k] = 0;
      for (int i=0; i<k-1; i++) {
        for (int j=0; j<i; j++) to[j] = from[j];
        for (int j=i; j<k-1; j++) to[j] = from[j+1];
        to[k-1] = from[i];
        edges.insert(edge(from,to));
      }
      set<int> unused_letters;
      for (int i=0; i<=d; i++) unused_letters.insert(i);
      for (int i=0; i<k; i++) unused_letters.erase(from[i]);
      for (int i=0; i<k-1; i++) to[i] = from[i+1];
      for (set<int>::iterator l = unused_letters.begin();
           l != unused_letters.end();
           l++) {
        to[k-1] = *l;
        edges.insert(edge(from,to));
      }
    }
  }

  static void
  compute_node_vertices_edges(int nnps,
                              const set<vertex> &all_vertices,
                              set<vertex> &nodes_, set<edge> &all_edges)
  {
    nodes_.clear();
    for (set<vertex>::iterator i = all_vertices.begin();
         i != all_vertices.end(); i++) {
      const vertex &sw = *i;
      vertex node = sw;
      for (int j=0; j<nnps; j++) {
        node[node.v().size()-1] = j + 1;
        nodes_.insert(node);
        all_edges.insert(edge(sw,node));
        all_edges.insert(edge(node,sw));
      }
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::gamma",
  //                               sstmac::xmlclass::ctor_template<gamma>);

  // gamma::gamma(TiXmlHandle h)
  // {
  //   int d = 4;
  //   int k = 2;
  //   h.ToElement()->QueryIntAttribute("d",&d);
  //   h.ToElement()->QueryIntAttribute("k",&k);
  //   init_gamma(d,k);
  // }

  gamma::gamma(int d, int k, int nnps, int widtharg)
  {
    init_gamma(d,k,nnps,widtharg);
  }

  void
  gamma::init_gamma(int d, int k, int nnps, int widtharg)
  {
    width = widtharg;
    d_ = d;
    k_ = k;
    if (d_ < k_) {
      throw std::invalid_argument("gamma: must have d >= k");
    }

    set<vertex> all_vertices;

    compute_network_vertices(d_, k_, all_vertices);

    set<edge> all_edges;
    compute_edges_1(all_vertices, d_, k_, all_edges);

    if (nnps == 0) {
      nodes_ = all_vertices;
    }
    else {
      compute_node_vertices_edges(nnps, all_vertices, nodes_, all_edges);
      std::copy(nodes_.begin(), nodes_.end(),
                std::inserter(all_vertices, all_vertices.begin()));
    }

    g_.init(all_vertices, all_edges);
    init();
    if (nnps > 0) init_node_to_switch();
  }

  boost::shared_ptr<gamma> gamma::construct(int degree, int diameter,
                                            int nnodeperswitch,
                                            int widtharg) {
    return boost::shared_ptr<gamma>(new gamma(degree, diameter,
                                              nnodeperswitch, widtharg));
  }

  gamma::~gamma() {
  }

  int64_t
  gamma::shortest_switch_path_length(const vertex &v1, const vertex &v2) const
  {
    int64_t j;

    // Phase I
    set<int> Tk;
    for (int i=0; i<k_; i++) Tk.insert(v1[i]);
    if (Tk.find(v2[k_-1]) == Tk.end()) {
      j = k_;
    }
    else {
      int c = c_Tk(k_,v1.v(),v2);
      int n = 0;
      for (int i=0; i<k_; i++) if (Tk.find(v2[i])==Tk.end()) n++;
      if (n > c) {
        j = k_;
      }
      else {
        const vector<int> &vv1 = v1.v();
        for (j=k_-1; j>=1; j--) {
          if (Tk.find(v2[j-1]) == Tk.end()) break;
          // Note that vv1 has an extra digit which gives the node
          // number (if 0, then this is a network element).  That is
          // why --vv1.end() are used below.
          if (find(find(vv1.begin(),--vv1.end(),v2[j-1]),
                   --vv1.end(),
                   v2[j]) == --vv1.end()) break;
        }
      }
    }

    return j;
  }

  int64_t
  gamma::shortest_path_length(const vertex &v1, const vertex &v2) const
  {
    int64_t j = shortest_switch_path_length(v1, v2);

    int end_hops = 0;
    if (!is_switch(g().vertex_to_index(v1))) end_hops++;
    if (!is_switch(g().vertex_to_index(v2))) end_hops++;
    return j + end_hops;
  }

  void
  gamma::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    vertex_index vi1 = g().vertex_to_index(v1);
    vertex_index vi2 = g().vertex_to_index(v2);

    if (vi1 == vi2) {
      w.ev().clear();
      return;
    }

    vector<vertex_index> vertices;

    vertex_index si1, si2;
    if (is_switch(vi1)) si1 = vi1;
    else                si1 = node_to_switch(vi1);
    if (is_switch(vi2)) si2 = vi2;
    else                si2 = node_to_switch(vi2);

    if (si1 == si2) {
      w.ev().resize(2);
      w.ev()[0] = lwedge(vi1,si1);
      w.ev()[1] = lwedge(si1,vi2);
      return;
    }

    const vertex &s1 = g().index_to_vertex(si1);
    const vertex &s2 = g().index_to_vertex(si2);

    int64_t j = shortest_switch_path_length(s1,s2);

//     std::cout << "routing from " << v1
//               << " to " << v2
//               << " in " << j << " backbone steps"
//               << std::endl;

    // hop from the node to the first switch
    if (si1 != vi1) {
//       std::cout << "next vertex = " << v1 << std::endl;
      vertices.push_back(vi1);
    }

    // Phase II
    vector<int> y(s1.v());
    if (j > 0) {
//       std::cout << "next vertex = " << y << std::endl;
      vertices.push_back(g().vertex_to_index(y));
    }
    for (int64_t i=j-1; i>=0; i--) {
      int x = s2[i];
      int loc;
      for (loc = 0; loc < y.size() - 1; loc++) {
        if (x == y[loc]) break;
      }
      // shift everything between 0 and loc - 1 to the right
      int ifence = ((loc < y.size() - 1)?loc:(loc-1));
      for (int i=ifence-1; i>=0; i--) y[i+1] = y[i];
      y[0] = x;
//       vector<int>::iterator loc = find(y.begin(), y.end(), x);
//       if (loc == y.end()) y.pop_back();
//       else y.erase(loc);
//       y.insert(y.begin(), x);
//       std::cout << "next vertex = " << y << std::endl;
      vertices.push_back(g().vertex_to_index(y));
    }

    // hop from the last switch to the node
    if (si2 != vi2) {
//       std::cout << "next vertex = " << v2 << std::endl;
      vertices.push_back(vi2);
    }

    w.init(vertices);
  }

  int
  gamma::d() const
  {
    return d_;
  }

  int64_t
  gamma::diameter() const
  {
    return k_;
  }

  bool
  gamma::vertex_symmetric() const
  {
    return true;
  }

  string
  gamma::name() const
  {
    std::ostringstream ostr;
    ostr << "Gamma_" << d_ << "(" << k_ << ")";
    return ostr.str();
  }

  bool
  gamma::is_switch(vertex_index v) const
  {
    const vertex &ver = g().index_to_vertex(v);
    if (ver[ver.v().size()-1] == 0) return true;
    return false;
  }

  int
  gamma::edge_width(edge_index e) const
  {
    const lwedge &lwe = g().index_to_edge(e);
    if (is_switch(lwe.from()) && is_switch(lwe.to())) return width;
    return 1;
  }

}
