/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// gamma.h: Interface for gamma networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_NETWORK_GAMMA_H
#define SSTMAC_NETWORK_GAMMA_H

#include <sstmac/network/network.h>

namespace sstmac {

  /** A gamma graph is a network with a relatively
      high node count and low diameter
      for a network elements of a given degree.  Each
      vertex in such a network can be represented as a word of length k
      formed from an alphabet having d + 1 letters: [0, 1, ..., d]. The
      vertices of the gamma graph are all permutations of k distinct letters
      of this alphabet, where k is the diameter and d is the degree. The
      network links are unidirectional, with each vertex having
      d inbound and d outbound edges. The degree must be greater than or
      equal to the diamater. The inputs for this network
      can specify the following attributes:

      <table border="1">

      <tr><td>Keyword<td>Type<td>Default<td>Description

      <tr><td><tt>d</tt><td>int<td>4<td>The degree of the switching elements.

      <tr><td><tt>k</tt><td>int<td>2<td>The diameter of the graph.

      <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.

      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.

      </table>

      The following table gives the number of nodes in
      gamma graphs of varying degree and diameter.

      <table border="1">

      <tr><td rowspan="2">Degree<td colspan="8" align="center">Number of Nodes
      <tr>     <td>k=1<td>k=2<td>k=3<td>k=4<td>k=5<td>k=6<td>k=7<td>k=8
      <tr><td>1<td>2<td>-<td>-<td>-<td>-<td>-<td>-<td>-
      <tr><td>2<td>3<td>6<td>-<td>-<td>-<td>-<td>-<td>-
      <tr><td>3<td>4<td>12<td>24<td>-<td>-<td>-<td>-<td>-
      <tr><td>4<td>5<td>20<td>60<td>120<td>-<td>-<td>-<td>-
      <tr><td>5<td>6<td>30<td>120<td>360<td>720<td>-<td>-<td>-
      <tr><td>6<td>7<td>42<td>210<td>840<td>2,520<td>5,040<td>-<td>-
      <tr><td>7<td>8<td>56<td>336<td>1,680<td>6,720<td>20,160<td>40,320<td>-
      <tr><td>8<td>9<td>72<td>504<td>3024<td>15,120<td>60,480<td>181,440<td>362,
880

      </table>
  */

  class gamma: public network {
  public:
    int width;
  private:
    int d_, k_;
    void init_gamma(int d, int k, int nnps, int widtharg);

  protected:
    //gamma(TiXmlHandle);
    // Create a gamma graph network with degree d, diameter k, and
    // nnps nodes per switch. If nnps is zero, then a direct network is
    // used.
    gamma(int d, int k, int nnps, int widtharg);

  public:
    /// Create a gamma graph network
    /// @param degree The degree of the switch elements.
    /// @param diameter The diameter of the graph core (not counting node links)
    /// @param nnodeperswitch The number of nodes connected to each switch.
    /// @param widtharg The width of the interswitch links
    /// If nnodeperswitch is zero, then a direct network is
    /// used. The width of links involving nodes is always one.
    static boost::shared_ptr<gamma> construct(int degree = 2, int diameter = 4,
                                              int nnodeperswitch = 0,
                                              int widtharg = 1);
    virtual ~gamma();

    int64_t shortest_switch_path_length(const vertex &v1, const vertex &v2) const;
    int64_t shortest_path_length(const vertex &v1, const vertex &v2) const;
    void shortest_path(const vertex &v1, const vertex &v2, walk &) const;
    int d() const;
    int64_t diameter() const;
    //double a() const;
    //int b() const;
    bool vertex_symmetric() const;
    std::string name() const;
    bool is_switch(vertex_index v) const;
    int edge_width(edge_index e) const;
  };

}

#endif
