/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// ring.h: Interface for ring networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_NETWORK_RING_H
#define SSTMAC_NETWORK_RING_H


#include <sstmac/network/network.h>

namespace sstmac {

  /** Construct a ring. The inputs for this network
      can specify the following attributes:

      <table border="1">
      <tr><td>Keyword<td>Type<td>Default<td>Description
      <tr><td>nsw<td>int<td>0<td>The number of switches in the ring.
      <tr><td>nndpersw<td>int<td>0<td>The number of nodes per switch.
      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
      </table>

      If the number of nodes per switch is zero, then a direct
      network is used. Otherwise, the given number of nodes
      is attached to the switch, each through their own link.
      The width of edges between a node and a switch is always one.
  */

  class ring: public network {
  public:
    bool mesh_routing;
    int width;
  private:
    int nx_;
    bool direct_;
    int nndpersw_;

    bool is_switch(const vertex_index &vi) const {
      if (direct_) return true;
      const vertex &v = g().index_to_vertex(vi);
      if (v[1] == nndpersw_) return true;
      return false;
    }
  protected:
    ring(int nsw, int nndpersw, int widtharg);
    //ring(TiXmlHandle);

  public:
    /** Construct a ring.
        @param nsw The number of switches in the ring.
        @param nndpersw The number of nodes per switch.
        @param widtharg The width of edges connecting two switches
        If the number of nodes per switch is zero, then a direct
        network is used. Otherwise, the given number of nodes
        is attached to the switch, each through their own link.
        The width of edges between a node and a switch is always one.
    */
    static boost::shared_ptr<ring> construct(int nsw, int nndpersw=0,
                                             int widtharg = 1);
    virtual ~ring();

    /** Return the width of the given edge. Edges involving nodes
        always have a width of one. The width of edges between two
        switches is specified in the constructor or by setting the
        public width member. */
    int edge_width(edge_index e) const;

    void shortest_path(const vertex &v1, const vertex &v2, walk &) const;
    void next_hop(vertex_index loc, vertex_index dst,
                  std::vector<lwedge> &nxt) const;
    int64_t diameter() const;
    int64_t b() const;
    bool vertex_symmetric() const;
    std::string name() const;
  };

}

#endif
