/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/actormagic.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Construct a new barrier handler -- private nested class method.
  //
  actormagic::barr::barr() :
    key_(key::construct())
  {}

  //
  // Join a barrier -- private nested class method.
  //
  void actormagic::barr::join(actorid id, const std::set<actorid> &group,
                            const shared_ptr<scheduler>& sched)
  {
    if(group.find(id) == group.end())
      throw illformederror("actormagic::barr::join:  Attempting to join a "
                           "barrier group without being a member of the group.");
    if(group_.empty()) {
      group_ = group_;
    }
    else {
      // Not high performance, but quick and easy to understand :)
      std::set<actorid> group_union(group_);
      group_union.insert(group.begin(), group.end());
      if(group_union.size() != group_.size())
        throw valueerror("actormagic::barr::join:  Input group differs from "
                         "other input groups in the same barrier.");
    }
    entered_.insert(id);
    if(entered_.size() < group_.size()) {
      // not done yet.
      sched->block(key_);
    }
    else {
      // Last entrant.
      sched->unblock(key_);
    }
  }

  //
  // Constructor.
  //
  actormagic::actormagic(const shared_ptr<scheduler>& sched) :
    sched_(sched)
  {}

  //
  // External objects can allocate a new object using this static method.
  //
  shared_ptr<actormagic> actormagic::construct(const shared_ptr<scheduler>& sched) {
    return shared_ptr<actormagic>(new actormagic(sched));
  }

  //
  // A process can be added to the world.
  //
  bool actormagic::add(const actorid &id) {
    if(! id.is_real())
      throw valueerror("actormagic::add:  Adding an invalid process id.");
    bool rv = (world_.find(id) == world_.end());
    if(rv) world_.insert(id);
    return rv;
  }

  //
  // A process can be removed from this world.
  //
  bool actormagic::remove(const actorid &id) {
    if(! id.is_real())
      throw valueerror("actormagic::remove:  Removing an invalid process id.");
    bool rv = (world_.find(id) != world_.end());
    if(rv) world_.erase(id);
    return rv;
  }

  //
  // Who is in our world?
  //
  const std::set<actorid>& actormagic::world() const {
    return world_;
  }

  //
  // Join a barrier.
  //
  void actormagic::barrier(int tag, const actorid &id, const std::set<actorid> &group)
  {
    barrier_[tag].join(id, group, sched_);
    // When any process gets here, the barrier is complete.
    barrier_t::iterator it = barrier_.find(tag);
    if(it != barrier_.end()) barrier_.erase(it);
  }

  //
  // Create or move a global counter.
  //
  void actormagic::
  move_global_counter(const std::string &label, int increment) {
    counter_[label] += increment;
  }

  //
  // Read a global counter.
  //
  int actormagic::read_global_counter(const std::string &label) const {
    counter_t::const_iterator it = counter_.find(label);
    if(it == counter_.end())
      throw valueerror(std::string("actormagic::read_global_counter:  Invalid"
                                   "counter label \"") + label + "\"");
    return it->second;
  }

} // end of namespace sstmac
