/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/applaunch.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/app.h>
#include <sstmac/process/applaunch-completion.h>
#include <sstmac/processor/computeapi.h>
#include <sstmac/simulator/environment.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Thread exit signal.
  //
  void applaunch::scheduler_exit(const timestamp&) {
    environment_->exit();
    environment_ = shared_ptr<environment>();
  }

  //
  // Protected constructor.
  //
  applaunch::applaunch() :
    scheduler_(scheduler::construct())
  { }
  
  //
  // Who am I?
  //
  void applaunch::set_self(const shared_ptr<applaunch>& self) {
    if(self.get() != this) {
      throw valueerror("applaunch::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<applaunch> applaunch::get_self() {
    return self_.lock();
  }

  //
  // Perform setup for a run
  //
  void applaunch::run_setup(const shared_ptr<environment>& my_world) {
    environment_ = my_world;
    scheduler_->clear();
    scheduler_->set_completion( completion::construct(get_self()) );
  }

  //
  // Public constructor
  //
  shared_ptr<applaunch> applaunch::construct()
  {
    shared_ptr<applaunch> retval(new applaunch);
    retval->set_self(retval);
    return retval;
  }

  //
  // Destructor
  //
  applaunch::~applaunch() throw() {}

  //
  // Set application map
  //
  void applaunch::set_applications(std::map<taskid,boost::shared_ptr<app> > apps)
  { apps_ = apps; }
  
  //
  // Access the scheduler.
  //
  shared_ptr<scheduler> applaunch::get_scheduler() {
    if( scheduler_ == 0 )
      throw nullerror("applaunch:get_scheduler: null pointer."
                      "  set_scheduler must be called on the launcher.");
    return scheduler_;
  }

  //
  // Define the scheduler.
  //
  void applaunch::set_scheduler(const shared_ptr<scheduler>& sched) {
    if(sched == 0) {
      throw valueerror("applaunch::set_scheduler: null pointer.");
    }
    scheduler_ = sched;
  }

  //
  // Access the environment.
  //
  shared_ptr<environment> applaunch::get_environment() {
    if(environment_ == 0)
      throw nullerror("applaunch::get_environment: null pointer."
                      "  run_setup should initialize the pointer.");
    return environment_;
  }

  //
  // Define the environment.
  //
  void applaunch::set_environment(const shared_ptr<environment>& env) {
    if(env == 0) {
      throw valueerror("applaunch::set_environment: null pointer.");
    }
    environment_ = env;
  }

  //
  // Description
  // Inherited from launcher.
  //
  const std::string& applaunch::description() const
  {
    static const std::string desc("applaunch");
    return desc;
  }

  //
  // Completed? 
  // Inherited from launcher.
  //
  bool applaunch::completed() const {
    if(scheduler_ != 0 && environment_ != 0)
      scheduler_->report_deadlock();
    return (environment_ == 0);
  }

  //
  // Fire this object off.
  // Inherited from launcher.
  //
  void applaunch::run(const timestamp &current_time,
                   const shared_ptr<environment>& my_world)
  {
    run_setup(my_world);
    shared_ptr<tasklist> tlist = environment_->get_tasklist();
    int64_t ntask = tlist->size();
    for(int64_t i = 0; i < ntask; ++i) {
      taskid id(i);
      shared_ptr<nodeinterface> the_node = environment_->get(id);
      shared_ptr<computeapi> compapi =
        computeapi::construct(apps_[id], the_node);
      apps_[id]->init_computeapi(compapi);
      scheduler_->add_thread(apps_[id]);
    }
    // Go.
    scheduler_->run_threads();
  }


} // end of namespace sstmac
