/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/context.h>

namespace sstmac {

#if defined(SSTMAC_USE_PTHREAD)
  
  static pthread_mutex_t context_switch_lock = PTHREAD_MUTEX_INITIALIZER;

  void *pthreadfunc(void*args)
  {
    threadargs *thread_info = (threadargs*)(args);
    
    // The following code handles the initial context switch
    // starting this thread. After that, swap_context
    // performs all context switches.
    
    // Grab the context lock.
    pthread_mutex_lock(&context_switch_lock);
    
    // Put this thread to sleep, if it has not
    // already been started.
    if (! thread_info->context->started) {
      pthread_cond_wait(&thread_info->context->ready, &context_switch_lock);
    }
        
    pthread_mutex_unlock(&context_switch_lock);
    
    // Run the thread
    (*thread_info->func)(thread_info->args);
    
    delete thread_info;
    
    return NULL;
  }
  
  /// This part of context initialization is common to
  /// init_context and start_context.
  void init_context_common(threadcontext_t &context) {
    pthread_cond_init(&context.ready, NULL);
    context.started = false;
  }
  
  /// Initialize the context to be that for the currently running thread.
  void init_context(threadcontext_t &context) {
    init_context_common(context);
    context.thread = pthread_self();
  }
  
  /// This tears down the context. It is only called from the scheduler's thread.
  void destroy_context(threadcontext_t &context) {
    pthread_cond_destroy(&context.ready);
    // This can be called for the main thread's context as well,
    // in which case we cannot join.
    if (!pthread_equal(pthread_self(),context.thread)) {
      pthread_join(context.thread, NULL);
    }
  }
  
  /// Start a new context. It does not start yet -- swap_context will start it. This is only called from the scheduler's thread.
  void start_context(threadcontext_t &context,
                     void *stack, size_t stacksize,
                     void(*func)(void*), void *args,
                     threadcontext_t *yield_to)
  {
    pthread_attr_t thread_attr;
    pthread_attr_init(&thread_attr);
    //pthread_attr_setstack(&thread_attr, stack, stacksize);
    init_context_common(context);
    
    threadargs *targs = new threadargs(func, args, &context);
    pthread_create(&context.thread, &thread_attr, &pthreadfunc, (void*)targs);
    pthread_attr_destroy(&thread_attr);
  }
  
  void complete_context(threadcontext_t&from, threadcontext_t &to)
  {
    static const std::string here("complete_context (pthread impl): ");
    if (!pthread_equal(pthread_self(), from.thread)) {
      throw schedulererror(here + "done from thread other than \"from\" thread");
    }
    pthread_cond_signal(&to.ready);
    pthread_exit(0);
  }
 
  /// Swap context. The from context is always the currently running context.
  void swap_context(threadcontext_t &from, threadcontext_t &to) {
    static const std::string here("swap_context (pthread impl): ");
    if (!pthread_equal(pthread_self(), from.thread)) {
      throw schedulererror(here + "done from thread other than \"from\" thread");
    }
    
    // incorrect ordering for c1->c2->c1 context switches:
    // c1 signals c2 to start
    // c2 signals c1 to start
    // c1 begins waiting on signals (deadlock)
    //
    // correct order:
    // (lock) c1 signals c2 to start
    // c1 begins waiting on signals (unlock)
    // (lock) c2 signals c1 to start
    // c2 begins waiting on signals (unlock)
    
    // Grab the context lock.
    pthread_mutex_lock(&context_switch_lock);
    
    to.started = true;

    // Wake up the to context.
    pthread_cond_signal(&to.ready);

    // Put the from thread to sleep.
    pthread_cond_wait(&from.ready, &context_switch_lock);
    
    pthread_mutex_unlock(&context_switch_lock);
  }
  
#endif

} // end of namespace sstmac

