/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/dumpitracer.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/cast.h>
#include <dumpi/common/types.h>
#include <dumpi/common/funcs.h>
#include <dumpi/common/dumpiio.h>
#include <dumpi/common/io.h>
#include <dumpi/libdumpi/libdumpi.h>
#include <dumpi/common/debugflags.h>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string.h>

namespace sstmac {

  using boost::shared_ptr;

  static const int thread = 0;

  //
  // Construct a dumpitracer and give it a file root.
  //
  dumpitracer::dumpitracer(const std::string &fileroot) :
    fileroot_(fileroot), active_(false), profile_(NULL),
    footer_(NULL), outputs_(NULL),
    next_comm_(DUMPI_FIRST_USER_COMM), next_type_(DUMPI_FIRST_USER_DATATYPE),
    next_op_(DUMPI_FIRST_USER_OP), next_request_(DUMPI_FIRST_USER_REQUEST)
  {
    // The profile is the main workhorse of the collection.
    // We set the target buffer much smaller than for regular dumpi runs,
    // since we need to worry about space constrictions.
    profile_ = dumpi_alloc_output_profile(0, 0, 4096);
    // The footer tracks cal counts, ignored calls, etc.
    // It's safe to just clear it completely
    footer_ = new dumpi_footer;
    memset(footer_, 0, sizeof(dumpi_footer));
    // libdumpi uses the outputs variables (timestamp, statuses, perfinfo
    // to set up a config mask of what gets written (and later read).
    outputs_ = new dumpi_outputs;
    outputs_->timestamps = DUMPI_TIME_WALL; // we don't have cpu time (yet)
    outputs_->statuses = DUMPI_SUCCESS;
    outputs_->perfinfo = 0;
    for(int i = 0; i < DUMPI_END_OF_STREAM; ++i)
      outputs_->function[i] = 1;
  }

  //
  // Private blocked copy constructor (not a safe container to copy).
  //
  dumpitracer::dumpitracer(const dumpitracer&) : mpitracer() {
    throw illformederror("dumpitracer:  Blocked copy constructor called.");
  }
   
  //
  // Private blocked assignment operator.
  //
  void dumpitracer::operator=(const dumpitracer&) {
    throw illformederror("dumpitracer:  Blocked assignment operator called.");
  }

  //
  // Set up.
  //
  void dumpitracer::setup(const timestamp &start, const mpicomm &world,
                          const mpicomm &self)
  {
    if(active_)
      throw illformederror("dumpitracer::setup called but profile is "
                           "already active.");
    //dumpi_debug = DUMPI_DEBUG_ALL;
    int rank = world.rank().id;
    std::stringstream ss;
    std::string myroot = fileroot_;
    if(myroot == "") myroot = "trace";
    ss << myroot << "-" << std::setw(4) << std::setfill('0') << rank << ".bin";
    profile_->file = dumpi_open_output_file(ss.str().c_str());
    if(! profile_->file)
      throw ioerror("dumpitracer:  Failed to open tracefile.");
    // Set up maps.
    // Communicators.
    commmap_[world] = DUMPI_COMM_WORLD;
    commmap_[mpicomm::comm_null] = DUMPI_COMM_NULL;
    commmap_[self] = DUMPI_COMM_SELF;
    // Types.
    typemap_[mpitype::mpi_null] = DUMPI_DATATYPE_NULL;
    typemap_[mpitype::mpi_char] = DUMPI_CHAR;
    typemap_[mpitype::mpi_unsigned_char] = DUMPI_UNSIGNED_CHAR;
    typemap_[mpitype::mpi_signed_char] = DUMPI_SIGNED_CHAR;
    typemap_[mpitype::mpi_wchar] = DUMPI_WCHAR;
    typemap_[mpitype::mpi_unsigned_long_long] = DUMPI_UNSIGNED_LONG_LONG;
    typemap_[mpitype::mpi_byte] = DUMPI_BYTE;
    typemap_[mpitype::mpi_short] = DUMPI_SHORT;
    typemap_[mpitype::mpi_unsigned_short] = DUMPI_UNSIGNED_SHORT;
    typemap_[mpitype::mpi_int] = DUMPI_INT;
    typemap_[mpitype::mpi_unsigned] = DUMPI_UNSIGNED;
    typemap_[mpitype::mpi_long] = DUMPI_LONG;
    typemap_[mpitype::mpi_unsigned_long] = DUMPI_UNSIGNED_LONG;
    typemap_[mpitype::mpi_float] = DUMPI_FLOAT;
    typemap_[mpitype::mpi_double] = DUMPI_DOUBLE;
    typemap_[mpitype::mpi_long_double] = DUMPI_LONG_DOUBLE;
    typemap_[mpitype::mpi_long_long_int] = DUMPI_LONG_LONG_INT;
    typemap_[mpitype::mpi_long_long] = DUMPI_LONG_LONG;
    typemap_[mpitype::mpi_packed] = DUMPI_PACKED;
    typemap_[mpitype::mpi_float_int] = DUMPI_FLOAT_INT;
    typemap_[mpitype::mpi_double_int] = DUMPI_DOUBLE_INT;
    typemap_[mpitype::mpi_long_int] = DUMPI_LONG_INT;
    typemap_[mpitype::mpi_short_int] = DUMPI_SHORT_INT;
    typemap_[mpitype::mpi_2int] = DUMPI_2INT;
    typemap_[mpitype::mpi_long_double_int] = DUMPI_LONG_DOUBLE_INT;
    // Operations
    opmap_[mpiop::max] = DUMPI_MAX;
    opmap_[mpiop::min] = DUMPI_MIN;
    opmap_[mpiop::sum] = DUMPI_SUM;
    opmap_[mpiop::prod] = DUMPI_PROD;
    opmap_[mpiop::land] = DUMPI_LAND;
    opmap_[mpiop::band] = DUMPI_BAND;
    opmap_[mpiop::lor] = DUMPI_LOR;
    opmap_[mpiop::bor] = DUMPI_BOR;
    opmap_[mpiop::lxor] = DUMPI_LXOR;
    opmap_[mpiop::bxor] = DUMPI_BXOR;
    opmap_[mpiop::minloc] = DUMPI_MINLOC;
    opmap_[mpiop::maxloc] = DUMPI_MAXLOC;
    opmap_[mpiop::replace] = DUMPI_REPLACE;
    opmap_[mpiop::op_null] = DUMPI_OP_NULL;
    // Rank 0 creates the metafile.
    if(rank == 0) {
      std::string metaname = myroot + ".meta";
      std::ofstream meta(metaname.c_str());
      if(! meta)
        throw ioerror("dumpitracer:  Failed to create metafile.");
      meta << "hostname=sstmacro\n";
      meta << "numprocs=" << world.size().id << "\n";
      meta << "username=none\n";
      meta << "startime=" << start.sec() << "\n";
      meta << "fileprefix=" << myroot << "\n";
      meta << "version=" << (int)dumpi_version << "\n";
      meta << "subversion=" << (int)dumpi_subversion << "\n";
      meta << "subsubversion=" << (int)dumpi_subsubversion << "\n";        
    }
    // Done setting up
    active_ = true;
  }
 
  //
  // Build a new dumpitracer with the given fileroot.
  //
  shared_ptr<dumpitracer> dumpitracer::construct(const std::string &fileroot) {
    return shared_ptr<dumpitracer>(new dumpitracer(fileroot));
  }

  //
  // Clean up.
  //
  dumpitracer::~dumpitracer() {
    finish();
  }

  /// Finish tracing.  Marks the tracer as inactive.
  void dumpitracer::finish() { 
    if(active_) {
      // Now we just need to handle the rest of the dumpi records.
      // We're not dealing with keyvals, perfctrs, function labels, etc.
      dumpi_header *header = dumpi_build_header();
      header->version[0] = dumpi_version;
      header->version[1] = dumpi_subversion;
      header->version[2] = dumpi_subsubversion;
      dumpi_write_header(profile_, header);
      dumpi_free_header(header);
      dumpi_write_footer(profile_, footer_);
      /* This closes the oprofile and frees up the memory buffer */
      dumpi_write_index(profile_);
      dumpi_free_output_profile(profile_);
      delete footer_;
      delete outputs_;
      commmap_.clear();
      typemap_.clear();
      opmap_.clear();
      active_ = false;
    }
  }

  /// Clone this tracer.
  shared_ptr<mpitracer> dumpitracer::clone() const {
    return construct(fileroot_);
  }

  inline dumpi_clock to_dumpi_clock(const timestamp &ts) {
    timestamp::tick_t ticks = ts.ticks();
    timestamp::tick_t space = timestamp::tick_interval();
    timestamp::tick_t flat = (ticks / space) / timestamp::tick_t(1000);
    timestamp::tick_t sec = flat / timestamp::tick_t(1e9);
    timestamp::tick_t frac = flat % timestamp::tick_t(1e9);
    dumpi_clock rv = dumpi_clock_init_time(timestamp::exact_ticks(sec).ticks_int64(),
                       timestamp::exact_ticks(frac).ticks_int64());
    return rv;
  }

  inline dumpi_time to_dumpi_time(const timestamp &start, const timestamp &stop)
  {
    dumpi_time rv = {to_dumpi_clock(start), to_dumpi_clock(stop)};
    return rv;
  }

  /// Generic handler.
  template <typename Argtype>
  void dumpitracer::generic_handler
  (const timestamp &start, const timestamp &finish,
   int label, const Argtype &prm,
   int (*func)(const Argtype *, uint16_t, const dumpi_time*, const dumpi_time*,
               const dumpi_perfinfo*, const dumpi_outputs*, dumpi_profile*))
  {
    if(outputs_->function[label]) {
      ++footer_->call_count[label];
      dumpi_time wall = to_dumpi_time(start, finish);
      func(&prm, thread, NULL, &wall, NULL, outputs_, profile_);
    }
    else {
      ++footer_->ignored_count[label];
    }    
  }

  // Make sure we have properly registered a communicator.
  void dumpitracer::check_comm(const timestamp &beg,
                                   const timestamp &end,
                                   const mpicomm &comm) {
    if(commmap_.find(comm) == commmap_.end()) {
      commmap_[comm] = next_comm_++;
      // Pretend to create this comm from the world comm.
      // First, we pretend to get the world group:
      dumpi_comm_group fakegroup = { DUMPI_COMM_WORLD, DUMPI_COMM_WORLD };
      generic_handler(beg, beg, DUMPI_Comm_group, fakegroup,
                      dumpi_write_comm_group);
      // Then, we pretend to create a new group for this communicator.
      const std::vector<nodeid> &nodes = comm.nodes();
      int64_t size = nodes.size();
      int *ranks = new int[size];
      for(int64_t i = 0; i < size; ++i) {
        ranks[i] = check_cast<int>(nodes.at(i).id);
      }
      dumpi_group_incl fakeincl = { DUMPI_COMM_WORLD,
                                    check_cast<int>(size), ranks,
                                    commmap_[comm] };
      generic_handler(beg, beg, DUMPI_Group_incl, fakeincl,
                      dumpi_write_group_incl);
      delete ranks;
      // and, finally, we pretend to have just created this new communicator.
      dumpi_comm_create fakecomm = { DUMPI_COMM_WORLD, commmap_[comm],
                                     commmap_[comm] };
      generic_handler(beg, end, DUMPI_Comm_create, fakecomm,
                      dumpi_write_comm_create);
    }
  }

  // Make sure we have a properly registered type.
  void dumpitracer::check_type(const timestamp &beg, const timestamp &end,
                               const mpitype &type)
  {
    if(typemap_.find(type) == typemap_.end()) {
      typemap_[type] = next_type_++;
      dumpi_type_contiguous prm = { type.size, DUMPI_BYTE,
                                    typemap_[type] };
      generic_handler(beg, beg, DUMPI_Type_contiguous, prm,
                      dumpi_write_type_contiguous);
      dumpi_type_commit com = { typemap_[type] };
      generic_handler(beg, end, DUMPI_Type_commit, com,
                      dumpi_write_type_commit);
    }
  }

  /// Get the rank of a node.
  void dumpitracer::comm_rank(const timestamp &start, const timestamp &finish,
                              const mpicomm &comm, int rank)
  {
    int label = DUMPI_Comm_rank;
    check_comm(start, start, comm);
    dumpi_comm_rank prm = { commmap_[comm], rank };
    generic_handler(start, finish, label, prm, dumpi_write_comm_rank);
  }

  /// The size of a communicator.
  void dumpitracer::comm_size(const timestamp &start, const timestamp &finish,
                         const mpicomm &comm, int size)
  {
    int label = DUMPI_Comm_size;
check_comm(start, start, comm);
    dumpi_comm_size prm = { commmap_[comm], size };
    generic_handler(start, finish, label, prm, dumpi_write_comm_size);
  }

  inline dumpi_status build_status(const dumpitracer::const_mpistatus_t &stat) {
    dumpi_status retval = {0, -1, 0, 0, 0};
    if(stat) {
      retval.bytes = check_cast<int32_t>(stat->bytes_received());
      retval.source = stat->source().id;
      retval.tag = stat->tag().id;
    }
    return retval;
  }

  /// Get the number of elements received assuming a specific type.
  void dumpitracer::get_count(const timestamp &start, const timestamp &finish,
                              const_mpistatus_t status,
                              const mpitype &type, int count)
  {
    int label = DUMPI_Get_count;
    dumpi_status stat = build_status(status);
    check_type(start, start, type);
    dumpi_get_count prm = { (stat.source >= 0 ? &stat : NULL),
                            typemap_[type], count };
    generic_handler(start, finish, label, prm, dumpi_write_get_count);
  }

  /// Split a communicator.
  void dumpitracer::comm_split(const timestamp &start, const timestamp &finish,
                               const mpicomm &incomm, int color, int key,
                               const mpicomm &outcomm)
  {
    int label = DUMPI_Comm_split;
    check_comm(start, start, incomm);
    commmap_[outcomm] = next_comm_++;
    dumpi_comm_split prm = { commmap_[incomm], color, key, commmap_[outcomm] };
    generic_handler(start, finish, label, prm, dumpi_write_comm_split);
  }

    /// Duplicate a type.
    void dumpitracer::type_dup(const timestamp &start, const timestamp &finish,
                               const mpitype &intype, const mpitype &outtype)
    {
      int label = DUMPI_Type_dup;
      check_type(start, start, intype);
      typemap_[outtype] = next_type_++;
      dumpi_type_dup prm = {typemap_[intype], typemap_[outtype]};
      generic_handler(start, finish, label, prm, dumpi_write_type_dup);
    }

    /// Commit a type.
    void dumpitracer::type_commit(const timestamp &start,
                                  const timestamp &finish,
                                  const mpitype& type)
    {
      int label = DUMPI_Type_commit;
      if(typemap_.find(type) == typemap_.end())
        typemap_[type] = next_type_++;
      dumpi_type_commit prm = { typemap_[type] };
      generic_handler(start, finish, label, prm, dumpi_write_type_commit);
    }

    /// Create a new type.
  void dumpitracer::type_contiguous(const timestamp &start,
                                    const timestamp &finish,
                                    int count, const mpitype &oldtype,
                                    const mpitype &newtype)
  {
    int label = DUMPI_Type_contiguous;
    check_type(start, start, oldtype);
    typemap_[newtype] = next_type_++;
    dumpi_type_contiguous prm = { count, typemap_[oldtype], typemap_[newtype] };
    generic_handler(start, finish, label, prm, dumpi_write_type_contiguous);
  }

  /// Free a type.
  void dumpitracer::type_free(const timestamp &start, const timestamp &finish,
                              const mpitype &type)
  {
    int label = DUMPI_Type_free;
    check_type(start, start, type);
    dumpi_type_free prm = { typemap_[type] };
    generic_handler(start, finish, label, prm, dumpi_write_type_free);
  }

  /// Get type size.
  void dumpitracer::type_size(const timestamp &start, const timestamp &finish,
                              const mpitype &type, int size)
  {
    int label = DUMPI_Type_size;
    check_type(start, start, type);
    dumpi_type_size prm = { typemap_[type], size };
    generic_handler(start, finish, label, prm, dumpi_write_type_size);
  }

  /// Test whether MPI has been initialized.
  void dumpitracer::initialized(const timestamp &start, const timestamp &finish, 
                                bool ini)
  {
    int label = DUMPI_Initialized;
    dumpi_initialized prm = { ini };
    generic_handler(start, finish, label, prm, dumpi_write_initialized);
  }
  
  /// Test whether MPI has been finalized.
  void dumpitracer::finalized(const timestamp &start, const timestamp &finish,
                              bool fini)
  {
    int label = DUMPI_Finalized;
    dumpi_finalized prm = { fini };
    generic_handler(start, finish, label, prm, dumpi_write_finalized);
  }
  
  /// Initialize MPI.  This call will also create the actual trace file.
  void dumpitracer::init(const timestamp &start, const timestamp &finish,
                         const mpicomm &world, const mpicomm &self)
  {
    int label = DUMPI_Init;
    this->setup(start, world, self);
    dumpi_init prm = {0, NULL};
    generic_handler(start, finish, label, prm, dumpi_write_init);
  }
  
  /// Finalize MPI.
  void dumpitracer::finalize(const timestamp &start, const timestamp &finish)
  {
    int label = DUMPI_Finalize;
    dumpi_finalize prm = {0};
    generic_handler(start, finish, label, prm, dumpi_write_finalize);
  }

  /// Duplicate a communicator.
  void dumpitracer::comm_dup(const timestamp &start, const timestamp &finish,
                             const mpicomm &input, const mpicomm &output)
  {
    int label = DUMPI_Comm_dup;
    check_comm(start, start, input);
    commmap_[output] = next_comm_++;
    dumpi_comm_dup prm = { commmap_[input], commmap_[output] };
    generic_handler(start, finish, label, prm, dumpi_write_comm_dup);
  }
  
  /// Create a communicator containing a subset of an existing comm.
  void dumpitracer::comm_create(const timestamp &start, const timestamp &finish,
                                const mpicomm &input,
                                const std::vector<mpiid> &members,
                                const mpicomm &output)
  {
    int grouplabel = DUMPI_Comm_group;
    int incllabel = DUMPI_Group_incl;
    int commlabel = DUMPI_Comm_create;
    check_comm(start, start, input);
    commmap_[output] = next_comm_++;
    dumpi_comm_group groupprm = { commmap_[input], commmap_[input] };
    generic_handler(start, start, grouplabel, groupprm, dumpi_write_comm_group);
    int64_t count = members.size();
    int *ranks = new int[count];
    for(int i = 0; i < count; ++i) ranks[i] = members.at(i).id;
    dumpi_group_incl inclprm = { commmap_[input], check_cast<int>(count),
                                 ranks, commmap_[output] };
    generic_handler(start, start, incllabel, inclprm, dumpi_write_group_incl);
    dumpi_comm_create commprm = { commmap_[input], commmap_[input],
                                  commmap_[output] };
    generic_handler(start, finish, commlabel, commprm, dumpi_write_comm_create);
    delete ranks;
  }

  //
  // Destroy a communicator.
  //
  void dumpitracer::comm_free(const timestamp &start, const timestamp &finish,
                              const mpicomm &input)
  {
    int label = DUMPI_Comm_free;
    check_comm(start, start, input);
    dumpi_comm_free prm = { commmap_[input] };
    generic_handler(start, finish, label, prm, dumpi_write_comm_free);
  }

  /// Blocking send.
  void dumpitracer::send(const timestamp &start, const timestamp &finish,
                         int count, const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm)
  {
    int label = DUMPI_Send;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_send prm = { count, typemap_[type], target.id,
                       tag.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_send);
  }

  /// Blocking send using a user-defined buffer.
  void dumpitracer::bsend(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type, const mpiid &target,
                          const mpitag &tag, const mpicomm &comm)
  {
    int label = DUMPI_Bsend;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_bsend prm = { count, typemap_[type], target.id,
                        tag.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_bsend);
  }

  /// Blocking syncronous send.
  void dumpitracer::ssend(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type, const mpiid &target,
                          const mpitag &tag, const mpicomm &comm)
  {
    int label = DUMPI_Ssend;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_ssend prm = { count, typemap_[type], target.id,
                        tag.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_ssend);
  }

  /// Ready send.
  void dumpitracer::rsend(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type, const mpiid &target,
                          const mpitag &tag, const mpicomm &comm)
  {
    int label = DUMPI_Rsend;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_rsend prm = { count, typemap_[type], target.id,
                        tag.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_rsend);
  }

  /// Nonblocking send.
  void dumpitracer::isend(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type, const mpiid &target,
                          const mpitag &tag, const mpicomm &comm,
                          const mpirequest_t &req)
  {
    int label = DUMPI_Isend;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_isend prm = { count, typemap_[type], target.id,
                        tag.id, commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_isend);
  }

  /// Nonblocking user-buffered send.
  void dumpitracer::ibsend(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiid &target,
                           const mpitag &tag, const mpicomm &comm,
                           const mpirequest_t &req)
  {
    int label = DUMPI_Ibsend;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_ibsend prm = { count, typemap_[type], target.id,
                         tag.id, commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_ibsend);
  }

    /// Nonblocking synchronous send.
    void dumpitracer::issend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req)
  {
    int label = DUMPI_Issend;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_issend prm = { count, typemap_[type], target.id,
                         tag.id, commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_issend);
  }

    /// Nonblocking ready send.
    void dumpitracer::irsend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req)
  {
    int label = DUMPI_Irsend;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_irsend prm = { count, typemap_[type], target.id,
                         tag.id, commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_irsend);
  }

  /// Blocking receive.
  void dumpitracer::recv(const timestamp &start, const timestamp &finish,
                         int count, const mpitype &type, const mpiid &source,
                         const mpitag &tag, const mpicomm &comm,
                         const const_mpistatus_t &status)
  {
    int label = DUMPI_Recv;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_status stat = build_status(status);
    dumpi_recv prm = { count, typemap_[type], source.id, tag.id,
                       commmap_[comm], (stat.source >= 0 ? &stat : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_recv);
  }

  /// Asynchronous receive.
  void dumpitracer::irecv(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type, const mpiid &source,
                          const mpitag &tag, const mpicomm &comm,
                          const mpirequest_t &req)
  {
    int label = DUMPI_Irecv;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_irecv prm = { count, typemap_[type], source.id,
                        tag.id, commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_irecv);
  }

  /// Send/receieve exchange.
  void dumpitracer::sendrecv(const timestamp &start, const timestamp &finish,
                             int sendcount, const mpitype &sendtype,
                             const mpiid &target, const mpitag &sendtag,
                             int recvcount, const mpitype &recvtype,
                             const mpiid &source, const mpitag &recvtag,
                             const mpicomm &comm,
                             const const_mpistatus_t &status)
  {
    int label = DUMPI_Sendrecv;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    dumpi_status stat = build_status(status);
    dumpi_sendrecv prm = { sendcount, typemap_[sendtype], target.id, sendtag.id,
                           recvcount, typemap_[recvtype], source.id, recvtag.id,
                           commmap_[comm], (stat.source >= 0 ? &stat : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_sendrecv);
  }
  
  /// Send and receive using a single buffer.
  void dumpitracer::sendrecv_replace(const timestamp &start,
                                     const timestamp &finish,
                                     int count, const mpitype &type,
                                     const mpiid &dest, const mpitag &sendtag,
                                     const mpiid &source, const mpitag &recvtag, 
                                     const mpicomm &comm,
                                     const const_mpistatus_t &status)
  {
    int label = DUMPI_Sendrecv_replace;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_status stat = build_status(status);
    dumpi_sendrecv_replace prm =
      { count, typemap_[type], dest.id, sendtag.id, source.id, recvtag.id,
        commmap_[comm], (stat.source >= 0 ? &stat : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_sendrecv_replace);
  }
  
  /// Initiate a send.
  void dumpitracer::send_init(const timestamp &start, const timestamp &finish,
                              int count, const mpitype &type,
                              const mpiid &target,
                              const mpitag &tag, const mpicomm &comm,
                              const mpirequest_t &req)
  {
    int label = DUMPI_Send_init;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_send_init prm = { count, typemap_[type], target.id, tag.id,
                            commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_send_init);
  }

  /// Initiate a buffered send.
  void dumpitracer::bsend_init(const timestamp &start, const timestamp &finish,
                               int count, const mpitype &type,
                               const mpiid &target,
                               const mpitag &tag, const mpicomm &comm,
                               const mpirequest_t &req)
  {
    int label = DUMPI_Bsend_init;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_bsend_init prm = { count, typemap_[type], target.id, tag.id,
                            commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_bsend_init);
  }

  /// Initiate a synchronous send.
  void dumpitracer::ssend_init(const timestamp &start, const timestamp &finish,
                               int count, const mpitype &type,
                               const mpiid &target,
                               const mpitag &tag, const mpicomm &comm,
                               const mpirequest_t &req)
  {
    int label = DUMPI_Ssend_init;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_ssend_init prm = { count, typemap_[type], target.id, tag.id,
                            commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_ssend_init);
  }

    /// Initiate a ready send.
  void dumpitracer::rsend_init(const timestamp &start, const timestamp &finish,
                               int count, const mpitype &type,
                               const mpiid &target,
                               const mpitag &tag, const mpicomm &comm,
                               const mpirequest_t &req)
  {
    int label = DUMPI_Rsend_init;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_rsend_init prm = { count, typemap_[type], target.id, tag.id,
                            commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_rsend_init);
  }
    
  /// Initiate a receive.
  void dumpitracer::recv_init(const timestamp &start, const timestamp &finish,
                              int count, const mpitype &type,
                              const mpiid &target,
                              const mpitag &tag, const mpicomm &comm,
                              const mpirequest_t &req)
  {
    int label = DUMPI_Recv_init;
    check_type(start, start, type);
check_comm(start, start, comm);
    requestmap_[req] = next_request_++;
    dumpi_recv_init prm = { count, typemap_[type], target.id, tag.id,
                            commmap_[comm], requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_recv_init);
  }

  /// Start a previously initiated request.
  void dumpitracer::start(const timestamp &start, const timestamp &finish,
                          const mpirequest_t &req)
  {
    int label = DUMPI_Start;
    if(requestmap_.find(req) == requestmap_.end())
      requestmap_[req] = next_request_++;
    dumpi_start prm = { requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_start);
  }

  /// Start a collection of previously initiated requests.
  void dumpitracer::startall(const timestamp &start, const timestamp &finish,
                             const std::vector< mpirequest_t > &requests)
  {
    int label = DUMPI_Startall;
    int64_t count = requests.size();
    dumpi_request *req = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(requests.at(i)) == requestmap_.end())
        requestmap_[requests.at(i)] = next_request_++;
      req[i] = requestmap_[requests.at(i)];
    }
    dumpi_startall prm = { check_cast<int>(count), req };
    generic_handler(start, finish, label, prm, dumpi_write_startall);
    delete req;
  }

  /// Cancel a request that had been previously initiated.
  void dumpitracer::cancel(const timestamp &start, const timestamp &finish,
                           const mpirequest_t &req)
  {
    int label = DUMPI_Cancel;
    if(requestmap_.find(req) == requestmap_.end())
      requestmap_[req] = next_request_++;
    dumpi_cancel prm = { requestmap_[req] };
    generic_handler(start, finish, label, prm, dumpi_write_cancel);
  }

  /// Test whether a request has been cancelled.
  void dumpitracer::test_cancelled(const timestamp &start,
                                   const timestamp &finish,
                                   const const_mpistatus_t &status,
                                   bool cancelled)
  {
    int label = DUMPI_Test_cancelled;
    dumpi_status stat = build_status(status);
    dumpi_test_cancelled prm = { &stat, cancelled };
    generic_handler(start, finish, label, prm, dumpi_write_test_cancelled);
  }

  /// Wait for a request to finish.
  void dumpitracer::wait(const timestamp &start, const timestamp &finish,
                         const mpirequest_t &req,
                         const const_mpistatus_t &status)
  {
    int label = DUMPI_Wait;
    if(requestmap_.find(req) == requestmap_.end())
      requestmap_[req] = next_request_++;
    dumpi_status stat = build_status(status);
    dumpi_wait prm = { requestmap_[req], &stat };
    generic_handler(start, finish, label, prm, dumpi_write_wait);
  }
  
  /// Wait for all the given requests to finish.
  void dumpitracer::waitall(const timestamp &start, const timestamp &finish,
                            const std::vector< mpirequest_t > &requests,
                            const std::vector< const_mpistatus_t > &statuses)
  {
    int label = DUMPI_Waitall;
    int64_t count = requests.size();
    dumpi_request *req = new dumpi_request[count];
    dumpi_status *stat = new dumpi_status[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(requests.at(i)) == requestmap_.end())
        requestmap_[requests.at(i)] = next_request_++;
      req[i] = requestmap_[requests.at(i)];
      stat[i] = build_status(statuses.at(i));
    }
    dumpi_waitall prm = { check_cast<int>(count), req, stat };
    generic_handler(start, finish, label, prm, dumpi_write_waitall);
    delete req;
    delete stat;
  }

  /// Wait for at least one of the given requests to complete.
  void dumpitracer::waitany(const timestamp &start, const timestamp &finish,
                            const std::vector< mpirequest_t > &requests,
                            int index, const const_mpistatus_t &status)
  {
    int label = DUMPI_Waitany;
    int64_t count = requests.size();
    dumpi_request *req = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(requests.at(i)) == requestmap_.end())
        requestmap_[requests.at(i)] = next_request_++;
      req[i] = requestmap_[requests.at(i)];
    }
    dumpi_status stat = build_status(status);
    dumpi_waitany prm = { check_cast<int>(count), req, index, (index >= 0 ? &stat : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_waitany);
    delete req;
  }

  /// Wait for some of the given requests to complete.
  void dumpitracer::waitsome(const timestamp &start, const timestamp &finish,
                             const std::vector<mpirequest_t > &requests,
                             const std::vector<int> &indices,
                             const std::vector< const_mpistatus_t > &statuses)
  {
    int label = DUMPI_Waitsome;
    int64_t count = requests.size();
    dumpi_request *req = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(requests.at(i)) == requestmap_.end())
        requestmap_[requests.at(i)] = next_request_++;
      req[i] = requestmap_[requests.at(i)];
    }
    int64_t outcount = indices.size();
    int *idx = NULL;
    dumpi_status *stat = NULL;
    if(outcount > 0) {
      stat = new dumpi_status[outcount];
      idx = new int[outcount];
      for(int i = 0; i < outcount; ++i) {
        idx[i] = indices.at(i);
        stat[i] = build_status(statuses.at(i));
      }
    }
    dumpi_waitsome prm = { check_cast<int>(count), req, check_cast<int>(outcount), idx, stat };
    generic_handler(start, finish, label, prm, dumpi_write_waitsome);
    delete req;
    if(stat) delete stat;
    if(idx) delete idx;
  }

  /// Test whether a request has finished.
  void dumpitracer::test(const timestamp &start, const timestamp &finish,
                         const mpirequest_t &req, bool flag, 
                         const const_mpistatus_t &stat)
  {
    int label = DUMPI_Test;
    if(requestmap_.find(req) == requestmap_.end())
      requestmap_[req] = next_request_++;
    dumpi_status status = build_status(stat);
    dumpi_test prm = { requestmap_[req], flag, 
                       ((flag && (status.source >= 0)) ? &status : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_test);
  }

  /// Test whether all the given requests have finished.
  void dumpitracer::testall(const timestamp &start, const timestamp &finish,
                            const std::vector< mpirequest_t > &req, bool flag,
                            const std::vector< const_mpistatus_t > &statuses)
  {
    int label = DUMPI_Testall;
    int64_t count = req.size();
    dumpi_request *requests = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(req.at(i)) == requestmap_.end())
        requestmap_[req.at(i)] = next_request_++;
      requests[i] = requestmap_[req.at(i)];
    }
    dumpi_status *stat = NULL;
    if(count > 0) {
      stat = new dumpi_status[count];
      for(int i = 0; i < count; ++i)
        stat[i] = build_status(statuses.at(i));
    }
    dumpi_testall prm = { check_cast<int>(count), requests, flag, stat };
    generic_handler(start, finish, label, prm, dumpi_write_testall);
    delete requests;
    if(stat) delete stat;
  }
    
  /// Test whether at least one of the given requests has completed.
  void dumpitracer::testany(const timestamp &start, const timestamp &finish,
                            const std::vector< mpirequest_t > &req, int index,
                            bool flag, const const_mpistatus_t &stat)
  {
    int label = DUMPI_Testany;
    int64_t count = req.size();
    dumpi_request *requests = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(req.at(i)) == requestmap_.end())
        requestmap_[req.at(i)] = next_request_++;
      requests[i] = requestmap_[req.at(i)];
    }
    dumpi_status status = build_status(stat);
    dumpi_testany prm = { check_cast<int>(count), requests, index, flag,
                          ((flag && (status.source >= 0)) ? &status : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_testany);
  }

  /// Test whether some of the given requests have completed.
  void dumpitracer::testsome(const timestamp &start, const timestamp &finish,
                             const std::vector<mpirequest_t > &req,
                             const std::vector<int> &indices,
                             const std::vector< const_mpistatus_t > &statuses)
  {
    int label = DUMPI_Testsome;
    int64_t count = req.size();
    dumpi_request *requests = new dumpi_request[count];
    for(int i = 0; i < count; ++i) {
      if(requestmap_.find(req.at(i)) == requestmap_.end())
        requestmap_[req.at(i)] = next_request_++;
      requests[i] = requestmap_[req.at(i)];
    }
    dumpi_status *stat = NULL;
    int *idx = NULL;
    int64_t outcount = indices.size();
    if(outcount > 0) {
      stat = new dumpi_status[outcount];
      idx = new int[outcount];
      for(int i = 0; i < outcount; ++i) {
        stat[i] = build_status(statuses.at(i));
        idx[i] = indices.at(i);
      }
    }
    dumpi_testsome prm = { check_cast<int>(count), requests, check_cast<int>(outcount), idx, stat };
    generic_handler(start, finish, label, prm, dumpi_write_testsome);
    delete requests;
    if(stat) delete stat;
    if(idx) delete idx;
  }

  /// Blocking probe for a message that matches the given signature.
  void dumpitracer::probe(const timestamp &start, const timestamp &finish,
                          const mpiid &source, const mpitag &tag,
                          const mpicomm &comm, const const_mpistatus_t &stat)
  {
    int label = DUMPI_Probe;
check_comm(start, start, comm);
    dumpi_status status = build_status(stat);
    dumpi_probe prm = { source.id, tag.id, commmap_[comm],
                        (status.source >= 0 ? &status : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_probe);
  }
  
  /// Non-blocking probe for messages.
  void dumpitracer::iprobe(const timestamp &start, const timestamp &finish,
                           const mpiid &source, const mpitag &tag,
                           const mpicomm &comm, bool &flag,
                           const const_mpistatus_t &stat)
  {
    int label = DUMPI_Iprobe;
check_comm(start, start, comm);
    dumpi_status status = build_status(stat);
    dumpi_iprobe prm = { source.id, tag.id, commmap_[comm], flag,
                         ((flag && (status.source >= 0)) ? &status : NULL) };
    generic_handler(start, finish, label, prm, dumpi_write_iprobe);
  }

  /// Attach a user-defined buffer for *bsend operations.
  void dumpitracer::buffer_attach(const timestamp &start,
                                  const timestamp &finish,
                                  int bytes)
  {
    int label = DUMPI_Buffer_attach;
    dumpi_buffer_attach prm = {bytes};
    generic_handler(start, finish, label, prm, dumpi_write_buffer_attach);
  }

  /// Detach the user-defined buffer.
  void dumpitracer::buffer_detach(const timestamp &start,
                                  const timestamp &finish,
                                  int bytes)
  {
    int label = DUMPI_Buffer_detach;
    dumpi_buffer_detach prm = {bytes};
    generic_handler(start, finish, label, prm, dumpi_write_buffer_detach);
  }

  /// Set a barrier.
  void dumpitracer::barrier(const timestamp &start, const timestamp &finish,
                            const mpicomm &comm)
  {
    int label = DUMPI_Barrier;
check_comm(start, start, comm);
    dumpi_barrier prm = { commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_barrier);
  }

  /// Broadcast data with no payload.
  void dumpitracer::bcast(const timestamp &start, const timestamp &finish,
                          int count, const mpitype &type,
                          const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Bcast;
    check_type(start, start, type);
check_comm(start, start, comm);
    dumpi_bcast prm = { count, typemap_[type], root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_bcast);
  }

  /// Scatter data uniformly.
  void dumpitracer::scatter(const timestamp &start, const timestamp &finish,
                            int sendcount, const mpitype &sendtype,
                            int recvcount, const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Scatter;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
    check_comm(start, start, comm);
    dumpi_scatter prm = { comm.rank().id, sendcount, typemap_[sendtype], 
                          recvcount, typemap_[recvtype],
                          root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_scatter);
  }

  /// Scatter data non-uniformly.
  void dumpitracer::scatterv(const timestamp &start, const timestamp &finish,
                             const std::vector<int> &sendcounts,
                             const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Scatterv;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    int64_t sc = sendcounts.size();
    int *sd = new int[sc];
    int *disp = new int[sc];
    int basedisp = 0;
    for(int i = 0; i < sc; ++i) {
      sd[i] = sendcounts.at(i);
      disp[i] = basedisp;
      basedisp += sd[i];
    }
    dumpi_scatterv prm = { comm.rank().id, comm.size().id,
                           sd, disp, typemap_[sendtype],
                           recvcount, typemap_[recvtype],
                           root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_scatterv);
    if(sd) delete sd;
    if(disp) delete disp;
  }
    
  /// Gather uniformly distributed data.
  void dumpitracer::gather(const timestamp &start, const timestamp &finish,
                           int sendcount, const mpitype &sendtype,
                           int recvcount, const mpitype &recvtype,
                           const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Gather;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    dumpi_gather prm = { comm.rank().id, sendcount, typemap_[sendtype], 
                         recvcount, typemap_[recvtype],
                         root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_gather);
  }

  /// Gather non-uniformly distributed data.
  void dumpitracer::gatherv(const timestamp &start, const timestamp &finish,
                            int sendcount, const mpitype &sendtype,
                            const std::vector<int> &recvcnts,
                            const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Gatherv;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    int64_t rc = recvcnts.size();
    int *rd = new int[rc];
    int *disp = new int[rc];
    int basedisp = 0;
    for(int i = 0; i < rc; ++i) {
      rd[i] = recvcnts.at(i);
      disp[i] = basedisp;
      basedisp += rd[i];
    }
    dumpi_gatherv prm = { comm.rank().id, comm.size().id,
                          sendcount, typemap_[sendtype],
                          rd, disp, typemap_[recvtype],
                          root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_gatherv);
    if(rd) delete rd;
    if(disp) delete disp;
  }

  /// Gather uniformly distributed data and make it available to all nodes.
  void dumpitracer::allgather(const timestamp &start, const timestamp &finish,
                              int sendcount, const mpitype &sendtype,
                              int recvcount, const mpitype &recvtype,
                              const mpicomm &comm)
  {
    int label = DUMPI_Allgather;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    dumpi_allgather prm = { sendcount, typemap_[sendtype],
                            recvcount, typemap_[recvtype],
                            commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_allgather);
  }
  
  /// Gather non-uniformly distributed data onto all nodes.
  void dumpitracer::allgatherv(const timestamp &start, const timestamp &finish,
                               int sendcount, const mpitype &sendtype,
                               const std::vector<int> &recvcnts,
                               const mpitype &recvtype, const mpicomm &comm)
  {
    int label = DUMPI_Allgather;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    int64_t rc = recvcnts.size();
    int *rd = new int[rc];
    int *disp = new int[rc];
    int basedisp = 0;
    for(int i = 0; i < rc; ++i) {
      rd[i] = recvcnts.at(i);
      disp[i] = basedisp;
      basedisp += rd[i];
    }
    dumpi_allgatherv prm = { comm.size().id, sendcount, typemap_[sendtype],
                             rd, disp, typemap_[recvtype], commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_allgatherv);
    if(rd) delete rd;
    if(disp) delete disp;
  }

  /// Send uniformly distributed data from all to all processes.
  void dumpitracer::alltoall(const timestamp &start, const timestamp &finish,
                             int sendcount, const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpicomm &comm)
  {
    int label = DUMPI_Alltoall;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    dumpi_alltoall prm = { sendcount, typemap_[sendtype],
                           recvcount, typemap_[recvtype], commmap_[comm] } ;
    generic_handler(start, finish, label, prm, dumpi_write_alltoall);
  }
  
  /// Send non-uniformly distributed data from all to all processors.
  void dumpitracer::alltoallv(const timestamp &start, const timestamp &finish,
                              const std::vector<int> &sendcnts,
                              const mpitype &sendtype,
                              const std::vector<int> &recvcnts,
                              const mpitype &recvtype,
                              const mpicomm &comm)
  {
    int label = DUMPI_Alltoallv;
    check_type(start, start, sendtype);
    check_type(start, start, recvtype);
check_comm(start, start, comm);
    int64_t scount = sendcnts.size();
    int *sdata = new int[scount];
    int *sdisp = new int[scount];
    int sbase = 0;
    for(int i = 0; i < scount; ++i) {
      sdata[i] = sendcnts.at(i);
      sdisp[i] = sbase;
      sbase += sdata[i];
    }
    int64_t rcount = recvcnts.size();
    int *rdata = new int[rcount];
    int *rdisp = new int[rcount];
    int rbase = 0;
    for(int i = 0; i < rcount; ++i) {
      rdata[i] = recvcnts.at(i);
      rdisp[i] = rbase;
      rbase += rdata[i];
    }
    dumpi_alltoallv prm = { comm.size().id,
                            sdata, sdisp, typemap_[sendtype],
                            rdata, rdisp, typemap_[recvtype],
                            commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_alltoallv);
    if(sdata) delete sdata;
    if(sdisp) delete sdisp;
    if(rdata) delete rdata;
    if(rdisp) delete rdisp;
  }

  /// Reduce data from all nodes onto the root node.
  void dumpitracer::reduce(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiop &op,
                           const mpiid &root, const mpicomm &comm)
  {
    int label = DUMPI_Reduce;
    check_type(start, start, type);
    if(opmap_.find(op) == opmap_.end())
      opmap_[op] = next_op_++;
check_comm(start, start, comm);
    dumpi_reduce prm = { count, typemap_[type], opmap_[op],
                         root.id, commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_reduce);
  }

  /// Reduce data from all nodes to all nodes.
  void dumpitracer::allreduce(const timestamp &start, const timestamp &finish,
                              int count, const mpitype &type,
                              const mpiop &op, const mpicomm &comm)
  {
    int label = DUMPI_Allreduce;
    check_type(start, start, type);
    if(opmap_.find(op) == opmap_.end())
      opmap_[op] = next_op_++;
check_comm(start, start, comm);
    dumpi_allreduce prm = { count, typemap_[type], opmap_[op], commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_allreduce);
  }
  
  /// Reduce data and scatter results.
  void dumpitracer::reduce_scatter(const timestamp &start,
                                   const timestamp &finish,
                                   const std::vector<int> &recvcnts,
                                   const mpitype &type, const mpiop &op,
                                   const mpicomm &comm)
  {
    int label = DUMPI_Reduce_scatter;
    check_type(start, start, type);
    if(opmap_.find(op) == opmap_.end())
      opmap_[op] = next_op_++;
check_comm(start, start, comm);
    int64_t rcnt = recvcnts.size();
    int *rdata = new int[rcnt];
    for(int i = 0; i < rcnt; ++i)
      rdata[i] = recvcnts.at(i);
    dumpi_reduce_scatter prm = { comm.size().id, rdata, typemap_[type],
                                 opmap_[op], commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_reduce_scatter);
    if(rdata) delete rdata;
  }

  /// Compute partial reductions on data.
  void dumpitracer::scan(const timestamp &start, const timestamp &finish,
                         int count, const mpitype &type,
                         const mpiop &op, const mpicomm &comm)
  {
    int label = DUMPI_Scan;
    check_type(start, start, type);
    if(opmap_.find(op) == opmap_.end())
      opmap_[op] = next_op_++;
check_comm(start, start, comm);
    dumpi_scan prm = { count, typemap_[type], opmap_[op], commmap_[comm] };
    generic_handler(start, finish, label, prm, dumpi_write_scan);
  }


} // end of namespace sstmac
