/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_DUMPITRACER_H
#define SSTMAC_PROCESS_DUMPITRACER_H

#include <sstmac/process/mpitracer.h>
#include <map>

// Forward declaration of important types.
class dumpi_profile;
class dumpi_footer;
class dumpi_outputs;
class dumpi_time;
class dumpi_perfinfo;

namespace sstmac {

  /**
   * A tracing facility that writes a DUMPI output file for each node
   * in a simulated MPI run.
   */
  class dumpitracer : public mpitracer {
  public:
    typedef mpitracer::mpirequest mpirequest;
    typedef mpitracer::mpirequest_t mpirequest_t;
    typedef mpitracer::const_mpistatus_t const_mpistatus_t;

  private:
    /// The fileroot used by this tracer.
    const std::string fileroot_;

    /// Track whether this tracer has been initialized.
    bool active_;

    /// The dumpi_profile that output information gets written to.
    dumpi_profile *profile_;

    /// The footer information (call counts etc.).
    dumpi_footer *footer_;

    /// The flags controlling what gets collected and how.
    dumpi_outputs *outputs_;

    /// Map sstmacro communicators to dumpi handles.
    /// We "cheat" by using the comm map also as a group map.
    /// The sstmacro simulator  doesn't currently use normal
    /// groups, but we have to fake it if we want to write sensible trace info
    /// for MPI_Comm_create and similar calls.
    typedef std::map<mpicomm, int> commmap_t;
    commmap_t commmap_;
    /// The index given to the next communicator.
    int next_comm_;
    
    /// Map sstmacro datatypes to dumpi handles.
    typedef std::map<mpitype, int> typemap_t;
    typemap_t typemap_;
    /// The index given to the next type.
    int next_type_;

    // sstmacro currently doesn't handle groups according to the
    // MPI standard. This is slated to be fixed in the upcoming rewrite of
    // the simulator mpi bindings.

    /// Map sstmacro operations to dumpi handles.
    typedef std::map<mpiop, int> opmap_t;
    opmap_t opmap_;
    /// The index given to the next op.
    int next_op_;

    /// Map sstmacro requests to dumpi handles.
    typedef std::map<mpirequest_t, int> requestmap_t;
    requestmap_t requestmap_;
    /// The index given to the next request.
    int next_request_;

    /// Construct a dumpitracer and give it a fileroot.
    dumpitracer(const std::string &fileroot);
    
    /// Private blocked copy constructor (not a safe container to copy).
    dumpitracer(const dumpitracer&);
   
    /// Private blocked assignment operator.
    void operator=(const dumpitracer&);

    /// Open files and do whatever other setup is required.
    void setup(const timestamp &start, const mpicomm &world,
               const mpicomm &self);

    /// Private generic handler for the boilerplate part of tracing.
    template <typename Argtype>
    void generic_handler(const timestamp &start, const timestamp &finish,
                         int label, const Argtype &arg,
                         int (*func)(const Argtype *, uint16_t,
                                     const dumpi_time*, const dumpi_time*,
                                     const dumpi_perfinfo*,
                                     const dumpi_outputs*, dumpi_profile*));

    /// Private method to ensure that a communicator is properly
    /// registered in the trace file.
    /// If this communicator is "new" to the tracer, a set of entries
    /// will be written to the trace file to pretend that this communicator
    /// was just created.
    void check_comm(const timestamp &beg, const timestamp &end,
                    const mpicomm &comm);
    
    /// Private method to ensure that a type has been properly registered.
   void check_type(const timestamp &beg, const timestamp &end,
                   const mpitype &type);

  public:
    /// Build a new dumpitracer with the given fileroot.
    /// If the config file is not found, a set of suitable defaults will be used.
    static boost::shared_ptr<dumpitracer>
    construct(const std::string &fileroot);

    /// Clean up.
    virtual ~dumpitracer();

    /// Clone this tracer.
    boost::shared_ptr<mpitracer> clone() const;

    /// Finish tracing.  Marks the tracer as inactive.
    virtual void finish();

    /// Get the rank of a node.
    virtual void comm_rank(const timestamp &start, const timestamp &finish,
                           const mpicomm &comm, int rank);

    /// The size of a communicator.
    virtual void comm_size(const timestamp &start, const timestamp &finish,
                           const mpicomm &comm, int size);

    /// Get the number of elements received assuming a specific type.
    virtual void get_count(const timestamp &start, const timestamp &finish,
                           const_mpistatus_t status,
                           const mpitype &type, int count);

    /// Split a communicator.
    virtual void comm_split(const timestamp &start, const timestamp &finish,
                            const mpicomm &incomm, int color, int key,
                            const mpicomm &outcomm);

    /// Duplicate a type.
    virtual void type_dup(const timestamp &start, const timestamp &finish,
                          const mpitype &intype, const mpitype &outtype);

    /// Commit a type.
    virtual void type_commit(const timestamp &start, const timestamp &finish,
                             const mpitype& type);

    /// Create a new type.
    virtual void type_contiguous(const timestamp &start, const timestamp &finish,
                                 int count, const mpitype &oldtype,
                                 const mpitype &newtype);

    /// Free a type.
    virtual void type_free(const timestamp &start, const timestamp &finish,
                           const mpitype &type);

    /// Get type size.
    virtual void type_size(const timestamp &start, const timestamp &finish,
                           const mpitype &type, int size);

    /// Test whether MPI has been initialized.
    virtual void
    initialized(const timestamp &start, const timestamp &finish, bool ini);

    /// Test whether MPI has been finalized.
    virtual void
    finalized(const timestamp &start, const timestamp &finish, bool fini);

    /// Initialize MPI.  This call will also create the actual trace file.
    virtual void init(const timestamp &start, const timestamp &finish,
                      const mpicomm &world, const mpicomm &self);

    /// Finalize MPI.
    virtual void finalize(const timestamp &start, const timestamp &finish);

    /// Duplicate a communicator.
    virtual void comm_dup(const timestamp &start, const timestamp &finish,
                          const mpicomm &input, const mpicomm &output);

    /// Create a communicator containing a subset of an existing comm.
    virtual void
    comm_create(const timestamp &start, const timestamp &finish,
                const mpicomm &input, const std::vector<mpiid> &members,
                const mpicomm &output);

    /// Destroy a communicator.
    virtual void comm_free(const timestamp &start, const timestamp &finish,
                           const mpicomm &input);

    /// Blocking send.
    virtual void send(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type, const mpiid &target,
                      const mpitag &tag, const mpicomm &comm);

    /// Blocking send using a user-defined buffer.
    virtual void bsend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm &comm);

    /// Blocking syncronous send.
    virtual void ssend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm &comm);

    /// Ready send.
    virtual void rsend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm &comm);

    /// Nonblocking send.
    virtual void isend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm &comm,
                       const mpirequest_t &req);

    /// Nonblocking user-buffered send.
    virtual void ibsend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req);

    /// Nonblocking synchronous send.
    virtual void issend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req);

    /// Nonblocking ready send.
    virtual void irsend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req);

    /// Blocking receive.
    virtual void recv(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type, const mpiid &source,
                      const mpitag &tag, const mpicomm &comm,
                      const const_mpistatus_t &stat);

    /// Asynchronous receive.
    virtual void irecv(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &source,
                       const mpitag &tag, const mpicomm &comm,
                       const mpirequest_t &req);

    /// Send/receieve exchange.
    virtual void
    sendrecv(const timestamp &start, const timestamp &finish,
             int sendcount, const mpitype &sendtype,
             const mpiid &target, const mpitag &sendtag,
             int recvcount, const mpitype &recvtype,
             const mpiid &source, const mpitag &recvtag,
             const mpicomm &comm, const const_mpistatus_t &stat);

    /// Send and receive using a single buffer.
    virtual void
    sendrecv_replace(const timestamp &start, const timestamp &finish,
                     int count, const mpitype &type,
                     const mpiid &dest, const mpitag &sendtag,
                     const mpiid &source, const mpitag &recvtag, 
                     const mpicomm &comm, const const_mpistatus_t &stat);

    /// Initiate a send.
    virtual void send_init(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiid &target,
                           const mpitag &tag, const mpicomm &comm,
                           const mpirequest_t &req);

    /// Initiate a buffered send.
    virtual void bsend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm &comm,
                            const mpirequest_t &req);

    /// Initiate a synchronous send.
    virtual void ssend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm &comm,
                            const mpirequest_t &req);

    /// Initiate a ready send.
    virtual void rsend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm &comm,
                            const mpirequest_t &req);
    
    /// Initiate a receive.
    virtual void recv_init(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiid &target,
                           const mpitag &tag, const mpicomm &comm,
                           const mpirequest_t &req);

    /// Start a previously initiated request.
    virtual void start(const timestamp &start, const timestamp &finish,
                       const mpirequest_t &req);

    /// Start a collection of previously initiated requests.
    virtual void startall(const timestamp &start, const timestamp &finish,
                          const std::vector< mpirequest_t > &requests);

    /// Cancel a request that had been previously initiated.
    virtual void cancel(const timestamp &start, const timestamp &finish,
                        const mpirequest_t &req);

    /// Test whether a request has been cancelled.
    virtual void test_cancelled(const timestamp &start, const timestamp &finish,
                                const const_mpistatus_t &req, bool cancelled);

    /// Wait for a request to finish.
    virtual void wait(const timestamp &start, const timestamp &finish,
                      const mpirequest_t &req, const const_mpistatus_t &stat);

    /// Wait for all the given requests to finish.
    virtual void waitall(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req,
                         const std::vector< const_mpistatus_t > &statuses);

    /// Wait for at least one of the given requests to complete.
    virtual void waitany(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req,
                         int index, const const_mpistatus_t &stat);

    /// Wait for some of the given requests to complete.
    virtual void waitsome(const timestamp &start, const timestamp &finish,
                          const std::vector<mpirequest_t > &req,
                          const std::vector<int> &indices,
                          const std::vector< const_mpistatus_t > &statuses);

    /// Test whether a request has finished.
    virtual void test(const timestamp &start, const timestamp &finish,
                      const mpirequest_t &req, bool flag, 
                      const const_mpistatus_t &stat);

    /// Test whether all the given requests have finished.
    virtual void testall(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req, bool flag,
                         const std::vector< const_mpistatus_t > &statuses);
    
    /// Test whether at least one of the given requests has completed.
    virtual void testany(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req, int index,
                         bool flag, const const_mpistatus_t &stat);

    /// Test whether some of the given requests have completed.
    virtual void testsome(const timestamp &start, const timestamp &finish,
                          const std::vector<mpirequest_t > &req,
                          const std::vector<int> &indices,
                          const std::vector< const_mpistatus_t > &statuses);

    /// Blocking probe for a message that matches the given signature.
    virtual void probe(const timestamp &start, const timestamp &finish,
                       const mpiid &source, const mpitag &tag,
                       const mpicomm &comm, const const_mpistatus_t &stat);

    /// Non-blocking probe for messages.
    virtual void iprobe(const timestamp &start, const timestamp &finish,
                        const mpiid &source, const mpitag &tag,
                        const mpicomm &comm, bool &flag,
                        const const_mpistatus_t &stat);

    /// Attach a user-defined buffer for *bsend operations.
    virtual void buffer_attach(const timestamp &start, const timestamp &finish,
                               int bytes);

    /// Detach the user-defined buffer.
    virtual void buffer_detach(const timestamp &start, const timestamp &finish,
                               int bytes);

    /// Set a barrier.
    virtual void barrier(const timestamp &start, const timestamp &finish,
                         const mpicomm &comm);

    /// Broadcast data with no payload.
    virtual void bcast(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type,
                       const mpiid &root, const mpicomm &comm);

    /// Scatter data uniformly.
    virtual void scatter(const timestamp &start, const timestamp &finish,
                         int sendcount, const mpitype &sendtype,
                         int recvcount, const mpitype &recvtype,
                         const mpiid &root, const mpicomm &comm);

    /// Scatter data non-uniformly.
    virtual void scatterv(const timestamp &start, const timestamp &finish,
                          const std::vector<int> &sendcounts,
                          const mpitype &sendtype,
                          int recvcount, const mpitype &recvtype,
                          const mpiid &root, const mpicomm &comm);
    
    /// Gather uniformly distributed data.
    virtual void gather(const timestamp &start, const timestamp &finish,
                        int sendcount, const mpitype &sendtype,
                        int recvcount, const mpitype &recvtype,
                        const mpiid &root, const mpicomm &comm);

    /// Gather non-uniformly distributed data.
    virtual void gatherv(const timestamp &start, const timestamp &finish,
                         int sendcount, const mpitype &sendtype,
                         const std::vector<int> &recvcnts,
                         const mpitype &recvtype,
                         const mpiid &root, const mpicomm &comm);

    /// Gather uniformly distributed data and make it available to all nodes.
    virtual void allgather(const timestamp &start, const timestamp &finish,
                           int sendcount, const mpitype &sendtype,
                           int recvcount, const mpitype &recvtype,
                           const mpicomm &comm);

    /// Gather non-uniformly distributed data onto all nodes.
    virtual void allgatherv(const timestamp &start, const timestamp &finish,
                            int sendcount, const mpitype &sendtype,
                            const std::vector<int> &recvcnts,
                            const mpitype &recvtype, const mpicomm &comm);

    /// Send uniformly distributed data from all to all processes.
    virtual void alltoall(const timestamp &start, const timestamp &finish,
                          int sendcount, const mpitype &sendtype,
                          int recvcount, const mpitype &recvtype,
                          const mpicomm &comm);

    /// Send non-uniformly distributed data from all to all processors.
    virtual void
    alltoallv(const timestamp &start, const timestamp &finish,
              const std::vector<int> &sendcnts, const mpitype &sendtype,
              const std::vector<int> &recvcnts, const mpitype &recvtype,
              const mpicomm &comm);

    /// Reduce data from all nodes onto the root node.
    virtual void reduce(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiop &op,
                        const mpiid &root, const mpicomm &comm);

    /// Reduce data from all nodes to all nodes.
    virtual void allreduce(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type,
                           const mpiop &op, const mpicomm &comm);

    /// Reduce data and scatter results.
    virtual void reduce_scatter(const timestamp &start, const timestamp &finish,
                                const std::vector<int> &recvcnts,
                                const mpitype &type, const mpiop &op,
                                const mpicomm &comm);

    /// Compute partial reductions on data.
    virtual void scan(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type,
                      const mpiop &op, const mpicomm &comm);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_DUMPITRACER_H
