/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiapi-persistent.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/key.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;

  mpiapi::persistent::persistent() :
    processrequest(key::construct())
  {}

  mpiapi::persistent::~persistent() throw() {
  }

  mpiapi::persistentsend::
  persistentsend(shared_ptr<mpiapi> api,
                 sendfunc_t func, int count, const mpitype &type,
                 const mpiid &target, const mpitag &tag, const mpicomm &comm,
                 const shared_ptr<const payload>& content) :
    parent_(api), func_(func), count_(count),
    type_(type), target_(target), tag_(tag), comm_(comm), content_(content)
  {
    if(parent_ == NULL || func_ == NULL)
      throw illformederror("mpiapi::persistentsend:  Invalid parent info.");
    // Start by marking this complete (needed for error checking).
    this->complete(parent_->wtime(), mpistatus::construct());
  }

  shared_ptr<mpiapi::persistentsend> mpiapi::persistentsend::
  construct(shared_ptr<mpiapi> api,
            sendfunc_t func, int count, const mpitype &type,
            const mpiid &target, const mpitag &tag, const mpicomm &comm,
            const shared_ptr<const payload>& content)
  {
    shared_ptr<persistentsend> retval
      (new persistentsend(api, func, count, type, target, tag, comm, content));
    retval->set_self(retval);
    return retval;
  }

  mpiapi::persistentsend::~persistentsend() throw() {
  }

  //
  // Start this request.
  //
  void mpiapi::persistentsend::start() {
    if(! this->is_complete())
      throw illformederror("mpiapi::persistensend::start:  This operation is "
                           "currently active (you need to call wait or test)");
    this->reset();
    ((*parent_).*func_)(count_, type_, target_, tag_, comm_,
                      self_.lock(), content_);
  }


  mpiapi::persistentrecv::
  persistentrecv(shared_ptr<mpiapi> api, int count, const mpitype &type,
                 const mpiid &dest, const mpitag &tag, const mpicomm &comm) :
    parent_(api), count_(count), type_(type),
    dest_(dest), tag_(tag), comm_(comm)
  {
    if(parent_ == NULL)
      throw illformederror("mpiapi::persistentrecv:  Invalid parent info.");
    // Start by marking this complete (needed for error checking).
    this->complete(parent_->wtime(), mpistatus::construct());
  }

  shared_ptr<mpiapi::persistentrecv> mpiapi::persistentrecv::
  construct(shared_ptr<mpiapi> api, int count, const mpitype &type,
            const mpiid &dest, const mpitag &tag, const mpicomm &comm)
  {
    shared_ptr<persistentrecv> retval
      (new persistentrecv(api, count, type, dest, tag, comm));
    retval->set_self(retval);
    return retval;
  }

  mpiapi::persistentrecv::~persistentrecv() throw() {
  }

  //
  // Start this request.
  //
  void mpiapi::persistentrecv::start() {
    if(! this->is_complete())
      throw illformederror("mpiapi::persistenrecv::start:  This operation is "
                           "currently active (you need to call wait or test)");
    this->reset();
    parent_->do_irecv(count_, type_, dest_, tag_, comm_, self_.lock());
  }

} // end of namespace sstmac
