/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIAPI_PERSISTENT_H
#define SSTMAC_PROCESS_MPIAPI_PERSISTENT_H

#include <sstmac/process/mpiapi.h>
#include <sstmac/process/processrequest.h>

namespace sstmac {

  /**
   * Base for persistent operations.
   */
  class mpiapi::persistent : public processrequest {
  public:
    persistent();
    virtual ~persistent() throw();
    virtual void start() = 0;
  };

  /**
   * Persistent send operations (send, bsend, rsend, ssend)
   */
  class mpiapi::persistentsend : public mpiapi::persistent {
  public:
    /// The function type used to decide what kind of send we do.
    typedef timestamp (mpiapi::*sendfunc_t)
    (int count, const mpitype &type, const mpiid &target,
     const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
     const boost::shared_ptr<const payload>& content);

  private:
    /// The parent object.
    boost::shared_ptr<mpiapi> parent_;
    sendfunc_t func_;
    /// This object.
    boost::weak_ptr<persistentsend> self_;
    /// The arguments.
    int count_;
    mpitype type_;
    mpiid target_;
    mpitag tag_;
    mpicomm comm_;
    boost::shared_ptr<const payload> content_;

  protected:
    persistentsend(boost::shared_ptr<mpiapi> api,
                   sendfunc_t func, int count, const mpitype &type,
                   const mpiid &target, const mpitag &tag, const mpicomm &comm,
                   const boost::shared_ptr<const payload>& content);


    void set_self(boost::shared_ptr<persistentsend> self) {
      self_ = self;
    }

  public:
    static boost::shared_ptr<persistentsend>
    construct(boost::shared_ptr<mpiapi> api,
              sendfunc_t func, int count, const mpitype &type,
              const mpiid &target, const mpitag &tag, const mpicomm &comm,
              const boost::shared_ptr<const payload>& content);

    virtual ~persistentsend() throw();

    /// Start this request.
    virtual void start();
  };

  /**
   * Persistent recv operations
   */
  class mpiapi::persistentrecv : public mpiapi::persistent {
  private:
    /// Parent object.
    boost::shared_ptr<mpiapi> parent_;
    /// This object.
    boost::weak_ptr<persistentrecv> self_;
    /// The arguments.
    int count_;
    mpitype type_;
    mpiid dest_;
    mpitag tag_;
    mpicomm comm_;

  protected:
    persistentrecv(boost::shared_ptr<mpiapi> api, int count, const mpitype &type,
                   const mpiid &target, const mpitag &tag, const mpicomm &comm);

    void set_self(boost::shared_ptr<persistentrecv> self) {
      self_ = self;
    }

  public:
    static boost::shared_ptr<persistentrecv>
    construct(boost::shared_ptr<mpiapi> api, int count, const mpitype &type,
              const mpiid &target, const mpitag &tag, const mpicomm &comm);

    virtual ~persistentrecv() throw();

    /// Start this request.
    virtual void start();
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIAPI_PERSISTENT_H
