/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiapi.h>
#include <sstmac/process/mpiapi-persistent.h>
#include <sstmac/process/mpiapp.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpitracer.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/process/mpistrategy.h>
#include <sstmac/process/mpicommfactory.h>
#include <sstmac/process/mpikernel.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/key.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/mpiprobe.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/process/processrequest.h>
#include <sstmac/process/threadwake.h>
#include <sstmac/process/multicomplete.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;

  // Catch ugly null pointers.
  template <typename T>
  void nonnull(const boost::shared_ptr<T> &p, const char *caller) {
    if(p == 0)
      throw nullerror(caller);
  }

  //
  // Get a unique tag for a collective operation.
  //
  inline mpitag mpiapi::next_collective_tag(const mpicomm &comm) {
    next_collective_tag_[comm.id()].id += 1;
    return next_collective_tag_[comm.id()];
  }

  //
  // Catch calls on an un-initialized MPI environment.
  //
  inline void mpiapi::assert_initialized(const char *caller) const {
    if(status_ != is_initialized) {
      throw illformederror(std::string(caller) +
                           ":  mpiapi is not active (initialized).");
    }
  }

  // Test whether a communicator is valid for sending or receiving.
  inline void validcomm(const mpicomm &comm, const char *here) {
    if(comm == mpicomm::comm_null)
      throw valueerror(std::string(here) + ": Can't communicate on MPI_COMM_NULL.");
  }

  // Test whether we are sending a valid number of elements.
  inline void validcount(int count, const char *here) {
    if(count < 0)
      throw valueerror(std::string(here) + ": Negative array length.");
  }

  // Test whether the send type is acceptable.
  inline void validtype(const mpitype &type, const char *here) {
    if(type.size <= 0)
      throw valueerror(std::string(here) + ":  MPI type \"" + type.label +
                       "\" does not have size greater than zero.");
  }

  // Test whether the given id is valid for sending.
  inline void validsendid(const mpiid &target, const mpicomm &comm,
                          const char *here)
  {
    if(target.id <  0 || target >= comm.size())
       throw valueerror(std::string(here) + ": Invalid destination rank "
                        + lexical_cast<std::string>(target));
  }

  // Test whether the given id is valid for receiving.
  inline void validrecvid(const mpiid &source, const mpicomm &comm,
                          const char *here)
  {
    if((source.id <  0 || source >= comm.size()) &&
       (source != mpiid::any_source))
      {
        std::cerr << "Failing on source rank" << source << "\n";
        throw valueerror(std::string(here) + ": Invalid source rank"
                         + lexical_cast<std::string>(source));
      }
  }

  // Test wheter the given tag is valid for sending.
  inline void validsendtag(const mpitag &tag, const char *here) {
    if(tag.id < 0)
      throw valueerror(std::string(here) + ": Send-tag must be >= 0.");
  }

  // Test wheter the given tag is valid for receiving.
  inline void validrecvtag(const mpitag &tag, const char *here) {
    if((tag.id < 0) && (tag != mpitag::any_tag))
      throw valueerror(std::string(here) +
                       ": Receive-tag must be >= 0 or mpitag::any_tag.");
  }

  // Test paramters for sending a single element.
  inline void validsend(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(target, comm, here);
    validsendtag(tag, here);
  }

  // Test paramters for receiving a single element.
  inline void validrecv(int count, const mpitype &type, const mpiid &source,
                        const mpitag &tag, const mpicomm &comm,
                        const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validrecvid(source, comm, here);
    validrecvtag(tag, here);
  }

  // Test paramters for sending a vector of elements.
  inline void validsends(const std::vector<int> &counts,
                         const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm,
                         const char *here)
  {
    std::vector<int>::const_iterator it, end = counts.end();
    for(it = counts.begin(); it != end; ++it)
      validcount(*it, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(target, comm, here);
    validsendtag(tag, here);
  }

  // Test paramters for receiving a vector of elements.
  inline void validrecvs(const std::vector<int> &counts,
                         const mpitype &type, const mpiid &source,
                         const mpitag &tag, const mpicomm &comm,
                         const char *here)
  {
    std::vector<int>::const_iterator it, end = counts.end();
    for(it = counts.begin(); it != end; ++it)
      validcount(*it, here);
    validtype(type, here);
    validcomm(comm, here);
    validrecvid(source, comm, here);
    validrecvtag(tag, here);
  }

  // Test arguments for a broadcast.
  inline void validbcast(int count, const mpitype &type, const mpiid &root,
                         const mpitag &tag, const mpicomm &comm,
                         const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(root, comm, here);
    validsendtag(tag, here);
  }

  // Test whether all the given ranks are valid in a communicator.
  inline void validranks(const std::vector<mpiid> &rank,
                         const mpicomm &comm, const char *here)
  {
    const int ub = comm.size().id;
    std::vector<mpiid>::const_iterator it, end = rank.end();
    for(it = rank.begin(); it != end; ++it) {
      if(it->id < 0 || it->id >= ub) {
        throw valueerror(std::string(here) + ":  Invalid rank found.");
      }
    }
  }

  //
  // Build a new mpiapi.
  //
  mpiapi::mpiapi(mpiapp_t my_mpiapp,
                 node_t my_node,
                 mpiqueue_t my_queue,
                 mpimagic_t my_magic,
                 mpiid my_world_id,
                 const_mpistrategy_t strategy,
                 mpicommfactory_t comm_factory,
                 tracer_t tracer) :
    mpiapp_(my_mpiapp), tracer_(tracer), node_(my_node),
    queue_(my_queue), magic_(my_magic), id_(my_world_id), strategy_(strategy),
    comm_factory_(comm_factory), status_(is_fresh)
  {
    nonnull(my_mpiapp, "mpiapi(...):  Null mpiapp pointer.");
    nonnull(my_node, "mpiapi(...):  Null nodeinterface pointer.");
    nonnull(my_queue, "mpiapi(...):  Null mpiqueue pointer.");
    nonnull(my_magic, "mpiapi(...):  Null mpimagic pointer.");
    nonnull(strategy, "mpiapi(...):  Null strategy pointer.");
    nonnull(comm_factory, "mpiapi(...):  Null mpicommfactory pointer.");
    worldcomm_ = comm_factory_->get_world(id_);
    selfcomm_ = comm_factory_->get_self(id_);
  }

  //
  // Who am I?
  //
  void mpiapi::set_self(const shared_ptr<mpiapi>& self) {
    if(self.get() != this) {
      throw valueerror("mpiapi::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpiapi> mpiapi::get_self() {
    return self_.lock();
  }

  //
  // Build a new object.
  //
  shared_ptr<mpiapi> mpiapi::construct(mpiapp_t my_mpiapp,
                                       node_t my_node,
                                       mpiqueue_t my_queue,
                                       mpimagic_t my_magic,
                                       mpiid my_world_id,
                                       const_mpistrategy_t strategy,
                                       mpicommfactory_t comm_factory,
                                       tracer_t tracer)
  {
    shared_ptr<mpiapi> retval(new mpiapi(my_mpiapp, my_node, my_queue,
                                         my_magic, my_world_id,
                                         strategy, comm_factory, tracer));
    retval->set_self(retval);
    return retval;
  }

  //
  // Adieu.
  //
  mpiapi::~mpiapi() {
  }

  //
  // Access the mpiapp we're attached to.
  //
  shared_ptr<mpiapp> mpiapi::app() {
    return mpiapp_.lock();
  }

  //
  // Get the node.
  //
  mpiapi::node_t mpiapi::get_node() {
    return node_;
  }

  //
  // Get the magic communicator.
  //
  mpiapi::mpimagic_t mpiapi::magic() {
    return magic_;
  }

  //
  // Get the world.
  //
  const mpicomm& mpiapi::comm_world() const {
    return worldcomm_;
  }

  //
  // Get self.
  //
  const mpicomm& mpiapi::comm_self() const {
    return selfcomm_;
  }

  /*
   * Methods exposing MPI calls and other key methods to
   * derived objects.
   */
   
  //
  // Pass time without doing anything constructive.
  //
  timestamp mpiapi::sleep(const timestamp &until) {
    shared_ptr<key> thekey = key::construct();
    shared_ptr<scheduler> sched = app()->get_scheduler();
    shared_ptr<threadwake> wakeup(threadwake::construct(sched, thekey));
    node_->schedule(until, wakeup);
    if(! wakeup->done())
      sched->block(thekey);
    return node_->now();
  }

  //
  // Pretend to compute.
  //
  timestamp mpiapi::compute(const timestamp &cputime) {
    shared_ptr<key> thekey = key::construct();
    shared_ptr<scheduler> sched = app()->get_scheduler();
    shared_ptr<threadwake> wakeup(threadwake::construct(sched, thekey));
    node_->compute(node_->now(), cputime, wakeup);
    if(! wakeup->done())
      sched->block(thekey);
    return node_->now();    
  }

  //
  // Perform a computation.  Work in progress.
  //
  timestamp mpiapi::compute(const shared_ptr<instructions>& instr) {
    shared_ptr<key> thekey = key::construct();
    shared_ptr<scheduler> sched = app()->get_scheduler();
    shared_ptr<threadwake> wakeup(threadwake::construct(sched, thekey));
    node_->compute(node_->now(), instr, wakeup);
    if(! wakeup->done())
      sched->block(thekey);
    return node_->now();
  }

  //
  // Get count.
  //
  int mpiapi::get_count(const_mpistatus_t status, const mpitype &type) {
    assert_initialized("mpiapi::get_count");
    nonnull(status, "mpiapi::get_count:  Null status pointer.");
    int count = status->bytes_received() / type.size;
    if(tracer_)
      tracer_->get_count(node_->now(), node_->now(), status, type, count);
    return count;
  }

  //
  // Split a communicator.  This one is a little weird.
  //
  timestamp mpiapi::comm_split(const mpicomm &incomm,
                               int color, int key,
                               mpicomm &outcomm)
  {
    static const char* here = "mpiapi::comm_split";
    //std::cerr << here << "(" << incomm << ", " << color << ", "
    //      << key << ", &outcomm)\n";
    assert_initialized(here);
    outcomm = comm_factory_->comm_split(incomm, color, key);
    if(tracer_)
      tracer_->comm_split(node_->now(), node_->now(), 
                          incomm, color, key, outcomm);
    return node_->now();
  }

  /* Set up and tear down */

  //
  // Test whether MPI has bee initialized.
  //
  bool mpiapi::initialized() {
    if(tracer_)
      tracer_->initialized(node_->now(), node_->now(), status_==is_initialized);
    return (status_ == is_initialized);
  }

  //
  // Test whether MPI has been finalized.
  //
  bool mpiapi::finalized() {
    if(tracer_)
      tracer_->finalized(node_->now(), node_->now(), status_ == is_finalized);
    return (status_ == is_finalized);
  }

  //
  // Initialize MPI.
  //
  timestamp mpiapi::init() {
    mpitag inittag = next_collective_tag(worldcomm_);
    timestamp entry = node_->now();
    if(status_ != is_fresh) {
      throw illformederror("mpiapi::init: Object previously initialized.");
    }
    magic_->barrier(inittag, comm_factory_->get_world(id_),
                    app()->get_scheduler());
    status_ = is_initialized;
    //std::cerr << "mpiapi[" << id_ << "]::init:  at " << node_->now() << "\n";
    if(tracer_) tracer_->init(entry, node_->now(), this->comm_world(),
                              this->comm_self());
    return node_->now();
  }

  //
  // Finalize MPI.
  //
  timestamp mpiapi::finalize() {
    mpitag finishtag = next_collective_tag(worldcomm_);
    timestamp entry = node_->now();
    if(debugger::active()) debug() << "mpiapi::finalize(): rank "<< id_ << "\n";
    status_ = is_finalizing;
    magic_->barrier(finishtag, comm_factory_->get_world(id_),
                    app()->get_scheduler());
    status_ = is_finalized;
    //std::cerr << "mpiapi[" << id_ << "]::finalize:  at "
    //          << node_->now() << "\n";
    if(tracer_) {
      tracer_->finalize(entry, node_->now());
      tracer_->finish();
    }
    return node_->now();
  }

  //
  // Get current time.
  //
  timestamp mpiapi::wtime() {
    return node_->now();
  }

  /* Create and destroy communicators. */
  
  //
  // Duplicate a communicator.
  //
  timestamp mpiapi::comm_dup(const mpicomm &input, mpicomm &output) {
    assert_initialized("mpiapi::comm_dup");
    timestamp entry = node_->now();
    output = comm_factory_->comm_dup(input);
    if(tracer_) tracer_->comm_dup(entry, node_->now(), input, output);
    return node_->now();
  }

  //
  // Create a communicator containing a subset of an existing comm.
  //
  timestamp mpiapi::comm_create(const mpicomm &input,
                                const std::vector<mpiid> &members,
                                mpicomm &output)
  {
    static const char* here = "mpiapi::comm_create";
    assert_initialized(here);
    timestamp entry = node_->now();
    validranks(members, input, here);
    output = comm_factory_->comm_create(input, members);
    if(tracer_)
      tracer_->comm_create(entry, node_->now(), input, members, output);
    return node_->now();
  }

  //
  // Destroy a communicator.
  //
  timestamp mpiapi::comm_free(mpicomm &comm) {
    assert_initialized("mpiapi::comm_create");
    if(tracer_) tracer_->comm_free(node_->now(), node_->now(), comm);
    return node_->now();
  }

  /* Basic point-to-point operations. */

  //
  // Blocking send.
  //
  timestamp mpiapi::send(int count, const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm,
                         const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::send";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    this->do_isend(count, type, target, tag, comm, req, content);
    timestamp finish = req->wait(app()->get_scheduler());
    if(tracer_) tracer_->send(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Blocking send using a user-defined buffer (see buffer_attach).
  //
  timestamp mpiapi::bsend(int count, const mpitype &type,
                          const mpiid &target,
                          const mpitag &tag, const mpicomm &comm,
                          const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::bsend";
    assert_initialized(here);
    validsend(count,type, target, tag, comm, here);
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    this->do_ibsend(count, type, target, tag, comm, req, content);
    timestamp finish = req->wait(app()->get_scheduler());
    if(tracer_) tracer_->bsend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Blocking syncronous send.
  //
  timestamp mpiapi::ssend(int count, const mpitype &type,
                          const mpiid &target,
                          const mpitag &tag, const mpicomm &comm,
                          const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::comm_ssend";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    mpirequest_t req = processrequest::construct(key::construct());
    timestamp entry = node_->now();
    this->do_issend(count, type, target, tag, comm, req, content);
    timestamp finish = req->wait(app()->get_scheduler());
    if(tracer_) tracer_->ssend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Ready send (as in "ready or not, here comes the data").
  timestamp mpiapi::rsend(int count, const mpitype &type,
                          const mpiid &target,
                          const mpitag &tag, const mpicomm &comm,
                          const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::rsend";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    mpirequest_t req = processrequest::construct(key::construct());
    timestamp entry = node_->now();
    this->do_irsend(count, type, target, tag, comm, req, content);
    timestamp finish = req->wait(app()->get_scheduler());
    if(tracer_) tracer_->ssend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Nonblocking send.
  //
  timestamp mpiapi::isend(int count, const mpitype &type,
                          const mpiid &dest, const mpitag &tag,
                          const mpicomm &comm, mpirequest_t &req,
                          const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::isend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = node_->now();
    req = processrequest::construct(key::construct());
    this->do_isend(count, type, dest, tag, comm, req, content);
    timestamp finish = node_->now();
    if(tracer_)
      tracer_->isend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  /// Nonblocking send.
  /// wait for this request has the same completion as send.
  /// \return time when the request has been registered (likely current time).
  timestamp mpiapi::do_isend(int count, const mpitype &type,
                            const mpiid &dest, const mpitag &tag,
                            const mpicomm &comm, const mpirequest_t &req,
                            const boost::shared_ptr<const payload>& content)
  {
    if(debugger::active())
      debug() << "mpiapi[" << this << "]::do_isend(" << count << "," << type
              << "," << dest << ", " << tag << ", " << comm << ", "
              << req << ")\n";
    if(debugger::active()) debug() << "    building a new send kernel\n";
    shared_ptr<mpikernel> ker =
      strategy_->send()->build(req, queue_, count, type,
                               dest, tag, comm, content, magic_);
    if(debugger::active()) debug() << "mpiapi::isend: adding send kernel.\n";
    node_->add(node_->now(), ker);
    return node_->now();
  }

  //
  // Nonblocking user-buffered send.
  //
  timestamp mpiapi::ibsend(int count, const mpitype &type,
                           const mpiid &dest, const mpitag &tag,
                           const mpicomm &comm, mpirequest_t &req,
                           const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::ibsend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = node_->now();
    req = processrequest::construct(key::construct());
    this->do_ibsend(count, type, dest, tag, comm, req, content);
    timestamp finish = node_->now();
    if(tracer_)
      tracer_->ibsend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking user-buffered send.
  //
  timestamp mpiapi::do_ibsend(int count, const mpitype &type,
                              const mpiid &dest, const mpitag &tag,
                              const mpicomm &comm, const mpirequest_t &req,
                              const boost::shared_ptr<const payload>& content)
  {
    std::cerr << "WARNING:  mpiapi::ibsend does not use "
              << "user-defined buffer\n";
    return this->do_isend(count, type, dest, tag, comm, req, content);
  }

  //
  // Nonblocking synchronous send.
  //
  timestamp mpiapi::issend(int count, const mpitype &type,
                           const mpiid &dest, const mpitag &tag,
                           const mpicomm &comm, mpirequest_t &req,
                           const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::issend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = node_->now();
    req = processrequest::construct(key::construct());
    this->do_issend(count, type, dest, tag, comm, req, content);
    timestamp finish = node_->now();
    if(tracer_)
      tracer_->issend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking synchronous send.
  //
  timestamp mpiapi::do_issend(int count, const mpitype &type,
                              const mpiid &dest, const mpitag &tag,
                              const mpicomm &comm, const mpirequest_t &req,
                              const boost::shared_ptr<const payload>& content)
  {
    shared_ptr<mpikernel> ker = 
      strategy_->ssend()->build(req, queue_, count, type,
                                dest, tag, comm, content, magic_);
    node_->add(node_->now(), ker);
    return node_->now();
  }

  //
  // Nonblocking ready send.
  //
  timestamp mpiapi::irsend(int count, const mpitype &type,
                           const mpiid &dest, const mpitag &tag,
                           const mpicomm &comm, mpirequest_t &req,
                           const boost::shared_ptr<const payload>& content)
  {
    static const char *here = "mpiapi::irsend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = node_->now();
    req = processrequest::construct(key::construct());
    this->do_irsend(count, type, dest, tag, comm, req, content);
    timestamp finish = node_->now();
    if(tracer_)
      tracer_->irsend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking ready send.
  //
  timestamp mpiapi::do_irsend(int count, const mpitype &type,
                              const mpiid &dest, const mpitag &tag,
                              const mpicomm &comm, const mpirequest_t &req,
                              const boost::shared_ptr<const payload>& content)
  {
    shared_ptr<mpikernel> ker =
      strategy_->rsend()->build(req, queue_, count, type,
                                dest, tag, comm, content, magic_);
    node_->add(node_->now(), ker);
    return node_->now();
  }

  //
  // Synchronous receive.
  //
  timestamp mpiapi::recv(int count, const mpitype &type, const mpiid &source,
                         const mpitag &tag, const mpicomm &comm,
                         const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::recv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    this->do_irecv(count, type, source, tag, comm, req);
    timestamp fini = req->wait(app()->get_scheduler());
    if(tracer_)
      tracer_->recv(entry, fini, count, type, source, tag, comm, stat);
    stat = dynamic_pointer_cast<const mpistatus>(req->get_status());
    return fini;
  }

  //
  // Nonblocking receive
  //
  timestamp mpiapi::irecv(int count, const mpitype &type,
                          const mpiid &source, const mpitag &tag,
                          const mpicomm &comm, mpirequest_t &req)
  {
    static const char *here = "mpiapi::irecv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);
    timestamp entry = node_->now();
    req = processrequest::construct(key::construct());
    this->do_irecv(count, type, source, tag, comm, req);
    timestamp finish = node_->now();
    if(tracer_)
      tracer_->irecv(entry, finish, count, type, source, tag, comm, req);
    return finish;
  }

  //
  // Asynchronous receive.
  //
  timestamp mpiapi::do_irecv(int count, const mpitype &type,
                             const mpiid &source, const mpitag &tag,
                             const mpicomm &comm, const mpirequest_t &req)
  {
    static const char *here = "mpiapi::irecv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);
    if(debugger::active())
      debug() << "mpiapi[" << this << "]::irecv(" << count << "," << type
              << "," << source << ", " << tag << ", " << comm << ", "
              << req << ")\n";
    if(debugger::active()) debug() << "    Building a new recv kernel\n";
    shared_ptr<mpikernel> ker =
      strategy_->recv()->build(req, queue_, count, type,
                               source, tag, comm, magic_);
    if(debugger::active()) debug() << "mpiapi[" << this << "]::irecv:  adding recv kernel.\n";
    node_->add(node_->now(), ker);
    return node_->now();
  }

  //
  // Send/receieve exchange.
  //
  timestamp mpiapi::sendrecv(int sendcount, const mpitype &sendtype,
                             const mpiid &target, const mpitag &sendtag,
                             int recvcount, const mpitype &recvtype,
                             const mpiid &source, const mpitag &recvtag,
                             const mpicomm &comm, const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::sendrecv";
    assert_initialized(here);
    validsend(sendcount, sendtype, target, sendtag, comm, here);
    validrecv(recvcount, recvtype, source, recvtag, comm, here);
    timestamp entry = node_->now();
    mpirequest_t req1 = processrequest::construct(key::construct());
    mpirequest_t req2 = processrequest::construct(key::construct());
    shared_ptr<payload> cont;
    this->do_isend(sendcount, sendtype, target, sendtag, comm, req1, cont);
    this->do_irecv(recvcount, recvtype, source, recvtag, comm, req2);
    req1->wait(app()->get_scheduler());
    if(! req2->is_complete())
      req2->wait(app()->get_scheduler());
    stat = dynamic_pointer_cast<const mpistatus>(req2->get_status());
    timestamp fini = node_->now();
    if(tracer_)
      tracer_->sendrecv(entry, fini, sendcount, sendtype, target, sendtag,
                        recvcount, recvtype, source, recvtag,
                        comm, stat);
    return fini;
  }

  //
  // Send and receive using a single buffer.
  //
  timestamp mpiapi::sendrecv_replace(int count,
                                     const mpitype &type,
                                     const mpiid &dest,
                                     const mpitag &sendtag,
                                     const mpiid &source,
                                     const mpitag &recvtag, 
                                     const mpicomm &comm,
                                     const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::sendrecv";
    assert_initialized(here);
    validsend(count, type, dest, sendtag, comm, here);
    validrecv(count, type, source, recvtag, comm, here);
    timestamp entry = node_->now();
    mpirequest_t req1 = processrequest::construct(key::construct());
    mpirequest_t req2 = processrequest::construct(key::construct());
;
    shared_ptr<payload> cont;
    this->do_isend(count, type, dest, sendtag, comm, req1, cont);
    this->do_irecv(count, type, source, recvtag, comm, req2);
    req1->wait(app()->get_scheduler());
    if(! req2->is_complete())
      req2->wait(app()->get_scheduler());
    stat = dynamic_pointer_cast<const mpistatus>(req2->get_status());
    timestamp fini = node_->now();
    if(tracer_)
      tracer_->sendrecv_replace(entry, fini, count, type, dest, sendtag,
                                source, recvtag, comm, stat);
    return fini;
  }

  /* Stepwise point-to-point (not implemented, but
   * should be fairly straight-forward to complete). */

  //
  // Initiate a send.
  //
  timestamp mpiapi::send_init(int count,
                              const mpitype &type, const mpiid &target,
                              const mpitag &tag, const mpicomm &comm,
                              mpirequest_t &req,
                              const shared_ptr<const payload> &content)
  {
    static const char *here = "mpiapi::send_init";
    assert_initialized(here);
    req = persistentsend::construct(self_.lock(), &mpiapi::do_isend, count,
                                    type, target, tag, comm, content);
    return this->wtime();
  }

  //
  // Initiate a buffered send.
  //
  timestamp mpiapi::bsend_init(int count, const mpitype &type,
                               const mpiid &target, const mpitag &tag,
                               const mpicomm &comm, mpirequest_t &req,
                               const shared_ptr<const payload> &content)
  {
    static const char *here = "mpiapi::bsend_init";
    assert_initialized(here);
    validsend(count,type, target, tag, comm, here);
    req = persistentsend::construct(self_.lock(), &mpiapi::do_ibsend, count,
                                    type, target, tag, comm, content);
    return this->wtime();
  }

  //
  // Initiate a synchronous send.
  //
  timestamp mpiapi::ssend_init(int count, const mpitype &type,
                               const mpiid &target,
                               const mpitag &tag, const mpicomm &comm,
                               mpirequest_t &req,
                               const shared_ptr<const payload> &content)
  {
    static const char *here = "mpiapli::ssend_init";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    req = persistentsend::construct(self_.lock(), &mpiapi::do_issend, count,
                                    type, target, tag, comm, content);
    return this->wtime();
  }

  //
  // Initiate a ready send.
  //
  timestamp mpiapi::rsend_init(int count, const mpitype &type,
                               const mpiid &target, const mpitag &tag,
                               const mpicomm &comm, mpirequest_t &req,
                               const shared_ptr<const payload> &content)
  {
    static const char *here = "mpiapi::rsend_init";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    req = persistentsend::construct(self_.lock(), &mpiapi::do_irsend, count,
                                    type, target, tag, comm, content);
    return this->wtime();
  }

  //
  // Initiate a receive.
  //
  timestamp mpiapi::recv_init(int count, const mpitype &type,
                              const mpiid &target, const mpitag &tag,
                              const mpicomm &comm, mpirequest_t &req)
  {
    static const char *here = "mpiapi::recv_init";
    assert_initialized(here);
    validrecv(count, type, target, tag, comm, here);
    req = persistentrecv::construct(self_.lock(), count, type,
                                    target, tag, comm);
    return this->wtime();
  }

  //
  // Start a previously initiated request.
  //
  timestamp mpiapi::start(mpirequest_t req) {
    shared_ptr<persistent> pers = dynamic_pointer_cast<persistent>(req);
    if(! pers)
      throw illformederror("mpiapi::start:  Incorrect request handle type.");
    pers->start();
    return this->wtime();
  }

  //
  // Start a collection of previously initiated requests.
  //
  timestamp mpiapi::startall(std::vector< mpirequest_t > &requests) {
    typedef std::vector<mpirequest_t>::iterator iterator_t;
    for(iterator_t it = requests.begin(); it != requests.end(); ++it)
      this->start(*it);
    return this->wtime();
  }

  //
  // Cancel a request that had been previously initiated.
  //
  timestamp mpiapi::cancel(mpirequest_t req) {
    static bool first = true;
    assert_initialized("mpiapi::cancel");
    //throw unimplementederror("mpiapi::cancel:  Not implemented yet.");
    if(first)
      std::cerr << "Warning:  mpiapi::cancel is not implemented "
                << "(but we can currently live without it)\n";
    first = false;
    return this->wtime();
  }

  //
  // Test whether a request has been cancelled (takes no time).
  //
  bool mpiapi::test_cancelled(const const_mpistatus_t &req) {
    assert_initialized("mpiapi::test_cancelled");
    throw unimplementederror("mpiapi::test_cancelled:  Not implemented yet.");
  }

  /* Completion of outstanding requests */

  //
  // Wait for a request to finish.
  //
  timestamp mpiapi::wait(mpirequest_t req, const_mpistatus_t &stat) {
    assert_initialized("mpiapi::wait");
    timestamp entry = node_->now();
    timestamp fini  = entry;
    if(req.get() == NULL) {
      // Null request is instantly complete.
      return node_->now();
    }
    if(! req->is_complete()) {
      fini = req->wait(app()->get_scheduler());
    }
    stat = dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(tracer_)
      tracer_->wait(entry, fini, req, stat);
    return fini;
  }

  //
  // Wait for all the given requests to finish.
  //
  timestamp mpiapi::waitall(std::vector< mpirequest_t > &req,
                            std::vector< const_mpistatus_t > &statuses)
  {
    assert_initialized("mpiapi::waitall");
    timestamp entry = node_->now(), fini = entry;
    if(! req.empty()) {
      typedef std::vector<mpirequest_t>::iterator iterator_t;
      shared_ptr<key> thekey = key::construct();
      shared_ptr<multicomplete> completer =
        multicomplete::construct(app()->get_scheduler(), thekey);
      fini = completer->complete_all(node_->now(), req);
      statuses.clear();
      for(iterator_t it = req.begin(); it != req.end(); ++it) {
        const_mpistatus_t stat;
        if(it->get() != NULL) {
          stat = dynamic_pointer_cast<const mpistatus>((*it)->get_status());
          if(stat == 0) {
            throw illformederror("mpiapi::waitall:  Wait request did not "
                                 "return a valid mpistatus object.");
          }
        }
        statuses.push_back(stat);
      }
    }
    else {
      statuses.clear();
    }
    if(tracer_)
      tracer_->waitall(entry, fini ,req, statuses);
    return fini;
  }

  //
  // Wait for at least one of the given requests to complete.
  //
  timestamp mpiapi::waitany(std::vector< mpirequest_t > &req,
                            int &index,
                            const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::waitany");
    timestamp entry = node_->now(), fini = entry;
    if(! req.empty()) {
      shared_ptr<key> thekey = key::construct();
      shared_ptr<multicomplete> completer =
        multicomplete::construct(app()->get_scheduler(), thekey);
      fini = completer->complete_one(node_->now(), req, index);
      // index is less than zero if no request is active.
      if(index >= 0) {
        stat = dynamic_pointer_cast<const mpistatus>(req.at(index)->get_status());
        if(stat == 0) {
          throw illformederror("mpiapi::waitany:  Wait request did not "
                               "return a valid mpistatus object.");
        }
      }
    }
    else {
      index = -1;
    }
    if(tracer_)
      tracer_->waitany(entry, fini, req, index, stat);
    return fini;
  }

  //
  // Wait for some of the given requests to complete.
  //
  timestamp mpiapi::waitsome(std::vector<mpirequest_t > &req,
                             std::vector<int> &indices,
                             std::vector< const_mpistatus_t > &statuses)
  {
    assert_initialized("mpiapi::waitsome");
    timestamp entry = node_->now(), fini = entry;
    if(! req.empty()) {
      shared_ptr<key> thekey = key::construct();
      shared_ptr<multicomplete> completer =
        multicomplete::construct(app()->get_scheduler(), thekey);
      fini = completer->complete_some(node_->now(), req, indices);
      statuses.clear();
      statuses.reserve(indices.size());
      for(size_t id = 0; id < indices.size(); ++id) {
        const_mpistatus_t stat;
        if(req.at(id).get() != NULL) { 
          stat = dynamic_pointer_cast<const mpistatus>(req.at(id)->get_status());
        }
        statuses.push_back(stat);
      }
    }
    else {
      statuses.clear();
    }
    if(tracer_)
      tracer_->waitsome(entry, fini, req, indices, statuses);
    return fini;
  }

  /* Testing of outstanding requests */

  /// Test whether a request has finished.
  /// \return the completion time for the call (current time).
  timestamp mpiapi::test(mpirequest_t req, bool &flag,
                         const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::test");
    timestamp entry = node_->now(), fini = entry;
    if(req.get() == NULL) {
      flag = true;
      stat = const_mpistatus_t();      
    }
    else {
      flag = req->is_complete();
      if(flag) {
        req->set_inactive();
        stat = dynamic_pointer_cast<const mpistatus>(req->get_status());
      }
    }
    if(tracer_)
      tracer_->test(entry, fini, req, flag, stat);
    return fini;
  }

  //
  // Test whether all the given requests have finished.
  //
  timestamp mpiapi::testall(std::vector< mpirequest_t > &req,
                            bool &flag,
                            std::vector< const_mpistatus_t > &stats)
  {
    assert_initialized("mpiapi::testall");
    timestamp entry = node_->now();
    if(! req.empty()) {
      flag = true;
      for(size_t id = 0; id < req.size(); ++id) {
        if(req.at(id).get() != NULL && ! req.at(id)->is_complete()) {
          flag = false;
          break;
        }
      }
      if(flag) {
        // All requests have completed.
        stats.resize(req.size());
        for(size_t id = 0; id < req.size(); ++id) {
          if(req.at(id).get() != NULL) {
            req.at(id)->set_inactive();
            stats.at(id) =
              dynamic_pointer_cast<const mpistatus>(req.at(id)->get_status());
          }
          else
            stats.at(id) = const_mpistatus_t();
        }
      }
    }
    else {
      stats.clear();
      flag = false;
    }
    timestamp fini = node_->now();
    if(tracer_)
      tracer_->testall(entry, fini, req, flag, stats);
    return fini;
  }

  //
  // Test whether at least one of the given requests has completed.
  //
  timestamp mpiapi::testany(std::vector< mpirequest_t > &req,
                            int &index,
                            bool &flag,
                            const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::testany");
    timestamp entry = node_->now();
    if(! req.empty()) {
      flag = false;
      for(size_t id = 0; id < req.size(); ++id) {
        if(req.at(id).get() != NULL && req.at(id)->is_active() && req.at(id)->is_complete()) {
          req.at(id)->set_inactive();
          index = check_cast<int>(id);
          flag = true;
          stat = dynamic_pointer_cast<const mpistatus>
            (req.at(id)->get_status());
          break;
        }
      }
    }
    else {
      flag = false;
    }
    timestamp fini = node_->now();
    if(tracer_)
      tracer_->testany(entry, fini, req, index, flag, stat);
    return fini;
  }

  //
  // Test whether some of the given requests have completed.
  //
  timestamp mpiapi::testsome(std::vector<mpirequest_t > &req,
                             std::vector<int> &indices,
                             std::vector< const_mpistatus_t > &statuses)
  {
    assert_initialized("mpiapi::testsome");
    timestamp entry = node_->now();
    if(! req.empty()) {
      indices.clear();
      statuses.clear();
      for(size_t id = 0; id < req.size(); ++id) {
        if(req.at(id).get() != NULL && req.at(id)->is_active() && req.at(id)->is_complete()) {
          req.at(id)->set_inactive();
          indices.push_back(check_cast<int>(id));
          statuses.push_back
            (dynamic_pointer_cast<const mpistatus>(req.at(id)->get_status()));
        }
      }
    }
    else {
      indices.clear();
      statuses.clear();
    }
    timestamp fini = node_->now();
    if(tracer_)
      tracer_->testsome(entry, fini, req, indices, statuses);
    return fini;
  }

  //
  // Blocking probe for a message that matches the given signature.
  //
  timestamp mpiapi::probe(const mpiid &source, const mpitag &tag,
                          const mpicomm &comm, const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::probe");
    timestamp entry = node_->now();
    shared_ptr<key> thekey = key::construct();
    shared_ptr<mpiprobe> pr=mpiprobe::construct(app()->get_scheduler(),thekey);
    queue_->probe(comm, source, tag, pr);
    if(! pr->is_done())
      app()->get_scheduler()->block(thekey);
    stat = pr->get_mpistatus();
    timestamp fini = node_->now();
    if(tracer_) tracer_->probe(entry, fini, source, tag, comm, stat);
    return fini;
  }

  //
  // Non-blocking probe for messages.
  //
  timestamp mpiapi::iprobe(const mpiid &source, const mpitag &tag,
                           const mpicomm &comm, bool &flag,
                           const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::iprobe");
    timestamp entry = node_->now();
    mpistatus_t statout = queue_->iprobe(comm, source, tag);
    if(statout != 0) {
      flag = true;
      stat = statout;
    }
    else {
      flag = false;
    }
    timestamp fini = node_->now();
    if(tracer_) tracer_->iprobe(entry, fini, source, tag, comm, flag, stat);
    return fini;
  }


  /* Mucking with MPI state. */

  //
  // Attach a user-defined buffer for *bsend operations.
  //
  timestamp mpiapi::buffer_attach(int bytes) {
    assert_initialized("mpiapi::buffer_attach");
    throw unimplementederror("mpiapi::buffer_attach:  Not implemented.");
  }

  //
  // Detach the user-defined buffer.
  //
  timestamp mpiapi::buffer_detach(int &bytes) {
    assert_initialized("mpiapi::buffer_detach");
    throw unimplementederror("mpiapi::buffer_detach:  Not implemented.");
  }

  /* Collective operations */

  //
  // Set a barrier.
  //
  timestamp mpiapi::barrier(const mpicomm &comm) {
    assert_initialized("mpiapi::barrier");
    mpitag barriertag = next_collective_tag(comm);
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->barrier()->build(req, queue_, barriertag, comm, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());    
    if(tracer_) tracer_->barrier(entry, fini, comm);
    return fini;
 }

  //
  // Broadcast data without payload
  //
  timestamp mpiapi::bcast(int count, const mpitype &type,
                          const mpiid &root, const mpicomm &comm)
  {
    boost::shared_ptr<const payload> content = payload::null();
    return this->bcast(count, type, root, comm, content);
  }

  //
  // Broadcast data with payload.
  //
  timestamp mpiapi::bcast(int count, const mpitype &type,
                          const mpiid &root, const mpicomm &comm,
                          shared_ptr<const payload> &content)
  {
    static const char *here = "mpiapi::bcast";
    mpitag bcasttag = next_collective_tag(comm);
    assert_initialized(here);
    validbcast(count, type, root, bcasttag, comm, here);
    timestamp entry = node_->now();
    shared_ptr<const payload> sendload;
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->bcast()->build(req, queue_, count, type,
                                root, bcasttag, comm, sendload, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");
    shared_ptr<const payload> cp = stat->content();
    if(tracer_) tracer_->bcast(entry, fini, count, type, root, comm);
    return fini;
  }

  //
  // Scatter data uniformly.
  //
  timestamp mpiapi::scatter(int sendcount, const mpitype &sendtype,
                            int recvcount, const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm)
  {
    size_t csize = comm.size().id;
    shared_ptr<mpicollpayload> data = mpicollpayload::construct(csize);
    if(comm.rank() == root)
      for(mpiid id = mpiid(0); id < comm.size(); ++id)
        data->set(id, payload::null());
    shared_ptr<const payload> result;
    timestamp fini = this->scatter(sendcount, sendtype, recvcount, recvtype,
                                   root, comm, data, result);
    return fini;
  }

  //
  // Scatter data with payload.
  //
  timestamp mpiapi::scatter(int sendcount, const mpitype &sendtype,
                            int recvcount, const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm,
                            const shared_ptr<const mpicollpayload>& cc,
                            shared_ptr<const payload> &result)
  {
    static const char *here = "mpiapi::scatter";
    mpitag scattertag = next_collective_tag(comm);
    assert_initialized(here);
    if(comm.rank() == root)
      validsend(sendcount, sendtype, root, scattertag, comm, here);
    else
      validrecv(recvcount, recvtype, root, scattertag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->scatter()->build(req, queue_, sendcount, sendtype,
                                  recvcount, recvtype, root, scattertag,
                                  comm, cc, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::scatter: NULL status pointer.");
    result = stat->content();
    if(tracer_)
      tracer_->scatter(entry, fini, sendcount, sendtype,
                       recvcount, recvtype, root, comm);
    return fini;
  }

  //
  // Scatter data non-uniformly.
  //
  timestamp mpiapi::scatterv(const std::vector<int> &sendcounts,
                             const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpiid &root, const mpicomm &comm)
  {
    size_t csize = comm.size().id;
    shared_ptr<mpicollpayload> data = mpicollpayload::construct(csize);
    if(comm.rank() == root)
      for(mpiid id = mpiid(0); id < comm.size(); ++id)
        data->set(id, payload::null());
    shared_ptr<const payload> rv;
    return this->scatterv(sendcounts, sendtype, recvcount, recvtype,
                          root, comm, data, rv);
  }
    
  //
  // Scatter data non-uniformly.
  //
  timestamp mpiapi::scatterv(const std::vector<int> &sendcounts,
                             const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpiid &root, const mpicomm &comm,
                             const shared_ptr<const mpicollpayload> &cc,
                             shared_ptr<const payload> &result)
  {
    static const char *here = "mpiapi::scatter";
    mpitag scattertag = next_collective_tag(comm);
    assert_initialized(here);
    if(comm.rank() == root)
      validsends(sendcounts, sendtype, root, scattertag, comm, here);
    else
      validrecv(recvcount, recvtype, root, scattertag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->scatterv()->build(req, queue_, sendcounts, sendtype,
                                   recvcount, recvtype, root, scattertag, comm,
                                   cc, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::scatterv: NULL status pointer.");
    result = stat->content();
    if(tracer_)
      tracer_->scatterv(entry, fini, sendcounts, sendtype,
                        recvcount, recvtype, root, comm);
    return fini;
  }

  //
  // Gather uniformly distributed data.
  //
  timestamp mpiapi::gather(int sendcount, const mpitype &sendtype,
                           int recvcount, const mpitype &recvtype,
                           const mpiid &root, const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> rv;
    return gather(sendcount, sendtype, recvcount, recvtype, root, comm,
                  payload::null(), rv);
  }

  //
  // Gather uniformly distributed data.
  //
  timestamp mpiapi::gather(int sendcount, const mpitype &sendtype,
                           int recvcount, const mpitype &recvtype,
                           const mpiid &root, const mpicomm &comm,
                           const shared_ptr<const payload>& content,
                           shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::gather";
    mpitag gathertag = next_collective_tag(comm);
    assert_initialized(here);
    if(comm.rank() == root)
      validrecv(recvcount, recvtype, root, gathertag, comm, here);
    else
      validsend(sendcount, sendtype, root, gathertag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->gather()->build(req, queue_, sendcount, sendtype,
                                 recvcount, recvtype, root, gathertag, comm,
                                 content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    if(comm.rank()== root) {
      boost::shared_ptr<const mpistatus> stat =
        dynamic_pointer_cast<const mpistatus>(req->get_status());
      if(! stat)
        throw illformederror("mpiapi::gather: NULL status pointer.");
      result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
      if(! result)
        throw illformederror("mpiapi::gather:  NULL mpicollpayload.");
    }
    if(tracer_)
      tracer_->gather(entry, fini, sendcount, sendtype,
                      recvcount, recvtype, root, comm);
    return fini;
  }

  //
  // Gather non-uniformly distributed data.
  //
  timestamp mpiapi::gatherv(int sendcount, const mpitype &sendtype,
                            const std::vector<int> &recvcnts,
                            const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm)
  {
    boost::shared_ptr<const mpicollpayload> rv;
    return gatherv(sendcount, sendtype, recvcnts, recvtype, root, comm,
                   payload::null(), rv);
  }

  //
  // Gather non-uniformly distributed data.
  //
  timestamp mpiapi::gatherv(int sendcount, const mpitype &sendtype,
                            const std::vector<int> &recvcnts,
                            const mpitype &recvtype,
                            const mpiid &root, const mpicomm &comm,
                            const shared_ptr<const payload>& content,
                            shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::gatherv";
    mpitag scattertag = next_collective_tag(comm);
    assert_initialized(here);
    if(comm.rank() == root)
      validrecvs(recvcnts, recvtype, root, scattertag, comm, here);
    else
      validsend(sendcount, sendtype, root, scattertag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->gatherv()->build(req, queue_, sendcount, sendtype,
                                  recvcnts, recvtype, root, scattertag,
                                  comm, content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    if(comm.rank()== root) {
      boost::shared_ptr<const mpistatus> stat =
        dynamic_pointer_cast<const mpistatus>(req->get_status());
      if(! stat)
        throw illformederror("mpiapi::gatherv: NULL status pointer.");
      result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
      if(! result)
        throw illformederror("mpiapi::gatherv:  NULL mpicollpayload.");
    }
    if(tracer_)
      tracer_->gatherv(entry, fini, sendcount, sendtype,
                       recvcnts, recvtype, root, comm);
    return fini;
  }

  //
  // Gather uniformly distributed data and make it available to all nodes.
  //
  timestamp mpiapi::allgather(int sendcount, const mpitype &sendtype,
                              int recvcount, const mpitype &recvtype,
                              const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> rv;
    return allgather(sendcount, sendtype, recvcount, recvtype, comm,
                     payload::null(), rv);
  }

  //
  // Gather uniformly distributed data and make it available to all nodes.
  //
  timestamp mpiapi::allgather(int sendcount, const mpitype &sendtype,
                              int recvcount, const mpitype &recvtype,
                              const mpicomm &comm,
                              const shared_ptr<const payload>& content,
                              shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::allgather";
    mpitag mytag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(sendcount, sendtype, comm.rank(), mytag, comm, here);
    validrecv(recvcount, recvtype, comm.rank(), mytag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->allgather()->build(req, queue_, sendcount, sendtype,
                                    recvcount, recvtype, mytag, comm,
                                    content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::allgather: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(! result)
      throw illformederror("mpiapi::allgather:  NULL mpicollpayload.");
    if(tracer_)
      tracer_->allgather(entry, fini, sendcount, sendtype,
                         recvcount, recvtype, comm);
    return fini;
  }

  //
  // Gather non-uniformly distributed data onto all nodes.
  //
  timestamp mpiapi::allgatherv(int sendcnt,
                               const mpitype &sendtype,
                               const std::vector<int> &recvcnts,
                               const mpitype &recvtype,
                               const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return allgatherv(sendcnt, sendtype, recvcnts, recvtype, comm,
                      payload::null(), result);
  }

  //
  // Gather non-uniformly distributed data onto all nodes.
  //
  timestamp mpiapi::allgatherv(int sendcnt,
                               const mpitype &sendtype,
                               const std::vector<int> &recvcnts,
                               const mpitype &recvtype,
                               const mpicomm &comm,
                               const shared_ptr<const payload>& content,
                               shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::allgatherv";
    mpitag mytag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(sendcnt, sendtype, comm.rank(), mytag, comm, here);
    validsends(recvcnts, recvtype, comm.rank(), mytag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->allgatherv()->build(req, queue_, sendcnt, sendtype,
                                     recvcnts, recvtype, mytag, comm,
                                     content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::allgatherv: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(! result)
      throw illformederror("mpiapi::allgatherv:  NULL mpicollpayload.");
    if(tracer_)
      tracer_->allgatherv(entry, fini, sendcnt, sendtype,
                          recvcnts, recvtype, comm);
    return fini;
  }

  //
  // Send uniformly distributed data from all to all processes.
  //
  timestamp mpiapi::alltoall(int sendcount, const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return this->alltoall(sendcount, sendtype, recvcount, recvtype, comm,
                          payload::null(), result);
  }

  //
  // Send uniformly distributed data from all to all processes.
  //
  timestamp mpiapi::alltoall(int sendcount, const mpitype &sendtype,
                             int recvcount, const mpitype &recvtype,
                             const mpicomm &comm,
                             const shared_ptr<const payload>& content,
                             shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::alltoall";
    mpitag a2atag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(sendcount, sendtype, comm.rank(), a2atag, comm, here);
    validrecv(recvcount, recvtype, comm.rank(), a2atag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->alltoall()->build(req, queue_, sendcount, sendtype,
                                   recvcount, recvtype, a2atag, comm,
                                   content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::alltoall: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(! result)
      throw illformederror("mpiapi::alltoall:  NULL mpicollpayload.");
    if(tracer_)
      tracer_->alltoall(entry, fini, sendcount, sendtype,
                        recvcount, recvtype, comm);
    return fini;
  }

  //
  // Send non-uniformly distributed data from all to all processors.
  //
  timestamp mpiapi::alltoallv(const std::vector<int> &sendcnts,
                              const mpitype &sendtype,
                              const std::vector<int> &recvcnts,
                              const mpitype &recvtype,
                              const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return alltoallv(sendcnts, sendtype, recvcnts, recvtype, comm,
                     payload::null(), result);
  }

  //
  // Send non-uniformly distributed data from all to all processors.
  //
  timestamp mpiapi::alltoallv(const std::vector<int> &sendcnts,
                              const mpitype &sendtype,
                              const std::vector<int> &recvcnts,
                              const mpitype &recvtype,
                              const mpicomm &comm,
                              const shared_ptr<const payload>& content,
                              shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::alltoallv";
    mpitag a2avtag = next_collective_tag(comm);
    assert_initialized(here);
    validsends(sendcnts, sendtype, comm.rank(), a2avtag, comm, here);
    validrecvs(recvcnts, recvtype, comm.rank(), a2avtag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->alltoallv()->build(req, queue_, sendcnts, sendtype,
                                    recvcnts, recvtype, a2avtag, comm,
                                    content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::alltoallv: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(! result)
      throw illformederror("mpiapi::alltoallv:  NULL mpicollpayload.");
    if(tracer_)
      tracer_->alltoallv(entry, fini, sendcnts, sendtype,
                         recvcnts, recvtype, comm);
    return fini;
  }

  //
  // Reduce data from all nodes onto the root node.
  //
  timestamp mpiapi::reduce(int count, const mpitype &type, const mpiop &op,
                           const mpiid &root, const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return reduce(count, type, op, root, comm, payload::null(), result);
  }

  //
  // Reduce data from all nodes onto the root node.
  //
  timestamp mpiapi::reduce(int count, const mpitype &type,
                           const mpiop &op,
                           const mpiid &root, const mpicomm &comm,
                           const boost::shared_ptr<const payload>& content,
                           boost::shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::reduce";
    mpitag reducetag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(count, type, comm.rank(), reducetag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->reduce()->build(req, queue_, count, type, op, root,
                                 reducetag, comm, content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    if(comm.rank() == root) {
      boost::shared_ptr<const mpistatus> stat =
        dynamic_pointer_cast<const mpistatus>(req->get_status());
      if(! stat)
        throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");
      result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    }
    else {
      result = boost::shared_ptr<const mpicollpayload>();
    }
    if(tracer_)
      tracer_->reduce(entry, fini, count, type, op, root, comm);
    return fini;
  }

  //
  // Reduce data from all nodes to all nodes.
  //
  timestamp mpiapi::allreduce(int count, const mpitype &type,
                              const mpiop &op,
                              const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return allreduce(count, type, op, comm, payload::null(), result);
  }

  timestamp mpiapi::allreduce(int count, const mpitype &type,
                              const mpiop &op,
                              const mpicomm &comm,
                              const boost::shared_ptr<const payload>& content,
                              boost::shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::allreduce";
    mpitag allreducetag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(count, type, comm.rank(), allreducetag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->allreduce()->build(req, queue_, count, type, op,
                                    allreducetag, comm, content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(tracer_)
      tracer_->allreduce(entry, fini, count, type, op, comm);
    return fini;
  }

  //
  // Reduce data and scatter results.
  //
  timestamp mpiapi::reduce_scatter(const std::vector<int> &recvcnts,
                                   const mpitype &type,
                                   const mpiop &op,
                                   const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return this->reduce_scatter(recvcnts, type, op, comm,
                                payload::null(), result);
  }

  //
  // Reduce data and scatter results.
  //
  timestamp mpiapi::reduce_scatter(const std::vector<int> &recvcnts,
                                   const mpitype &type,
                                   const mpiop &op,
                                   const mpicomm &comm,
                                   const shared_ptr<const payload>& content,
                                   shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::reduce_scatter";
    static mpitag rscattag(0xFFAA1A);
    assert_initialized(here);
    validsends(recvcnts, type, comm.rank(), rscattag, comm, here);    
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->reduce_scatter()->build(req, queue_, recvcnts, type,
                                         op, rscattag, comm, content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::alltoallv: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(! result)
      throw illformederror("mpiapi::alltoallv:  NULL mpicollpayload.");
    if(tracer_)
      tracer_->reduce_scatter(entry, fini, recvcnts, type, op, comm);
    return fini;
  }

  //
  // Compute partial reductions on data.
  //
  timestamp mpiapi::scan(int count, const mpitype &type,
                         const mpiop &op, const mpicomm &comm)
  {
    shared_ptr<const mpicollpayload> result;
    return this->scan(count, type, op, comm, payload::null(), result);
  }

  //
  // Compute partial reductions and aggregate data.
  //
  timestamp mpiapi::scan(int count, const mpitype &type,
                         const mpiop &op, const mpicomm &comm,
                         const shared_ptr<const payload>& content,
                         shared_ptr<const mpicollpayload> &result)
  {
    static const char *here = "mpiapi::scan";
    mpitag scantag = next_collective_tag(comm);
    assert_initialized(here);
    validsend(count, type, comm.rank(), scantag, comm, here);
    //
    timestamp entry = node_->now();
    mpirequest_t req = processrequest::construct(key::construct());
    shared_ptr<mpikernel> ker =
      strategy_->scan()->build(req, queue_, count, type,
                               op, scantag, comm, content, magic_);
    node_->add(node_->now(), ker);
    timestamp fini = req->wait(app()->get_scheduler());
    boost::shared_ptr<const mpistatus> stat =
      dynamic_pointer_cast<const mpistatus>(req->get_status());
    if(! stat)
      throw illformederror("mpiapi::scan: NULL status pointer.");
    result = dynamic_pointer_cast<const mpicollpayload>(stat->content());
    if(tracer_)
      tracer_->scan(entry, fini, count, type, op, comm);
    return fini;
  }

} // end of namespace sstmac

