/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIAPI_H
#define SSTMAC_PROCESS_MPIAPI_H

#include <sstmac/process/mpicomm.h>
#include <sstmac/process/payload.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpiop.h>
#include <boost/weak_ptr.hpp>
#include <map>

namespace sstmac {

  // Forward declaration of relevant types.
  class mpiqueue;
  class mpistatus;
  class nodeinterface;
  class mpistrategy;
  class mpicommfactory;
  class mpimagic;
  class mpicollpayload;
  class eventhandler;
  class mpiapp;
  class instructions;
  class processrequest;
  class key;
  class mpitracer;
  
  /**
   * All the joy of using MPI from an application (or rather, a thread).
   */
  class mpiapi {
    /// This friendship declaration can probably be eliminated later.
    friend class mpilaunch;

    /// Nested classes to take care of persistent communications.
    class persistent;
    class persistentsend;
    class persistentrecv;

  public:
    /// Utility definition.
    typedef processrequest mpirequest;

    // Some of these shared pointer types are just outrageously long.
    typedef boost::shared_ptr<nodeinterface> node_t;
    typedef boost::shared_ptr<const mpistrategy> const_mpistrategy_t;
    typedef boost::shared_ptr<mpicommfactory> mpicommfactory_t;
    typedef boost::shared_ptr<const mpicommfactory> const_mpicommfactory_t;
    typedef boost::shared_ptr<mpiapp> mpiapp_t;
    typedef boost::shared_ptr<mpirequest> mpirequest_t;
    typedef boost::shared_ptr<const mpirequest> const_mpirequest_t;
    typedef boost::shared_ptr<mpistatus> mpistatus_t;
    typedef boost::shared_ptr<const mpistatus> const_mpistatus_t;
    typedef boost::shared_ptr<mpimagic> mpimagic_t;
    typedef boost::shared_ptr<mpiqueue> mpiqueue_t;
    typedef boost::shared_ptr<mpitracer> tracer_t;

  private:
    /// This object.
    boost::weak_ptr<mpiapi> self_;

    /// The app this api is attached to.  This is used to block and unblock
    /// the application when needed (must be a weak pointer, since the app
    /// is what keeps this object alive).
    boost::weak_ptr<mpiapp> mpiapp_;

    /// The MPI call tracer.  If this is NULL, no tracing will be performed.
    boost::shared_ptr<mpitracer> tracer_;

    /// My node.
    node_t node_;

    /// The MPI server.
    mpiqueue_t queue_;

    /// A "magic" communication route.
    mpimagic_t magic_;

    //// My MPI index in the world.
    mpiid id_;

    /// The strategy object that gets consulted for how to do anything.
    const_mpistrategy_t strategy_;

    /// The builder for mpi communicators.
    mpicommfactory_t comm_factory_;

    /// Ensure that collective operations are given unique tags
    /// to protect shared back-end data (magic payload info).
    std::map<mpicommid, mpitag> next_collective_tag_;
    inline mpitag next_collective_tag(const mpicomm &comm);

    /// The state of this object (initialized or not).
    enum { is_fresh, is_initialized, is_finalizing, is_finalized } status_;
    
    /// MPI_COMM_WORLD.
    mpicomm worldcomm_;

    /// MPI_COMM_SELF.
    mpicomm selfcomm_;

    /// Ensure that MPI is properly initialized.
    inline void assert_initialized(const char *caller) const;

    /// Nonblocking send -- actual implementation.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp do_isend(int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm &comm,
                       const mpirequest_t &req,
                       const boost::shared_ptr<const payload>& content);

    /// Nonblocking send -- actual implementation.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp do_ibsend(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req,
                        const boost::shared_ptr<const payload>& content);

    /// Nonblocking send -- actual implementation.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp do_issend(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req,
                        const boost::shared_ptr<const payload>& content);

    /// Nonblocking send -- actual implementation.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp do_irsend(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        const mpirequest_t &req,
                        const boost::shared_ptr<const payload>& content);

    /// Asynchronous receive -- actual implementation.
    /// wait for this request has the same completion as recv.
    /// \return time at which the receive request has been registered.
    timestamp do_irecv(int count, const mpitype &type, const mpiid &source,
                       const mpitag &tag, const mpicomm &comm,
                       const mpirequest_t &req);

  protected:
    /// Build a new mpiapi.
    mpiapi(mpiapp_t my_mpiapp, node_t my_node, mpiqueue_t my_queue,
           mpimagic_t my_magic, mpiid my_world_id, const_mpistrategy_t strategy,
           mpicommfactory_t comm_factory, tracer_t tracer);

    /// Define who I am.
    void set_self(const boost::shared_ptr<mpiapi>& self);

    /// Override the pure virtual method from thread.
    boost::shared_ptr<mpiapi> get_self();

  public:
    /// Hello.
    static boost::shared_ptr<mpiapi>
    construct(mpiapp_t my_mpiapp, node_t my_node, mpiqueue_t my_queue,
              mpimagic_t my_magic, mpiid world_id,
              const_mpistrategy_t strategy, mpicommfactory_t comm_factory,
              tracer_t tracer = tracer_t());

    /// Adieu.
    virtual ~mpiapi();

    /*
     * Access basic runtime information.
     */

    /// The node assigned to this run.
    node_t get_node();

    /// The mpiapp we are attached to.
    /// This is used to get access to the scheduler.
    boost::shared_ptr<mpiapp> app();

    /// "Magic" mpi communications.
    mpimagic_t magic();

    /// Direct access to the mpiqueue
    mpiqueue_t queue() { return queue_; }

    /*
     * Methods exposing MPI calls and other key methods to
     * derived objects.
     */

    /// Get the world communicator.
    const mpicomm& comm_world() const;

    /// Get the self communicator.
    const mpicomm& comm_self() const;
    
    /// Pass time without doing anything constructive.
    /// Returns time when sleep was released.
    timestamp sleep(const timestamp &until);

    /// Pass time while pretending to compute.
    /// Returns time when the given cpu seconds has been collected.
    timestamp compute(const timestamp &cputime);

    /// Perform a computation.  Work in progress.
    timestamp compute(const boost::shared_ptr<instructions>& instr);

    /* Utility functions to mirror MPI API functionality */

    /// The rank of this node in a communicator.
    inline mpiid comm_rank(const mpicomm &comm) {
      return comm.rank();
    }

    /// The size of a communicator.
    inline mpiid comm_size(const mpicomm &comm) {
      return comm.size();
    }

    /// Get the number of elements received assuming a specific type.
    int get_count(const_mpistatus_t status, const mpitype &type);

    /// Split a communicator.  This one is a little weird.
    timestamp comm_split(const mpicomm &incomm,
                         int color, int key,
                         mpicomm &outcomm);

    /// Duplicate a type.  Easy enough.
    inline timestamp type_dup(const mpitype &intype, mpitype &outtype) {
      outtype = intype;
      return this->wtime();
    }

    /// Commit a type.   This is a no-op.
    inline timestamp type_commit(const mpitype&) {
      return this->wtime();
    }

    /// Create a new type.
    inline timestamp type_contiguous(int count, const mpitype &oldtype,
                                     mpitype &newtype)
    {
      newtype = mpitype("contigous", count * oldtype.size);
      return this->wtime();
    }

    /// Free a type.  This is a no-op.
    inline timestamp type_free(mpitype&) {
      return this->wtime();
    }

    /// Get type size.
    inline int type_size(const mpitype &type) {
      return type.size;
    }

    /* Set up and tear down */

    /// Test whether MPI has bee initialized.
    bool initialized();

    /// Test whether MPI has been finalized.
    bool finalized();

    /// Initialize MPI.  Must be the first MPI call made other than
    /// initialized or finalized.
    timestamp init();

    /// Finalize MPI.  Must be the last MPI call other than
    /// initialized or finalized.
    timestamp finalize();

    /// Get current time.
    timestamp wtime();

    /* Create and destroy communicators. */
    
    /// Duplicate a communicator.
    timestamp comm_dup(const mpicomm &input, mpicomm &output);

    /// Create a communicator containing a subset of an existing comm.
    timestamp comm_create(const mpicomm &input,
                          const std::vector<mpiid> &members,
                          mpicomm &output);

    /// Destroy a communicator.  This is currently a noop, but should later
    /// mark the communicator invalid so erroneous program behavior can be
    /// detected.
    timestamp comm_free(mpicomm &input);

    /* Basic point-to-point operations. */

    /// Blocking send.
    /// \return time at which the (imaginary) input buffer can be reused.
    timestamp send(int count, const mpitype &type, const mpiid &target,
                   const mpitag &tag, const mpicomm &comm,
                   const boost::shared_ptr<const payload>& content = payload::null());

    /// Blocking send using a user-defined buffer (see buffer_attach).
    /// \return time at which it is safe to use the (hypothetical) input array.
    timestamp bsend(int count, const mpitype &type, const mpiid &target,
                    const mpitag &tag, const mpicomm &comm,
                    const boost::shared_ptr<const payload>& content = payload::null());

    /// Blocking syncronous send.
    /// \return time when a matching recv (or irecv) has acknowledged receipt.
    timestamp ssend(int count, const mpitype &type, const mpiid &target,
                    const mpitag &tag, const mpicomm &comm,
                    const boost::shared_ptr<const payload>& content = payload::null());

    /// Ready send (as in "ready or not, here comes the data").
    /// \return time when it is safe to use the input array.
    timestamp rsend(int count, const mpitype &type, const mpiid &target,
                    const mpitag &tag, const mpicomm &comm,
                    const boost::shared_ptr<const payload>& content = payload::null());

    /// Nonblocking send.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp isend(int count, const mpitype &type, const mpiid &target,
                    const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
                    const boost::shared_ptr<const payload>& content = payload::null());

    /// Nonblocking user-buffered send.
    /// wait for this request has the same completion as bsend.
    /// \return time when the request has been registered (likely current time).
    timestamp ibsend(int count, const mpitype &type, const mpiid &target,
                     const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
                     const boost::shared_ptr<const payload>& content = payload::null());

    /// Nonblocking synchronous send.
    /// wait for this request has the same completion as ssend.
    timestamp issend(int count, const mpitype &type, const mpiid &target,
                     const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
                   const boost::shared_ptr<const payload>& content = payload::null());

    /// Nonblocking ready send.
    /// wait for this request has the same completion as rsend.
    timestamp irsend(int count, const mpitype &type, const mpiid &target,
                     const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
                     const boost::shared_ptr<const payload>& content =payload::null());

    /// Blocking receive.
    /// Note that the message content (may be null) is available
    /// via stat.content().
    /// \return time at which the input data is available for use.
    timestamp recv(int count, const mpitype &type, const mpiid &source,
                   const mpitag &tag, const mpicomm &comm,
                   const_mpistatus_t &stat);

    /// Asynchronous receive.
    /// wait for this request has the same completion as recv.
    /// \return time at which the receive request has been registered.
    timestamp irecv(int count, const mpitype &type, const mpiid &source,
                    const mpitag &tag, const mpicomm &comm,
                    mpirequest_t &req);

    /// Send/receieve exchange.
    /// \return time at which the request has been completed.
    timestamp sendrecv(int sendcount, const mpitype &sendtype,
                       const mpiid &target, const mpitag &sendtag,
                       int recvcount, const mpitype &recvtype,
                       const mpiid &source, const mpitag &recvtag,
                       const mpicomm &comm, const_mpistatus_t &stat);

    /// Send and receive using a single buffer.
    timestamp sendrecv_replace(int count, const mpitype &type,
                               const mpiid &dest, const mpitag &sendtag,
                               const mpiid &source, const mpitag &recvtag, 
                               const mpicomm &comm, const_mpistatus_t &stat);

    /* Stepwise point-to-point (not implemented, but
     * should be fairly straight-forward to complete). */

    /// Initiate a send.
    timestamp send_init(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        mpirequest_t &req,
                        const boost::shared_ptr<const payload> &content = payload::null());

    /// Initiate a buffered send.
    timestamp bsend_init(int count, const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm,
                         mpirequest_t &req,
                         const boost::shared_ptr<const payload> &content = payload::null());

    /// Initiate a synchronous send.
    timestamp ssend_init(int count, const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm,
                         mpirequest_t &req,
                         const boost::shared_ptr<const payload> &content = payload::null());

    /// Initiate a ready send.
    timestamp rsend_init(int count, const mpitype &type, const mpiid &target,
                         const mpitag &tag, const mpicomm &comm,
                         mpirequest_t &req,
                         const boost::shared_ptr<const payload> &content = payload::null());

    /// Initiate a receive.
    timestamp recv_init(int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm &comm,
                        mpirequest_t &req);

    /// Start a previously initiated request.
    timestamp start(mpirequest_t req);

    /// Start a collection of previously initiated requests.
    timestamp startall(std::vector< mpirequest_t > &requests);

    /// Cancel a request that had been previously initiated.
    timestamp cancel(mpirequest_t req);

    /// Test whether a request has been cancelled (takes no time).
    bool test_cancelled(const const_mpistatus_t &req);

    /* Completion of outstanding requests */

    /// Wait for a request to finish.
    /// \return the completion time for the call (current time).
    timestamp wait(mpirequest_t req, const_mpistatus_t &stat);

    /// Wait for all the given requests to finish.
    /// \return the completion time for the call (current time).
    timestamp waitall(std::vector< mpirequest_t > &req,
                      std::vector< const_mpistatus_t > &statuses);

    /// Wait for at least one of the given requests to complete.
    /// \param index is the index of the first request identified as complete.
    ///        This does not have to be the request with the earliest completion
    /// \return the completion time for the call (current time).
    timestamp waitany(std::vector< mpirequest_t > &req,
                      int &index,
                      const_mpistatus_t &stat);

    /// Wait for some of the given requests to complete.
    /// \param indices are the request indices that have completed.
    /// \return the completion time for the call (current time).
    timestamp waitsome(std::vector<mpirequest_t > &req,
                       std::vector<int> &indices,
                       std::vector< const_mpistatus_t > &statuses);

    /* Testing of outstanding requests */

    /// Test whether a request has finished.
    /// \return the completion time for the call (current time).
    timestamp test(mpirequest_t req, bool &flag, const_mpistatus_t &stat);

    /// Test whether all the given requests have finished.
    /// \return the completion time for the call (current time).
    timestamp testall(std::vector< mpirequest_t > &req,
                      bool &flag,
                      std::vector< const_mpistatus_t > &statuses);

    /// Test whether at least one of the given requests has completed.
    /// \param index is the index of the first request identified as complete.
    ///        This does not have to be the request with the earliest completion
    /// \return the completion time for the call (current time).
    timestamp testany(std::vector< mpirequest_t > &req,
                      int &index,
                      bool &flag,
                      const_mpistatus_t &stat);

    /// Test whether some of the given requests have completed.
    /// \param indices are the request indices that have completed.
    /// \return the completion time for the call (current time).
    timestamp testsome(std::vector<mpirequest_t > &req,
                       std::vector<int> &indices,
                       std::vector< const_mpistatus_t > &statuses);

    /// Blocking probe for a message that matches the given signature.
    /// \return current time when the probe succeeds.
    timestamp probe(const mpiid &source, const mpitag &tag,
                    const mpicomm &comm, const_mpistatus_t &stat);

    /// Non-blocking probe for messages.
    /// \return current time when the call completes.
    timestamp iprobe(const mpiid &source, const mpitag &tag,
                     const mpicomm &comm, bool &flag,
                     const_mpistatus_t &stat);

    /* Mucking with MPI state. */

    /// Attach a user-defined buffer for *bsend operations.
    timestamp buffer_attach(int bytes);

    /// Detach the user-defined buffer.
    /// \param bytes is the previously assigned buffer size.
    timestamp buffer_detach(int &bytes);

    // I am omitting all the keyval stuff; it doesn't have much function
    // in a simulator and has a significant implementation complexity.

    /* Collective operations */

    /// Set a barrier.
    /// \return the time at which this node was released from the barrier.
    timestamp barrier(const mpicomm &comm);

    /// Broadcast data with no payload.
    timestamp bcast(int count, const mpitype &type,
                    const mpiid &root, const mpicomm &comm);

    /// Broadcast data and send/receive payload.
    /// On root, the content will be unchanged.
    /// Other nodes will get a copy of root's payload.
    timestamp bcast(int count, const mpitype &type,
                    const mpiid &root, const mpicomm &comm,
                    boost::shared_ptr<const payload> &content);

    /// Scatter data uniformly.
    timestamp scatter(int sendcount, const mpitype &sendtype,
                      int recvcount, const mpitype &recvtype,
                      const mpiid &root, const mpicomm &comm);

    /// Scatter data uniformly and send payload.
    /// The content at root will be sent to all nodes.
    /// The 'result' pointer will be updated on all nodes to
    /// point to a copy of root's content.
    timestamp scatter(int sendcount, const mpitype &sendtype,
                      int recvcount, const mpitype &recvtype,
                      const mpiid &root, const mpicomm &comm,
                      const boost::shared_ptr<const mpicollpayload>& content,
                      boost::shared_ptr<const payload> &result);

    /// Scatter data non-uniformly.
    timestamp scatterv(const std::vector<int> &sendcounts,
                       const mpitype &sendtype,
                       int recvcount, const mpitype &recvtype,
                       const mpiid &root, const mpicomm &comm);
    
    /// Scatter data non-uniformly.
    /// The content at root will be sent to all nodes.
    /// The 'result' pointer will be updated on all nodes to
    /// point to a copy of root's content.
    timestamp scatterv(const std::vector<int> &sendcounts,
                       const mpitype &sendtype,
                       int recvcount, const mpitype &recvtype,
                       const mpiid &root, const mpicomm &comm,
                       const boost::shared_ptr<const mpicollpayload>& content,
                       boost::shared_ptr<const payload> &result);
    
    /// Gather uniformly distributed data.
    timestamp gather(int sendcount, const mpitype &sendtype,
                     int recvcount, const mpitype &recvtype,
                     const mpiid &root, const mpicomm &comm);

    /// Gather uniformly distributed data and payload.
    /// The 'result' at root will have pointers to copies of
    /// all node's content.
    timestamp gather(int sendcount, const mpitype &sendtype,
                     int recvcount, const mpitype &recvtype,
                     const mpiid &root, const mpicomm &comm,
                     const boost::shared_ptr<const payload>& content,
                     boost::shared_ptr<const mpicollpayload> &result);

    /// Gather non-uniformly distributed data.
    timestamp gatherv(int sendcount, const mpitype &sendtype,
                      const std::vector<int> &recvcnts,
                      const mpitype &recvtype,
                      const mpiid &root, const mpicomm &comm);

    /// Gather non-uniformly distributed data and payload.
    /// The 'result' at root will have pointers to copies of
    /// all node's content.
    timestamp gatherv(int sendcount, const mpitype &sendtype,
                      const std::vector<int> &recvcnts,
                      const mpitype &recvtype,
                      const mpiid &root, const mpicomm &comm,
                      const boost::shared_ptr<const payload>& content,
                      boost::shared_ptr<const mpicollpayload> &result);

    /// Gather uniformly distributed data and make it available to all nodes.
    timestamp allgather(int sendcount, const mpitype &sendtype,
                        int recvcount, const mpitype &recvtype,
                        const mpicomm &comm);

    /// Gather uniformly distributed data and make it available to all nodes.
    timestamp allgather(int sendcount, const mpitype &sendtype,
                        int recvcount, const mpitype &recvtype,
                        const mpicomm &comm,
                        const boost::shared_ptr<const payload>& content,
                        boost::shared_ptr<const mpicollpayload> &result);

    /// Gather non-uniformly distributed data onto all nodes.
    timestamp allgatherv(int sendcount,
                         const mpitype &sendtype,
                         const std::vector<int> &recvcnts,
                         const mpitype &recvtype,
                         const mpicomm &comm);

    /// Gather non-uniformly distributed data onto all nodes.
    timestamp allgatherv(int sendcount,
                         const mpitype &sendtype,
                         const std::vector<int> &recvcnts,
                         const mpitype &recvtype,
                         const mpicomm &comm,
                         const boost::shared_ptr<const payload>& content,
                         boost::shared_ptr<const mpicollpayload> &result);

    /// Send uniformly distributed data from all to all processes.
    timestamp alltoall(int sendcount, const mpitype &sendtype,
                       int recvcount, const mpitype &recvtype,
                       const mpicomm &comm);

    /// Send uniformly distributed data from all to all processes.
    timestamp alltoall(int sendcount, const mpitype &sendtype,
                       int recvcount, const mpitype &recvtype,
                       const mpicomm &comm,
                       const boost::shared_ptr<const payload>& content,
                       boost::shared_ptr<const mpicollpayload> &result);

    /// Send non-uniformly distributed data from all to all processors.
    timestamp alltoallv(const std::vector<int> &sendcnts,
                        const mpitype &sendtype,
                        const std::vector<int> &recvcnts,
                        const mpitype &recvtype,
                        const mpicomm &comm);

    /// Send non-uniformly distributed data from all to all processors.
    timestamp alltoallv(const std::vector<int> &sendcnts,
                        const mpitype &sendtype,
                        const std::vector<int> &recvcnts,
                        const mpitype &recvtype,
                        const mpicomm &comm,
                        const boost::shared_ptr<const payload>& content,
                        boost::shared_ptr<const mpicollpayload> &result);

    /// Reduce data from all nodes onto the root node.
    timestamp reduce(int count, const mpitype &type,
                     const mpiop &op,
                     const mpiid &root, const mpicomm &comm);

    /// Reduce data from all nodes onto the root node and give root node
    /// access to everybody's payload.
    /// On root node, the result pointer will be rewritten to point to
    /// an object of type mpicollpayload.  On other nodes, this
    /// pointer will be set to null.
    timestamp reduce(int count, const mpitype &type,
                     const mpiop &op,
                     const mpiid &root, const mpicomm &comm,
                     const boost::shared_ptr<const payload>& content,
                     boost::shared_ptr<const mpicollpayload> &result);

    /// Reduce data from all nodes to all nodes.
    timestamp allreduce(int count, const mpitype &type,
                        const mpiop &op,
                        const mpicomm &comm);

    /// Reduce data from all nodes to all nodes and give all nodes access
    /// to the payload posted by each node.  The result pointer will point to
    /// an object of type mpicollpayload.
    timestamp allreduce(int count, const mpitype &type,
                        const mpiop &op,
                        const mpicomm &comm,
                        const boost::shared_ptr<const payload>& content,
                        boost::shared_ptr<const mpicollpayload> &result);

    /// Reduce data and scatter results.
    timestamp reduce_scatter(const std::vector<int> &recvcnts,
                             const mpitype &type,
                             const mpiop &op,
                             const mpicomm &comm);

    /// Reduce data and scatter results.
    /// All nodes will get a pointer to all payload.
    timestamp reduce_scatter(const std::vector<int> &recvcnts,
                             const mpitype &type,
                             const mpiop &op,
                             const mpicomm &comm,
                             const boost::shared_ptr<const payload>& content,
                             boost::shared_ptr<const mpicollpayload> &result);

    /// Compute partial reductions on data.
    timestamp scan(int count, const mpitype &type,
                   const mpiop &op, const mpicomm &comm);

    /// Compute partial reductions on data and share data.
    /// Each node will have as a result its own payload and that of
    /// all lower-ranked nodes
    timestamp scan(int count, const mpitype &type,
                   const mpiop &op, const mpicomm &comm,
                   const boost::shared_ptr<const payload>& content,
                   boost::shared_ptr<const mpicollpayload> &result);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIAPI_H
