/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOLLECTIVE_H
#define SSTMAC_PROCESS_MPICOLLECTIVE_H

#include <sstmac/process/mpikernel.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpistatus.h>

namespace sstmac {

  /// Important types.
  class mpiqueue;
  class mpimagic;
  class collectivestrategy;
  class mpiqueue;
  class mpipayload;
  class mpicollpayload;

  /**
   * Base type for mpikernel objects that do collective operations.
   */
  class mpicollective : public mpikernel {
    /// Nested type to handle the actual send operations.
    class sender;
    /// Nested type to handle the actual receive operations.
    class recver;
    /// Nested type to handle funneled (one-at-a-time) copy operations.
    class funneledcopy;
    /// Nested type to handle concurrent (start-at-once) copy operations.
    class immediatecopy;

    /// This object.
    boost::weak_ptr<mpicollective> self_;
    /// The node we are running on.
    boost::shared_ptr<nodeinterface> node_;
    /// Copy object for funneled copies.
    boost::shared_ptr<funneledcopy> funnel_;
    /// Information on total bytes sent and received.
    boost::shared_ptr<mpistatus> status_;

    /// Our queue.
    boost::shared_ptr<mpiqueue> queue_;
    /// Our communicator.
    mpicomm comm_;

  protected:
    /// Hi.
    mpicollective(const boost::shared_ptr<processrequest>& when_complete,
                  const boost::shared_ptr<mpiqueue>& queue,
                  const mpicomm &comm);


    /// Who am I?
    void set_self(const boost::shared_ptr<mpicollective>& self);

    /// This is who I am.
    boost::shared_ptr<mpicollective> get_self();

    /// You still need to fill in the pure virtual method from kernel.
    using mpikernel::start;

    /*
     * Callbacks for the derived type to request send or receive operations.
     */

    /// Called by the derived type to request a send.
    void start_send(int count,
                    const mpitype &type,
                    const mpitag &tag,
                    const mpiid &dest,
                    const boost::shared_ptr<const payload>& content =
                         boost::shared_ptr<const payload>());

    /// Called by the derived type to request a receive.
    void start_recv(int count,
                    const mpitype &type,
                    const mpitag &tag,
                    const mpiid &dest);

    /// Called by the derived type to request a funneled copy operation
    /// (into a shared buffer).  Calls copy_complete when complete.
    void start_funneled_copy(int count,
                             const mpitype &type,
                             const mpitag &tag,
                             const mpiid &partner);

    /// Called by the derived type to request an immediate copy operation
    /// (not into a shared buffer).  Calls copy_complete when complete.
    void start_immediate_copy(int count,
                              const mpitype &type,
                              const mpitag &tag,
                              const mpiid &partner);

    /*
     * Callbacks that the derived type needs to provide to complete
     * a send or receive operation.
     */

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest) = 0;

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now, int count,
                               const boost::shared_ptr<const mpistatus>& stat) =0;

    /// Callback method to indicate that a copy operation has completed.
    virtual void copy_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &source) = 0;

  public:
    /// Goodbye.
    virtual ~mpicollective() throw();

    /// Complete.  Calls mpikernel::complete.
    void complete(const timestamp &now,
                  const boost::shared_ptr<const payload>& content)
    {
      status_->set_content(content);
      mpikernel::complete(now, status_);
    }

    /// If you override this method, you should remember to call
    /// mpicollective::start before you try any sends, receives, or copies.
    virtual void start(const timestamp &start_time,
                       const boost::shared_ptr<nodeinterface>& my_node);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPICOLLECTIVE_H
