/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOMM_H
#define SSTMAC_PROCESS_MPICOMM_H

#include <sstmac/process/mpiid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/process/mpicommid.h>
#include <sstmac/util/errors.h>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <vector>
#include <iosfwd>

#include <sstmac/process/mpicommdata.h>

namespace sstmac {

  // Forward declaration of important types.
  //class mpicommdata;

  /**
   * An MPI communicator handle.
   */
  class mpicomm {
#ifndef SWIG // Suppress a swig warning.
    friend std::ostream& operator<<(std::ostream &os, const mpicomm &comm);
#endif

    /// My rank in this communicator.
    mpiid rank_;
    /// The data (possibly shared) associated with this communicator.
    boost::shared_ptr<const mpicommdata> data_;

  public:
    /// This is the null communicator.
    static const mpicomm comm_null;

    /// Construct this as a null communicator.
    mpicomm() : rank_(-1), data_() {}
    
    /// Hello.
    mpicomm(const mpiid &rank, const boost::shared_ptr<const mpicommdata>& data) :
      rank_(rank), data_(data)
    {}

    /// Goodbye.
    virtual ~mpicomm() {
    }

    /// The rank of this peer in the communicator.
    const mpiid& rank() const {
      return rank_;
    }

    /// The size of the communicator.
    mpiid size() const {
      return mpiid(data_->peers().size());
    }

    /// The identifier for this communicator.
    /// To be used to tag messages to/from this communicator.
    const mpicommid& id() const {
      static const mpicommid invalid(-1);
      if(data_ != 0)
        return data_->id();
      else
        return invalid;
    }

    /// The task index of the caller.
    taskid my_task() const {
      if(data_ == 0)
        throw illformederror("mpicomm::my_task:  NULL communicator.");
      if(rank_.id < 0 || size_t(rank_.id) >= data_->peers().size())
        throw illformederror(std::string("mpicomm::my_task:  Invalid rank ")
                             + boost::lexical_cast<std::string>(rank_));
      return data_->peers().at(rank_.id);
    }

    /// The task index of the given peer.
    taskid peer_task(const mpiid &rank) const {
      return data_->peers().at(rank.id);
    }
    
    /// The list of tasks involved in this communicator.
    /// Indexing is done by mpiid::rank().id.
    const std::vector<taskid>& peers() const {
      return data_->peers();
    }

    nodeid my_node() const {
      return data_->nodes().at(rank_.id);
    }

    /// The list of nodes involved in this communicator.
    /// Indexing is done by mpiid::rank().id.
    const std::vector<nodeid>& nodes() const {
      return data_->nodes();
    }

    /// Equality comparison.
    inline bool operator==(const mpicomm &other) const {
      return ((rank_ == other.rank_) && (data_ == other.data_));
    }

    /// Inequality comparison.
    inline bool operator!=(const mpicomm &other) const {
      return !this->operator==(other);
    }
  };

  /// Allow communicators in sorted containers.
  inline bool operator<(const mpicomm &a, const mpicomm &b) {
    return (a.id() == b.id() ? a.rank() < b.rank() : a.id() < b.id());
  }

  /// Fairly self-explanatory.
  std::ostream& operator<<(std::ostream &os, const mpicomm &comm);

}

#endif // ! SSTMAC_PROCESS_MPICOMM_H
