/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpicommfactory-pending-keylist.h>
#include <sstmac/process/mpicommdata.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Base constructor.
  //
  mpicommfactory::pending::pending(const mpicommid &new_id,
                                   const boost::shared_ptr<scheduler>& sched) :
    key_(key::construct()), scheduler_(sched), id_(new_id)
  {}

  //
  // Block current caller.
  //
  void mpicommfactory::pending::block() {
      scheduler_->block(key_);
  }

  //
  // Unblock all callers.
  //
  void mpicommfactory::pending::unblock() {
    scheduler_->unblock(key_);
  }

  //
  // Construct a new comm_create handler.
  //
  mpicommfactory::pending_create::pending_create(const mpicommid &new_id,
                                                 const shared_ptr<scheduler>& sched) :
    pending(new_id, sched)
  {}
  
  //
  // Construct a new comm_create handler.
  //
  boost::shared_ptr<mpicommfactory::pending_create>
  mpicommfactory::pending_create::construct(const mpicommid &new_id,
                                            const shared_ptr<scheduler>& sched)
  {
    return shared_ptr<pending_create>(new pending_create(new_id, sched));
  }

  //
  // Add the given caller to the membership.
  //
  mpicomm mpicommfactory::pending_create::
  build(const mpicomm &caller, const std::vector<mpiid> &members)
  {
    const std::vector<taskid> &caller_peers = caller.peers();
    const std::vector<nodeid> &caller_nodes = caller.nodes();
    if(target_peers_.empty()) {
      // This is the first node to enter -- figure out index mappings.
      template_ = caller;
      // Initialize all nodes to be number -1.
      index_mapping_.resize(caller_peers.size(), mpiid(-1));
      // Figure out which nodes are active in the new group
      // and how their indices map.      
      for(size_t pos = 0; pos < members.size(); ++pos) {
        mpiid member = members.at(pos);
        const int id = member.id;
        if(index_mapping_.at(member.id).id >= 0) {
          // A node index is appearing more than once
          throw valueerror("mpicommfactory::pending::build:  "
                           "Node index appears more than once.");
        }
        else if(id < 0 || id >= int(caller_peers.size())) {
          throw valueerror("mpicommfactory::pending::build:  "
                           "Node index out of range.");
        }
        // Store the index mapping
        index_mapping_.at(id) = mpiid(pos);
        // Store the taskid values for the members we want to keep.
        target_peers_.push_back(caller_peers.at(id));
        target_nodes_.push_back(caller_nodes.at(id));
      }
      added_ = 0;
    }
    else {
      // The group is under construction.
      // Make sure this node is working from the same communicator
      // and requesting the same nodes as everybody else.
      if(caller.id() != template_.id()) {
        throw valueerror("mpicommfactory::pending::build:  Non-matching "
                         "communicator id.");
      }
      if(members.size() != target_peers_.size()) {
        /*
        std::cerr << "mpicommfactory::pending:  Error:  Group " << caller.id() << " under construction with " << target_peers_.size() << " nodes, but new entrant " << caller << " has specified " << members.size() << " neighbors\n";
        */
        throw valueerror("mpicommfactory::pending::build:  Request for a "
                         "different number of active (target) nodes.");
      }
      std::vector<mpiid>::const_iterator it, end = members.end();
      for(it = members.begin(); it != end; ++it) {
        taskid tid = caller_peers.at(it->id);
        std::vector<taskid>::const_iterator found =
          std::find(target_peers_.begin(), target_peers_.end(), tid);
        if(found == target_peers_.end()) {
          throw valueerror("mpicommfactory::pending::build:  Non-matching node "
                           "membership in request.");
        }
      }
    }
    ++added_;
    if(added_ < target_peers_.size()) {
      // This is not the last one in.
      this->block();
    }
    else {
      // Construct the mpicommdata object.
      commdata_ = mpicommdata::construct(id_, target_peers_, target_nodes_);
      //std::cerr << "mpicommfactory::pending_create::build:  Communicator id " << id_ << " now points to data object " << commdata_ << "\n";
      // Let everybody roam free.
      this->unblock();
    }
    // We get here when everybody is ready.
    return mpicomm(index_mapping_.at(caller.rank().id), commdata_);
  }

  //
  // Construct a new handler for comm_split.
  //
  mpicommfactory::pending_split::pending_split(const mpicommid &new_id,
                                               int maxsize,
                                               const shared_ptr<scheduler>& sched) :
    pending(new_id, sched), pending_(-1), maxsize_(maxsize), current_id_(new_id)
  {}

  //
  // Construct handler for comm_split.
  //
  shared_ptr<mpicommfactory::pending_split>
  mpicommfactory::pending_split::construct(const mpicommid &new_id,
                                           int maxsize,
                                           const shared_ptr<scheduler>& sched)
  {
    return shared_ptr<pending_split>(new pending_split(new_id, maxsize, sched));
  }

  //
  // Add the given caller to the membership.
  //
  mpicomm mpicommfactory::pending_split::build(const mpicomm &caller,
                                               int color, int key)
  {
    if(colors_.empty()) {
      // First one in.  Initialize stuff.
      template_ = caller;
      pending_ = caller.size().id;
    }
    else {
      if(caller.id() != template_.id())
        throw valueerror("mpicommfactory::pending_split:  "
                         "Not the correct input communicator.");
    }
    // If this is the first time we see this color, we need to assign
    // a communicator id.
    if(! colors_[color]) {
      if(current_id_.id >= this->id_.id + maxsize_)
        throw valueerror("mpicommfactory::pending_split::build:  "
                         "MPI communicator index exceeded allowed bounds.");
      //std::cerr << "mpicommfactory::pending_split::build:  Creating a new keylist with id " << current_id_ << " for color " << color << "\n";
      colors_[color] = keylist::construct(current_id_);
      current_id_.id++;
    }
    // Add this rank to the colored communicator.
    colors_[color]->insert(key, caller.my_task(), caller.my_node());
    // Decide whether we're done
    --pending_;
    if(pending_ > 0) {
      // Still waiting for others to join the collective.
      //std::cerr << "Peer joined comm_split.  Still " << pending_
      //        << " pending members\n";
      this->block();
    }
    else {
      //std::cerr << "Last peer joined comm_split.  Unblocking\n";
      // Done. Shared data gets constructed by the keylist when it's needed.
      // Set all blocked contexts free
      this->unblock();
    }
    // And now we are no longer context-switched out.
    // Note that if this peer has key less than zero, it is inactive
    // and gets comm_null as a return value.
    mpicomm retval = colors_[color]->get_mpicomm(key);
    if(debugger::active())
      debug() << "mpicommfactory::pending_split::build:  Returning "
              << retval << " for caller " << caller
              << " and key " << key << "\n";
    // All done.
    return retval;
  }
  
} // end of namespace sstmac
