/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpicommfactory.h>
#include <sstmac/process/mpicommfactory-pending.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/simulator/environment.h>
#include <sstmac/process/mpicommdata.h>
#include <sstmac/process/request.h>
#include <sstmac/util/counter.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iterator>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Build comm_world using information retrieved from the environment.
  //
  mpicommfactory::mpicommfactory() :
    mpirun_np_(0), next_id_(1)
  {}

  //
  // Who am I?
  //
  void mpicommfactory::set_self(const boost::shared_ptr<mpicommfactory>& self) {
    if(self.get() != this) {
      throw valueerror("mpicommfactory::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // Build a new factory.
  //
  boost::shared_ptr<mpicommfactory> mpicommfactory::construct() {
    shared_ptr<mpicommfactory> retval(new mpicommfactory);
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpicommfactory::~mpicommfactory() {
  }

  //
  // Initialize.
  //
  void mpicommfactory::init(const boost::shared_ptr<scheduler>& sched,
                            const boost::shared_ptr<const environment>& env,
                            int mpirun_np)
  {
    scheduler_ =  sched;
    mpirun_np_ = mpirun_np;
    next_id_.id = 1;
    // Build peer ids.
    int64_t target_peers = env->get_tasklist()->size();
    if(mpirun_np_ > 0 && mpirun_np_ < target_peers)
      target_peers = mpirun_np_;
    std::vector<taskid> peers;
    peers.reserve(target_peers);
    // Fill the vector with sequential numbers ffrom 0 to the target size.
    std::generate_n(std::back_inserter(peers), target_peers,
                    counter<taskid>(taskid(0)));
    // Collect node ids.
    std::vector<nodeid> nodes;
    nodes.reserve(peers.size());
    std::vector<taskid>::const_iterator it;
    for(it = peers.begin(); it != peers.end(); ++it) {
      nodes.push_back(env->get_tasklist()->node(*it));
    }
    // Build the world data.
    world_data_ = mpicommdata::construct(mpicommid(0), peers, nodes);
  }
  
  //
  // Get the world communicator.
  //
  const mpicomm& mpicommfactory::get_world(const mpiid &caller) {
    if(worldcomms_.find(caller) == worldcomms_.end()) {
      if(caller.id < 0 || caller.id >= int(world_data_->peers().size())) {
        throw valueerror("mpicommfactory::get_world:  caller id out of range.");
      }
      worldcomms_[caller] = mpicomm(caller, world_data_);
    }
    return worldcomms_[caller];
  }

  //
  // Get the self communicator.
  //
  const mpicomm& mpicommfactory::get_self(const mpiid &caller) {
    if(selfcomms_.find(caller) == selfcomms_.end()) {
      if(caller.id < 0 || caller.id >= int(world_data_->peers().size())) {
        throw valueerror("mpicommfactory::get_world:  caller id out of range.");
      }
      std::vector<taskid> peers(1);
      peers[0] = world_data_->peers().at(caller.id);
      std::vector<nodeid> nodes(1);
      nodes[0] = world_data_->nodes().at(caller.id);
      shared_ptr<mpicommdata> new_world =
        mpicommdata::construct(next_id_, peers, nodes);
      next_id_.id++;
      selfcomms_[caller] = mpicomm(mpiid(0), new_world);
    }
    return selfcomms_[caller];
  }
  
  //
  // Duplicate a communicator.
  //
  mpicomm mpicommfactory::comm_dup(const mpicomm &caller) {
    //std::cerr << "mpicommfactory::comm_dup(" << caller << ").  Input comm id=" << caller.id() << ", size=" << caller.size() << "\n";
    std::vector<mpiid> participants;
    participants.reserve(caller.peers().size());
    std::generate_n(std::back_inserter(participants), caller.peers().size(),
                    counter<mpiid>(mpiid(0)));
    return this->comm_create(caller, participants);
  }

  //
  // Make the given mpiid refer to a newly created communicator.
  //
  mpicomm mpicommfactory::comm_create(const mpicomm &caller,
                                      const std::vector<mpiid> &participants)
  {
    //std::cerr << "mpicommfactory::comm_create(" << caller << ", " << participants.size() << ") : id=" << caller.id() << "\n";
    mpicommid id = caller.id();
    if(under_construction_.find(id) == under_construction_.end()) {
      under_construction_[id] = pending_create::construct(next_id_, scheduler_);
      next_id_.id++;
    }
    shared_ptr<pending_create> pend =
      dynamic_pointer_cast<pending_create>(under_construction_[id]);
    mpicomm retval = pend->build(caller, participants);
    // When this call returns (unblocks), we erase the constructor
    // from the map (to make sure we can build more communicators from
    // the same input comm).
    //std::cerr << "mpicommfactory::comm_create(" << caller << ", " << participants.size() << "):  Group construction for " << id << " complete\n";
    under_construction_.erase(id);
    scheduler_->yield();
    return retval;
  }

  //
  // MPI_Comm_split.
  //
  mpicomm mpicommfactory::comm_split(const mpicomm &caller, int color, int key)
  {
    //if(debugger::active())
      debug() << "mpicommfactory::comm_split(" << caller << ", "
              << color << ", " << key << "):  id=" << caller.id() << "\n";
    mpicommid id = caller.id();
    if(under_construction_.find(id) == under_construction_.end()) {
      under_construction_[id] =
        pending_split::construct(next_id_, caller.size().id, scheduler_);
      // MPI_Comm_split may generate up to MPI_Comm_size new communicators.
      // Unfortunately, we don't know how many until after construction.
      next_id_.id += caller.size().id;
    }
    shared_ptr<pending_split> pend =
      dynamic_pointer_cast<pending_split>(under_construction_[id]);
    if(! pend)
      throw valueerror("mpicommfactory::comm_split:  Communicator under "
                       "construction is not compatible with comm_split");
    mpicomm retval = pend->build(caller, color, key);
    // When this call returns (unblocks), we erase the constructor
    // from the map (to make sure we can build more communicators from
    // the same input comm).
    /*
    std::cerr << "mpicommfactory::comm_split(" << caller << ", " << color 
              << ", " << key << "):  New communicator is " << retval 
              << " with size " << retval.size() << " and id " << retval.id()
              << "\n";
    */
    under_construction_.erase(id);
    scheduler_->yield();
    return retval;
  }

} // end of namespace sstmac
