/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOMMFACTORY_H
#define SSTMAC_PROCESS_MPICOMMFACTORY_H

#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpicommid.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <vector>
#include <set>
#include <map>

namespace sstmac {

  // Important types.
  class scheduler;
  class environment;

  /**
   * Construct mpi communicators.
   */
  class mpicommfactory : virtual public nocopy {
    /// The private type used to manage communicators under construction.
    class pending;
    class pending_create;
    class pending_split;

    /// This object.
    boost::weak_ptr<mpicommfactory> self_;

    /// The scheduler used to block and unblock tasks.
    boost::shared_ptr<scheduler> scheduler_;

    /// We can restrict our run to use fewer than the nodes allocated.
    int mpirun_np_;

    /// The shared data for the world communicator.
    boost::shared_ptr<mpicommdata> world_data_;

    /// The next available communicator index.
    mpicommid next_id_;

    /// Keyring(s) for pending comm_dup or comm_create requests.
    /// The signature for each pending request is determined by the
    /// input communicator (construction of communicators is global (blocking)
    /// across the communicator).
    typedef std::map< mpicommid, boost::shared_ptr<pending> > buildmap_t;
    buildmap_t under_construction_;

    /// The world communicators for all nodes.
    std::map<mpiid, mpicomm> worldcomms_;

    /// The self communicators for all nodes.
    std::map<mpiid, mpicomm> selfcomms_;

  protected:
    /// Build comm_world using information retrieved from the environment.
    mpicommfactory();

    /// Who am I?
    void set_self(const boost::shared_ptr<mpicommfactory>& self);

  public:
    /// Build a new factory.
    static boost::shared_ptr<mpicommfactory> construct();

    /// Goodbye.
    virtual ~mpicommfactory();

    /// Initialize the object.
    void init(const boost::shared_ptr<scheduler>& scheduler,
              const boost::shared_ptr<const environment>& env,
              int mpirun_np);

    /// Get the world communicator for the given node.
    /// In this communicator, mpiid indices are the same as the taskid
    /// indices given by the environment.  Each node will only have one world.
    const mpicomm& get_world(const mpiid &caller);

    /// Get a 'self' communicator.  Each node will have a unique index
    /// for its self.  Each node will only have one self.
    const mpicomm& get_self(const mpiid &caller);

    /// Duplicate the given communicator.
    /// Blocks unti lall nodes have entered the call.
    /// Eventually we may opt to deal with the extended attributes (keyval)
    /// stuff in the MPI standard, even though I don't know anybody
    /// who actually uses it.
    mpicomm comm_dup(const mpicomm &caller);

    /// Make the given mpiid refer to a newly created communicator.
    /// Blocks the caller until all nodes have entered the call.
    /// This call must be performed by all nodes in the caller communicator
    /// a creating any new communicators.  Returns mpicomm::null_comm
    /// on all nodes that are not participants the new group.
    mpicomm comm_create(const mpicomm &caller,
                        const std::vector<mpiid> &participants);

    /// MPI_Comm_split -- collective operation.
    mpicomm comm_split(const mpicomm &caller, int color, int key);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPICOMMFACTORY_H
