/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpidirectalltoall.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpidirectalltoall::mpidirectalltoall(const shared_ptr<processrequest>& thekey,
                                       const shared_ptr<mpiqueue>& queue,
                                       const std::vector<int> &sendcnt,
                                       const mpitype &sendtype,
                                       const std::vector<int> &recvcnt,
                                       const mpitype &recvtype,
                                       const mpitag &tag,
                                       const mpicomm &comm,
                                       const shared_ptr<const payload>& content,
                                       const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm), sendcnt_(sendcnt), sendtype_(sendtype),
    recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag), comm_(comm),
    magic_(magic), max_concurrent_ops_(5)
  {
    if(sendcnt_.size() != size_t(comm_.size().id) ||
       recvcnt_.size() != size_t(comm_.size().id))
      throw valueerror("mpidirectalltoall(...):  Length of send- and/or recv- "
                       "arrays is not the same as communicator size.");
    // Clone the content if it is non-null.
    magic_->add_collective_payload(typeid(*this), tag_, comm_,
                                   (content ? content->clone() : content),
                                   true);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpidirectalltoall::send_complete(const timestamp &now,
                                        int count,
                                        const mpitype &type,
                                        const mpitag &tag,
                                        const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpidirectalltoall[" << this << "]::send_complete( " << now
              << ", " << count << ", " << type << ", " << tag << ", "
              << dest << ")\n";
    if(! pendsend_.empty()) {
      pendsend snd = pendsend_.back();
      pendsend_.pop_back();
      this->start_send(snd.cnt, sendtype_, tag_, snd.dest);
    }
    else {
      --pending_sends_;
      if((pending_sends_ == 0) && (pending_recvs_ == 0)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpidirectalltoall::send_complete:  Transfers "
                                 "done but not all payload has been gathered.");
        shared_ptr<const payload> content =
          magic_->get_collective_payload(typeid(*this), tag_, comm_);
        mpicollective::complete(now, content);
      }
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpidirectalltoall::recv_complete(const timestamp &now,
                                        int count,
                                        const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpidirectalltoall[" << this << "]::recv_complete( " << now
              << ", " << count << ", " << stat << ")\n";
    this->start_immediate_copy(count, stat->recv_type(),
                               stat->tag(), stat->source());
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void mpidirectalltoall::copy_complete(const timestamp &now,
                                        int count,
                                        const mpitype &type,
                                        const mpitag &tag,
                                        const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpidirectalltoall[" << this << "]::copy_complete( " << now
              << ", " << count << ", " << type << ", " << tag << ", "
              << source << ") -- pending recvs: " << pending_recvs_ << "\n";
    if(! pendrecv_.empty()) {
      pendrecv rcv = pendrecv_.back();
      pendrecv_.pop_back();
      this->start_recv(rcv.cnt, recvtype_, tag_, rcv.source);
    }
    else {
      --pending_recvs_;
      if((pending_sends_ == 0) && (pending_recvs_ == 0)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpidirectalltoall::copy_complete:  Transfers "
                               "done but not all payload has been gathered.");
        shared_ptr<const payload> content =
          magic_->get_collective_payload(typeid(*this), tag_, comm_);
        mpicollective::complete(now, content);
      }
    }
  }

  //
  // Hello.
  //
  shared_ptr<mpidirectalltoall>
  mpidirectalltoall::construct(const shared_ptr<processrequest>& thekey,
                               const shared_ptr<mpiqueue>& queue,
                               const std::vector<int> &sendcnt,
                               const mpitype &sendtype,
                               const std::vector<int> &recvcnt,
                               const mpitype &recvtype,
                               const mpitag &tag,
                               const mpicomm &comm,
                               const shared_ptr<const payload>& content,
                               const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpidirectalltoall> retval
      (new mpidirectalltoall(thekey, queue, sendcnt, sendtype,
                             recvcnt, recvtype, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectalltoall::~mpidirectalltoall() throw() {
  }

  //
  // Set the maximum number of concrurrent sends.
  //
  void mpidirectalltoall::set_max_concurrent_ops(int maxval) {
    max_concurrent_ops_ = maxval;
  }

  // Utility routine.
  template <typename T>
  T min(T a, T b) {
    return (a < b ? a : b);
  }

  //
  // Start this kernel.
  //
  void mpidirectalltoall::start(const timestamp &now,
                                const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpidirectalltoall[" << this << "]::start( " << now
              << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    // Prepare to fire off all my sends and receives.
    mpiid rank = comm_.rank();
    mpiid size = comm_.size();
    // Catch the case where comm_.size() was 1.
    if(size.id <= 1) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpidirectalltoall::start:  Transfers "
                             "done but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
    else {
      pending_sends_ = pending_recvs_ = 
        (max_concurrent_ops_ > 0 ?
         min(max_concurrent_ops_, size.id-1) :
         size.id - 1);
      // Stagger the sends so not everybody hammers node 0 at the same time.
      for(int relrank = 1; relrank < size.id; ++relrank) {
        mpiid peer((size.id + rank.id - relrank) % size.id);
        if(relrank <= pending_sends_)
          this->start_send(sendcnt_.at(peer.id), sendtype_, tag_, peer);
        else
          pendsend_.push_back(pendsend(sendcnt_.at(peer.id), peer));
      }
      // Stagger the receives in the opposite order from the sends
      // to avoid deadlock.
      for(int relrank = 1; relrank <= pending_recvs_; ++relrank){
        mpiid peer((size.id + rank.id - relrank) % size.id);
        if(relrank <= pending_sends_)
          this->start_recv(sendcnt_.at(peer.id), sendtype_, tag_, peer);
        else
          pendrecv_.push_back(pendrecv(recvcnt_.at(peer.id), peer));
      }
    }
  }

} // end of namespace sstmac
