/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpidirectgatherv.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <deque>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Hi.
  //
  mpidirectgatherv::mpidirectgatherv(const shared_ptr<processrequest>& thekey,
                                     const shared_ptr<mpiqueue>& queue,
                                     int sendcount, const mpitype &sendtype,
                                     const std::vector<int> &recvcount,
                                     const mpitype &recvtype,
                                     const mpiid &root, const mpitag &tag,
                                     const mpicomm &comm,
                                     const shared_ptr<const payload>& content,
                                     const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm),
    sendcount_(sendcount), sendtype_(sendtype),
    recvcount_(recvcount), recvtype_(recvtype),
    root_(root), tag_(tag), comm_(comm), magic_(magic),
    pending_sends_(0), pending_recvs_(0)
  {
    bool is_root = comm_.rank() == root_;
    // Clone the content if it is non-null.
    magic_->add_collective_payload(typeid(*this), tag, comm,
                                   (content ? content->clone() : content),
                                   is_root);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpidirectgatherv::send_complete(const timestamp &now, int count,
                                       const mpitype &type, const mpitag &tag,
                                       const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpidirectgatherv::send_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_sends_) {
      shared_ptr<const payload> content;
      if((comm_.rank() == root_)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpidirectgatherv::send_complete:  Root is done "
                               "but not all payload has been gathered.");
        content = magic_->get_collective_payload(typeid(*this), tag_, comm_);
      }
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpidirectgatherv::recv_complete(const timestamp &now, int count,
                                       const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpidirectgatherv::recv_complete(" << now << ", " << count
              << ", " << stat << ")\n";
    this->start_funneled_copy(count, stat->recv_type(),
                              stat->tag(), stat->source());
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void mpidirectgatherv::copy_complete(const timestamp &now, int count,
                                       const mpitype &type, const mpitag &tag,
                                       const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpidirectgatherv::copy_done(" << now << ", " << count << ", "
              << type << ", " << tag << ", " << source << ")\n";
    --pending_recvs_;
    if(! pending_recvs_) {
      shared_ptr<const payload> content;
      if((comm_.rank() == root_)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpidirectgatherv::copy_complete:  Root is done "
                               "but not all payload has been gathered.");
        content = magic_->get_collective_payload(typeid(*this), tag_, comm_);
      }
      mpicollective::complete(now, content);
    }
  }

  //
  // Hello.
  //
  shared_ptr<mpidirectgatherv>
  mpidirectgatherv::construct(const shared_ptr<processrequest>& thekey,
                              const shared_ptr<mpiqueue>& queue,
                              int sendcount, const mpitype &sendtype,
                              const std::vector<int> &recvcount,
                              const mpitype &recvtype,
                              const mpiid &root, const mpitag &tag,
                              const mpicomm &comm,
                              const shared_ptr<const payload>& content,
                              const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpidirectgatherv> retval
      (new mpidirectgatherv(thekey, queue, sendcount, sendtype, recvcount,
                            recvtype, root, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectgatherv::~mpidirectgatherv() throw() {
  }

  //
  // Start this kernel.
  //
  void mpidirectgatherv::start(const timestamp &now,
                                const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpidirectgatherv::start(" << now << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    if(comm_.rank() == root_) {
      // Root node -- fire off a bunch of receives.
      mpiid size = comm_.size();
      pending_recvs_ = comm_.size().id - 1;
      if(pending_recvs_ == 0) {
        shared_ptr<const payload> content;
        if((comm_.rank() == root_)) {
          if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
            throw illformederror("mpidirectgatherv::start:  Root is done "
                                 "but not all payload has been gathered.");
          content = magic_->get_collective_payload(typeid(*this), tag_, comm_);
        }
        mpicollective::complete(now, content);
      }
      else {
        for(mpiid dest(0); dest < size; ++dest) {
          if(dest != root_) {
            this->start_recv(recvcount_.at(dest.id), recvtype_, tag_, dest);
          }
        }
      }
    }
    else {
      // Client node -- fire off a single send.
      pending_sends_ = 1;
      this->start_send(sendcount_, sendtype_, tag_, root_);
    }
  }

} // end of namespace sstmac
