/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIDIRECTGATHERV_H
#define SSTMAC_PROCESS_MPIDIRECTGATHERV_H

#include <sstmac/process/mpicollective.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpicomm.h>
#include <vector>

namespace sstmac {

  // Important types.
  class mpiqueue;
  class mpistatus;
  class mpimagic;

  /**
   * Perform an MPI gatherv using direct sends to the root node.
   *
   * This is the way mpich2 does gatherv.  The problem is that the
   * mpi standard specifies that the array of send lengths only needs
   * to be valid at root.  This means that that you either need to send
   * the length array separately (using a gather, bcast, or whatever)
   * or you just bite the bullet (like here) and send the data all
   * directly from the root node.
   */
  class mpidirectgatherv : public mpicollective {
    /// The number of elements of type sendtype sent to root node.
    int sendcount_;
    /// The send type
    mpitype sendtype_;
    /// The number of elements of type recvtype received from each node.
    std::vector<int> recvcount_;
    /// The receive type
    mpitype recvtype_;
    /// The root node.
    mpiid root_;
    /// The tag associated with our communications.
    mpitag tag_;
    /// Store a copy of the communicator.
    mpicomm comm_;
    /// The magic bypass for handling data gathering.
    boost::shared_ptr<mpimagic> magic_;
    /// Number of pending sends (only on clients).
    int pending_sends_;
    /// Number of pending receives (only on root).
    int pending_recvs_;

  protected:
    /// Hi.
    mpidirectgatherv(const boost::shared_ptr<processrequest>& thekey,
                     const boost::shared_ptr<mpiqueue>& queue,
                     int sendcount,const mpitype &sendtype,
                     const std::vector<int> &recvcount, const mpitype &recvtype,
                     const mpiid &root, const mpitag &tag,
                     const mpicomm &comm,
                     const boost::shared_ptr<const payload>& content,
                     const boost::shared_ptr<mpimagic>& magic);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now,
                               int count,
                               const boost::shared_ptr<const mpistatus>& stat);

    /// Callback to signal copy completion (only on root).
    void copy_complete(const timestamp &now,
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    /// Hello.
    static boost::shared_ptr<mpidirectgatherv>
    construct(const boost::shared_ptr<processrequest>& thekey,
              const boost::shared_ptr<mpiqueue>& queue,
              int sendcount, const mpitype &sendtype,
              const std::vector<int> &recvcount, const mpitype &recvtype,
              const mpiid &root, const mpitag &tag,
              const mpicomm &comm, const boost::shared_ptr<const payload>& content,
              const boost::shared_ptr<mpimagic>& magic);

    /// Goodbye.
    virtual ~mpidirectgatherv() throw();

    /// Start this kernel.
    virtual void start(const timestamp &now,
                       const boost::shared_ptr<nodeinterface>& my_node);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPIDIRECTGATHERV_H
