/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIDIRECTSCATTERV_H
#define SSTMAC_PROCESS_MPIDIRECTSCATTERV_H

#include <sstmac/process/mpicollective.h>
#include <vector>

namespace sstmac {

  // Important types.
  class mpiqueue;
  class mpistatus;

  /**
   * Perform an MPI scatterv using direct sends from the root node.
   *
   * This is the way mpich2 does scatterv.  The problem is that the
   * mpi standard specifies that the array of send lengths only needs
   * to be valid at root.  This means that that you either need to send
   * the length array separately (using a scatter, bcast, or whatever)
   * or you just bite the bullet (like here) and send the data all
   * directly from the root node.
   */
  class mpidirectscatterv : public mpicollective {
    /// The number of elements of type sendtype sent to each node.
    std::vector<int> sendcount_;
    /// The send type
    mpitype sendtype_;
    /// The number of elements of type recvtype received by each node.
    int recvcount_;
    /// The receive type
    mpitype recvtype_;
    /// The root node.
    mpiid root_;
    /// The tag associated with our communications.
    mpitag tag_;
    /// Store a copy of the communicator.
    mpicomm comm_;
    /// The content we send and receive.
    boost::shared_ptr<mpicollpayload> content_;
    /// Number of pending sends (only on root).
    int pending_sends_;
    /// Number of pending receives (only on clients).
    int pending_recvs_;

  protected:
    /// Hi.
    mpidirectscatterv(const boost::shared_ptr<processrequest>& thekey,
                      const boost::shared_ptr<mpiqueue>& queue,
                      const std::vector<int> &sendcount,const mpitype &sendtype,
                      int recvcount, const mpitype &recvtype,
                      const mpiid &root, const mpitag &tag,
                      const mpicomm &comm,
                      const boost::shared_ptr<const mpicollpayload>& content);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now,
                               int count,
                               const boost::shared_ptr<const mpistatus>& stat);

    /// Callback to signal copy completion (only on clients).
    void copy_complete(const timestamp &now,
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    /// Hello.
    static boost::shared_ptr<mpidirectscatterv>
    construct(const boost::shared_ptr<processrequest>& thekey,
              const boost::shared_ptr<mpiqueue>& queue,
              const std::vector<int> &sendcount, const mpitype &sendtype,
              int recvcount, const mpitype &recvtype,
              const mpiid &root, const mpitag &tag,
              const mpicomm &comm,
              const boost::shared_ptr<const mpicollpayload>& content);

    /// Goodbye.
    virtual ~mpidirectscatterv() throw();

    /// Start this kernel.
    virtual void start(const timestamp &now,
                       const boost::shared_ptr<nodeinterface>& my_node);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPIDIRECTSCATTERV_H
