/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIKERNEL_H
#define SSTMAC_PROCESS_MPIKERNEL_H

#include <sstmac/simulator/kernel.h>
#include <sstmac/simulator/taskid.h>

namespace sstmac {

  // Forward declaration of important types.
  class mpistatus;
  class processrequest;
  class key;

  /**
   * A kernel for mpi communications.  Adds a way to get an request
   * handle for this kernel -- the request shall be notified when the
   * local chain of operations prescribed by this kernel is complete.
   *
   * The request is responsible for blocking and unblocking
   * individual tasks (mpiapp objects) in collaboration with the
   * scheduler; the mpikernel doesn't have to worry about that stuff.
   */
  class mpikernel : public kernel {
    /// This request will be notified when we're done.
    boost::shared_ptr<processrequest> when_done_;

  protected:
    /// Constructor.  Build a kernel around the given request.
    mpikernel(const boost::shared_ptr<processrequest> &req);

  public:
    /// Virtual destructor is essential.
    virtual ~mpikernel() throw() {}

    /// You still need to fill in the pure virtual method from kernel.
    using kernel::start;

    /// Get a request for this kernel.
    boost::shared_ptr<processrequest> get_request();

    /// kernel has a different signature for complete which we would like
    /// to overload (rather than hide).
    using kernel::complete;

  protected:
    /// Mark this kernel as complete.
    /// Notifies the request object prior to calling kernel::complete.
    void complete(const timestamp &current_time,
                  const boost::shared_ptr<mpistatus>& stat);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIKERNEL_H
