/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpimagic-payloaddata.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/util/errors.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  using boost::shared_ptr;
  using boost::lexical_cast;

  //
  // Initialize a new object.
  //
  mpimagic::payloaddata::payloaddata() :
    target_size_(-1)
  {}

  //
  // Get a new reference counted object.
  //
  shared_ptr<mpimagic::payloaddata> mpimagic::payloaddata::construct() {
    return shared_ptr<payloaddata>(new payloaddata);
  }

  //
  // Clean up.
  //
  mpimagic::payloaddata::~payloaddata() {
  }

  //
  // Add a new member to the collection.
  //
  void mpimagic::payloaddata::add(mpicomm comm, const shared_ptr<const payload>& data,
                                  bool consumer)
  {
    static const std::string here("mpimagic::payloaddata::add");
    //std::cerr << here << "(" << comm << ", " << data << ", "
    //          << consumer << ")\n";
    if(result_ != 0 || present_.find(comm.rank()) != present_.end())
      throw valueerror(here + ":  Attempt to redefine payload for " +
                       lexical_cast<std::string>(comm.rank()));
    if(target_size_ < 0) target_size_ = comm.size().id;
    else if(comm.size().id != target_size_)
      throw illformederror(here + ":  Incorrect communicator size.");
    present_[comm.rank()] = data;
    //std::cerr << "  " << here << ":  target size == " << target_size_
    //          << ", present size == " << present_.size() << "\n";
    if(consumer)
      retrieve_.insert(comm.rank());
    if(present_.size() == size_t(target_size_)) {
      result_ = mpicollpayload::construct(target_size_);
      present_t::iterator it, end = present_.end();
      for(it = present_.begin(); it != end; ++it)
        result_->set(it->first, it->second);
    }
  }

  //
  // Retrieve the collected data for the given member.
  //
  shared_ptr<mpicollpayload>  mpimagic::payloaddata::retrieve(mpicomm comm) {
    static const std::string here("mpimagic::payloaddata::retrieve");
    if(result_ == 0)
      throw illformederror(here + ":  Payload container is not " +
                           "fully populated.");
    retrieve_t::iterator it = retrieve_.find(comm.rank());
    if(it == retrieve_.end())
      throw illformederror(here + ":  Payload either not requested or "
                           "has already been retrieved for " +
                           lexical_cast<std::string>(comm.rank()));
    retrieve_.erase(it);
    return result_;
  }

  //
  // Test whether the given members has defined its payload.
  //
  bool mpimagic::payloaddata::rank_populated(mpiid rank) const {
    return (present_.find(rank) != present_.end());
  }

  //
  // Test whether all members have defined their payload.
  //
  bool mpimagic::payloaddata::fully_populated() const {
    //std::cerr << "payloaddata::fully_populated():  present size == "
    //          << present_.size() << ", target size == " << target_size_
    //          << "\n";
    return (present_.size() == size_t(target_size_));
  }

  //
  // Test whether all members have retrieved their data.
  //
  bool mpimagic::payloaddata::fully_retrieved() const {
    //std::cerr << "payloaddata::fully_retrieved():  result_ == "
    //          << result_ << ", retrieve_.size() == " << retrieve_.size()
    //          << "\n";
    return ((result_ != 0) && (retrieve_.empty()));
  }

} // end of namespace sstmac
