/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIQUEUE_H
#define SSTMAC_PROCESS_MPIQUEUE_H

#include <sstmac/process/mpimessage.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/payload.h>
#include <sstmac/process/messageid.h>
#include <sstmac/util/nocopy.h>
#include <list>
#include <map>
#include <set>

namespace sstmac {

  // Important types.
  class mpistatus;
  class nodeinterface;
  class networkinterface;
  class mpiserver;
  class mpisend;
  class mpirecv;
  class mpiprobe;
  class payload;
  class eventhandler;

  /**
   * A one-per-peer queue that handles inbound and outbound messages.
   * 
   * TODO:  Fix the server to re-arrange messages that arrive out-of-order.
   */
  class mpiqueue : virtual public nocopy {
    /// The mpiserver is allowed to access the private incoming_message method.
    friend class mpiserver;

    /// Nested types.
    class recvrequest;
    class sendrequest;
    class proberequest;
    
    /// Utility class to sort arriving messages in arrival order.
    struct sortbyid {
      bool operator()(const boost::shared_ptr<mpimessage> &a,
                      const boost::shared_ptr<mpimessage> &b) const;
    };
    
    /// This object.
    boost::weak_ptr<mpiqueue> self_;

    /// The shared backend (currently likely to result in a memory leak).
    boost::shared_ptr<mpiserver> server_;

    /// The node.
    boost::shared_ptr<nodeinterface> node_;

    /// The cutoff at which we switch between direct (ready send)
    /// and handshake.
    int64_t handsize_;
    
    /// Extent of envelope information.
    int64_t envelope_;

    /// Minimum transmission size.
    int64_t mintrans_;

    /// The sequence number for our next outbound transmission.
    std::map<taskid, int> next_outbound_;

    /// The sequence number expected for our next inbound transmission.
    std::map<taskid, int> next_inbound_;

    /// Hold messages that arrived out of order.
    typedef std::set< boost::shared_ptr<mpimessage>, sortbyid > hold_list_t;
    std::map<taskid, hold_list_t> held_;

    /// The (locally unique) id that will be given to the next message.
    messageid next_id_;

    /// Receive requests waiting for a matching message (or envelope).
    typedef std::list< boost::shared_ptr<recvrequest> > pending_message_t;
    pending_message_t pending_message_;

    /// Inbound messages waiting for a matching receive request.
    typedef std::list< boost::shared_ptr<mpimessage> > need_recv_t;
    need_recv_t need_recv_;

    /// Inbound messages that arrived out-of-order and are waiting for
    /// the message sequence to be complete.
    typedef std::map< int, boost::shared_ptr<mpimessage> > reorderlist_t;
    reorderlist_t reorderlist_;

    /// Requests waiting for an ack.
    typedef std::map< messageid, boost::shared_ptr<sendrequest> > ack_needed_t;
    ack_needed_t ack_needed_;

    /// Requests that sent an ack and are waiting for the rest of their data.
    typedef std::map< messageid, boost::shared_ptr<recvrequest> > ack_sent_t;
    ack_sent_t ack_sent_;

    /// Probe requests watching for a given envelope.
    typedef std::list< boost::shared_ptr<proberequest> > probelist_t;
    probelist_t probelist_;

  protected:
    /// Hi there.
    mpiqueue(const boost::shared_ptr<mpiserver>& server,
             const boost::shared_ptr<nodeinterface>& node,
             int64_t handshake_cutoff,
             int64_t envelope_bytes,
             int64_t smallest_transmission_unit);

    /// Who am I?
    void set_self(const boost::shared_ptr<mpiqueue>& self);

    /// This is who I am.
    boost::shared_ptr<mpiqueue> get_self() {
      return self_.lock();
    }

  public:
    /// \param server:   The MPI server through which we route messages.
    /// \param node:     The node to which we post copy and compute requests.
    /// \param envelope: Number of bytes prefixed to each MPI message
    /// \param mintrans: Smallest MPI message.
    static boost::shared_ptr<mpiqueue>
    construct(const boost::shared_ptr<mpiserver>& server,
              const boost::shared_ptr<nodeinterface>& node,
              int64_t handshake_cutoff = 1<<16,
              int64_t envelope = 24,
              int64_t mintrans = 0);

    /// Goodbye.
    virtual ~mpiqueue() throw();

    /// Send data.  All sends are triggered at time nodeinterface::now().
    /// Upon completion, the completion object have an updated
    /// status object relating to the send operation.
    void send(int count, const mpitype &type, const mpiid &dest,
              const mpitag &tag, const mpicomm &comm,
              bool rsend, bool ssend,
              mpimessage::category message_type,
              const boost::shared_ptr<mpisend>& completion,
              const boost::shared_ptr<const payload>& content = payload::null());

    /// Receive data.  All receives are triggered at time nodeinterface::now().
    /// Upon completion, the receiver object will be updated to
    /// hold a copy of the inbound mpimessage with updated status information
    /// including any inbound content (null or not).
    void recv(int count, const mpitype &type, const mpiid &source,
              const mpitag &tag, const mpicomm &comm,
              mpimessage::category message_type,
              const boost::shared_ptr<mpirecv>& completion);

    /// Ask for a notification when a message with the given signature arrives.
    /// Probing is only enabled for message_type==mpimessage::user.
    /// Upon completion, the status object held by the completion object
    /// will be updated to reflect the envelope information for the incoming
    /// message, and the completion object will be notified 
    /// (using eventhandler::complete).
    void probe(const mpicomm &comm, const mpiid &source, const mpitag &tag,
               const boost::shared_ptr<mpiprobe>& completion);
    
    /// Immediate-mode probe for whether a message with the given signature
    /// has arrived.
    /// Probing is only enabled for message_type==mpimessage::user.
    /// \return null if no such message is pending.
    boost::shared_ptr<mpistatus>
    iprobe(const mpicomm &comm, const mpiid &source, const mpitag &tag);

    /// Cancel all pending operations associated with the given event handler.
    /// Work in progress.
    void cancel(const boost::shared_ptr<eventhandler>& handler);


    /// CALLBACK used by mpiserver when this object has a message.
    void incoming_message(const timestamp &current_time,
                          const boost::shared_ptr<mpimessage>& message);
  protected:
    /// The incoming_message method forwards to here when the message is
    /// a response to a handshake request.
    void incoming_ack_response(const timestamp &current_time,
                               const boost::shared_ptr<mpimessage> &message);

    /// The incoming_message method forwards to here when the message is
    /// data delivered following a handshake request this queue ack-ed.
    void incoming_data_payload(const timestamp &current_time,
                               const boost::shared_ptr<mpimessage> &message);

    /// The incoming_message method forwards to here when the message is
    /// a new message, whehter it's an eager send or a new handshake request.
    void incoming_new_message(const timestamp &current_time,
                              const boost::shared_ptr<mpimessage> &message);

    /// The incoming_new_message method calls here to complete an
    /// inbound message.  At this point, arrival ordering has been ensured.
    void handle_new_message(const timestamp &current_time,
                            const boost::shared_ptr<mpimessage> &message);

    /// Used when a receive operation sends an ack (either ready-to-receive
    /// or message received).
    /// The ack is keyed to the sequence number associated with the mpimessage.
    void send_ack(const boost::shared_ptr<mpimessage>& message,
                  const boost::shared_ptr<recvrequest>& operation);

    /// Used when a send operation requests an ack for the given message.
    /// The ack is keyed to the sequence number associated with the mpimessage.
    /// When an ack arrives for this message_id, the sendrequest::ack
    /// method will be called.
    void request_ack(const boost::shared_ptr<mpimessage>& message,
                     const boost::shared_ptr<sendrequest>& operation);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIQUEUE_H
