/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIRABENSEIFNER_H
#define SSTMAC_PROCESS_MPIRABENSEIFNER_H

#include <sstmac/process/mpicollective.h>
#include <sstmac/process/mpiop.h>
#include <deque>

namespace sstmac {

  /**
   * Rolf Rabenseifner's MPI reduce algorithm.
   *
   * Re-implemented from https://fs.hlrs.de/projects/par/mpi//myreduce.c
   * Copyright: Rolf Rabenseifner, 1997
   *            Computing Center University of Stuttgart
   *            rabenseifner@rus.uni-stuttgart.de 
   *
   * Which has the license terms:
   *   The usage of this software is free, 
   *   but this header must not be removed.
   */
  class mpirabenseifner : public mpicollective {
    /// We pre-compute all the sends, receives, and reduces that
    /// will have to be done on our data.
    /// If either send_count or recv_count is <= 0, then send or receive
    /// is skipped in the current cycle.
    struct sendrecv {
      mpiid partner;
      int send_count, recv_count;
      bool recv_reduce;
      sendrecv();
      sendrecv(mpiid partner, int sendcnt, int recvcnt, bool reduce);
    };
    /// This will be the order of event.  Push back when populating,
    /// pop front when dispatching.
    std::deque<sendrecv> queue_;

    /// Flag indicating whether we're doing an allreduce or a reduce.
    bool allreduce_;
    /// Number of elements in our complete data array.
    int count_;
    /// The datatype we're sending and receiving.
    mpitype type_;
    /// The operation we're applying to our data.
    mpiop op_;
    /// The root node -- only used if allreduce_ is false.
    mpiid root_;
    /// The tag used for our sends and receives.
    mpitag tag_;
    /// The communicator we're using.
    mpicomm comm_;
    /// A magic shortcut to allow payload objects to be shared across all nodes
    boost::shared_ptr<mpimagic> magic_;
    /// We can have any number of pending sends, but the collective op
    /// is not over until all the sends have completed.
    int pending_sends_;
    /// We will not dispatch the next operation (send or receive) until
    /// a pending receive is retired.  There should only be zero or one.
    int pending_recvs_;
    /// We have a one-deep stack telling us whether we need to reduce
    /// the results from the next receive or not.
    bool reduce_this_recv_;
    /// Keep track of whether this node will report back reduced data or not.
    bool get_result_;
    bool got_result_;

    /// Private method that handles the actual send/receive exchanges.
    void do_sendrecv(const timestamp &now);
    
  protected:
    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now, int count,
                               const boost::shared_ptr<const mpistatus>& stat);

    /// Callback method to indicate that a copy operation has completed.
    virtual void copy_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &source);

    /// Hello.
    mpirabenseifner(bool allreduce,
                    const boost::shared_ptr<processrequest>& thekey,
                    const boost::shared_ptr<mpiqueue>& queue, int count,
                    const mpitype &datatype, const mpiop &op, const mpiid &root,
                    const mpitag &tag, const mpicomm &comm,
                    const boost::shared_ptr<const payload>& content,
                    const boost::shared_ptr<mpimagic>& magic);

  public:
    /// Hi.
    static boost::shared_ptr<mpirabenseifner>
    construct(bool allreduce, const boost::shared_ptr<processrequest>& thekey,
              const boost::shared_ptr<mpiqueue>& queue, int count,
              const mpitype &datatype, const mpiop &op, const mpiid &root,
              const mpitag &tag, const mpicomm &comm,
              const boost::shared_ptr<const payload>& content,
              const boost::shared_ptr<mpimagic>& magic);

    /// Goodbye.
    virtual ~mpirabenseifner() throw();

    /// Get busy child.
    virtual void start(const timestamp &now,
                       const boost::shared_ptr<nodeinterface>& my_node);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIRABENSEIFNER_H
