/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpirecvrequest.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/process/mpimessage.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hello.
  mpirecvrequest::mpirecvrequest(mpimessage::category cat,
                                 const mpicommid &commid,
                                 const mpiid &source, const mpitag &tag, 
                                 const mpitype &type, int64_t bytes,
                                 const boost::shared_ptr<eventhandler>& finally) :
    category_(cat), commid_(commid), source_(source), tag_(tag),
    type_(type), bytes_(bytes), finally_(finally)
  {}

  //
  // Who am I?
  //
  void mpirecvrequest::set_self(const boost::shared_ptr<mpirecvrequest>& self) {
    if(self.get() != this) {
      throw valueerror("mpirecvrequest::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<mpirecvrequest> mpirecvrequest::get_self() {
    return self_.lock();
  }

  //
  // Give me an object.
  //
  boost::shared_ptr<mpirecvrequest>
  mpirecvrequest::construct(mpimessage::category cat,  const mpicommid &commid,
                            const mpiid &source, const mpitag &tag, 
                            const mpitype &type, int64_t bytes,
                            const boost::shared_ptr<eventhandler>& finally)
  {
    shared_ptr<mpirecvrequest> retval
      (new mpirecvrequest(cat, commid, source, tag, type, bytes, finally));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpirecvrequest::~mpirecvrequest() throw() {
    if(finally_ != 0) {
      // We're not done.
      std::cerr << "WARNING:  mpirecvrequest going out of scope with "
                << "pending event handler.\n";
      finally_->handle(timestamp(-1));
    }
  }

  //
  // Return true if this request has been fulfilled.
  //
  bool mpirecvrequest::is_complete() const {
    return (data_ != 0);
  }

  //
  // Get the data received by this request.
  //
  boost::shared_ptr<const mpimessage> mpirecvrequest::message() {
    if(data_ == 0) {
      throw illformederror("mpirecvrequest::message: Request is not complete.");
    }
    return data_;
  }

  //
  // Method required by the recvrequest parent class.
  //
  bool mpirecvrequest::matches(const boost::shared_ptr<const networkdata>& message) {
    bool matches = false;
    // We only match mpimessage objects.
    shared_ptr<const mpimessage> mess =
      boost::dynamic_pointer_cast<const mpimessage>(message);
    if(mess != 0) {
      matches = ((mess->cat() == category_) &&
                 (mess->commid() == commid_) &&
                 source_.matches(mess->source()) &&
                 tag_.matches(mess->tag()));
      if(matches) {
        if(type_ != mess->type()) {
          std::cerr << "WARNING:  mpirecvrequest:  Expected message with "
                    << type_ << " but received " << mess->type() << "\n";
        }
        if(bytes_ != mess->bytes()) {
          std::cerr << "WARNING:  mpirecvrequest:  Expected message size "
                    << bytes_ << " bytes but received " << mess->bytes() <<"\n";
        }
      }
    }
    return matches;
  }

  //
  // Complete a receive.
  //
  void mpirecvrequest::complete(const timestamp &completion_time,
                                const boost::shared_ptr<const networkdata>& message)
  {
    if(! this->matches(message)) {
      throw valueerror("mpirecvrequest::complete:  Message mismatch.");
    }
    data_ = boost::dynamic_pointer_cast<const mpimessage>(message);    
    finally_->handle(completion_time);
    finally_ = shared_ptr<eventhandler>();
  }

} // end of namespace sstmac.

