/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiringallgather.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpiringallgather::mpiringallgather(const shared_ptr<processrequest>& thekey,
                                     const shared_ptr<mpiqueue>& queue,
                                     int sendcnt, const mpitype &sendtype,
                                     int recvcnt, const mpitype &recvtype,
                                     const mpitag &tag, const mpicomm &comm,
                                     const shared_ptr<const payload>& content,
                                     const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm), sendcnt_(sendcnt), sendtype_(sendtype),
    recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag), comm_(comm), 
    magic_(magic), iteration_(0), target_iterations_(comm.size().id - 1),
    source_((comm.size().id + comm.rank().id - 1) % comm.size().id),
    dest_((comm.rank().id + 1) % comm.size().id),
    pending_sends_(0), pending_recvs_(0)
  {
    // Clone the content if it is non-null.
    //std::cerr << "mpiringallgather(...):  " << comm << " of size " << comm.size() << ": adding collective payload for tag " << tag << "\n";
    magic_->add_collective_payload(typeid(*this), tag, comm,
                                   (content ? content->clone() : content),
                                   true);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpiringallgather::send_complete(const timestamp &now,
                                       int count,
                                       const mpitype &type,
                                       const mpitag &tag,
                                       const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpiringallgather::send_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      //std::cerr << "mpiringallgather::send_complete:  No pending messages on " << comm_ << " with tag " << tag_ << "\n";
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgather::send_complete:  Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpiringallgather::recv_complete(const timestamp &now,
                                       int count,
                                       const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpiringallgather::recv_complete(" << now << ", " << count
              << ", " << stat << ")\n";
    --pending_recvs_;
    ++iteration_;
    if(iteration_ < target_iterations_) {
      // Next send/receive pair.
      ++pending_sends_;
      ++pending_recvs_;
      this->start_recv(recvcnt_, recvtype_, tag_, source_);
      this->start_send(sendcnt_, sendtype_, tag_, dest_);
    }
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgather::recv_complete:  Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback to signal copy completion (ignored).
  //
  void mpiringallgather::copy_complete(const timestamp &now,
                                       int count,
                                       const mpitype &type,
                                       const mpitag &tag,
                                       const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpiringallgather::copy_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << source << ")\n";
  }

  //
  // Hello.
  //
  shared_ptr<mpiringallgather>
  mpiringallgather::construct(const shared_ptr<processrequest>& thekey,
                              const shared_ptr<mpiqueue>& queue,
                              int sendcnt, const mpitype &sendtype,
                              int recvcnt, const mpitype &recvtype,
                              const mpitag &tag, const mpicomm &comm,
                              const shared_ptr<const payload>& content,
                              const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpiringallgather> retval
      (new mpiringallgather(thekey, queue, sendcnt, sendtype,
                            recvcnt, recvtype, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiringallgather::~mpiringallgather() throw() {
  }

  //
  // Get busy child.
  //
  void mpiringallgather::start(const timestamp &now,
                               const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpiringallgather::start(" << now << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    // Catch situations where comm_.size() is 1.
    if(target_iterations_ == 0) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgather::start:  Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
    else {
      // Set up the first send/receive pair.
      iteration_ = 0;
      pending_sends_ = pending_recvs_ = 1;
      this->start_recv(recvcnt_, recvtype_, tag_, source_);
      this->start_send(sendcnt_, sendtype_, tag_, dest_);
    }
  }

} // end of namespace sstmac
