/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIRINGALLGATHER_H
#define SSTMAC_PROCESS_MPIRINGALLGATHER_H

#include <sstmac/process/mpicollective.h>

namespace sstmac {

  /**
   * Perform an allgather using a ring.
   * Each node performs size-1 iterations where it receives data from node
   * rank-1 and sends to node rank+1 (where both ranks wrap around).
   */
  class mpiringallgather : public mpicollective {
    /// The number of elements sent.
    int sendcnt_;
    /// The type we're sending.
    mpitype sendtype_;
    /// The number of elements received.
    int recvcnt_;
    /// The type we are receiving.
    mpitype recvtype_;
    /// Our tag.
    mpitag tag_;
    /// The communiator.
    mpicomm comm_;
    /// The magic shortcut for sharing payload data between all nodes
    /// (helps this part of memory use scale as N rather than N**2)
    boost::shared_ptr<mpimagic> magic_;
    /// The current iteration round (we perform size-1 iterations).
    int iteration_;
    /// The target iterations.
    const int target_iterations_;
    /// The node we will be receiving from.
    const mpiid source_;
    /// The node we will be sending to.
    const mpiid dest_;
    /// The number of pending sends (usually zero or one).
    int pending_sends_;
    /// The number of pending receives (usually zero or one).
    int pending_recvs_;

  protected:
    /// Hello.
    mpiringallgather(const boost::shared_ptr<processrequest>& thekey,
                     const boost::shared_ptr<mpiqueue>& queue,
                     int sendcnt, const mpitype &sendtype,
                     int recvcnt, const mpitype &recvtype,
                     const mpitag &tag, const mpicomm &comm,
                     const boost::shared_ptr<const payload>& content,
                     const boost::shared_ptr<mpimagic>& magic);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now,
                               int count,
                               const boost::shared_ptr<const mpistatus>& stat);

    /// Callback to signal copy completion.
    void copy_complete(const timestamp &now,
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    /// Hello.
    static boost::shared_ptr<mpiringallgather>
    construct(const boost::shared_ptr<processrequest>& thekey,
              const boost::shared_ptr<mpiqueue>& queue,
              int sendcnt, const mpitype &sendtype,
              int recvcnt, const mpitype &recvtype,
              const mpitag &tag, const mpicomm &comm,
              const boost::shared_ptr<const payload>& content,
              const boost::shared_ptr<mpimagic>& magic);

    /// Goodbye.
    virtual ~mpiringallgather() throw();

    /// Get busy child.
    virtual void start(const timestamp &now,
                       const boost::shared_ptr<nodeinterface>& my_node);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIRINGALLGATHER_H
