/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiringallgatherv.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpiringallgatherv::mpiringallgatherv(const shared_ptr<processrequest>& thekey,
                                       const shared_ptr<mpiqueue>& queue,
                                       int sendcnt, const mpitype &sendtype,
                                       const std::vector<int> &recvcnt,
                                       const mpitype &recvtype,
                                       const mpitag &tag,
                                       const mpicomm &comm,
                                       const shared_ptr<const payload>& content,
                                       const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm), sendcnt_(sendcnt), sendtype_(sendtype),
    recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag), comm_(comm),
    magic_(magic),
    source_((comm.size().id + comm.rank().id - 1) % comm.size().id),
    dest_((comm.rank().id + 1) % comm.size().id),
    pending_sends_(0), pending_recvs_(0)
  {
    // Clone the content if it is non-null.
    magic_->add_collective_payload(typeid(*this), tag, comm,
                                   (content ? content->clone() : content),
                                   true);
    const size_t size = comm_.size().id;
    if(recvcnt_.size() != size) {
      std::cerr << "ERROR:  mpiringallgatherv called with a receive count of "
                << recvcnt_.size() << " on a communicator of size "
                << size << "\n";
      throw valueerror("mpiringallgatherv(...):  Incorrect recv count.");
    }
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpiringallgatherv::send_complete(const timestamp &now,
                                        int count,
                                        const mpitype &type,
                                        const mpitag &tag,
                                        const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpiringallgatherv::send_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgatherv::send_complete: Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpiringallgatherv::recv_complete(const timestamp &now,
                                        int count,
                                        const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpiringallgatherv::recv_complete(" << now << ", " << count
              << ", " << stat << ")\n";
    --pending_recvs_;
    if(! iteration_.empty()) {
      // Next send/receive pair.
      srpair sr = iteration_.front();
      iteration_.pop_front();
      ++pending_sends_;
      ++pending_recvs_;
      this->start_recv(sr.recvcnt, recvtype_, tag_, source_);
      this->start_send(sr.sendcnt, sendtype_, tag_, dest_);
    }
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgatherv::recv_complete: Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback to signal copy completion (ignored).
  //
  void mpiringallgatherv::copy_complete(const timestamp &now,
                                        int count,
                                        const mpitype &type,
                                        const mpitag &tag,
                                        const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpiringallgatherv::copy_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << source << ")\n";
  }

  //
  // Hello.
  //
  shared_ptr<mpiringallgatherv>
  mpiringallgatherv::construct(const shared_ptr<processrequest>& thekey,
                               const shared_ptr<mpiqueue>& queue,
                               int sendcnt, const mpitype &sendtype,
                               const std::vector<int> &recvcnt,
                               const mpitype &recvtype,
                               const mpitag &tag, const mpicomm &comm,
                               const shared_ptr<const payload>& content,
                               const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpiringallgatherv> retval
      (new mpiringallgatherv(thekey, queue, sendcnt, sendtype,
                             recvcnt, recvtype, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiringallgatherv::~mpiringallgatherv() throw() {
  }

  //
  // Get busy child.
  //
  void mpiringallgatherv::start(const timestamp &now,
                                const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpiringallgatherv::start(" << now << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    // Catch situations where comm_.size() is 1.
    if(comm_.size().id <= 1) {
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpiringallgatherv::start: Transfers done "
                             "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
    else {
      // Set up the iteration_ queue.
      const int size = comm_.size().id;
      const int rank = comm_.rank().id;
      for(int offset = 1; offset < size; ++offset) {
        int receiving_data_owner = (size + rank - offset) % size;
        int sending = sendcnt_;
        int recving = recvcnt_.at(receiving_data_owner);
        iteration_.push_back(srpair(sending, recving));
      }
      // start the first send/receive pair.
      srpair first = iteration_.front();
      iteration_.pop_front();
      pending_sends_ = pending_recvs_ = 1;
      this->start_recv(first.recvcnt, recvtype_, tag_, source_);
      this->start_send(first.sendcnt, sendtype_, tag_, dest_);
    }
  }

} // end of namespace sstmac
