/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPISCANENGINE_H
#define SSTMAC_PROCESS_MPISCANENGINE_H

#include <sstmac/process/mpicollective.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpiop.h>

namespace sstmac {

  /// Important types.
  class mpiqueue;
  class mpistatus;
  class nodeinterface;
  class topologyiterator;

  /**
   * An mpikernel to performs an mpi_scan
   */
  class mpiscanengine : public mpicollective {
    /// The iterator to whom we entrust our communication patterns.
    boost::shared_ptr<topologyiterator> iter_;
    /// The lenght of our data array.
    int count_;
    /// The type we are sending and receiving.
    mpitype type_;
    /// The operation we are applying to our data.
    mpiop op_;
    /// The tag associated with this communication.
    mpitag tag_;
    /// My communicator.
    mpicomm comm_;
    /// The data we are collecting.
    boost::shared_ptr<mpicollpayload> content_;

    /// The number of pending sends.
    int pending_sends_;
    /// The number of pending receives.
    int pending_recvs_;

    /// Fire off the next set of send and receive operations.
    void sendrecv(const timestamp &now);

    /// Private method used by the completer to indicate that our work is done
    void completer_done(const timestamp &now,
                        const boost::shared_ptr<mpicollpayload>& content);

  protected:
    /// Hi.
    mpiscanengine(const boost::shared_ptr<topologyiterator>& iter,
                  const boost::shared_ptr<processrequest>& thekey,
                  const boost::shared_ptr<mpiqueue>& queue,
                  int count, const mpitype &type, 
                  const mpiop &op,
                  const mpitag &tag, const mpicomm &comm,
                  const boost::shared_ptr<const payload>& content);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(const timestamp &now,
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(const timestamp &now,
                               int count,
                               const boost::shared_ptr<const mpistatus>& stat);

    /// Callback to to signal copy completion.
    void copy_complete(const timestamp &now,
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    /// Make a new engine.
    static boost::shared_ptr<mpiscanengine>
    construct(const boost::shared_ptr<topologyiterator>& iter,
              const boost::shared_ptr<processrequest>& thekey,
              const boost::shared_ptr<mpiqueue>& queue,
              int count, const mpitype &type, 
              const mpiop &op, const mpitag &tag, const mpicomm &comm,
              const boost::shared_ptr<const payload>& content);

    /// Goodbye.
    virtual ~mpiscanengine() throw();

    /// Capture the 'start' method so we can set up our reducer.
    virtual void start(const timestamp &start_time,
                       const boost::shared_ptr<nodeinterface>& my_node);

  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPISCANENGINE_H
