/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpisendengine.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/payload.h>
#include <sstmac/util/alloc.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <vector>
#include <stdlib.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  // Yet another class of short-lived objects.  Let's reduce the load
  // on the memory manager.
  namespace mspool {
    static std::vector<mpisendengine*> *chunks = 0;

    static mpisendengine* alloc_chunk() {
      if(! chunks) {
        chunks = new std::vector<mpisendengine*>;
        chunks->reserve(1024);
      }
      if(! chunks->empty()) {
        mpisendengine *memptr = chunks->back();
        chunks->pop_back();
        return memptr;
      }
      else {
        mpisendengine *memptr = allocate_aligned<mpisendengine>();
        return memptr;
      }
    }

    static void free_chunk(mpisendengine *chunk) {
      chunk->~mpisendengine();
      chunks->push_back(chunk);
    }

  } // end of namespace mspool.

  //
  // Constructor -- sets up our payload to send.
  //
  mpisendengine::mpisendengine(const boost::shared_ptr<processrequest>& thekey,
                               const shared_ptr<mpiqueue>& queue,
                               int count, const mpitype &type,
                               const mpiid &dest, const mpitag &tag,
                               const mpicomm &comm, bool rsend, bool ssend,
                               mpimessage::category cat,
                               const boost::shared_ptr<const payload>& content) :
    mpikernel(thekey),
    queue_(queue), count_(count), type_(type), dest_(dest), tag_(tag),
    comm_(comm), rsend_(rsend), ssend_(ssend), cat_(cat)
  {
    if(debugger::active())
      debug() << "mpisendengine[" << this << "]::mpisendengine(" << thekey
              << ", " << queue << ", " << count << ", " << type << ", " << dest
              << ", " << tag << ", " << comm << ", " << rsend << ", " << ssend
              << ", " << cat << ", " << content << ")\n";
    if(content)
      content_ = content->clone();
  }

  //
  // Who am I?
  //
  void mpisendengine::set_self(const shared_ptr<mpisendengine>& self) {
    if(self.get() != this) {
      throw valueerror("mpisendengine::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<mpisendengine> mpisendengine::get_self() {
    return self_.lock();
  }

  //
  // Build me a sharable object.
  //
  shared_ptr<mpisendengine>
  mpisendengine::construct(const boost::shared_ptr<processrequest>& thekey,
                           const boost::shared_ptr<mpiqueue>& queue,
                           int count, const mpitype &type, const mpiid &dest, 
                           const mpitag &tag, const mpicomm &comm,
                           bool rsend, bool ssend, mpimessage::category cat,
                           const shared_ptr<const payload>& content)
  {
    shared_ptr<mpisendengine> retval
      (new(mspool::alloc_chunk()) mpisendengine(thekey, queue, count, type,dest,
                                                tag, comm, rsend, ssend, cat,
                                                content),
       mspool::free_chunk);
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpisendengine::~mpisendengine() throw() {
  }

  //
  // This is the method from the kernel.
  //
  void mpisendengine::start(const timestamp & start_time,
                            const boost::shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpisendengine[" << this << "]::start(" << start_time
              << ", " << my_node << ")\n";
    queue_->send(count_, type_, dest_, tag_, comm_, rsend_,
                 ssend_, cat_, this->get_self(), content_);
  }

  //
  // Grab the completion message.
  //
  void mpisendengine::handle(const timestamp &current_time) {
    if(debugger::active())
      debug() << "mpisendengine[" << this << "]::handle("
              << current_time << ")\n";
    mpisend::get_mpistatus()->set_stop_time(current_time);
    mpikernel::complete(current_time, mpisend::get_mpistatus());
  }

} // end of namespace sstmac
