/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPISERVER_H
#define SSTMAC_PROCESS_MPISERVER_H

#include <sstmac/process/server.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/ppid.h>
#include <sstmac/util/nocopy.h>
#include <map>

namespace sstmac {

  /// Important types.
  class mpiqueue;
  class mpimessage;
  class tasklist;
  class node;
  class networkinterface;

  /**
   * A server that performs tag matching (including wildcards),
   * allows probing for message arrival, and network bypass for
   * communication between tasks sharing a node.
   */
  class mpiserver : public server, virtual public nocopy {
    /// A nested type to handle a local transfer event.
    class localtransfer;

    /// This object.
    boost::weak_ptr<mpiserver> self_;
    
    /// My ppid.
    ppid ppid_;

    /// The list of tasks in my ppid.
    boost::shared_ptr<tasklist> tasklist_;

    /// The node I am associated with.
    boost::shared_ptr<node> node_;

    /// My network.
    boost::shared_ptr<networkinterface> nic_;

    /// The message queue for each of the peers residing on this node
    /// (indexed by their task id).
    /// These objects handle unfulfilled inbound and outbound requests,
    /// ack-ing messages, etc.
    typedef std::map< taskid, boost::shared_ptr<mpiqueue> > queue_t;
    queue_t queue_;

  protected:
    /// Hello.
    mpiserver(const boost::shared_ptr<tasklist>&,
              const ppid &,
              const boost::shared_ptr<node> &nodearg,
              const boost::shared_ptr<networkinterface> &nic);

    /// Who am I?
    void set_self(const boost::shared_ptr<mpiserver>& self);

    /// This is who I am.
    boost::shared_ptr<mpiserver> get_self();

  public:
    /// Hello.
    static boost::shared_ptr<mpiserver>
    construct(const boost::shared_ptr<tasklist>&,
              const ppid &,
              const boost::shared_ptr<node> &nodearg,
              const boost::shared_ptr<networkinterface> &nic);

    /// Goodbye.
    virtual ~mpiserver() throw();

    /// Define a task for this server.  Each task can only be defined once
    virtual void add_task(const taskid &tid, const boost::shared_ptr<mpiqueue>& peer);

    /// Test whether the given task is defined on this node.
    virtual bool has_task(const taskid &tid) const;

    /// Send data.  All sends are triggered at time nodeinterface::now().
    /// The backend has the "right" to handle local transfers differently
    /// than network transfers.
    /// The data will be sent exactly as is, it is up to the queue to
    /// decide on when to do handshakes, ready-sends, etc.
    virtual void send(const boost::shared_ptr<mpimessage>& payload,
                      const boost::shared_ptr<eventhandler>& completion);

  protected:
    /// CALLBACK:  Complete a receive.
    /// Forwarded by the node when an incoming message arrives.
    /// It is the responsibility of the server to figure out how to
    /// understand any envelope information in the networkdata.
    /// This method will forward the message to the corresponding queue
    /// using mpiqueue::incoming_message.
    virtual void incoming_message(const timestamp &start_time,
                                  const nodeid &sender,
                                  const boost::shared_ptr<networkdata>& load);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPISERVER_H
