/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/multicomplete.h>
#include <sstmac/process/multicomplete-handler.h>
#include <sstmac/process/processrequest.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/cast.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Private method used by a handler upon completion.
  void multicomplete::handler_done(const timestamp &current_time,
                                   const boost::shared_ptr<handler>& the_handler,
                                   int64_t index)
  {
    if(debugger::active()) debug() << "multicomplete[" << this << "]::handler_done("
            << current_time << ", " << the_handler << ", " << index << ")\n";
    typedef handlermap_t::iterator iterator;
    iterator it = handlers_.find(the_handler);
    if(it == handlers_.end()) {
      throw illformederror("multicomplete::handler_done:  "
                           "No match for handler.");
    }
    completion_time_ = current_time;
    last_completion_ = index;
    handlers_.erase(it);
    if(--required_completions_ <= 0) {
      // We don't need to wait for those other requests.
      for(iterator hand = handlers_.begin(); hand != handlers_.end(); ++hand) {
        hand->second->erase_listener(hand->first);
      }
      // Wake the caller up again.
      scheduler_->unblock(key_);
    }
  }

  //
  // Hello.
  //
  multicomplete::multicomplete(const shared_ptr<scheduler>& sched,
                               const shared_ptr<key>& thekey) :
    key_(thekey), scheduler_(sched), required_completions_(0),
    last_completion_(-1), completion_time_(-1)
  {}

  //
  // Who am I?
  //
  void multicomplete::set_self(const shared_ptr<multicomplete>& self) {
    if(self.get() != this) {
      throw valueerror("multicomplete::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<multicomplete> multicomplete::get_self() {
    return self_.lock();
  }

  //
  // Hello.
  //
  shared_ptr<multicomplete>
  multicomplete::construct(const shared_ptr<scheduler>& sched,
                           const shared_ptr<key>& thekey)
  {
    shared_ptr<multicomplete> retval(new multicomplete(sched, thekey));
    retval->set_self(retval);
    return retval;
  }
  
  //
  // Goodbye.
  multicomplete::~multicomplete() throw() {
    if(! handlers_.empty()) {
      std::cerr << "WARNING:  ~multicomplete:  Still waiting for "
                << handlers_.size() << " handles to complete.  "
                << "Active processes: [ ";
      handlermap_t::iterator it;
      for(it=handlers_.begin(); it!=handlers_.end(); ++it) {
        std::cerr << it->second << " ";
      }
      std::cerr << "]\n";
    }
  }

  //
  // Wait for all the given requests to complete.
  //
  timestamp multicomplete::complete_all(const timestamp &start_time,
                                      std::vector< shared_ptr<processrequest> > &req)
  {
    if(debugger::active()) debug() << "multicomplete[" << this << "]::complete_all(" << start_time
            << ", <vector>.size()=" << req.size() << ")\n";
    //
    if(! handlers_.empty()) {
      throw illformederror("multicomplete::complete_all:  Current pending "
                           "event handlers.");
    }
    completion_time_ = start_time;
    // Walk the requests and register as a listener for the ones not completed.
    for(size_t id = 0; id < req.size(); ++id) {
      shared_ptr<processrequest> it = req.at(id);
      if((it != 0) && (it->is_active())) {
        if(it->is_complete()) {
          it->set_inactive();
        }
        else {
          // We need to wait for the completion of this request.
          shared_ptr<handler> hh = handler::construct(this->get_self(), id);
          it->add_listener(hh);
          handlers_[hh] = it;
        }
      }
    }
    // Now test whether we have any incomplete handlers.
    while(! handlers_.empty()) {
      required_completions_ = handlers_.size();
      if(debugger::active()) debug() << "multicomplete[" << this << "]::complete_all:  Currently "
              << handlers_.size() << " completions pending\n";
      scheduler_->block(key_);
    }
    // All done.
    if(debugger::active()) debug() << "multicomplete[" << this << "]::complete_all:  Completed.\n";
    return completion_time_;
  }

  //
  // Wait for one of the given requests to complete.
  //
  timestamp multicomplete::complete_one(const timestamp &start_time,
                                      std::vector< shared_ptr<processrequest> > &req,
                                      int &index)
  {
    if(! handlers_.empty()) {
      throw illformederror("multicomplete::complete_one:  Current pending "
                           "event handlers.");
    }
    completion_time_ = start_time;
    index = -1; // sentinel.
    // See if we can get out easy; test whether any of the input requests
    // is already finished.
    for(size_t id = 0; id < req.size(); ++id) {
      shared_ptr<processrequest> it = req.at(id);
      if((it != 0) && (it->is_active())) {
        if(it->is_complete()) {
          // Well that was easy.
          it->set_inactive();
          handlers_.clear();
          index = check_cast<int>(id);
          break;
        }
        else {
          shared_ptr<handler> hh = handler::construct(this->get_self(), check_cast<int>(id));
          it->add_listener(hh);
          handlers_[hh] = it;
        }
      }
    }
    // Now we wait.
    if(index < 0 && !handlers_.empty()) {
      required_completions_ = 1;
      scheduler_->block(key_);
      index = check_cast<int>(last_completion_);
    }
    return completion_time_;
  }

  //
  // Wait for one or more of the given requests to complete.
  //
  timestamp multicomplete::complete_some(const timestamp &start_time,
                                       std::vector< shared_ptr<processrequest> > &req,
                                       std::vector<int> &indices)
  {
    if(! handlers_.empty()) {
      throw illformederror("multicomplete::complete_some:  Current pending "
                           "event handlers.");
    }
    completion_time_ = start_time;
    indices.clear();
    // Gather info on any completed requests.
    for(size_t id = 0; id < req.size(); ++id) {
      shared_ptr<processrequest> it = req.at(id);
      if((it != 0) && (it->is_active())) {
        if(it->is_complete()) {
          // We're getting off easy today.
          indices.push_back(id);
        }
      }
    }
    // Now we see whether we need to do anything.
    if(indices.empty()) {
      int first_to_finish = -1;
      completion_time_ = this->complete_one(start_time, req, first_to_finish);
      if(first_to_finish >= 0) {
        indices.push_back(first_to_finish);
      }
    }
    return completion_time_;
  }

} // end of namespace sstmac
