/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_REQUEST_H
#define SSTMAC_PROCESS_REQUEST_H

#include <sstmac/util/nocopy.h>
#include <sstmac/simulator/timestamp.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <vector>

namespace sstmac {

  // Important types.
  class eventhandler;
  class status;

  /**
   * A handle to wait for the completion of a kernel.
   *
   * When the kernel completes, it will tell the request about the completion
   * and give it information about the operation through a status object.
   */
  class request : virtual public nocopy {
    /// This object.
    boost::weak_ptr<request> self_;
    
    /// Any meta-requests that are waiting for our completion.
    std::vector< boost::shared_ptr<eventhandler> > listeners_;

    /// The status associated with the completion of this request.
    boost::shared_ptr<const status> status_;

    /// Flag to monitor whether this request is active or not.
    bool active_;

  protected:
    /// Hello.
    request();

    /// Who am I?
    void set_self(const boost::shared_ptr<request>& self);

    /// This is who I am.
    boost::shared_ptr<request> get_self();

  public:
    /// Construct a new request object.
    static boost::shared_ptr<request> construct();

    /// Virtual destructor is essential.
    /// Alerts all listeners that this request is exiting.
    virtual ~request() throw();

    /// Reset this request to be incomplete.  This allows recycling of
    /// requests for persistent handles and similar calls.
    virtual void reset();

    /// Test whether the machinery underlying this request has been completed.
    virtual bool is_complete() const;

    /// Test whether this request is active (can be waited on or retrieved).
    virtual bool is_active() const;
    
    /// Tag this request as inactive.  After this point it is an error
    /// to add a listener for this request.
    virtual void set_inactive();

    /// Get the status of this completed request.
    /// \throw illformederror if the request is not complete.
    boost::shared_ptr<const status> get_status();

    /// Perform a callback to the given eventhandler using the alert method
    /// when this request completes.  After the callback, this request
    /// should release the reference counted pointer to the request.
    /// This method is effective immediately and does not require a
    /// separate event to be scheduled.
    /// \param listener is a non-null pointer to a request object.
    virtual void add_listener(const boost::shared_ptr<eventhandler>& lis);

    /// Un-register from a callback to the given request.
    /// This must release any reference counted pointer to this request.
    /// This method is effective immediately and does not require a
    /// separate event to be scheduled.
    virtual void erase_listener(const boost::shared_ptr<eventhandler>& lis);

    /// Callback used by the kernel upon its completion.
    /// Alerts any eventhandlers that this request is done.
    void complete(const timestamp &current_time, const boost::shared_ptr<const status>&);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_REQUEST_H
