/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/simplemailserver-mbox.h>
#include <sstmac/process/simplemboxapp.h>
#include <sstmac/process/actormagic.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/threadwake.h>
#include <sstmac/simulator/node.h>
#include <boost/lexical_cast.hpp>
#include <algorithm>
#include <iostream>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::lexical_cast;

  //
  // Constructor.
  //
  simplemailserver::mbox::mbox(const actorid &id,
                               const shared_ptr<simplemailserver>& parent) :
    id_(id), parent_(parent), active_(false)
  {
    parent_->magic_->add(id_);
  }

  //
  // Get us a new mbox.
  //
  shared_ptr<simplemailserver::mbox>
  simplemailserver::mbox::construct(const actorid &id, 
                                    const shared_ptr<simplemailserver>& parent)
  {
    return shared_ptr<mbox>(new mbox(id, parent));
  }

  //
  // Done.
  //
  simplemailserver::mbox::~mbox() {
    if(parent_ && parent_->magic_)
      parent_->magic_->remove(id_);
  }

  //
  // Allow the parent to send us a message to handle.
  //
  void simplemailserver::mbox::
  handle(const actorid &sender, const shared_ptr<actormsg>& payload) {
    unhandled_.push_back(std::make_pair(sender, payload));
    // And finally test whether there is a pending receive for this payload.
    for(blocked_t::iterator it = blocked_.begin(); it != blocked_.end(); ++it){
      if(it->first.first.matches(sender)&&it->first.second->matches(payload)) {
        shared_ptr<key> thekey = it->second;    
        blocked_.erase(it);
        parent_->sched_->unblock(thekey);
        break;
      }
    }
  }

    /* Methods from here on down are part of the simplembox interface */

  //
  // What's my pid.
  //
  actorid simplemailserver::mbox::mypid() const {
    return id_;
  }

  //
  // What nodes are known to the current system?
  //
  const std::vector<nodeid>& simplemailserver::mbox::nodes() const {
    return parent_->nodes_;
  }

  //
  // All peer identifiers (magically, we know about everybody).
  //
  const std::set<actorid>& simplemailserver::mbox::world() const {
    return parent_->magic_->world();
  }

  //
  // Register the given tag mapping in the registry.
  //
  void simplemailserver::mbox::set_tag(int key, actorid value) {
    tagmap_t &tm = parent_->magic_->tagmap();
    if(tm.find(key) != tm.end())
      throw valueerror(std::string("simplemailserver::mbox::set_tag:  "
                                   "Attempt to redefine tag ") 
                       + lexical_cast<std::string>(key));
    tm[key] = value;
  }

  //
  // Test whether the given key is defined.
  //
  bool simplemailserver::mbox::has_tag(int key) const {
    const tagmap_t &tm = parent_->magic_->tagmap();
    return tm.find(key) != tm.end();
  }

  //
  // Access the value of the given key.
  //
  actorid simplemailserver::mbox::get_tag(int key) const {
    const tagmap_t &tm = parent_->magic_->tagmap();
    tagmap_t::const_iterator it = tm.find(key);
    if(it == tm.end())
      throw valueerror(std::string("simplemailserver::mbox::get_tag:  "
                                   "No match for the tag ")
                       + lexical_cast<std::string>(key));
    return it->second;
  }

  //
  // Access the tag mapping in a const context.
  //
  const simplemailserver::mbox::tagmap_t&
  simplemailserver::mbox::tag_map() const {
    return parent_->magic_->tagmap();
  }

  //
  // Global counters.
  //
  void simplemailserver::mbox::move_global_counter(const std::string &label,
                                                   int increment)
  {
    parent_->magic_->move_global_counter(label, increment);
  }

  //
  // Read the local state of the given global counter.
  //
  int simplemailserver::mbox::
  read_global_counter(const std::string &label) const {
    return parent_->magic_->read_global_counter(label);
  }

  //
  // Initialize.  This method is collective and "magic".
  //
  timestamp simplemailserver::mbox::init() {
    if(active_)
      throw illformederror("simplemailserver::mbox::init:  Already active\n");
    parent_->magic_->barrier(1, id_, parent_->magic_->world());
    return parent_->node_->now();
  }

  //
  // Finalize.  This is a magic operation.
  //
  timestamp simplemailserver::mbox::finalize() {
    if(active_) 
      throw illformederror("simplemailserver::mbox::finalize:  Not active\n");
    parent_->magic_->barrier(2, id_, parent_->magic_->world());
    return parent_->node_->now();
  }

  //  
  // Spawn the given task on the given node and return its pid.
  //
  actorid simplemailserver::mbox::
  spawn(nodeid target,
        const boost::shared_ptr<simplemboxapp>& entity,
        int64_t communication_size)
  {
    if(! entity)
      throw nullerror("simplemailserver::mbox::spawn:  NULL app pointer.");
    shared_ptr<simplembox> the_mbox = parent_->new_mbox();
    if(! the_mbox)
      throw illformederror("simplemailserver::mbox::spawn: Got NULL "
                           "mbox from parent.");
    shared_ptr<simplemboxapp> the_app = entity->clone();
    if(! the_app)
      throw illformederror(std::string("simplemailserver::mbox::spawn: "
                                       "'clone()' method returns NULL for ")
                           + typeid(*entity).name());
    the_app->init_mbox(the_mbox);
    parent_->sched_->add_thread(the_app);
    return the_mbox->mypid();
  }

  //
  // Send a message to the given pid.
  //
  void simplemailserver::mbox::
    send(const actorid &target, const shared_ptr<const actormsg>& msg, int64_t size)
  {
    if(! msg)
      throw nullerror("simplemailserver::mbox::send: "
                      "NULL messages are not allowed.");
    parent_->send(id_, target, msg, size);
  }

  //  
  // Test whether there is any message that matches the given pattern.
  bool simplemailserver::mbox::
  test(const actorid &source, const shared_ptr<const actorpattern>& pat) const
  {
    unhandled_t::const_iterator it, end = unhandled_.end();
    for(it = unhandled_.begin(); it != end; ++it) {
      if(source.matches(it->first) && pat->matches(it->second))
        return true;
    }
    return false;
  }

  //
  // Block until we get a message matching the given pattern.
  //
  std::pair< actorid, shared_ptr<actormsg> > simplemailserver::mbox::
    recv(const actorid &source, const shared_ptr<const actorpattern>& pat)
  {
    shared_ptr<key> thekey;
    std::pair< actorid, shared_ptr<actormsg> > retval;
    unhandled_t::iterator it;
    while(1) {
      for(it = unhandled_.begin(); it != unhandled_.end(); ++it) {
        if(source.matches(it->first) && pat->matches(it->second)) {
          retval.first = it->first;
          retval.second = it->second;
          unhandled_.erase(it);
          break;
        }
      }
      if(! retval.second) {
        // We get here because there has been no match yet.
        if(! thekey) {
          thekey = key::construct();
          actorp_t matcher = std::make_pair(source, pat);
          blocked_.push_back(std::make_pair(matcher, thekey));
        }
        parent_->sched_->block(thekey);
      }
      else {
        break;
      }
    }
    // and we get here when we have a match.
    // The match routine has already erased our 'blocked_' entry.
    return retval;
  }

  //
  // Pretend to compute for the given amount of time.
  //
  void simplemailserver::mbox::compute(timestamp cputime, int threads) {
    if(cputime.ticks() > 0) {
      shared_ptr<key> the_key = key::construct();
      shared_ptr<threadwake> wake =
        threadwake::construct(parent_->sched_, the_key);
      parent_->node_->compute(parent_->node_->now(), parent_->ppid_, cputime,
                                threads, wake);
      parent_->sched_->block(the_key);
    }
  }

  //
  // Pretend to be idle (block) for the given amount of time.
  //
  void simplemailserver::mbox::sleep(timestamp cputime) {
    if(cputime.ticks() > 0) {
      shared_ptr<key> the_key = key::construct();
      shared_ptr<threadwake> wake =
        threadwake::construct(parent_->sched_, the_key);
      parent_->node_->get_eventmanager()->schedule(parent_->node_->now() + cputime, wake);
      parent_->sched_->block(the_key);
    }
  }

  //
  // Pretend to copy the given number of bytes.
  //
  void simplemailserver::mbox::copy(int64_t bytes) {
    if(bytes > 0) {
      shared_ptr<key> the_key = key::construct();
      shared_ptr<threadwake> wake =
        threadwake::construct(parent_->sched_, the_key);
      parent_->node_->copy(parent_->node_->now(), parent_->ppid_, bytes, wake);
      parent_->sched_->block(the_key);
    }
  }

  //
  // Get current time.
  //
  timestamp simplemailserver::mbox::now() const {
    return parent_->node_->now();
  }

} // end of namespace sstmac
