/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_SIMPLEMAILSERVER_H
#define SSTMAC_PROCESS_SIMPLEMAILSERVER_H

#include <sstmac/process/server.h>
#include <sstmac/process/actorid.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/simulator/ppid.h>
#include <boost/weak_ptr.hpp>
#include <vector>
#include <map>
#include <set>
#include <stdint.h>

namespace sstmac {

  // Forward declaration of mportant types.
  class simplembox;
  class scheduler;
  class actormagic;
  class tasklist;
  class actormsg;
  class node;
  class networkinterface;

  /**
   * A very simple dataflow process network server (one per node).
   */
  class simplemailserver : public server {
    /// The actual implementation of a simplembox -- we make this a private
    /// nested type to make it easier to safely manage internal state.
    class mbox;

    /// Private networkdata type where we can package the info we need.
    class data;
    
    /// This object.  We need to be able to hand out reference counted pointers
    /// to child processes.
    boost::weak_ptr<simplemailserver> self_;

    /// The scheduler we use to block calling threads.
    boost::shared_ptr<scheduler> sched_;

    /// The "magic" bypass for barriers (used for init and finalize)
    /// and for tagging specific actorid objects with flat indices.
    boost::shared_ptr<actormagic> magic_;

    /// The identifiers of nodes we were given.  Stored in a vector to allow
    /// the user to query for nodes.
    std::vector<nodeid> nodes_;
    /// Nodes are also stored in a set to ensure that we only take valid ids.
    std::set<nodeid> valid_nodes_;

    /// The node id that this server resides on.
    nodeid my_node_id_;

    /// The rank id that we will give to our next spawned process.
    int32_t next_child_;

    /// Our node.
    boost::shared_ptr<node> node_;

    /// The network we communicate over.
    boost::shared_ptr<networkinterface> nic_;
    
    /// The client processes we have (weak pointers, because clients are
    /// welcome to go out of scope any time they want to).
    typedef std::map<actorid, boost::weak_ptr<mbox> > mboxes_t;
    mboxes_t mboxes_;

    /// Our parallel process id.
    ppid ppid_;

    /// Private method to send data.
    void send(const actorid &source, const actorid &target,
              const boost::shared_ptr<const actormsg>& msg, int64_t size);

    /// Private constructor.
    simplemailserver(nodeid my_node_id,
                     const boost::shared_ptr<actormagic>& magic,
                     const boost::shared_ptr<tasklist>& tlist,
                     const boost::shared_ptr<scheduler>& sched,
                     const ppid &ppidarg,
                     const boost::shared_ptr<node> &nodearg,
                     const boost::shared_ptr<networkinterface> &nic);

  public:
    /// Get us a new, heap-allocated object.
    static boost::shared_ptr<simplemailserver>
    construct(nodeid my_node_id,
              const boost::shared_ptr<actormagic>& magic,
              const boost::shared_ptr<tasklist>& tlist,
              const boost::shared_ptr<scheduler>& sched,
              const ppid &ppidarg,
              const boost::shared_ptr<node> &nodearg,
              const boost::shared_ptr<networkinterface> &nic);

    /// Clean up.
    virtual ~simplemailserver();

    /// Create a new mbox interface that can talk to this object.
    boost::shared_ptr<simplembox> new_mbox();

    /* Interface required by the 'server' base type. */

    /// Complete a receive.
    /// Forwarded by the node when an incoming message arrives.
    /// It is the responsibility of the server to figure out how to
    /// understand any envelope information in the networkdata.
    virtual void incoming_message(const timestamp &start_time,
                                  const nodeid &sender,
                                  const boost::shared_ptr<networkdata>& load);

  };
  
} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_SIMPLEMAILSERVER_H
