/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_SIMPLEMBOX_H
#define SSTMAC_PROCESS_SIMPLEMBOX_H

#include <sstmac/process/actorid.h>
#include <sstmac/process/actormsg.h>
#include <sstmac/process/actorpattern.h>
#include <sstmac/simulator/timestamp.h>
#include <utility>
#include <vector>
#include <map>
#include <set>

namespace sstmac {

  // Forward declaration of important types.
  class simplemboxapp;

  /**
   * A very simplified version of a distributed a Kahn process network.
   *
   * Perform non-blocking sends and blocking receives.
   */
  class simplembox {
  public:
    /// Required support for knowing about all peer identifiers in our world.
    typedef std::set<actorid> world_t;

    /// Required support for globally mapping integer tags to actorid
    /// objects.
    typedef std::map<int, actorid> tagmap_t;

    /// Return type for recv calls.
    typedef std::pair< actorid, boost::shared_ptr<actormsg> > recvresult_t;

    /// Goodbye.
    virtual ~simplembox() {}

    /// What's my pid.
    virtual actorid mypid() const = 0;

    /// What nodes are known to the current system?
    virtual const std::vector<nodeid>& nodes() const = 0;

    /// All peer identifiers (magically, we know about everybody).
    virtual const world_t& world() const = 0;

    /// Register the given tag mapping in the registry.
    /// A key cannot be redefined unless it is first erased.
    virtual void set_tag(int key, actorid value) = 0;

    /// Test whether the given key is defined.
    virtual bool has_tag(int key) const = 0;

    /// Access the value of the given key.
    virtual actorid get_tag(int key) const = 0;

    /// Access the tag mapping in a const context.
    virtual const tagmap_t& tag_map() const = 0;

    /// Update (or define) the given global counter.
    /// In a real system, this would have
    /// roughly log(N) overhead (so use sparingly).  Also left as an
    /// exercise for the implementor is how global counters account for
    /// node faults (for example probably via linked processes and by 
    /// having the peer process handle decrementing the counter as needed.
    virtual void move_global_counter(const std::string &label,
                                     int increment) = 0;

    /// Read the local state of the given global counter.
    /// This would (in a real system) be assumed to be a purely
    /// local operation.
    /// \throw valueerror if the given global counter is not defined.
    virtual int read_global_counter(const std::string &label) const = 0;

    /// Initialize.  This method is collective and "magic" -- it ensures
    /// that all nodes have joined the party before anybody starts.
    virtual timestamp init() = 0;

    /// Finalize.  This is a magic operation.
    virtual timestamp finalize() = 0;

    /// Spawn the given task on the given node and return its pid.
    /// This can only be called between 'init()' and 'finalize()'.
    virtual actorid spawn(nodeid target,
                        const boost::shared_ptr<simplemboxapp>& entity,
                        int64_t communication_size) = 0;

    /// Send a message to the given pid.
    /// This call always succeeds except for invalid pid.
    /// \throw valueerror if the pid is not valid.
    /// \throw nullerror if sending a NULL message (reserved for recv timeout).
    virtual void send(const actorid &target,
                      const boost::shared_ptr<const actormsg>& msg,
                      int64_t size) = 0;

    /// Test whether there is any message that matches the given pattern.
    /// This has the same effect as calling recv with a zero-length timeout.
    /// \param source  Our desired source (may be actorid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    virtual bool test(const actorid &source,
                      const boost::shared_ptr<const actorpattern>& pat) const = 0;
    
    /// Block until we get a message matching the given pattern.
    /// \param source  Our desired source (may be actorid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \return        The received object (sorry about the dynamic casts)
    ///                or std::make_pair(actorid::invalid, NULL) if timed out.
    virtual recvresult_t
    recv(const actorid &source, const boost::shared_ptr<const actorpattern>& pat) = 0;

    /// Pretend to compute for the given amount of time.
    virtual void compute(timestamp cputime_per_thread, int threads=1) = 0;

    /// Pretend to be idle (block) for the given amount of time.
    virtual void sleep(timestamp cputime) = 0;

    /// Pretend to copy the given number of bytes.
    virtual void copy(int64_t bytes) = 0;

    /// Get current time.
    virtual timestamp now() const = 0;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_SIMPLEMBOX_H
