/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/thread.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/util/errors.h>
#include <iostream>
#include <exception>
#include <unistd.h>  // getpagesize()
#include <sys/mman.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
//#include <dlfcn.h> // dladdr (not a POSIX call)
//#include <cxxabi.h> // Demangle C++ names (not a POSIX call)

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Private method that gets called by the scheduler.
  //
  void thread::init(void *stack, int64_t stacksize,
                    const boost::shared_ptr<scheduler>& sched,
                    threadcontext_t *yield_to)
  {
    stack_ = stack;
    stacksize_ = stacksize;
    scheduler_ = sched;
    start_context(context_, stack, stacksize, run_routine, this, yield_to);
  }
  
  void thread::destroy()
  {
    destroy_context(context_);
  }
  
  //
  // Run routine that defines the initial context for this task.
  //
  void thread::run_routine(void *threadptr) {
    thread *task = (thread*)threadptr;
    assert(task != 0);
    /// This locks the reference counted pointer, ensuring that this
    /// object will not be erased before its time.
    shared_ptr<thread> self = task->get_self();
    if(self == 0) {
      // You silly person -- you probably forgot to call thread::set_self.
      throw nullerror("thread::init:  thread::get_self returns a null pointer."
                      "  You need to call thread::set_self when you construct "
                      "a new (derived) thread object.");
    }
    // Go.
    if(self->initialized()) {
      self->state_ = ACTIVE;
      self->run();
      // We are done, ask the scheduler to remove this task from the
      // list of active tasks.
      self->state_ = DONE;
      self->get_scheduler()->unblock(self->key_);
      self->get_scheduler()->complete(self->stack_);
    }
    else {
      throw schedulererror("thread::run_routine:  Task has not "
                           "been initialized.");
    }
  }

  //
  // Constuctor.
  //
  thread::thread() :
    state_(PENDING), key_(key::construct())
  {}

  //
  // Who am I?
  //
  void thread::set_self(const shared_ptr<thread>& self) {
    if(self.get() != this) {
      throw valueerror("thread::set_self:  'self' pointer does not refer "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<thread> thread::get_self() {
    return self_.lock();
  }

  //
  // Goodbye.
  //
  thread::~thread() {
  }

  //
  // Start a new thread.
  //
  void thread::start_thread(const boost::shared_ptr<thread>& thr) {
    this->get_scheduler()->add_thread(thr);
  }

  //
  // Join threads.
  //
  void thread::join() {
    if(! this->initialized()) {
      // We can't context switch the caller out without a scheduler.
      throw illformederror("thread::join():  Target thread "
                           "has not been initialized.");
    }
    const threadcontext_t *current = this->get_scheduler()->current_context();
    if(current == &this->context_) {
      // A thread cannot join itself -- leads to deadlock.
      throw illformederror("thread::join():  Calling thread "
                           "attempting to join itself.");
    }
    this->get_scheduler()->block(key_);
  }

  //
  // Access the scheduler for this thread.
  //
  shared_ptr<scheduler> thread::get_scheduler() {
    return scheduler_.lock();
  }

} // end of namespace sstmac
