/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/process/key.h>
#include <sstmac/process/app.h>
#include <sstmac/process/processrequest.h>
#include <sstmac/processor/computeapi.h>
#include <sstmac/processor/computekernel.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/threadwake.h>
#include <sstmac/process/multicomplete.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  // Catch ugly null pointers.
  template <typename T>
  void nonnull(const boost::shared_ptr<T> &p, const char *caller) {
    if(p == 0)
      throw nullerror(caller);
  }

  /// Protected constructor
  computeapi::computeapi(app_t my_app, node_t my_node) :
    app_(my_app), node_(my_node)
  { }

  /// Define who I am.
  void computeapi::set_self(const boost::shared_ptr<computeapi>& self) {
    self_ = self;
  }

  /// Override the pure virtual method from thread.
  boost::shared_ptr<computeapi> computeapi::get_self() {
    return self_.lock();
  }

  boost::shared_ptr<computeapi>
  computeapi::construct(app_t my_app, node_t my_node) {
    boost::shared_ptr<computeapi> obj( new computeapi(my_app, my_node));
    obj->set_self(obj);
    return obj;
  }

  computeapi::~computeapi() {}

  // The node assigned to this run.
  computeapi::node_t computeapi::get_node() { return node_; }

  // The app we are attached to.
  // This is used to get access to the scheduler.
  boost::shared_ptr<app> computeapi::application() { return app_.lock(); }

  //
  // Perform a computation.
  //
  timestamp computeapi::compute(const shared_ptr<instructions>& inst) {
    shared_ptr<key> thekey = key::construct();
    shared_ptr<computekernel> ker = computekernel::construct(thekey, inst);
    shared_ptr<processrequest> req = ker->get_request();
    node_->add(node_->now(), ker);
    return req->wait(this->application()->get_scheduler());
  }

} // end of namespace sstmac

