/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESSOR_COMPUTEAPI_H
#define SSTMAC_PROCESSOR_COMPUTEAPI_H

#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Forward declaration of relevant types.
  class nodeinterface;
  class eventhandler;
  class instructions;
  class processrequest;
  class key;
  class app;
  
  /**
   * Simulating computation from an application (or rather, a thread).
   */
  class computeapi {

  public:

    // Some of these shared pointer types are just outrageously long.
    typedef boost::shared_ptr<nodeinterface> node_t;
    typedef boost::shared_ptr<app> app_t;

  private:
    /// This object.
    boost::weak_ptr<computeapi> self_;

    /// The app this api is attached to.  This is used to block and unblock
    /// the thread when needed (must be a weak pointer, since the thread
    /// is what keeps this object alive).
    boost::weak_ptr<app> app_;

    /// The key we use to put our thread to sleep.
    boost::shared_ptr<key> key_;
    
    /// My node.
    node_t node_;

  protected:
    /// Protected constructor
    computeapi(app_t my_app, node_t my_node);

    /// Define who I am.
    void set_self(const boost::shared_ptr<computeapi>& self);

    /// Override the pure virtual method from thread.
    boost::shared_ptr<computeapi> get_self();

  public:
    /// Public constructor
    static boost::shared_ptr<computeapi>
      construct(app_t my_app, node_t my_node);
  
    /// Adieu.
    virtual ~computeapi();

    /// Process a compute kernel.
    timestamp compute(const boost::shared_ptr<instructions>& inst);

    /*
     * Access basic runtime information.
     */

    /// The node assigned to this run.
    node_t get_node();

    /// The app we are attached to.
    /// This is used to get access to the scheduler.
    boost::shared_ptr<app> application();

  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESSOR_COMPUTEAPI_H
