/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/processor/instructionprocessor-config.h>
#include <sstmac/util/errors.h>

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

namespace sstmac {

  //
  // Protected constructor.
  //
  instructionprocessor_config::
  instructionprocessor_config() : processor_config(processor_config::PROC_INSTRUCTION),
    flop_("PAPI_FP_INS"), tflop_(timestamp::exact_nsec(10)),
    memop_("PAPI_LST_INS"), tmemop_(timestamp::exact_nsec(250))
  {}

  //
  // Protected constructor using strings.
  //
  instructionprocessor_config::
  instructionprocessor_config(const string &flop_label,
                         const timestamp &tflop,
                         const string &memop_label,
                         const timestamp &tmemop,
                         const string &memop_local_label,
                         const timestamp &tmemop_local,
                         const string &memop_rand_label,
                         const timestamp &tmemop_rand
                         ) : processor_config(processor_config::PROC_INSTRUCTION),
    flop_(flop_label), tflop_(tflop),
    memop_(memop_label), tmemop_(tmemop),
    memop_local_(memop_local_label), tmemop_local_(tmemop_local),
    memop_rand_(memop_rand_label), tmemop_rand_(tmemop_rand)
  {}

  //
  // Protected constructor from file.
  //
  instructionprocessor_config::
  instructionprocessor_config(const string &filename) : processor_config(processor_config::PROC_INSTRUCTION),
    filename_(filename)
  {
    ifstream infile(filename.c_str());

    if (!infile.is_open()) {
      stringstream error;
      error << "problem opening instructionprocessor config file: "
            << filename;
      throw valueerror(error.str());
    }

    string key, val;
    while( infile >> key ) {
      infile >> val;
      if(key=="flop") set_flop_label(val);
      else if(key=="memop") set_memop_label(val);
      else if(key=="memop_local") set_memop_local_label(val);
      else if(key=="memop_rand") set_memop_rand_label(val);
      else if(key=="tflop") set_tflop(timestamp::exact_nsec(strtoul(val.c_str(),0,10)));
      else if(key=="tmemop") set_tmemop(timestamp::exact_nsec(strtoul(val.c_str(),0,10)));
      else if(key=="tmemop_local") set_tmemop_local(timestamp::exact_nsec(strtoul(val.c_str(),0,10)));
      else if(key=="tmemop_rand") set_tmemop_rand(timestamp::exact_nsec(strtoul(val.c_str(),0,10)));
      else {
        stringstream error;
        error << "simple processor: unrecognized option in config file: "
              << key;
        throw valueerror(error.str());
      }
    }
  }

  //
  // Public default constructor.
  //
  boost::shared_ptr<instructionprocessor_config>
  instructionprocessor_config::construct()
  {
    boost::shared_ptr<instructionprocessor_config>
      retval(new instructionprocessor_config());
    return retval;
  }

  //
  // Public constructor.
  //
  boost::shared_ptr<instructionprocessor_config>
  instructionprocessor_config::construct(const string flop_label,
                                    const timestamp tflop,
                                    const string memop_label,
                                    const timestamp tmemop,
                                    const string memop_local_label, 
                                    const timestamp tmemop_local,
                                    const string memop_rand_label,
                                    const timestamp tmemop_rand) 
  {
    boost::shared_ptr<instructionprocessor_config>
      retval(new instructionprocessor_config(flop_label, tflop,
                                        memop_label, tmemop,
                                        memop_local_label, tmemop_local, 
                                        memop_rand_label, tmemop_rand));
    return retval;
  }

  //
  // Public constructor using a file.
  //
  boost::shared_ptr<instructionprocessor_config>
  instructionprocessor_config::construct(const string &filename)
  {
    boost::shared_ptr<instructionprocessor_config>
      retval(new instructionprocessor_config(filename));
    return retval;
  }


} // end of namespace sstmac
