/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESSOR_PROCSCHEDULER_LOADMONITOR_H
#define SSTMAC_PROCESSOR_PROCSCHEDULER_LOADMONITOR_H

#include <sstmac/processor/procscheduler.h>
#include <sstmac/simulator/timestamp.h>
#include <boost/shared_ptr.hpp>
#include <map>
#include <set>
#include <vector>

namespace sstmac {

  // Forward declaration of important types.
  class instructions;
  class eventmanager;

  /**
   * Processor scheduler's load monitor
   */
  class procscheduler::loadmonitor {

  private:
    struct segment {
      timestamp start;
      timestamp end;
      int active;
    };

    boost::shared_ptr<eventmanager> manager_;
    std::map<int64_t,timestamp> start_time_;
    std::map<int64_t,timestamp> end_time_;
    std::map<int64_t,timestamp> target_time_;
    int iter_;
    int n_core_;

    void make_segment_list(std::vector<segment> &segments,int64_t compute_id);
    void print_segment_list(std::vector<segment> &segments);
    timestamp accumulate_time(const std::vector<segment> &segments, 
                              const timestamp &start,
                              const timestamp &end,
                              const int &extra_load);
    
    void cleanup();

  protected:
    /// Protected constructor
    loadmonitor(int n_core);

  public:
    /// Public constructor
    static boost::shared_ptr<loadmonitor>
    construct(int n_core);

    /// Goodbye.
    virtual ~loadmonitor() throw();

    /// Add instructions to monitor
    void add_instructions( int64_t compute_id,
                           const timestamp &start_time,
                           const timestamp &first_end_time );

    /// Check if kernel is complete 
    bool is_complete(int64_t compute_id);

    /// Compute next possible end time
    timestamp next_end_time(int64_t compute_id);

    /// Test whether this monitor has a manger.
    bool has_manager() const {
      return (manager_ != 0);
    }

    /// Define a manager.
    void set_manager(const boost::shared_ptr<eventmanager>& mgr) {
      manager_ = mgr;
    }

    /// Access the manager.
    const boost::shared_ptr<eventmanager> get_manager() const {
      return manager_;
    }

  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESSOR_PROCSCHEDULER_LOADMONITOR_H
