/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/processor/procscheduler.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/processor/processor.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/processor/procscheduler-loadmonitor.h>
#include <sstmac/processor/procscheduler-loadhandler.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <boost/shared_ptr.hpp>
#include <iostream>

#include <sstmac/processor/simpleprocessor.h>
#include <sstmac/processor/instructionprocessor.h>

#include <sstmac/processor/simpleprocessor-config.h>
#include <sstmac/processor/instructionprocessor-config.h>

namespace sstmac
{

  int64_t procscheduler::next_id_ = 0;

  //
  // Constructor
  //
  procscheduler::procscheduler(int n_core)
  {
    monitor_ = loadmonitor::construct(n_core);

    processor_
        = simpleprocessor::construct(simpleprocessor_config::construct());

  }

  //
  // Constructor
  //
  procscheduler::procscheduler(int n_core,
      const boost::shared_ptr<processor_config>& cfg)
  {
    monitor_ = loadmonitor::construct(n_core);

    switch (cfg->proctype)
      {

    case processor_config::PROC_SIMPLE:
      {
        boost::shared_ptr<simpleprocessor_config> cfgcast = boost::dynamic_pointer_cast<
        simpleprocessor_config>(cfg);
        processor_ = simpleprocessor::construct(cfgcast);
        break;
      }

    case processor_config::PROC_INSTRUCTION:
      {
        boost::shared_ptr<instructionprocessor_config> cfgcast = boost::dynamic_pointer_cast<
        instructionprocessor_config>(cfg);
        processor_ = instructionprocessor::construct(cfgcast);
        break;
      }
    default:
      throw "procscheduler: unknown proc type";
      }
  }

  //
  // Who am I?
  //
  void
  procscheduler::set_self(const boost::shared_ptr<procscheduler>& self)
  {
    if (self.get() != this)
      {
        throw valueerror(
            "procscheduler::set_self:  'self' pointer does not refer to 'this' object.");
      }
    self_ = self;
  }

  //
  // Construct a new scheduler
  //
  boost::shared_ptr<procscheduler>
  procscheduler::construct(int n_core)
  {
    boost::shared_ptr<procscheduler> retval(new procscheduler(n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Construct a new scheduler
  //
  boost::shared_ptr<procscheduler>
  procscheduler::construct(int n_core,
      const boost::shared_ptr<processor_config>& cfg)
  {
    boost::shared_ptr<procscheduler> retval(
        new procscheduler(n_core, cfg));
    retval->set_self(retval);
    return retval;
  }

  void
  procscheduler::compute(const timestamp &start_time,
      const boost::shared_ptr<instructions>& inst,
      const boost::shared_ptr<eventhandler>& completion)
  {
    if (debugger::active())
      debug() << "procscheduler::compute(" << inst << ")\n";

    timestamp comp_time = processor_->compute(inst);

    ///GH: use of loadhandler is commented out.  testparsedumpi fails while using this.
    //monitor_->add_instructions(next_id_, start_time, start_time + comp_time);

    //mgr_->schedule(start_time + comp_time,
    //    loadhandler::construct(next_id_, monitor_, mgr_, completion));


    mgr_->schedule(start_time + comp_time, completion);
    ++next_id_;
  }

  /// Define a manager.
  void
  procscheduler::set_manager(const boost::shared_ptr<eventmanager>& mgr)
  {
    mgr_ = mgr;
    monitor_->set_manager(mgr);
  }

} // end of namespace sstmac
