/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESSOR_PROCSCHEDULER_H
#define SSTMAC_PROCESSOR_PROCSCHEDULER_H

#include <sstmac/simulator/timestamp.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <sstmac/processor/processor_config.h>
#include <sstmac/processor/processor.h>

namespace sstmac
{

  // Forward declaration of important types.
  class instructions;
  // class processor;
  class eventmanager;
  class eventhandler;

  /**
   * A scheduler for processors
   */
  class procscheduler
  {

  private:

    class loadmonitor;
    class loadhandler;

    /// Id number for compute calls
    static int64_t next_id_;
    /// Self pointer
    boost::weak_ptr<procscheduler> self_;
    /// Processor model  
    boost::shared_ptr<processor> processor_;
    /// Event manager 
    boost::shared_ptr<eventmanager> mgr_;
    /// CPU load monitor
    boost::shared_ptr<loadmonitor> monitor_;

    void
    set_self(const boost::shared_ptr<procscheduler>& self);

  protected:

    procscheduler(int n_core);
    procscheduler(int n_core, const boost::shared_ptr<processor_config>& cfg);

  public:
    /// Create processor scheduler
    static boost::shared_ptr<procscheduler>
    construct(int n_core);

    static boost::shared_ptr<procscheduler>
    construct(int n_core, const boost::shared_ptr<processor_config>& cfg);

    /// Goodbye.
    ~procscheduler()
    {
    }

    /// Process a compute kernel.
    /// \param inst instructions object that is processed.
    void
    compute(const timestamp &start_time,
        const boost::shared_ptr<instructions>& inst,
        const boost::shared_ptr<eventhandler>& completion);

    /// Test whether this scheduler has a processor.
    bool
    has_processor() const
    {
      return (processor_ != 0);
    }

    /// Test whether this scheduler has a manger.
    bool
    has_manager() const
    {
      return (mgr_ != 0);
    }

    /// Define a manager.
    void
    set_manager(const boost::shared_ptr<eventmanager>& mgr);

    /// Access the manager.
    const boost::shared_ptr<eventmanager>
    get_manager() const
    {
      return mgr_;
    }

    boost::shared_ptr<processor_config>
    getProcCfg()
    {
      return processor_->getProcCfg();
    }

    boost::shared_ptr<processor>
    getProcessor()
    {
      return processor_;
    }

  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESSOR_PROCSCHEDULER_H
