/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <boost/shared_ptr.hpp>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <sstmac/processor/simpleprocessor.h>
#include <sstmac/processor/eventdata.h>
#include <sstmac/processor/simpleprocessor-config.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/processor/simpleinstruction.h>
#include <iostream>

namespace sstmac
{

  //
  // Constructor
  //
  simpleprocessor::simpleprocessor(
      const boost::shared_ptr<simpleprocessor_config>& cfg) : processor(cfg),
    config_(cfg)
  {


    procpower = powerstat::construct(powerstat::PROCESSOR);
    pstats = procstats::construct();

    procpower->trackStaticPower(config_->get_pwr_static());


  }

  //
  // Who am I?
  //
  void
  simpleprocessor::set_self(const boost::shared_ptr<simpleprocessor>& self)
  {
    if (self.get() != this)
      {
        throw valueerror("simpleprocessor::set_self:  'self' pointer does not "
          "refer to 'this' object.");
      }
    self_ = self;
  }

  //
  // Construct a new processor with the given average service times
  //
  boost::shared_ptr<simpleprocessor>
  simpleprocessor::construct(
      const boost::shared_ptr<simpleprocessor_config>& cfg)
  {
    boost::shared_ptr<simpleprocessor>
        retval(new simpleprocessor(cfg));
    retval->set_self(retval);
    return retval;
  }

  boost::shared_ptr<processor>
  simpleprocessor::clone() const
  {
    if (debugger::active())
      debug() << "simpleprocessor::clone()\n";
    boost::shared_ptr<simpleprocessor> retval(
        new simpleprocessor(config_));
    return retval;
  }

  timestamp
  simpleprocessor::compute(const boost::shared_ptr<instructions>& inst)
  {
    if (debugger::active())
      debug() << "simpleprocessor::compute(" << inst << ")\n";
    boost::shared_ptr<simpleinstruction> data = boost::dynamic_pointer_cast<
        simpleinstruction>(inst);


    procpower->trackEnergy(data->comptime.sec() * config_->get_per_sec());
    pstats->trackFlop(data->comptime.sec() * config_->get_fps());

   // std::cerr << "simpleproc: compute: returning time " << data->comptime.sec() << endl;
    return data->comptime;
  }




} // end of namespace sstmac
