/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


#include <exception>
#include <sstmac/simulator/allocationstrategy.h>
#include <sstmac/util/errors.h>
#include <iterator>

namespace sstmac {

  //////////////////////////////////////////////////////////////////////
  // allocationstrategy

  allocationstrategy::~allocationstrategy() throw()
  {
  }

  void
  allocationstrategy::check_request(int nnode_requested,
                                    std::vector<nodeid> &allocation,
                                    std::set<nodeid> &available,
                                    std::set<nodeid> &allocated)
  {
    if(nnode_requested <= 0
       || nnode_requested > available.size()) {
      std::cerr << "allocationstrategy:  Request for "
                << nnode_requested
                << " nodes when " << available.size()
                << " are available.\n";
      throw valueerror("allocationstrategy: Not enough nodes available.");
    }
  }

  //////////////////////////////////////////////////////////////////////
  // firstavailableallocation

  firstavailableallocation::firstavailableallocation()
  {
  }

  firstavailableallocation::~firstavailableallocation() throw()
  {
  }

  boost::shared_ptr<firstavailableallocation>
  firstavailableallocation::construct()
  {
    return boost::shared_ptr<firstavailableallocation>(new firstavailableallocation());
  }

  void
  firstavailableallocation::allocate(int nnode_requested,
                                     std::vector<nodeid> &allocation,
                                     std::set<nodeid> &available,
                                     std::set<nodeid> &allocated)
  {
    check_request(nnode_requested,allocation,available,allocated);

    allocation.resize(nnode_requested);
    for (int i=0; i<nnode_requested; i++) {
      std::set<nodeid>::const_iterator nextnode = available.begin();
      allocation[i] = *nextnode;
      allocated.insert(*nextnode);
      available.erase(nextnode);
    }
  }

  //////////////////////////////////////////////////////////////////////
  // randomallocation

  randomallocation::randomallocation(unsigned seed)
  {
    rng_ = RNG::SimpleCombo::construct(seed);
  }

  randomallocation::randomallocation
  (const boost::shared_ptr<RNG::UniformInteger> &rng)
  {
    rng_ = rng;
  }

  randomallocation::~randomallocation() throw()
  {
  }

  boost::shared_ptr<randomallocation>
  randomallocation::construct(unsigned seed)
  {
    return boost::shared_ptr<randomallocation>(new randomallocation(seed));
  }

  boost::shared_ptr<randomallocation>
  randomallocation::construct(const boost::shared_ptr<RNG::UniformInteger> &rng)
  {
    return boost::shared_ptr<randomallocation>(new randomallocation(rng));
  }

  void
  randomallocation::allocate(int nnode_requested,
                             std::vector<nodeid> &allocation,
                             std::set<nodeid> &available,
                             std::set<nodeid> &allocated)
  {
    check_request(nnode_requested,allocation,available,allocated);
    /*
    std::cout << "randomallocation::allocate:"
              << " nnode_requested = " << nnode_requested
              << " available.size() = " << available.size()
              << " allocated.size() = " << allocated.size()
              << std::endl;
    */
    allocation.resize(nnode_requested);
    std::vector<nodeid> availvec(available.size());
    std::copy(available.begin(), available.end(), availvec.begin());
    RNG::UniformInteger_functor rngf(rng_);
    std::random_shuffle(availvec.begin(), availvec.end(), rngf);
    for (int i=0; i<nnode_requested; i++) {
      nodeid node = availvec[i];
      allocation[i] = node;
      allocated.insert(node);
      available.erase(node);
    }
    // we only want to randomly pick nodes, not randomize the nodes
    // picked, so sort them
    std::sort(allocation.begin(), allocation.end());
  }
  
  //////////////////////////////////////////////////////////////////////
  // fixedallocation
  
  fixedallocation::fixedallocation(const std::vector<nodeid> &nodes):
  nodes_(nodes)
  {
  }
  
  fixedallocation::~fixedallocation() throw()
  {
  }
  
  boost::shared_ptr<fixedallocation>
  fixedallocation::construct(const std::vector<nodeid> &nodes)
  {
    return boost::shared_ptr<fixedallocation>(new fixedallocation(nodes));
  }

  void
  fixedallocation::allocate(int nnode_requested,
                            std::vector<nodeid> &allocation,
                            std::set<nodeid> &available,
                            std::set<nodeid> &allocated)
  {
    static std::string here("fixedallocation::allocate: ");
    check_request(nnode_requested,allocation,available,allocated);
    if (nnode_requested != nodes_.size()) {
      throw illformederror(here + "nnode_requested"
                           " does not match size of node vector");
    }
    for (int i=0; i<nnode_requested; i++) {
      std::set<nodeid>::iterator node = available.find(nodes_[i]);
      if (node == available.end()) {
        throw illformederror(here + "required node not available");
      }
      available.erase(node);
      allocated.insert(nodes_[i]);
    }
    allocation = nodes_;
  }

}
