/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_ALLOCATIONSTRATEGY_H
#define SSTMAC_SIMULATOR_ALLOCATIONSTRATEGY_H

#include <set>
#include <vector>
#include <sstmac/util/rng.h>
#include <sstmac/simulator/nodeid.h>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  /**
   * Strategy type for assigning processes to nodes in a parallel run.
   *
   */
  class allocationstrategy {
  protected:
    /// Check the arguments for calls to allocate and throw if they
    /// are bad.
    void check_request(int nnode,
                       std::vector<nodeid> &allocation,
                       std::set<nodeid> &available,
                       std::set<nodeid> &allocated);
  public:
    virtual ~allocationstrategy() throw();

    /// Get nodes.
    /// @param nnode number of nodes requested
    /// @param allocation returns the nodes that have been allocated
    /// @param available the available set of nodes
    /// @param allocated the allocated set of nodes
    /// Nodes are removed from available and put into allocated.
    /// If not enough nodes are available, then this will throw.
    virtual void allocate(int nnode,
                          std::vector<nodeid> &allocation,
                          std::set<nodeid> &available,
                          std::set<nodeid> &allocated) = 0;
  };

  class firstavailableallocation: public allocationstrategy {
  protected:
    firstavailableallocation();
  public:
    static boost::shared_ptr<firstavailableallocation> construct();
    virtual ~firstavailableallocation() throw();
    void allocate(int nnode_requested,
                  std::vector<nodeid> &allocation,
                  std::set<nodeid> &available,
                  std::set<nodeid> &allocated);
  };

  class randomallocation: public allocationstrategy {
    boost::shared_ptr<RNG::UniformInteger> rng_;
  protected:
    randomallocation(unsigned seed);
    randomallocation(const boost::shared_ptr<RNG::UniformInteger> &rng);
  public:
    static boost::shared_ptr<randomallocation> construct(unsigned seed=1337);
    static boost::shared_ptr<randomallocation>
    construct(const boost::shared_ptr<RNG::UniformInteger> &rng);
    virtual ~randomallocation() throw();
    void allocate(int nnode_requested,
                  std::vector<nodeid> &allocation,
                  std::set<nodeid> &available,
                  std::set<nodeid> &allocated);
  };
  
  class fixedallocation: public allocationstrategy {
    std::vector<nodeid> nodes_;
  protected:
    fixedallocation(const std::vector<nodeid> &nodes);
  public:
    /// Allocates a fixed set of nodes.
    /// @param nodes a vector of nodes to use for allocation
    static boost::shared_ptr<fixedallocation>
    construct(const std::vector<nodeid>&nodes);
    virtual ~fixedallocation() throw();
    void allocate(int nnode_requested,
                  std::vector<nodeid> &allocation,
                  std::set<nodeid> &available,
                  std::set<nodeid> &allocated);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_ALLOCATIONSTRATEGY_H
