/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/circuit.h>
#include <sstmac/simulator/circuit-edge.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/constantlatency.h>
#include <sstmac/network/transfer_time.h>
#include <sstmac/network/network.h>
#include <sstmac/network/walk.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac
{

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::lexical_cast;

  //
  // Hi.
  //
  circuit::circuit(const shared_ptr<latencymodel> &latency, bandwidth bw,
      const boost::shared_ptr<sstmac::network> &net,
      const boost::shared_ptr<power_config> &pwrcfg) :
    network_(net), latency_(latency), bandwidth_(bw), timing_model_(ORIGINAL)
  {
    if (debugger::active())
      debug() << "circuit(latency=" << latency << ", bandwidth=" << bandwidth_
          << ", net=" << net << ")\n";
    if (latency_ == 0)
      {
        throw nullerror("circuit(...):  Null latencymodel pointer.");
      }
    if (bw.bits_per_second() <= 0)
      throw valueerror("circuit(...):  Zero or negative bandwidth.");
    if (network_ == 0)
      {
        throw nullerror("circuit(...):  null network pointer.");
      }

    //std::cerr << "circuit:  Making room for "
    //          << network_->g().n_edges() << " edges\n";

    edge_.resize(network_->g().n_edges());

    for (size_t i = 0; i < edge_.size(); i++)
      {
        sstmac::edge_index ei(i);
        edge_[i].set_nchannel(network_->edge_width(ei));
      }

    pwr_config_ = pwrcfg;
    pwrstat = powerstat::construct(powerstat::NETWORK);
    pwrstat->trackStaticPower(pwr_config_->get_rtr_static() * network_->n());

    nstat = netstat::construct();

    /*
     if(debugger::active()) {
     size_t i = 0;
     sstmac::graph::edge_iterator it, end=network_->g().edges_end();
     if(debugger::active()) {
     for(it = network_->g().edges_begin(); it != end; ++it) {
     debug() << "circuit[" << this << "]::circuit:  edge " << &edge_.at(i)
     << " connects " << it->from() << " and " << it->to() << "\n";
     ++i;
     }
     }
     }
     */
  }

  //
  // Who am I?
  //
  void
  circuit::set_self(const shared_ptr<circuit> &self)
  {
    if (self.get() != this)
      {
        throw valueerror("circuit::set_self:  'self' pointer does not "
          "point to 'this' object.");
      }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<circuit>
  circuit::get_self()
  {
    return self_.lock();
  }

  //
  // New circuit network.
  //
  shared_ptr<circuit>
  circuit::construct(const timestamp latency, double bw,
      const shared_ptr<sstmac::network> &net)
  {

    return construct(latency, bandwidth(bw, bandwidth::Bps), net);
  }

  //
  // New circuit network
  //
  shared_ptr<circuit>
  circuit::construct(const timestamp latency, bandwidth bw,
      const shared_ptr<sstmac::network> &net)
  {

    if (latency < timestamp(0))
      {
        throw valueerror("circuit(...):  negative latency.");
      }
    shared_ptr<constantlatency> lat = constantlatency::construct(latency);
    return construct(lat, bw, net, power_config::construct());
  }

  //
  // New circuit with a user-defined latency model - no power config
  //
  shared_ptr<circuit>
  circuit::construct(const boost::shared_ptr<latencymodel> &lat, bandwidth bw,
      const shared_ptr<network> &net)
  {

    shared_ptr<circuit> retval(
        new circuit(lat, bw, net, power_config::construct()));
    retval->set_self(retval);
    return retval;
  }

  //
  // New circuit with a user-defined latency model.
  //
  shared_ptr<circuit>
  circuit::construct(const boost::shared_ptr<latencymodel> &lat, bandwidth bw,
      const shared_ptr<network> &net,
      const boost::shared_ptr<power_config> &pwrcfg)
  {

    shared_ptr<circuit> retval(new circuit(lat, bw, net, pwrcfg));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  circuit::~circuit() throw ()
  {
  }

  //
  // Required method from parent.
  //
  int64_t
  circuit::slots() const
  {
    return network_->n();
  }

  //
  // My event manager.
  //
  void
  circuit::set_eventmanager(const boost::shared_ptr<eventmanager>& mgr)
  {
    if (mgr == 0)
      {
        throw nullerror(
            "circuit::set_eventmanager:  null eventmanager pointer.");
      }
    mgr_ = mgr;
  }



  //
  // Send the given message at the given time.
  //
  void
  circuit::send(const timestamp &time, int64_t bytes, nodeid sender,
      nodeid recver, const shared_ptr<eventhandler>& sendhandler,
      const shared_ptr<eventhandler>& recvhandler)
  {


    std::string me("circuit");
    me += lexical_cast<std::string> (size_t(this)) + "::send";
    if (debugger::active())
      debug() << me << "(time=" << time << ", bytes=" << bytes << ", sender="
          << sender << ", recver=" << recver << ", sendhandler=" << sendhandler
          << ", recvhandler=" << recvhandler << ")\n";
    sstmac::walk the_walk;
    the_walk.ev().reserve(network_->diameter());
    network_->route(sender.id, recver.id, the_walk);

    if (debugger::active())
      debug() << me << the_walk.length() << " step walk from " << sender
          << " to " << recver << ": [";
    for (int i = 0; i < the_walk.length(); ++i)
      {
        if (debugger::active())
          debug() << the_walk.e(i).from() << " --> ";
        if (i > 0)
          {
            if (the_walk.e(i).from() != the_walk.e(i - 1).to())
              {
                if (debugger::active())
                  debug() << "ERROR:  Non-contiguous walk\n";
                abort();
              }
          }
      }


    pwrstat->trackEnergy(
        the_walk.length() * pwr_config_->get_rtr_hop_per_byte() * bytes);
    nstat->trackBytes(bytes);
    nstat->trackHops(the_walk.length());

    if (debugger::active())
      {
        if (the_walk.length() > 0)
          {
            debug() << the_walk.e(the_walk.length() - 1).to() << "]\n";
          }
        else
          {
            debug() << "]\n";
          }
      }

    // Timing model stuff shifted down from constructor to permit
    // noisy latency (and eventually bandwidth) models.
    timestamp start, stime, rtime, ttime;
    data payload(bytes, data::Bytes);
    switch (timing_model_)
      {
    case ORIGINAL:
      original_timing_model(payload, start, stime, rtime, ttime);
      break;
    case SPLIT:
      split_timing_model(payload, start, stime, rtime, ttime);
      break;
    default:
      throw valueerror("circuit::send:  Invalid timing model.");
      }
    // Ready to roll
    shared_ptr<transfer> trans = transfer::construct(this->get_self(), bytes,
        the_walk, stime, rtime, ttime, sendhandler, recvhandler);

    if (debugger::active())
      debug() << me << ":  Scheduling send for " << (time + start) << "\n";
    mgr_->schedule(time + start, trans);
  }

  void
  circuit::original_timing_model(data payload, timestamp &start_time,
      timestamp &send_time, timestamp &recv_time, timestamp &trans_time)
  {
    // Original timing model has no startup cost.
    start_time = timestamp(0);
    // send time is delayed by full latency.
    // The latency model decides what the latency is for any given transfer
    timestamp lat = latency_->latency(payload);
    send_time = lat + payload / bandwidth_;

    /// The model assumes that the sender and receiver finish at the same time.
    recv_time = send_time;

    /// Transfer begins immediately (although it should include a delay).
    trans_time = send_time;
  }

  void
  circuit::split_timing_model(data payload, timestamp &start_time,
      timestamp &send_time, timestamp &recv_time, timestamp &trans_time)
  {
    /// Split timing model divides the latency onto the sender and receiver.
    timestamp lat = latency_->latency(payload);
    timestamp halflat = lat / 2;

    // Start after half latency.
    start_time = halflat;
    // Transfer time is pure bandwidth contribution.
    trans_time = payload / bandwidth_;
    send_time = trans_time;
    // and receive time is after transfer and half latency.
    recv_time = trans_time + halflat;
  }

} // end of namespace sstmac
