/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/data.h>
#include <sstmac/util/errors.h>
#include <iostream>
#include <limits>
#include <math.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  //
  // The static mapping from string names to unit labels.
  //
  data::unitmap_t *data::unitmap_ = NULL;

  //
  // Initialize the static unit map.
  //
  void data::init_unitmap() {
    if(unitmap_ == NULL) {
      unitmap_ = new unitmap_t;
#define SSTMAC_ADDUNIT(VAL) (*unitmap_)[#VAL] =  VAL
      SSTMAC_ADDUNIT(b);      SSTMAC_ADDUNIT(bits);
      SSTMAC_ADDUNIT(Kb);     SSTMAC_ADDUNIT(Kib);
      SSTMAC_ADDUNIT(Mb);     SSTMAC_ADDUNIT(Mib);
      SSTMAC_ADDUNIT(Gb);     SSTMAC_ADDUNIT(Gib);
      SSTMAC_ADDUNIT(Tb);     SSTMAC_ADDUNIT(Tib);
      SSTMAC_ADDUNIT(Pb);     SSTMAC_ADDUNIT(Pib);
      SSTMAC_ADDUNIT(Eb);     SSTMAC_ADDUNIT(Eib);
      SSTMAC_ADDUNIT(B);      SSTMAC_ADDUNIT(Bytes);
      SSTMAC_ADDUNIT(KB);     SSTMAC_ADDUNIT(KiB);
      SSTMAC_ADDUNIT(MB);     SSTMAC_ADDUNIT(MiB);
      SSTMAC_ADDUNIT(GB);     SSTMAC_ADDUNIT(GiB);
      SSTMAC_ADDUNIT(TB);     SSTMAC_ADDUNIT(TiB);
      SSTMAC_ADDUNIT(PB);     SSTMAC_ADDUNIT(PiB);
      SSTMAC_ADDUNIT(EB);     SSTMAC_ADDUNIT(EiB);
#undef SSTMAC_ADDUNIT
    }
  }

  //
  // Convert unit enum to multiplier.
  //
  inline data::data_t conversion_factor(data::units unit) {
    static data::data_t *conv = NULL;
    if(conv == NULL) {
      conv = new data::data_t[data::EiB+1];
      for(int i = 0; i <= int(data::EiB); ++i) {
        conv[i] = 0;
      }
#ifndef SSTMAC_USE_GMPXX
      conv[data::b]     = 1ULL;
      conv[data::bits]  = conv[data::b];
      conv[data::Kb]    = 1000ULL;
      conv[data::Kib]   = 1024ULL;
      conv[data::Mb]    = 1000000ULL;
      conv[data::Mib]   = 1048576ULL;
      conv[data::Gb]    = 1000000000ULL;
      conv[data::Gib]   = 1073741824ULL;
      conv[data::Tb]    = 1000000000000ULL;
      conv[data::Tib]   = 1099511627776ULL;
      conv[data::Pb]    = 1000000000000000ULL;
      conv[data::Pib]   = 1125899906842624ULL;
      conv[data::Eb]    = 1000000000000000000ULL;
      conv[data::Eib]   = 1152921504606846976ULL;
      conv[data::B]     = 8ULL;
      conv[data::Bytes] = conv[data::B];
      conv[data::KB]    = 8000ULL;
      conv[data::KiB]   = 8192ULL;
      conv[data::MB]    = 8000000ULL;
      conv[data::MiB]   = 8388608ULL;
      conv[data::GB]    = 8000000000ULL;
      conv[data::GiB]   = 8589934592ULL;
      conv[data::TB]    = 8000000000000ULL;
      conv[data::TiB]   = 8796093022208ULL;
      conv[data::PB]    = 8000000000000000ULL;
      conv[data::PiB]   = 9007199254740992ULL;
      conv[data::EB]    = 8000000000000000000ULL;
      conv[data::EiB]   = 9223372036854775808ULL;
#else
      conv[data::b]     = "1";
      conv[data::bits]  = conv[data::b];
      conv[data::Kb]    = "1000";
      conv[data::Kib]   = "1024";
      conv[data::Mb]    = "1000000";
      conv[data::Mib]   = "1048576";
      conv[data::Gb]    = "1000000000";
      conv[data::Gib]   = "1073741824";
      conv[data::Tb]    = "1000000000000";
      conv[data::Tib]   = "1099511627776";
      conv[data::Pb]    = "1000000000000000";
      conv[data::Pib]   = "1125899906842624";
      conv[data::Eb]    = "1000000000000000000";
      conv[data::Eib]   = "1152921504606846976";
      conv[data::B]     = "8";
      conv[data::Bytes] = conv[data::B];
      conv[data::KB]    = "8000";
      conv[data::KiB]   = "8192";
      conv[data::MB]    = "8000000";
      conv[data::MiB]   = "8388608";
      conv[data::GB]    = "8000000000";
      conv[data::GiB]   = "8589934592";
      conv[data::TB]    = "8000000000000";
      conv[data::TiB]   = "8796093022208";
      conv[data::PB]    = "8000000000000000";
      conv[data::PiB]   = "9007199254740992";
      conv[data::EB]    = "8000000000000000000";
      conv[data::EiB]   = "9223372036854775808";
#endif
    };
    if(unit < 0 || unit > data::EiB)
      throw valueerror("data::units: Invalid unit (out of valid range).");
    data::data_t value = conv[unit];
    if(value == 0)
      throw valueerror("data::units: Invalid unit (uninitialized conversion).");
    return value;
  }

  //
  // Private initialization method used by constructors.
  //
  void data::set_size(double bw, units unit) {
#ifndef SSTMAC_USE_GMPXX
    static const double minsize = double(std::numeric_limits<data_t>::min());
    static const double maxsize = double(std::numeric_limits<data_t>::max());
    double scaled = nearbyint(bw * conversion_factor(unit));
    if(scaled < minsize || scaled > maxsize) {
      throw valueerror(std::string("data(...):  Requested data ") +
                       boost::lexical_cast<std::string>(scaled) + 
                       " bps is outside the valid range of " +
                       boost::lexical_cast<std::string>(minsize) + " to " +
                       boost::lexical_cast<std::string>(maxsize) +
                       " bits per second (inclusive).");
    }
    data_ = data_t(scaled);
#else
    data_ = data_t(bw) * conversion_factor(unit);
#endif
  }

  //
  // Default constructor.
  //
  data::data() {
    set_size(0, bits);
  }

  //
  // Build a data object with exactly the given number of bits.
  //
  data::data(const data_t &size) :
    data_(size)
  {}

  //
  // Build a new data object.
  //
  data::data(double bw, units unit) {
    set_size(bw, unit);
  }

  //
  // Alternate convenience way of defining the units (as a string).
  //
  data::data(double bw, const std::string &unit) {
    if(unitmap_ == NULL) {
      init_unitmap();
    }
    unitmap_t::const_iterator it = unitmap_->find(unit);
    if(it == unitmap_->end())
      throw valueerror(std::string("data(...):  Invalid unit string \"") +
                       unit + "\"");
    set_size(bw, it->second);
  }

  //
  // Return the data using the given units.
  //
  double data::get_data(units unit) const {
    if(unit < 0 || unit > EiB)
      throw valueerror("data::get_data: Invalid unit.");
#ifndef SSTMAC_USE_GMPXX
    return (data_ / double(conversion_factor(unit)));
#else
    return (data_.get_d() / conversion_factor(unit).get_d());
#endif
  }

  //
  // A convenience function to specify units using a string value.
  //
  double data::get_data(const std::string &unit) const {
    if(unitmap_ == NULL) {
      init_unitmap();
    }
    unitmap_t::const_iterator it = unitmap_->find(unit);
    if(it == unitmap_->end()) {
      throw valueerror(std::string("data::get_data: Invalid unit \"")+
                       unit + "\"");
    }
    return get_data(it->second);
  }

  //
  // Increase data.
  //
  data operator+(const data &a, const data &b) {
    const data::data_t abps = a.get_bits(), bbps = b.get_bits();
    data::data_t sum = abps + bbps;
#ifndef SSTMAC_USE_GMPXX
    // Detect overflow -- only tested against two's complement.
    if((sum < abps) || (sum < bbps)) {
      throw valueerror
        (std::string("operator+(const data&, const data&) :  ") +
         "Adding " + boost::lexical_cast<std::string>(a) + " and " +
         boost::lexical_cast<std::string>(b) + " results in overflow.");
    }
#endif
    return data(sum);
  }

  //
  // Decrease data.
  //
  data operator-(const data &a, const data &b) {
    const data::data_t abps = a.get_bits(), bbps = b.get_bits();
    data::data_t del = abps - bbps;
    // Detect overflow -- only tested against two's complement.
#ifndef SSTMAC_USE_GMPXX
    if((del > abps) || (del > bbps)) {
      throw valueerror
        (std::string("operator+(const data&, const data&) :  ") +
         "Subtracting " + boost::lexical_cast<std::string>(a) + " and " +
         boost::lexical_cast<std::string>(b) + " results in overflow.");
    }
#endif
    return data(del);
  }

  //
  // Scale data.
  //
  data operator*(const data &t, double scaling) {
#ifndef SSTMAC_USE_GMPXX
    // Let's chicken out and just go with floating point ops.
    double scaled = nearbyint(t.get_bits() * scaling);
    if(scaled < double(std::numeric_limits<data::data_t>::min()) ||
       scaled > double(std::numeric_limits<data::data_t>::max())) {
      throw valueerror(std::string("operator*(const data&, double):  ") +
                       "Multiplying " + boost::lexical_cast<std::string>(t) +
                       " with " + boost::lexical_cast<std::string>(scaling) +
                       " results in value outside allowable data range.");
    }
#else
    // Stupid gmpxx promotion rules prioritize integer over float
    // (and operator* is not overloaded for floats).
    data::data_t scaled = t.get_bits().get_d() * scaling;
#endif
    return data(scaled);
  }

  //
  // Scale data.
  //
  data operator*(double scaling, const data &t) {
    return (t * scaling);
  }

  //
  // Scale data.
  //
  data operator/(const data &t, double scaling) {
#ifndef SSTMAC_USE_GMPXX
    double scaled = nearbyint(t.get_bits() / scaling);
    if(scaled < double(std::numeric_limits<data::data_t>::min()) ||
       scaled > double(std::numeric_limits<data::data_t>::max())) {
      throw valueerror(std::string("operator*(const data&, double):  ") +
                       "Dividing " + boost::lexical_cast<std::string>(t) +
                       " with " + boost::lexical_cast<std::string>(scaling) +
                       " results in value outside allowable data range.");
    }
#else
    // again, a huge potential trap with GMP promotion rules.
    data::data_t scaled = t.get_bits().get_d() / scaling;
#endif
    return data(scaled);
  }

  //
  // Very, very convenient.
  //
  std::ostream& operator<<(std::ostream &os, const data &t) {
    os << "data(" << t.get_bits() << "bps)";
    return os;
  }

} // end of namespace sstmac

