/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/eventmap.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  //
  // Hello.
  //
  eventmap::eventmap(const timestamp &tv) :
    running_(false)
  {
    set_now(tv);
  }

  //
  // Hello.
  //
  shared_ptr<eventmap> eventmap::construct(const timestamp &tv) {
    shared_ptr<eventmap> retval(new eventmap(tv));
    return retval;
  }

  //
  // Goodbye.
  //
  eventmap::~eventmap() throw() {
    if(running_) {
      if(debugger::active()) debug() << "Warning: eventmap (eventmanager) exiting while running\n.";
    }
  }

  //
  // Clear all events and set time back to a zero of your choice.
  //
  void eventmap::clear(const timestamp &zero_time) {
    if(running_) {
      throw illformederror("eventmap::clear:  Event manager is running.");
    }
    queue_.clear();
    set_now(zero_time);
  }

  //
  // Run the eventmanager.
  //
  void eventmap::run() {
    if(running_) {
      throw illformederror("eventmap::run:  Event manager already running.");
    }
    running_ = true;
    while(! queue_.empty()) {
      queue_t::iterator it = queue_.begin();
      set_now(it->first);
      shared_ptr<eventhandler> ev = it->second;
      queue_.erase(it);
      if(debugger::active())
        debug() << "eventmap[" << this << "]::run:  Processing event " << ev
                << " of type " << typeid(*ev).name()
                << " at time " << now() << "\n";
      ev->handle(now());
      if(debugger::active()) debug() << "eventmap[" << this << "]::run:  Event " << ev 
              << " complete, " << queue_.size() << " events pending\n";
    }
    if(debugger::active()) debug() << "eventmap[" << this << "]::run:  All out of events.\n";
    running_ = false;
  }

  //
  // Set off the given eventhandler at the given time.
  //
  void eventmap::schedule(const timestamp &start_time,
                          const shared_ptr<eventhandler> &handler)
  {
    if(start_time < now()) {
      std::cerr << "start_time " << start_time << " is before current time "
                << now() << "\n";
      throw valueerror("eventmap::schedule:  Start time before current time.");
    }
    if(handler == 0) {
      throw nullerror("eventmap::schedule:  Null eventhandler pointer.");
    }
    queue_.insert(std::make_pair(start_time, handler));
    if(debugger::active())
      debug() << "eventmap[" << this << "]::schedule:  Scheduling event "
              << handler << " of type " << typeid(*handler).name() << " for "
              << start_time << ", "  << queue_.size() << " events pending.\n";
  }

} // end of namespace sstmac
