/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/kernel.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iostream>

namespace sstmac {

  //
  // Hello.
  //
  kernel::kernel() :
    done_(false)
  {}

  //
  // Virtual destructor is essential.
  //
  kernel::~kernel() throw() {
    if(! complist_.empty()) {
      std::cerr << "kernel going out of scope with pending event handlers.\n";
      timestamp faulty(-1);
      this->complete(faulty);
    }
  }

  //
  // Test whether this kernel has been completed.
  //
  bool kernel::is_complete() const {
    return done_;
  }

  //
  // Notify an event handler when we complete.
  //
  void kernel::add_listener(const boost::shared_ptr<eventhandler> &handler) {
    if(this->is_complete()) {
      throw illformederror("kernel::add_listener: kernel is already complete.");
    }
    complist_.push_back(handler);
  }

  //
  // A listener is unsubscribing.
  //
  void kernel::erase_listener(const boost::shared_ptr<eventhandler> &handler) {
    complist_t::iterator it =
      std::find(complist_.begin(), complist_.end(), handler);
    if(it != complist_.end()) {
      complist_.erase(it);
    }
  }

  //
  // Mark this kernel as complete.
  //
  void kernel::complete(const timestamp &current_time) {
    if(debugger::active())
      debug() << "kernel[" << this << "]::complete(" << current_time << ")\n";
    complist_t::iterator it;
    for(it = complist_.begin(); it != complist_.end(); ++it) {
      if(debugger::active())
        debug() << "kernel[" << this << "]::complete:  Routing completion to "
                << (*it) << "\n";
      (*it)->handle(current_time);
    }
    complist_.clear();
  }

} // end of namespace sstmac
