/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/manager-env.h>
#include <sstmac/simulator/nodelist.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/kernelhandler.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Construct a delegate.
  //
  manager::env::env(const boost::shared_ptr<manager> &my_mgr,
                    const boost::shared_ptr<launcher> &my_job,
                    const boost::shared_ptr<nodelist> &my_nlist,
                    const std::vector<nodeid>& my_nodes,
                    ppid my_ppid) :
    manager_(my_mgr), launcher_(my_job), nodelist_(my_nlist),
    tasklist_(tasklist::construct(my_nodes)), ppid_(my_ppid)
  {}

  //
  // Who am I?
  //
  void manager::env::set_self(const shared_ptr<env> &self) {
    if(self.get() != this) {
      throw valueerror("manager::env::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<manager::env> manager::env::get_self() {
    return self_.lock();
  }

  //
  // Construct an object.
  //
  shared_ptr<manager::env>
  manager::env::construct(const boost::shared_ptr<manager> &my_mgr,
                          const boost::shared_ptr<launcher> &my_job,
                          const boost::shared_ptr<nodelist> &my_nlist,
                          const std::vector<nodeid>& my_nod,
                          ppid my_ppid)
  {
    shared_ptr<env> retval(new env(my_mgr, my_job, my_nlist, my_nod, my_ppid));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  manager::env::~env() throw() {
  }
    
  //
  // Get the parallel process id associated with this run.
  //
  ppid manager::env::get_ppid() const {
    return ppid_;
  }

  //
  // Get the tasklist.
  //
  shared_ptr<tasklist> manager::env::get_tasklist() const {
    return tasklist_;
  }

  //
  // Get a node delegate.
  //
  shared_ptr<nodeinterface> manager::env::get(const taskid &id) const {
    nodeid nid = tasklist_->node(id);
    return nodeinterface::construct(nodelist_->get(nid), ppid_, id);
  }
  
  shared_ptr<node> manager::env::get_node(const taskid &id) const {
    nodeid nid = tasklist_->node(id);
    return nodelist_->get(nid);
  }

  //
  // Process exiting.
  //
  void manager::env::exit() {
    shared_ptr<manager> mgr = manager_.lock();
    if(mgr) {
      mgr->launcher_exiting(launcher_.lock(), tasklist_->nodes());
    }
    manager_ = weak_ptr<manager>();
  }

} // end of namespace sstmac
