/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_MANAGER_ENV_H
#define SSTMAC_SIMULATOR_MANAGER_ENV_H

#include <sstmac/simulator/manager.h>
#include <sstmac/simulator/environment.h>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Important types.
  class nodelist;
  class tasklist;
  class launcher;

  /**
   * The nested delegate class used to permit communications between processes
   * and the manager and to handle parallel process ids and other
   * management-related activities.
   */
  class manager::env : public environment {
  private:
    /// This object.
    boost::weak_ptr<manager::env> self_;
    
    /// The manager that owns this delegate.
    /// The only time this pointer gets used is to post an exit call.
    boost::weak_ptr<manager> manager_;

    /// The launcher associated with this environment.
    /// The only time this pointer gets used is to post an exit call.
    boost::weak_ptr<launcher> launcher_;

    /// The (shared) list of nodes in the simulation environment,
    /// maps node indices onto node pointers.
    boost::shared_ptr<nodelist> nodelist_;
    
    /// The (shared) list matching task indices with node indices.
    boost::shared_ptr<tasklist> tasklist_;
    
    /// The parallel process id assigned to this run.
    /// This id tags all kernels and communications
    /// associated with the process.
    ppid ppid_;
    
  protected:
    /// Construct a delegate and assign it a manager, process id,
    /// and group of nodes.
    env(const boost::shared_ptr<manager> &my_mgr,
        const boost::shared_ptr<launcher> &my_job,
        const boost::shared_ptr<nodelist> &my_nlist,
        const std::vector<nodeid>& my_nodes,
        ppid my_ppid);

    /// Who am I?
    void set_self(const boost::shared_ptr<env> &self);
    
    /// This is who I am.
    boost::shared_ptr<env> get_self();

  public:
    /// Construct a new environment.
    static boost::shared_ptr<env>
    construct(const boost::shared_ptr<manager> &my_mgr,
              const boost::shared_ptr<launcher> &my_job,
              const boost::shared_ptr<nodelist> &my_nlist,
              const std::vector<nodeid>& my_nodes,
              ppid my_ppid);
    
    /// Goodbye.
    virtual ~env() throw();
    
    /// Get the parallel process id associated with this run.
    ppid get_ppid() const;
    
    /// Get the tasklist.
    virtual boost::shared_ptr<tasklist> get_tasklist() const;

    /// Get a delegate for the given node.
    boost::shared_ptr<nodeinterface> get(const taskid &id) const;
    
    /// Get the node corresponding to the given task.
    /// \param tid is the task id
    /// \return the node object for the task
    boost::shared_ptr<node> get_node(const taskid &tid) const;

    /// Process exiting.
    void exit();
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_MANAGER_ENV_H
