/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/manager-env.h>
#include <sstmac/simulator/manager.h>
#include <sstmac/simulator/nodelist.h>
#include <sstmac/simulator/node.h>
#include <sstmac/simulator/simplenic.h>
#include <sstmac/simulator/allocationstrategy.h>
#include <sstmac/simulator/launcher.h>
#include <sstmac/simulator/eventmap.h>
#include <sstmac/simulator/interconnect.h>
#include <sstmac/util/counter.h>
#include <sstmac/simulator/blockindexing.h>
#include <sstmac/simulator/launchhandler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <iterator>
#include <cstdlib>

namespace sstmac
{

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Private method to indicate that a launcher is exiting.
  //
  void
  manager::launcher_exiting(const boost::shared_ptr<launcher> &launch,
      const std::vector<nodeid> &release)
  {
    launchers_t::iterator it = launchers_.find(launch);
    if (it != launchers_.end())
      {
        std::vector<nodeid>::const_iterator nid;
        for (nid = release.begin(); nid != release.end(); ++nid)
          {
            allocated_.erase(*nid);
            available_.insert(*nid);
          }
        launchers_.erase(it);
      }
  }

  //
  // Default constructor.
  //
  manager::manager() :
    nodes_(nodelist::construct()), running_(false), next_ppid_(0),
        default_indexing_(blockindexing::construct()),
        default_allocation_(firstavailableallocation::construct()),
        stats(statscollector::construct("stats.log"))
  {
    if (debugger::active())
      debug() << "manager::manager()\n";
  }

  //
  // Set self.
  //
  void
  manager::set_self(const shared_ptr<manager> &self)
  {
    if (self.get() != this)
      {
        throw valueerror("manager::set_self:  'self' pointer does not point "
          "to 'this' object.");
      }
    self_ = self;
  }

  //
  // Return self.
  //
  shared_ptr<manager>
  manager::get_self()
  {
    return self_.lock();
  }

  //
  // Construct a manager.
  //
  shared_ptr<manager>
  manager::construct()
  {
    shared_ptr<manager> retval(new manager);
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  manager::~manager() throw ()
  {
    if (debugger::active())
      debug() << "manager::~manager()\n";
    if (this->running_)
      {
        std::cerr
            << "FATAL:  manager going out of scope while still running.\n";
        abort();
      }
    if (!launchers_.empty())
      {
        std::cerr << "WARNING: manager exiting with active launchers.\n"
            << "         This is normally caused by application deadlock.\n";
        launchers_t::const_iterator it, end = launchers_.end();
        for (it = launchers_.begin(); it != end; ++it)
          {
            if (!(*it)->completed())
              {
                std::cerr << "    Launcher " << (*it)->description()
                    << " is still reported as active\n";
              }
          }
        launchers_.clear();
        nodes_ = shared_ptr<nodelist> ();
      }
  }

  //
  // Define an event manager.
  //
  void
  manager::set_eventmanager(const shared_ptr<eventmanager> &mgr)
  {
    if (debugger::active())
      debug() << "manager::set_eventmanager(" << mgr << ")\n";
    if (running_)
      {
        throw illformederror(
            "manager::set_eventmanager: Simulation is running.");
      }
    eventmanager_ = mgr;
  }

  //
  // Access the event manager.
  //
  shared_ptr<eventmanager>
  manager::get_eventmanager() const
  {
    return eventmanager_;
  }

  //
  // Define a network.
  //
  void
  manager::set_interconnect(const shared_ptr<interconnect> &net)
  {
    if (debugger::active())
      debug() << "manager::set_interconnect(" << net << ")\n";
    if (running_)
      {
        throw illformederror(
            "manager::set_interconnect: Simulation is running.");
      }
    interconnect_ = net;

    stats->addStats(interconnect_->getStats());

  }

  //
  // Access the interconnect.
  //
  shared_ptr<const interconnect>
  manager::get_interconnect() const
  {
    return interconnect_;
  }

  //
  // Number of nodes currently defined.
  //
  int64_t
  manager::nodes() const
  {
    if (debugger::active())
      debug() << "manager::nodes() const\n";
    return nodes_->size();
  }

  //
  // Number of open network slots.
  //
  int64_t
  manager::open_slots() const
  {
    if (interconnect_ == 0)
      {
        throw nullerror("manager::open_slots:  null interconnect pointer.");
      }
    if (interconnect_->slots() <= 0)
      {
        return -1;
      }
    else
      {
        return (interconnect_->slots() - this->nodes());
      }
  }

  boost::shared_ptr<node>
  manager::get_node(nodeid id) const
  {
    return nodes_->get(id);
  }

  //
  // Add one or more copies of the given node type to the manager.
  //
  int64_t
  manager::add_nodes(const shared_ptr<const node> &type, int64_t count)
  {
    if (debugger::active())
      debug() << "manager::add_nodes(" << type << ", " << count << ")\n";
    if (interconnect_ == 0)
      {
        throw nullerror("manager::add_nodes:  No interconnect defined yet.");
      }
    if (type == 0)
      {
        throw nullerror("manager::add_nodes:  Null node pointer\n");
      }
    if ((interconnect_->slots() > 0) && (count > int(
        interconnect_->slots() - this->nodes())))
      {
        throw valueerror("manager::add_nodes:  Insufficient slots available.");
      }
    if (count < 0)
      {
        if (interconnect_->slots() <= 0)
          {
            std::cerr << "manager::add_nodes:  Simulator has unlimited size, "
                << "so an unlimited node count is not possible\n";
            throw illformederror("manager::add_nodes:  Need infinite memory.");
          }
        count = check_cast<int>(interconnect_->slots()) - this->nodes();
      }
    for (int i = 0; i < count; ++i)
      {
        shared_ptr<node> newnode = type->clone();

        shared_ptr<has_stats> sget = boost::dynamic_pointer_cast<has_stats>(
            newnode);
        stats->addStats(sget->getStats());

        nodeid id = nodes_->add(newnode);
        available_.insert(id);
      }
    return this->open_slots();
  }

  //
  // Accept a parallel job to be instantiated on the given number of nodes
  // at or after the given start time.
  //
  void
  manager::add_launcher(const timestamp &start_time,
      const shared_ptr<launcher> &the_launcher, int nodereq, int ppn,
      const shared_ptr<indexstrategy> &istrategy,
      const shared_ptr<allocationstrategy> &astrategy, int nproc)
  {
    if (debugger::active())
      debug() << "manager::add_launcher(" << start_time << ", " << the_launcher
          << ", " << nodereq << ", " << ppn << ")\n";
    if (running_)
      {
        throw illformederror("manager::add_launcher:  Simulation is running.");
      }
    if (interconnect_ == 0)
      {
        throw nullerror("manager::add_launcher:  No interconnect defined yet.");
      }
    if (the_launcher == 0)
      {
        throw nullerror("manager::add_launcher:  Null launcher pointer.");
      }

    // If nproc was not specified, then each node gets ppn procs.
    if (nproc == -1)
      nproc = nodereq * ppn;

    //
    // Assign a parallel launcher id to this job.
    ppid pid = next_ppid_++;
    //
    // Grab nodes for this job
    std::vector<nodeid> nodelist;
    if (astrategy != 0)
      {
        astrategy->allocate(nodereq, nodelist, available_, allocated_);
      }
    else
      {
        default_allocation_->allocate(nodereq, nodelist, available_, allocated_);
      }
    //
    // Expand the node list to have the desired number of launchers per node.
    std::vector<nodeid> allocation;
    if (istrategy != 0)
      {
        istrategy->allocate(nodelist, ppn, allocation, nproc);
      }
    else
      {
        default_indexing_->allocate(nodelist, ppn, allocation, nproc);
      }
    //
    // Build a delegate for this run.
    shared_ptr<env> gate(
        env::construct(self_.lock(), the_launcher, nodes_, allocation, pid));
    //
    // Add this launcher to the launcher list.
    launchers_.insert(the_launcher);
    //
    // Schedule for this launcher to be fired up at the desired time.
    shared_ptr<eventhandler> hand(launchhandler::construct(the_launcher, gate));
    startup_.push_back(std::make_pair(start_time, hand));
  }

  //
  // Setup the simulation.
  //
  void
  manager::setup()
  {
    if (debugger::active())
      debug() << "manager::setup: entered\n";
    if (interconnect_ == 0)
      {
        throw nullerror("manager::setup:  Null interconnect pointer.");
      }
    if (running_)
      {
        throw illformederror("manager::setup:  Simulation already running.");
      }
    // Initial setup.
    if (eventmanager_ == 0)
      {
        eventmanager_ = eventmap::construct();
      }
    // Make sure the network simulator knows about our event manager.
    interconnect_->set_eventmanager(eventmanager_);
    // Initialize the nodes.
    for (int i = 0; i < nodes_->size(); ++i)
      {
        nodeid id(i);
        shared_ptr<node> the_node = nodes_->get(id);
        the_node->initialize(id, get_self());
        the_node->add_networkinterface(simplenic::construct(the_node,
                                                            interconnect_));
      }

    // Schedule job startups.
    for (startup_t::iterator it = startup_.begin(); it != startup_.end(); ++it)
      {
        eventmanager_->schedule(it->first, it->second);
      }
    startup_.clear();
    running_ = true;
  }

  //
  // Start the simulation.
  //
  timestamp
  manager::run()
  {

    // setup the simulation before starting 
    setup();

    if (debugger::active())
      debug() << "manager::run: calling eventmanager::run\n";

    eventmanager_->run();

    // Now call done routine to end simulation and print Stats.
    done();

    return eventmanager_->now();

  }

  void
  manager::done()
  {
    if (debugger::active())
      debug() << "manager::done: entered\n";

    running_ = false;
    if (allocated_.size() > 0)
      {
        std::cout << "manager no longer running but " << allocated_.size()
		  << " nodes are still allocated--deallocating all nodes."
		  << std::endl;
        std::copy(allocated_.begin(), allocated_.end(),
		  std::inserter(available_, available_.begin()));
        allocated_.clear();
      }

    stats->printStats(eventmanager_->now());
  }

} // end of namespace sstmac
