/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_MANAGER_H
#define SSTMAC_SIMULATOR_MANAGER_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/ppid.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <vector>
#include <set>
#include <sstmac/simulator/statscollector.h>

namespace sstmac
{

  // Forward declaration of relevant types.
  class node;
  class nodelist;
  class launcher;
  class kernel;
  class interconnect;
  class eventmanager;
  class eventhandler;
  class indexstrategy;
  class allocationstrategy;

  /**
   * The owner of a discrete-event network simulation.
   * 
   * All time progression is handled by this object
   * and messages between nodes are managed here as well.
   */
  class manager : virtual public nocopy
  {
  private:
    /// The nested specialization of the environment type.
    /// One of these gets handed to each launcher started under this manager.
    class env;

    /// The reference counted version of the 'this' pointer.
    boost::weak_ptr<manager> self_;

    /// The event manager.
    boost::shared_ptr<eventmanager> eventmanager_;

    /// The network simulator.
    boost::shared_ptr<interconnect> interconnect_;

    /// The node objects.
    boost::shared_ptr<nodelist> nodes_;

    /// Keep track of parallel processes.
    typedef std::set<boost::shared_ptr<launcher> > launchers_t;
    launchers_t launchers_;

    /// The set of nodes that are available.
    std::set<sstmac::nodeid> available_;

    /// The set of nodes that are allocated.
    std::set<sstmac::nodeid> allocated_;

    /// Monitor whether the simulator is currently running.
    bool running_;

    /// Next parallel process id.
    ppid next_ppid_;

    /// Default allocation strategy.
    boost::shared_ptr<allocationstrategy> default_allocation_;

    /// Default index strategy.
    boost::shared_ptr<indexstrategy> default_indexing_;

    /// Power manager, collects power stats.
    boost::shared_ptr<statscollector> stats;

    /// The processes to be fired up at startup.
    typedef boost::shared_ptr<eventhandler> event_t;
    typedef std::vector<std::pair<timestamp, event_t> > startup_t;
    startup_t startup_;

    /// Private method to signal that a launcher has called exit.
    void
    launcher_exiting(const boost::shared_ptr<launcher> &,
        const std::vector<nodeid> &released_nodes);

  protected:
    /// Default constructor.
    /// By default the manager has no nodes or network; it can only
    /// be run after one or more nodes have been added.
    manager();

    /// Who am I?
    void
    set_self(const boost::shared_ptr<manager> &self);

    /// This is who I am.
    boost::shared_ptr<manager>
    get_self();

  public:
    /// Construct a new manager.
    static boost::shared_ptr<manager>
    construct();

    /// Goodbye.
    virtual
    ~manager() throw ();

    /// Define an event manager for the simulator.
    /// Defaults to eventmap.
    void
    set_eventmanager(const boost::shared_ptr<eventmanager> &mgr);

    /// Access the event manager.
    boost::shared_ptr<eventmanager>
    get_eventmanager() const;



    /// Define a interconnect for this manager.
    void
    set_interconnect(const boost::shared_ptr<interconnect> &net);

    /// Access the simulator used by this manager.
    boost::shared_ptr<const interconnect>
    get_interconnect() const;

    /// Access the number of nodes added to date.
    /// This can be any value <= get_simulator()->size().
    int64_t
    nodes() const;
    
    /// Get the node corresponding to nodeid.
    /// \param id is the nodeid of the desired node.
    /// \return the node object correspondint to nodeid.
    boost::shared_ptr<node> get_node(nodeid id) const;

    /// Find out how many slots in the network are free (after subtracting
    /// nodes added to date).  Return -1 if an unlimited
    /// number of slots are free.
    int64_t
    open_slots() const;

    /// Add one or more copies of the given node type to the manager.
    /// sequentially numbered starting at this->nodes().
    /// \param type The type of nodes added (replicated using node::clone()).
    /// \param count The number of instances of this type to add.
    ///        If count < 0, fill all available network connections with
    ///        this node type.
    /// \return the number of unfilled connections on the network.
    /// \throw nullerror if simulator has not been defined.
    /// \throw nullerror if type is a NULL pointer.
    /// \throw valueerror if this->nodes() is less than count.
    int64_t
    add_nodes(const boost::shared_ptr<const node> &type, int64_t count = -1);

    /// Schedule the launch of a parallel job on the given number of nodes
    /// at or after the given start time.
    /// The launcher will be given a list of nodes to run on (via the
    /// launcher::start method).  The launcher is free to set up shared
    /// resources between tasks running on a shared physical computer,
    /// but should not assume that all virtual nodes are running on the
    /// same physical computer.
    ///
    /// At startup, the launcher will be given a manager::delegate object
    /// that permits calls to add_launcher, schedule_trigger, schedule_send,
    /// and complete_launcher (these methods are not publically accessible).
    ///
    /// Please try to resist the urge to perform out-of-band communications
    /// directly between nodes, since this will harm future attempts to
    /// parallelize the simulator.
    /// \param start_time  the time at which the job is to be started (>= 0).
    /// \param launcher the object that gets run at the given time.
    ///        Note that this object is not cloned -- at the end of the run
    ///        it will reflect any changes performed during the execution.
    /// \param nodes the number of nodes for the simulation.
    /// \param ppn the number of launchers per node.
    /// \param is the strategy for assigning node indices.  Defaults to
    ///        blockindexing if no index strategy is specified.
    /// \param as the strategy for allocating nodes.  Defaults to
    ///        firstavailableallocationstrategy.
    /// \param nproc optional parameter giving the total number of processes. If nproc is not given or -1, then nodes*ppn processes are allocated.
    /// \throw timeerror if start_time is less than zero or out of range.
    /// \throw valueerror if nodes <= 0 or nodes > this->nodes().
    /// \throw valueerror if ppn <= 0.
    void
    add_launcher(
        const timestamp &start_time,
        const boost::shared_ptr<launcher> &the_launcher,
        int nodes,
        int ppn,
        const boost::shared_ptr<indexstrategy> &is = boost::shared_ptr<
            indexstrategy>(),
        const boost::shared_ptr<allocationstrategy> &as = boost::shared_ptr<
            allocationstrategy>(), int nproc = -1);

    // NOTES:
    //     What should we do about launchers added after the simulation
    //     has started with a start-time in the future (when more or fewer
    //     nodes may be available)?  How much of a "queueing system" 
    //     do we want to work in here?

    // TODO:
    // 1)  Add a way to specify which nodes get used for a given job
    //     (to facilitate simulating non-uniform node/network topologies).
    // 2)  Make it easier to name (or label) nodes, either individually by
    //     class (same reasoning as above).
    // 3)  Decide on a more final strategy for handling tiered topologies
    //     such as systems with GPGPUs or accelerators (e.g. RoadRunner).

    /// Setup the simulation.
    /// Initializes eventmanager, interconnect, nodes, and scheduler.
    void setup();

    /// Run the simulation.
    /// \return the time at the end of the simulation.
    /// \throw illformederror if the simulation cannot be run
    /// (e.g. if no nodes have been added).
    timestamp run();

    /// End the simulation.
    /// Does error checking and prints statistics.
    void done();

  };
} // end of namespace sstmac.

#endif // ! SSTMAC_SIMULATOR_MANAGER_H
