/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_NETWORKINTERFACE_H
#define SSTMAC_SIMULATOR_NETWORKINTERFACE_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/process/server.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/simulator/ppid.h>
#include <boost/shared_ptr.hpp>
#include <map>

namespace sstmac {

  // Important types.
  class networkdata;
  class eventhandler;

  /**
   * A networkinterface is a delegate between a node and a server module.
   * This object helps ornament network operations with information about
   * the process (ppid) involved.
   */
  class networkinterface {
    
    /// The network servers for parallel processes running on this node.
    typedef std::map<ppid, boost::shared_ptr<server> > servermap_t;
    servermap_t servers_;

  public:
    /// Goodbye.
    virtual ~networkinterface();
    
    /// Registers a server for the given ppid.
    /// \param ppidarg is the ppid for the new server
    /// \param srvr is the new server
    /// \throws illformederror if a server already exists for the ppid
    virtual void
    register_server(ppid caller,
                    const boost::shared_ptr<server> &srvr);
    
    /// Unregisters a server for the given ppid.
    /// \param ppidarg is the ppid for the new server
    /// \throws nullerror if a server does not already exist for the ppid
    virtual void
    unregister_server(ppid caller);
    
    /// Test whether the given ppid has defined a server on this node.
    /// \param caller is the ppid for the sought-after server
    /// \return true if and only if a server is registered for caller
    virtual bool
    has_server(ppid caller) const;
    
    /// Get the network server for the given ppid.
    /// \param caller is the desired ppid
    /// \throws nullerror no server is registered for the ppid
    virtual boost::shared_ptr<server>
    get_server(ppid caller);
    
    /// Schedule a send operation to happen at the given time.
    /// \param start_time is the time when the send is started.
    /// \param ppidarg is the parallel process id of the source and target.
    /// \param recver is the global index of the receiver node.
    /// \param payload is the data sent over the network
    /// \param completion is the eventhandler signaled when the send is complete
    virtual void send(const timestamp &start_time,
                      ppid ppidarg,
                      nodeid recver,
                      const boost::shared_ptr<networkdata> &payload,
                      const boost::shared_ptr<eventhandler> &completion) = 0;
    
    /// Incoming network transfer.
    /// Originates from a network simulator and is forwarded to the
    /// appropriate server.
    /// \param start_time is the time of message arrival
    /// \param caller is the ppid of the target server
    /// \param sender is the id of the sending node
    /// \param load is the network data
    /// \throws nullerror if the server is not registered
    virtual void
    incoming_message(const timestamp &start_time, ppid caller, nodeid sender,
                     const boost::shared_ptr<networkdata> &load);
    
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_SIMULATOR_NETWORKINTERFACE_H
