/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_NODE_H
#define SSTMAC_SIMULATOR_NODE_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/ppid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/simulator/powermanager.h>
#include <sstmac/processor/processor.h>
#include <boost/shared_ptr.hpp>
#include <sstmac/simulator/has_stats.h>
#include <sstmac/simulator/manager.h>

namespace sstmac
{

  // Forward declarations of relevant types.
  class server;
  class eventmanager;
  class eventhandler;
  class networkdata;
  class instructions;
  class nodeinterface;
  class networkinterface;

  /**
   * An interface for "physical" nodes -- may eventually be decorated
   * with additional functionality (OS/cpu modeling or whatever).
   */
  class node : public has_stats
  {

  protected:
    
    /// The owner of this object.
    boost::weak_ptr<manager> parent_;
    
    /// The id for this node.
    nodeid nodeid_;
    
    /// The network interfaces installed in this node.
    std::vector<boost::shared_ptr<networkinterface> > nics_;
    
    boost::shared_ptr<powermanager> pwr_mgr_;
    
    /// The initialize member in derived classes must call this
    /// to set the nodeid of this object.
    /// \param id is this node's id.
    void set_nodeid(const nodeid &id) { nodeid_ = id; }
  public:
    /// Goodbye.
    virtual
    ~node()
    {
    }

    /// Get a new node of the same type as the current one.
    virtual boost::shared_ptr<node>
    clone() const = 0;

    /// Assign a node index and event call interface for this node.
    /// \throw nullerror if ev is null.
    virtual void
    initialize(nodeid my_id, const boost::shared_ptr<manager> &parent);
    
    /// Returns the number of network interface objects.
    size_t n_networkinterface() const { return nics_.size(); }
    
    /// Add a network interface to this node.
    void add_networkinterface(const boost::shared_ptr<networkinterface> &nic);
    
    /// Find the networkinterface on which the server is running.
    /// \param ppidarg the parallel process id of the server
    /// \return the networkinterface object for ppidarg
    /// \throws nullerror if no server is registered for ppidarg
    boost::shared_ptr<networkinterface>
    networkinterface_for_server(ppid ppidarg);
    
    /// Get a networkinterface object.
    /// \param i is the index of the object desired
    /// \throws std::out_of_range if i is equal to or greater than the number of network interfaces
    const boost::shared_ptr<networkinterface> &get_networkinterface(size_t i) const
    {
      return nics_.at(i);
    }

    /// Access the event manager.
    boost::shared_ptr<eventmanager>
    get_eventmanager() const { return parent_.lock()->get_eventmanager(); }
    
    boost::shared_ptr<manager>
    get_parent() const { return parent_.lock(); }
    
    /// Returns the nodeid for this node.
    nodeid get_nodeid() const { return nodeid_; }

    /// Current time.  To be pulled from the eventmanager.
    virtual const timestamp&
    now() const = 0;

    /// Get a nodeinterface bridging the given ppid and taskid to this node.
    virtual boost::shared_ptr<nodeinterface>
    get_nodeinterface(const ppid &the_ppid, const taskid &the_taskid) = 0;

    /// Set a simple trigger (wake-up call).
    /// \param start_time is the time at which the trigger shall be fulfilled.
    /// \param completion is the eventhandler that gets signaled at start_time.
    virtual void
    schedule(const timestamp &start_time, ppid caller,
        const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to compute for the given number of cpu seconds.
    virtual void
    compute(const timestamp &start_time, ppid caller, const timestamp &cputime,
        const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to compute for the given number of cpu seconds.
    virtual void
    compute(const timestamp &start_time, ppid caller,
        const timestamp &cputime_per_thread, int threads,
        const boost::shared_ptr<eventhandler> &completion) = 0;

    /// I am not sure yet how best to handle this one :)
    /// Eventually, this is where we would bridge over to SST/microscale
    /// or other processor models.
    /// \param start_time is the time when the computation should start.
    /// \param instr is the instruction object that will be run.
    /// \param completion is the eventhandler that gets signaled when
    ///        the instructions have been run.
    virtual void
    compute(const timestamp &start_time, ppid caller,
        const boost::shared_ptr<instructions> &instr,
        const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to copy the given number of bytes between two address spaces.
    /// Used by messaging models to assess cost of non-network 
    /// data transfers.
    /// Completion shall be signaled using the
    /// delegate::kernel()->complete_copy call.
    /// \param start_time time when copy is initiated.
    /// \param bytes number of bytes to copy.
    /// \param completion is the eventhandler signaled upon completion.
    /// \throw valueerror if bytes < 0.
    virtual void
    copy(const timestamp &start_time, ppid caller, int64_t bytes,
        const boost::shared_ptr<eventhandler> &completion) = 0;
    
    /// Estimate the time spent copying data.
    virtual timestamp
    copycost(int64_t bytes) const = 0;

  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_NODE_H
