/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/noisylatency.h>
#include <sstmac/util/mersennetwister.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using RNG::normaldist;
  using RNG::boxmuller;
  using RNG::mersennetwister;

  //
  // One noisy latency generator coming up.
  //
  noisylatency::noisylatency(const timestamp &mean, const timestamp &variance,
                             uint64_t rngseed, bool noisemodel) :
    mean_(mean), variance_(variance), using_noise_model_(noisemodel)
  {
    if((mean.ticks() < 0) || (variance_.ticks() < 0))
      throw valueerror("noisylatency:  Mean and variance must not be negative.");
    // Construct the RNG.
    shared_ptr<mersennetwister> twist = mersennetwister::construct(rngseed);
    rng_ = boxmuller::construct(twist);
  }

  //
  // May we offer you some noisy latency?
  //
  shared_ptr<noisylatency>
  noisylatency::construct(const timestamp &mean, const timestamp &variance,
                          uint64_t rngseed, bool noisemodel)
  {
    return shared_ptr<noisylatency>
      (new noisylatency(mean, variance, rngseed, noisemodel));
  }

  //
  // Done.
  //
  noisylatency::~noisylatency() {
  }

  //
  // Provide a latency estimate for the next transfer.
  //
  timestamp noisylatency::latency(data) const {
    if(using_noise_model_) 
      return rng_->halfnormal(mean_, variance_);
    else
      return rng_->normal(mean_, variance_);
  }

} // end of namespace sstmac
