/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/ordered_flower.h>
#include <sstmac/util/errors.h>

#include <boost/lexical_cast.hpp>
#include <iterator>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <stdlib.h>

using boost::shared_ptr;

namespace sstmac {

  shared_ptr<ordered_flower>
  ordered_flower::construct(const boost::shared_ptr<sstmac::network> &net,
                               double node_router_bw, double router_router_bw)
  {
    shared_ptr<ordered_flower> self(new ordered_flower(net,
                                       node_router_bw, router_router_bw));
    return self;
  }

  ordered_flower::ordered_flower(const boost::shared_ptr<sstmac::network> &net,
                                       double node_router_bw, double router_router_bw):
    typical_flower(net, node_router_bw, router_router_bw),
    n_rebalance_max(-1)
  {
  }

  ordered_flower::~ordered_flower() throw()
  {
  }

  void
  ordered_flower::rebalance_flow_start(flow_id_t f)
  {
    // The new flow has a lower priority than all existing flows.
    // If the edges it uses have available bandwidth then it
    // can use that, but the bandwidth will not be decreased
    // on any existing flows.

    const std::vector<edge_id_t> &edges = flow_edges_[f];

    double headroom = max_bw_;
    for (unsigned long i=0; i<edges.size(); i++) {
      edge_id_t edge = edges[i];
      double edge_headroom = edge_bw_[edge] - edge_bw_utilization(edge);
      if (edge_headroom < 0.0) {
        throw illformederror("impossible headroom");
      }
      if (edge_headroom < headroom) headroom = edge_headroom;
    }
    set_flow_bw(f, headroom);
    if (debug) std::cout << "  allocating flow " << f << " bandwidth " << headroom
                         << std::endl;
  }

  void
  ordered_flower::rebalance_flow_stop(flow_id_t oldf)
  {
    if(debug) std::cout << "ordered_flower::rebalance_flow_stop: oldf = " << oldf
                        << ", n_rebalance_max = " << n_rebalance_max
                        << "\n";
    if (n_rebalance_max == 0) {
      throw valueerror("should rebalance at least one flow");
    }
    else if (n_rebalance_max == -1) {
      // Flows with priority higher than oldf are not affected. Reset
      // the bw used by all other flows (those starting with oldf) to
      // zero.
      for (flow_bw_t::const_iterator iter = flow_bw().lower_bound(oldf);
           iter != flow_bw().end(); iter++) {
        set_flow_bw(iter->first, 0.0);
      }

      // Now reallocate potentially affected flows in order of
      // decreasing priority. The reallocation happens to be exactly the
      // same operation as rebalance_flow_start.
      long i=0;
      for (flow_bw_t::const_iterator iter = flow_bw().lower_bound(oldf);
           iter != flow_bw().end(); iter++,i++) {
        rebalance_flow_start(iter->first);
      }
    }
    else {
      long i=0;
      for (flow_bw_t::const_iterator iter = flow_bw().lower_bound(oldf);
           iter != flow_bw().end(); iter++,i++) {
        if (i > n_rebalance_max) break;
        flow_id_t f = iter->first;
        const std::vector<edge_id_t> &edges = flow_edges_[f];

        double flow_current_bw = get_flow_bw(f);
        double flow_max_bw = max_bw_;
        for (unsigned long j=0; j<edges.size(); j++) {
          double edge_max_bw = edge_bw_[edges[j]];
          if (edge_max_bw < flow_max_bw) flow_max_bw = edge_max_bw;
          if (flow_current_bw >= flow_max_bw) break;
        }
        if (flow_current_bw >= flow_max_bw) continue; // if are at max bw already, then go to next flow

        // Find out how much bandwidth is available to steal from
        // lower priority flows.
        double headroom = max_bw_;
        for (unsigned long j=0; j<edges.size(); j++) {
          edge_id_t edge = edges[j];
          double edge_headroom = edge_bw_[edge] - edge_bw_equal_and_higher_priority_utilization(edge,f);
          if (edge_headroom < 0.0) {
            throw illformederror("impossible headroom");
          }
          if (edge_headroom < headroom) headroom = edge_headroom;
        }

        if(debug) std::cout << "  processing flow " << f
                            << ", current bw = " << get_flow_bw(f)
                            << ", headroom = " << headroom
                            << "\n";

        // For each edge in the current flow, steal bandwidth from lower
        // priority flows.
        for (unsigned long j=0; j<edges.size(); j++) {
          edge_id_t e = edges[j];
          double unused_bw = edge_bw_[e] - edge_bw_utilization(e);
          // nothing needs to be stolen if this link has enough available bw
          if (unused_bw >= headroom) continue;
          // needed_bw is the amount we must take from other flows
          double needed_bw = headroom - unused_bw;
          const std::set<flow_id_t> &flows = edge_flows_[e];
          // Steal bandwidth from lower priority flows, starting with the
          // lowest priority.
          for (std::set<flow_id_t>::const_reverse_iterator iter = flows.rbegin();
               *iter>f; iter++) {
            if (get_flow_bw(*iter) > 0.0) {
              double new_flow_bw;
              if (get_flow_bw(*iter) >= needed_bw) {
                new_flow_bw = get_flow_bw(*iter) - needed_bw;
                needed_bw = 0.0;
              }
              else {
                new_flow_bw = 0.0;
                needed_bw -= get_flow_bw(*iter);
              }
              if (debug) std::cout << "setting bw for victim flow " << *iter
                                   << " to " << new_flow_bw << std::endl;
              set_flow_bw(*iter, new_flow_bw);
              if (needed_bw <= 0.0) break;
            }
          }
          if (needed_bw > 0.0) {
            throw illformederror("could not get needed bw");
          }
        }

        // Add the available headroom to the current flow.
        if(debug) std::cout << "  setting bw for current flow " << f
                            << " to " << get_flow_bw(f) << " + " << headroom << "\n";
        add_to_flow_bw(f, headroom);

        if (get_flow_bw(f) > max_bw_) {
          throw illformederror("internal error: flow_bw_[f] > max_bw_");
        }
      }
    }
  }

} // end of namespace sstmac
