/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/original_flower.h>
#include <sstmac/util/errors.h>

#include <boost/lexical_cast.hpp>
#include <iterator>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <stdlib.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::lexical_cast;
  template <typename T> std::string strfy(T value) {
    return lexical_cast<std::string>(value);
  }

  shared_ptr<original_flower>
  original_flower::construct(const boost::shared_ptr<sstmac::network> &net,
                    double node_router_bw, double router_router_bw,
                    netsharing algo)
  {
    shared_ptr<original_flower> self(new original_flower(net,
                                       node_router_bw, router_router_bw,
                                       algo));
    return self;
  }

  original_flower::original_flower(const boost::shared_ptr<sstmac::network> &net,
                 double node_router_bw, double router_router_bw,
                 netsharing algo):
    flower(net),
    n_rebalance_max(-1),
    algorithm_(algo),
    next_flow_id_(1),
    stopping_flow_(0)
  {
    long nedge = net->g().n_edges();
    edge_bw_.resize(nedge);
    edge_flows_.resize(nedge);

    for (int i=0; i<nedge; i++) {
      sstmac::edge_index ei(i);
      const sstmac::lwedge &e = net->g().index_to_edge(ei);
      if (net->is_node(e.from()) || net->is_node(e.to())) {
        edge_bw_[i] = node_router_bw;
      }
      else {
        edge_bw_[i] = router_router_bw;
      }
    }

    max_bw_ = std::max(node_router_bw, router_router_bw);
  }

  timestamp
  original_flower::next_stopping_time()
  {
    if (stopping_flow_ == 0) {
      throw illformederror("no next stopping flow");
    }
    return stopping_time_;
  }

  flow_id_t
  original_flower::next_stopping_flow()
  {
    return stopping_flow_;
  }

  void
  original_flower::advance_time(timestamp tnew)
  {
    timestamp deltat = tnew - t_;
    t_ = tnew;

    for (flow_bw_t::iterator i=flow_bw_.begin(); i!=flow_bw_.end(); i++) {
      flow_id_t f = i->first;
      double bw = i->second;
      double &nbyte = flow_nbyte_[f];
      nbyte = nbyte - deltat.sec()*bw;
      if (nbyte < 0) nbyte = 0;
    }
  }

  void
  original_flower::find_next_stop()
  {
    if(debug) std::cout << "original_flower::find_next_stop: entered\n";
    stopping_flow_ = 0;
    for (flow_bw_t::iterator i=flow_bw_.begin(); i!=flow_bw_.end(); i++) {
      flow_id_t f = i->first;
      double bw = i->second;
      if(debug) std::cout << "  f = " << f << ", bw = " << bw;
      if(bw > 0) {
        double nbyte = flow_nbyte_[f];
        timestamp stopping_time = t_ + timestamp(nbyte/bw);
        if(debug) std::cout << ", stopping_time = " << stopping_time;
        if (stopping_flow_ == 0 || stopping_time_ > stopping_time) {
          stopping_flow_ = f;
          stopping_time_ = stopping_time;
        }
      }
      if(debug) std::cout << "\n";
    }
    if (stopping_flow_ == 0 && flow_bw_.size() > 0) {
      throw illformederror("flows exist but next stopping flow not found");
    }
    if(debug) std::cout << "  stopping_flow = " << stopping_flow_ << "\n";
  }

  double
  original_flower::edge_bw_utilization(edge_id_t e)
  {
    double bw = 0.0;
    const std::set<flow_id_t> &flows = edge_flows_[e];
    for (std::set<flow_id_t>::const_iterator i = flows.begin();
         i != flows.end(); i++) {
      bw += flow_bw_[*i];
    }
    return bw;
  }

  double
  original_flower::edge_bw_equal_and_higher_priority_utilization(edge_id_t e, flow_id_t f)
  {
    if(debug) std::cout << "original_flower::edge_bw_equal_and_higher_priority_utilization: e = " << e
                        << ", nflow = " << edge_flows_[e].size()
                        << ", bw = ";
    double bw = 0.0;
    const std::set<flow_id_t> &flows = edge_flows_[e];
    for (std::set<flow_id_t>::const_iterator i = flows.begin();
         i != flows.end(); i++) {
      // See if this (and, thus, all remaining) flows are lower
      // priority (higher flow_id). If so, break out of loop.
      if (*i > f) break;
      if(debug) std::cout << "+" << flow_bw_[*i];
      bw += flow_bw_[*i];
    }
    if(debug) std::cout << "\n";
    return bw;
  }

  void
  original_flower::rebalance_flow_start_overcommit(flow_id_t f)
  {
    flow_bw_[f] = max_bw_;
  }

  void
  original_flower::rebalance_flow_stop_overcommit(flow_id_t oldf)
  {
  }

  void
  original_flower::rebalance_flow_start_throttle_overcommits(flow_id_t f)
  {
    flow_bw_[f] = max_bw_;
    const std::vector<edge_id_t> &edges = flow_edges_[f];
    for (unsigned long i=0; i<edges.size(); i++) {
      edge_id_t edge = edges[i];
      double bw_req = edge_bw_utilization(edge);
      if (edge_bw_[edge] < bw_req) {
        double scale_factor = edge_bw_[edge]/bw_req;
        // Slow all of the flows through the overcommitted edge.  Note
        // that the flow rates satisfying the constraints are not
        // unique, and the result of this algorithm depends on the
        // order with which we iterate through the edges.
        for (flow_set_t::iterator iter=edge_flows_[edge].begin();
             iter!=edge_flows_[edge].end(); iter++) {
          if(debug) std::cout << "Edge " << edge
                    << " adjusting flow " << *iter
                    << " old rate = " << flow_bw_[*iter];
          flow_bw_[*iter] *= scale_factor;
          if(debug) std::cout << " new rate = " << flow_bw_[*iter]
                    << std::endl;
        }
      }
    }
  }

  void
  original_flower::rebalance_flow_stop_throttle_overcommits(flow_id_t oldf)
  {
  }

  void
  original_flower::rebalance_flow_start_ordered(flow_id_t f)
  {
    // The new flow has a lower priority than all existing flows.
    // If the edges it uses have available bandwidth then it
    // can use that, but the bandwidth will not be decreased
    // on any existing flows.

    const std::vector<edge_id_t> &edges = flow_edges_[f];

    double headroom = max_bw_;
    for (unsigned long i=0; i<edges.size(); i++) {
      edge_id_t edge = edges[i];
      double edge_headroom = edge_bw_[edge] - edge_bw_utilization(edge);
      if (edge_headroom < 0.0) {
        throw illformederror("impossible headroom");
      }
      if (edge_headroom < headroom) headroom = edge_headroom;
    }
    flow_bw_[f] = headroom;
    if (debug) std::cout << "  allocating flow " << f << " bandwidth " << headroom
                         << std::endl;
  }

  void
  original_flower::rebalance_flow_stop_ordered(flow_id_t oldf)
  {
    if(debug) std::cout << "original_flower::rebalance_flow_stop_ordered: oldf = " << oldf
                        << ", n_rebalance_max = " << n_rebalance_max
                        << ", flow_bw_.size() = " << flow_bw_.size()
                        << "\n";
    if (n_rebalance_max == 0) {
      throw valueerror("should rebalance at least one flow");
    }
    else if (n_rebalance_max == -1) {
      // Flows with priority higher than oldf are not affected. Reset
      // the bw used by all other flows (those starting with oldf) to
      // zero.
      for (flow_bw_t::iterator iter = flow_bw_.lower_bound(oldf);
           iter != flow_bw_.end(); iter++) {
        iter->second = 0.0;
      }

      // Now reallocate potentially affected flows in order of
      // decreasing priority. The reallocation happens to be exactly the
      // same operation as rebalance_flow_start_ordered.
      long i=0;
      for (flow_bw_t::iterator iter = flow_bw_.lower_bound(oldf);
           iter != flow_bw_.end(); iter++,i++) {
        rebalance_flow_start_ordered(iter->first);
      }
    }
    else {
      long i=0;
      for (flow_bw_t::iterator iter = flow_bw_.lower_bound(oldf);
           iter != flow_bw_.end(); iter++,i++) {
        if (i > n_rebalance_max) break;
        flow_id_t f = iter->first;
        const std::vector<edge_id_t> &edges = flow_edges_[f];

        double flow_current_bw = flow_bw_[f];
        double flow_max_bw = max_bw_;
        for (unsigned long j=0; j<edges.size(); j++) {
          double edge_max_bw = edge_bw_[edges[j]];
          if (edge_max_bw < flow_max_bw) flow_max_bw = edge_max_bw;
          if (flow_current_bw >= flow_max_bw) break;
        }
        if (flow_current_bw >= flow_max_bw) continue; // if are at max bw already, then go to next flow

        // Find out how much bandwidth is available to steal from
        // lower priority flows.
        double headroom = max_bw_;
        for (unsigned long j=0; j<edges.size(); j++) {
          edge_id_t edge = edges[j];
          double edge_headroom = edge_bw_[edge] - edge_bw_equal_and_higher_priority_utilization(edge,f);
          if (edge_headroom < headroom) headroom = edge_headroom;
        }

        if(debug) std::cout << "  processing flow " << f
                            << ", current bw = " << flow_bw_[f]
                            << ", headroom = " << headroom
                            << "\n";

        // For each edge in the current flow, steal bandwidth from lower
        // priority flows.
        for (unsigned long j=0; j<edges.size(); j++) {
          edge_id_t e = edges[j];
          double unused_bw = edge_bw_[e] - edge_bw_utilization(e);
          // nothing needs to be stolen if this link has enough available bw
          if (unused_bw >= headroom) continue;
          // needed_bw is the amount we must take from other flows
          double needed_bw = headroom - unused_bw;
          const std::set<flow_id_t> &flows = edge_flows_[e];
          // Steal bandwidth from lower priority flows, starting with the
          // lowest priority.
          for (std::set<flow_id_t>::const_reverse_iterator iter = flows.rbegin();
               *iter>f; iter++) {
            if (flow_bw_[*iter] > 0.0) {
              double new_flow_bw;
              if (flow_bw_[*iter] >= needed_bw) {
                new_flow_bw = flow_bw_[*iter] - needed_bw;
                needed_bw = 0.0;
              }
              else {
                new_flow_bw = 0.0;
                needed_bw -= flow_bw_[*iter];
              }
              flow_bw_[*iter] = new_flow_bw;
              if (needed_bw <= 0.0) break;
            }
          }
          if (needed_bw > 0.0) {
            throw illformederror("could not get needed bw");
          }
        }

        // Add the available headroom to the current flow.
        if(debug) std::cout << "  setting bw for current flow " << f
                            << " to " << flow_bw_[f] << " + " << headroom << "\n";
        flow_bw_[f] += headroom;

        if (flow_bw_[f] > max_bw_) {
          throw illformederror("internal error: flow_bw_[f] > max_bw_");
        }
      }
    }
  }

  void
  original_flower::rebalance_flow_start(flow_id_t f)
  {
    switch(algorithm_) {
    case OVERCOMMIT:
      rebalance_flow_start_overcommit(f);
      break;
    case THROTTLE_OVERCOMMITS:
      rebalance_flow_start_throttle_overcommits(f);
      break;
    case ORDERED:
      rebalance_flow_start_ordered(f);
      break;
    default:
      throw valueerror("original_flower::rebalance_flow_start:  Unknown target algorithm.");
    }
  }

  void
  original_flower::rebalance_flow_stop(flow_id_t oldf)
  {
    switch(algorithm_) {
    case OVERCOMMIT:
      rebalance_flow_stop_overcommit(oldf);
      break;
    case THROTTLE_OVERCOMMITS:
      rebalance_flow_stop_throttle_overcommits(oldf);
      break;
    case ORDERED:
      rebalance_flow_stop_ordered(oldf);
      break;
    default:
      throw valueerror("original_flower::rebalance_throw_stop:  Unknown target algorithm.");
    }
  }

  flow_id_t
  original_flower::start_flow(timestamp tstart, const std::vector<edge_id_t> &edges, long nbyte)
  {
    if (tstart < t_) {
      throw illformederror("inserting a flow at an earlier time");
    }

    advance_time(tstart);

    flow_id_t current_flow_id = next_flow_id_++;
  
    flow_edges_[current_flow_id] = edges;
    if(flow_edges_.find(current_flow_id) == flow_edges_.end()) {
      throw illformederror("flow::start_flow: flow_edges_ missing current_flow_id");
    }

    for (unsigned long i=0; i<edges.size(); i++) {
      if (debug) std::cout << "adding flow " << current_flow_id << " to edge " << edges[i] << "\n";
      edge_flows_[edges[i]].insert(current_flow_id);
    }

    flow_nbyte_[current_flow_id] = nbyte;

    flow_bw_[current_flow_id] = 0.0;

    rebalance_flow_start(current_flow_id);

    find_next_stop();

    if(debug) std::cout << "Started flow " << current_flow_id << "\n";
    return current_flow_id;
  }

  void
  original_flower::stop_flow(timestamp tstop, flow_id_t f)
  {
    if(debug) std::cout << "Stopping flow " << f << "\n";
    if (tstop < t_) {
      throw illformederror("stopping a flow at an earlier time");
    }
    if (flow_edges_.find(f) == flow_edges_.end()) {
      throw illformederror(std::string("stopping flow number ") +
                           strfy(f) + " that does not exist");
    }

    advance_time(tstop);

    const std::vector<edge_id_t> &edges = flow_edges_[f];

    for (unsigned long i=0; i<edges.size(); i++) {
      if(debug) std::cout << "removing flow " << f << " from edge " << edges[i] << "\n";
      edge_flows_[edges[i]].erase(f);
      if(debug) std::cout << "  edge " << edges[i]
                << " erasing flow " << f << std::endl;
    }

    flow_edges_.erase(f);
    flow_nbyte_.erase(f);
    flow_bw_.erase(f);
    if(debug) std::cout << "erasing flow " << f << std::endl;

    rebalance_flow_stop(f);

    find_next_stop();

  }

} // end of namespace sstmac
