/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/powermanager.h>

namespace sstmac
{

  powermanager::powermanager(std::string fname) : statsmanager(fname)
  {

  }

  powermanager::~powermanager()
  {

  }

  /// Calculates the average power across the duration of the simulation
  double
  powermanager::getAveragePower(const timestamp &simtime)
  {
    list<boost::shared_ptr<powerstat> >::iterator it;
    double ret = 0;


    for (it = stats.begin(); it != stats.end(); it++)
      {

        ret += (*it)->getAveragePower(simtime);
      }
    return ret;
  }

  /// Calculates the average power across the duration of the simulation for a
  /// specific type of stat
  double
  powermanager::getAveragePower(const timestamp &simtime,
      powerstat::PWR_TYPE type)
  {
    list<boost::shared_ptr<powerstat> >::iterator it;
    double ret = 0;
    for (it = stats.begin(); it != stats.end(); it++)
      {


        if ((*it)->getType() == type)
          ret += (*it)->getAveragePower(simtime);
      }
    return ret;
  }

  /// Calculates the total number of joules spent during the simulation
  double
  powermanager::getTotalEnergy(const timestamp &simtime)
  {
    list<boost::shared_ptr<powerstat> >::iterator it;
    double ret = 0;
    for (it = stats.begin(); it != stats.end(); it++)
      {
        ret += (*it)->getTotalEnergy(simtime);
      }

    return ret;
  }

  void
  powermanager::registerPowerStat(boost::shared_ptr<powerstat> &stat)
  {
    stats.push_back(stat);
  }

  double
  powermanager::getTotalStatic()
  {
    list<boost::shared_ptr<powerstat> >::iterator it;
    double ret = 0;
    for (it = stats.begin(); it != stats.end(); it++)
      {
        ret += (*it)->getStatic();
      }
    return ret;
  }

  double
  powermanager::getTotalDynamic()
  {
    list<boost::shared_ptr<powerstat> >::iterator it;
    double ret = 0;
    for (it = stats.begin(); it != stats.end(); it++)
      {
        ret += (*it)->getDynamic();
      }

    return ret;
  }

  void
  powermanager::printSummary(const timestamp &t)
  {

    ostream &out = getOutputStream();

    out << "------ Power (t=" << t.sec() << ") -------" << std::endl;
    out << "Totals:" << std::endl;
    out << "\t average power is: " << getAveragePower(t) << " W" << std::endl;
    out << "\t static power: " << getTotalStatic() << " W" << std::endl;
    out << "\t dynamic power: " << getTotalDynamic() / t.sec() << " W"
        << std::endl;
    out << "Breakdown:" << std::endl;
    out << "\t Processor: " << getAveragePower(t, powerstat::PROCESSOR)
        << std::endl;
    out << "\t Memory: " << getAveragePower(t, powerstat::MEMORY) << std::endl;
    out << "\t Network: " << getAveragePower(t, powerstat::NETWORK)
        << std::endl;
    out << "------------------" << std::endl;
  }

}
