/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


#include <sstmac/simulator/randomindexing.h>
#include <sstmac/util/errors.h>
#include <algorithm>
#include <stdlib.h>

namespace sstmac {

  using boost::shared_ptr;

  randomindexing::randomindexing()
  {
    rng_ = RNG::SimpleCombo::construct();
  }

  // Hi.
  randomindexing::randomindexing(unsigned seed)
  {
    rng_ = RNG::SimpleCombo::construct(seed);
  }

  randomindexing::randomindexing
  (const boost::shared_ptr<RNG::UniformInteger> &rng)
  {
    rng_ = rng;
  }

  shared_ptr<randomindexing> randomindexing::construct() {
    return shared_ptr<randomindexing>(new randomindexing);
  }

  shared_ptr<randomindexing> randomindexing::construct(unsigned seed) {
    return shared_ptr<randomindexing>(new randomindexing(seed));
  }

  shared_ptr<randomindexing> randomindexing::construct
  (const boost::shared_ptr<RNG::UniformInteger> &rng) {
    return shared_ptr<randomindexing>(new randomindexing(rng));
  }

  randomindexing::~randomindexing() throw() {
  }

  void randomindexing::allocate(const std::vector<nodeid> &nodes,
                                int ppn,
                                std::vector<nodeid> &result,
                                int nproc)
  {
    if(nodes.empty()) {
      throw valueerror("randomindexing::allocate:  Empty node vector.");
    }
    if(ppn <= 0) {
      throw valueerror("randomindexing::allocate:  ppn must be > 0.");
    }
    if(ppn*nodes.size() < nproc) {
      throw valueerror("randomindexing::allocate: require ppn*nodes.size()>=nproc");
    }

    result = nodes;
    for(int i = 1; i < ppn; ++i) {
      result.insert(result.end(), nodes.begin(), nodes.end());
    }

    
    RNG::UniformInteger_functor rngf(rng_);
    std::random_shuffle(result.begin(), result.end(), rngf);
    //std::random_shuffle(result.begin(), result.end());
    
    result.resize(nproc);
  }

} // end of namespace sstmac
