/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_SHAREDCIRCUIT_H
#define SSTMAC_SIMULATOR_SHAREDCIRCUIT_H

#include <sstmac/simulator/interconnect.h>
#include <sstmac/simulator/bandwidth.h>
#include <sstmac/simulator/powerstat.h>
#include <sstmac/simulator/power_config.h>

namespace sstmac
{

  // Important types in the sstmac namespace
  class flower;
  class network;
  class latencymodel;

  /**
   * A specialization of circuit that shares circuit bandwidth between
   * concurrent transfer requests.
   */
  class sharedcircuit : public interconnect
  {
    /// Private nested type to handle stopping flows and notifying listeners.
    class stopper;
    /// Private nested type to schedule starting a flow.
    class starter;

    /// The flower handles the actual sharing of the circuit.
    boost::shared_ptr<flower> flower_;
    /// The sharedcircuit has exactly one stopper.
    boost::shared_ptr<stopper> stopper_;
    /// Network latency.
    boost::shared_ptr<latencymodel> latency_;
    /// Network bandwidth.
    bandwidth bandwidth_;
    /// Our network.
    boost::shared_ptr<sstmac::network> net_;
    /// Our event manager.
    boost::shared_ptr<eventmanager> mgr_;



  protected:
    /// Construct a new circuit.
    sharedcircuit(const boost::shared_ptr<latencymodel> &latency, bandwidth bw,
        const boost::shared_ptr<sstmac::network> &net,
        const boost::shared_ptr<flower> &,
        const boost::shared_ptr<power_config> &pwrcfg);

  public:
    /// Construct a new circuit with constant latency - no power config
    static boost::shared_ptr<sharedcircuit>
    construct(const timestamp latency, double bandwidth,
        const boost::shared_ptr<sstmac::network> &net,
        const boost::shared_ptr<flower> &,
        const boost::shared_ptr<power_config> &pwrcfg);

    /// Construct a new circuit with constant latency.
    static boost::shared_ptr<sharedcircuit>
    construct(const timestamp latency, double bandwidth,
        const boost::shared_ptr<sstmac::network> &net,
        const boost::shared_ptr<flower> &);

    /// Construct a new circuit with user-defined latency modeling.
    static boost::shared_ptr<sharedcircuit>
    construct(const boost::shared_ptr<latencymodel> &latency, bandwidth bw,
        const boost::shared_ptr<sstmac::network> &net,
        const boost::shared_ptr<flower> &,
        const boost::shared_ptr<power_config> &pwrcfg);

    /// So long, Mr. Despair.
    virtual
    ~sharedcircuit() throw ();

    /// Required method from parent.
    virtual int64_t
    slots() const;

    /// My event manager.
    virtual void
    set_eventmanager(const boost::shared_ptr<eventmanager> &mgr);


    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void
    send(const timestamp &start_time, int64_t bytes, nodeid sender,
        nodeid recver, const boost::shared_ptr<eventhandler> &sendhandler,
        const boost::shared_ptr<eventhandler> &recvhandler);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_SHAREDCIRCUIT_H
