/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/simplenic.h>
#include <sstmac/simulator/networkdata.h>
#include <sstmac/process/recvhandler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hi.
  //
  simplenic::simplenic(const shared_ptr<node> &parent,
                       const shared_ptr<interconnect> &interconn) :
    parent_(parent), interconn_(interconn)
  {}

  //
  // Hello.
  //
  shared_ptr<simplenic>
  simplenic::construct(const shared_ptr<node> &parent,
                       const shared_ptr<interconnect> &interconn)
  {
    return shared_ptr<simplenic>(new simplenic(parent, interconn));
  }

  //
  // Goodbye.
  //
  simplenic::~simplenic() throw() {
  }

  //
  // Schedule a send operation to happen at the given time.
  //
  void simplenic::send(const timestamp &start_time,
                       ppid ppidarg,
                       nodeid recver,
                       const boost::shared_ptr<networkdata> &payload,
                       const boost::shared_ptr<eventhandler> &completion)
  {
    if (debugger::active()) {
      debug() << "simplenic::send(" << start_time << ", " << ppidarg << ", "
      << recver << ", " << payload << ", " << completion << ")\n";
    }
    
    if (payload == 0) {
      throw nullerror("simplenic::send: null payload pointer.");
    }

    nodeid sender = parent_.lock()->get_nodeid();

    // Get the node pointer for the receiver (needed for callback operations).
    shared_ptr<node> recvnode = parent_.lock()->get_parent()->get_node(recver);

    shared_ptr<networkinterface> recvnic = recvnode->networkinterface_for_server(ppidarg);
    
    // Set up the event associated with completion of the receive
    shared_ptr<eventhandler> recvevent =
    recvhandler::construct(recvnic, ppidarg, sender, payload);

    // Fire off the actual send operation
    interconn_->send(start_time, payload->bytes(), sender, recver,
                     completion, recvevent);
  }

} // end of namespace sstmac.
