/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/simplenode.h>
#include <sstmac/processor/simpleinstruction.h>
#include <sstmac/simulator/simplenic.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/simulator/networkdata.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/process/server.h>
#include <sstmac/processor/procscheduler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <sstream>

namespace sstmac
{

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Construct a new node.
  //
  simplenode::simplenode(const timestamp &latency, double memory_bandwidth,
      boost::shared_ptr<processor_config> proccfg, int n_core) :
    memlatency_(latency), membandwidth_(memory_bandwidth), n_core_(n_core)
  {
    scheduler_ = procscheduler::construct(n_core, proccfg);

  }

  //
  // Construct a new node.
  //
  simplenode::simplenode(const timestamp &latency, double memory_bandwidth,
      int n_core) :
    memlatency_(latency), membandwidth_(memory_bandwidth), n_core_(n_core)
  {
    scheduler_ = procscheduler::construct(n_core);

  }

  //
  // Construct a new node.
  //
  simplenode::simplenode(double memory_bandwidth, int n_core) :
    memlatency_(0), membandwidth_(memory_bandwidth), n_core_(n_core)
  {
    scheduler_ = procscheduler::construct(n_core);

  }

  //
  // Who am I?
  //
  void
  simplenode::set_self(const boost::shared_ptr<simplenode> &self)
  {
    if (self.get() != this)
      {
        throw valueerror(
            "simplenode::set_self:  'self' pointer does not refer "
              "to 'this' object.");
      }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<simplenode>
  simplenode::get_self()
  {
    return self_.lock();
  }

  //
  // Construct a new node with the given memory latency and bandwidth.
  //
  boost::shared_ptr<simplenode>
  simplenode::construct(const timestamp &latency, double bandwidth,
      boost::shared_ptr<processor_config> proccfg, int n_core)
  {
    shared_ptr<simplenode> retval(
        new simplenode(latency, bandwidth, proccfg, n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Construct a new node with the given memory latency and bandwidth.
  //
  boost::shared_ptr<simplenode>
  simplenode::construct(const timestamp &latency, double bandwidth, int n_core)
  {
    shared_ptr<simplenode> retval(new simplenode(latency, bandwidth, n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Construct a new node with the given memory bandwidth.
  //
  boost::shared_ptr<simplenode>
  simplenode::construct(double memory_bandwidth, int n_core)
  {
    shared_ptr<simplenode> retval(new simplenode(memory_bandwidth, n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  simplenode::~simplenode() throw ()
  {
  }

  //
  // Get a new node of the same type as the current one.
  //
  boost::shared_ptr<node>
  simplenode::clone() const
  {
    if (debugger::active())
      debug() << "simplenode::clone()\n";

    shared_ptr<simplenode> retval(
        new simplenode(this->memlatency_, this->membandwidth_,
            this->scheduler_->getProcCfg(), this->n_core_));
    retval->server_.clear();
    //retval->iface_.clear();
    retval->set_self(retval);
    return retval;
  }

  void
  simplenode::initialize(nodeid my_id,
                         const boost::shared_ptr<manager> &parent)
  {
    node::initialize(my_id, parent);
    scheduler_->set_manager(parent->get_eventmanager());
  }

  //
  // Current time.
  //
  const timestamp&
  simplenode::now() const
  {
    return get_eventmanager()->now();
  }

  //
  // Access a node interface for the given ppid.
  //
  shared_ptr<nodeinterface>
  simplenode::get_nodeinterface(const ppid &pid, const taskid &tid)
  {
    return nodeinterface::construct(self_.lock(), pid, tid);
  }

  //
  // Set a simple trigger (wake-up call).
  //
  void
  simplenode::schedule(const timestamp &start_time, ppid caller,
      const boost::shared_ptr<eventhandler> &completion)
  {
    if (debugger::active())
      debug() << "simplenode::schedule(" << start_time << ", " << caller
          << ", " << completion << ")\n";
    get_eventmanager()->schedule(start_time, completion);
  }

  //
  // Pretend to compute.
  //
  void
  simplenode::compute(const timestamp &start_time, ppid caller,
      const timestamp &cpu_time,
      const boost::shared_ptr<eventhandler> &completion)
  {
    //std::cerr << "compute1" << endl;
    compute(start_time, caller, cpu_time, 1, completion);
  }

  //
  // Pretend to compute.
  //
  void
  simplenode::compute(const timestamp &start_time, ppid caller,
      const timestamp &cpu_time_per_thread, int threads,
      const boost::shared_ptr<eventhandler> &completion)
  {

    //std::cerr << "compute2" << endl;
    static bool first_entry = true;
    if (debugger::active())
      debug() << "simplenode::compute(" << start_time << ", " << caller << ", "
          << cpu_time_per_thread << ", " << threads << ", " << completion
          << ")\n";
    if (threads < 1)
      throw valueerror("simplenode::compute:  Can't compute on less than "
        "one thread.");
    // FIXME:   Implement load monitor.
    if (first_entry)
      {
        std::cerr << "WARNING:  simplenode::compute:  "
            << "Load monitor not implemented\n";
        first_entry = false;
      }

    boost::shared_ptr<instructions> inst = simpleinstruction::construct(
        cpu_time_per_thread);
    // for (int i = 0; i < threads; i++)
    compute(start_time, caller, inst, completion);

    //mgr_->schedule(start_time + cpu_time_per_thread, completion);
  }


  //
  // Compute request
  //
  void
  simplenode::compute(const timestamp &start_time, ppid caller,
      const boost::shared_ptr<instructions> &instructions,
      const boost::shared_ptr<eventhandler> &completion)
  {

    if (instructions->type == instructions::I_DETAILED
        && scheduler_->getProcCfg()->proctype
            != processor_config::PROC_INSTRUCTION)
      throw valueerror(
          "simplenode: use of eventdata instructions requires an instructionprocessor, please configure simplenode to have one.");

    if (instructions->type == instructions::I_TIME
        && scheduler_->getProcCfg()->proctype != processor_config::PROC_SIMPLE)
      throw valueerror(
          "simplenode: use of simple instructions requires an simpleprocessor, please configure simplenode to have one.");

    // std::cerr << "compute3" << endl;
    if (debugger::active())
      debug() << "simplenode::compute(" << start_time << ", " << caller << ", "
          << instructions << ", " << completion << ")\n";
    scheduler_->compute(start_time, instructions, completion);

    // old stuff
    //timestamp ctime = 
    //  processor_->compute(instructions);
    //mgr_->schedule( start_time + ctime, completion );
  }

  //
  // Pretend to copy data.
  //
  void
  simplenode::copy(const timestamp &start_time, ppid p, int64_t bytes,
      const boost::shared_ptr<eventhandler> &completion)
  {
    if (debugger::active())
      debug() << "simplenode::copy(" << start_time << ", " << p << ", "
          << bytes << ", " << completion << ")\n";
    if (bytes < 0)
      {
        throw valueerror("simplenode::copy:  negative number of bytes.");
      }
    timestamp copylag = memlatency_ + this->copycost(bytes);
    get_eventmanager()->schedule(start_time + copylag, completion);
  }

  //
  // Private method to estimate copy cost.
  //
  timestamp
  simplenode::copycost(int64_t bytes) const
  {
    if (debugger::active())
      debug() << "simplenode::copycost(" << bytes << ")\n";
    if (bytes < 0)
      {
        throw valueerror("simplenode::copycost:  negative number of bytes.");
      }
    if (membandwidth_ > 0)
      {
        timestamp copytime(bytes / membandwidth_);
        return copytime;
      }
    else
      {
        return timestamp(0);
      }
  }

  list<boost::shared_ptr<stats> >
  simplenode::getStats()
  {
    return scheduler_->getProcessor()->getStats();
  }

} // end of namespace sstmac.
