/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_SIMPLENODE_H
#define SSTMAC_SIMULATOR_SIMPLENODE_H

#include <sstmac/simulator/node.h>
#include <sstmac/util/nocopy.h>
#include <boost/weak_ptr.hpp>
#include <map>

#include <sstmac/processor/processor_config.h>

namespace sstmac
{

  // Forward declaration of important types.
  class procscheduler;

  /**
   * A stand-in node.  This node type passes all requests directly to
   * the simulator and has no resource contention (infinite processors),
   */
  class simplenode : public node, virtual public nocopy
  {
    /// This object.
    boost::weak_ptr<simplenode> self_;

    /// The memory transfer latency.
    timestamp memlatency_;

    /// The memory transfer bandwidth in bytes/sec.
    double membandwidth_;

    /// The network servers for parallel processes running on this node.
    typedef std::map<ppid, boost::shared_ptr<server> > server_t;
    server_t server_;

    /// Processor model for this node.
    typedef boost::shared_ptr<processor> proc_t;
    //proc_t processor_;

    /// Processor scheduler for this node
    boost::shared_ptr<procscheduler> scheduler_;

    /// The number of processor cores
    int n_core_;

    /// The callback interface for each ppid to this node
    /// (helps break cyclic reference-count relationships).
    //typedef std::map< ppid, boost::shared_ptr<nodeinterface> > iface_t;
    //iface_t iface_;

  protected:
    /// Construct a new node with the given latency and bandwidth.
    simplenode(const timestamp &memory_latency, double memory_bandwidth,

    boost::shared_ptr<processor_config> proccfg, int n_core);

    simplenode(const timestamp &memory_latency, double memory_bandwidth,
        int n_core);

    /// Construct a new node.  A memory bandwidth of 0 or less
    /// translates into infinite memory bandwidth.
    simplenode(double memory_bandwidth, int n_core);

    /// Who am I?
    void
    set_self(const boost::shared_ptr<simplenode> &self);

    /// This is who I am.
    boost::shared_ptr<simplenode>
    get_self();

  public:

    /// Construct a new node with the given memory bandwidth.
    static boost::shared_ptr<simplenode>
    construct(const timestamp &memory_latency, double memory_bandwidth,

    boost::shared_ptr<processor_config> proccfg, int n_core = 1);

    /// Construct a new node with the given memory bandwidth.
    static boost::shared_ptr<simplenode>
    construct(const timestamp &memory_latency, double memory_bandwidth,
        int n_core = 1);

    /// Construct a new node with the given memory bandwidth and 0 latency.
    static boost::shared_ptr<simplenode>
    construct(double memory_bandwidth, int n_core = 1);

    /// Goodbye.
    virtual
    ~simplenode() throw ();

    /* The rest of this file is the methods specified by the node interface. */

    /// Get a new node of the same type as the current one.
    virtual boost::shared_ptr<node>
    clone() const;

    /// Current time.
    virtual const timestamp&
    now() const;

    /// Get a nodeinterface bridging the given ppid and taskid to this node.
    virtual boost::shared_ptr<nodeinterface>
    get_nodeinterface(const ppid &the_ppid, const taskid &the_taskid);

    /// Set a simple trigger (wake-up call).
    /// \param start_time is the time at which the trigger shall be fulfilled.
    /// \param completion is the eventhandler that gets signaled at start_time.
    virtual void
    schedule(const timestamp &start_time, ppid caller,
        const boost::shared_ptr<eventhandler> &completion);

    /// Pretend to compute for the given number of cpu seconds.
    virtual void
    compute(const timestamp &start_time, ppid caller, const timestamp &cputime,
        const boost::shared_ptr<eventhandler> &completion);

    /// Pretend to compute for the given number of cpu seconds.
    virtual void
    compute(const timestamp &start_time, ppid caller,
        const timestamp &cputime_per_thread, int threads,
        const boost::shared_ptr<eventhandler> &completion);

    /// \throw unimplementederror (well, we've got to do something).
    /// Eventually, this is where we would bridge over to SST/microscale
    /// or other processor models.
    /// \param start_time is the time when the computation should start.
    /// \param instr is the instruction object that will be run.
    /// \param completion is the eventhandler that gets signaled when
    ///        the instructions have been run.
    virtual void
    compute(const timestamp &start_time, ppid caller,
        const boost::shared_ptr<instructions> &instr,
        const boost::shared_ptr<eventhandler> &completion);

    /// Pretend to copy the given number of bytes between two address spaces.
    /// Used by messaging models to assess cost of non-network 
    /// data transfers.
    /// Completion shall be signaled using the
    /// delegate::kernel()->complete_copy call.
    /// \param start_time time when copy is initiated.
    /// \param bytes number of bytes to copy.
    /// \param completion is the eventhandler signaled upon completion.
    virtual void
    copy(const timestamp &start_time, ppid caller, int64_t bytes,
        const boost::shared_ptr<eventhandler> &completion);

    /// Estimate the time spent copying data.
    timestamp
    copycost(int64_t bytes) const;
    
    virtual void
    initialize(nodeid my_id, const boost::shared_ptr<manager> &parent);

  private:

    virtual list<boost::shared_ptr<stats> > getStats();
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_SIMPLENODE_H
