/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/util/cast.h>
#include <sstmac/skeleton/actormatmul/actormatmul.h>
#include <sstmac/skeleton/actormatmul/actormatmul-messages.h>
#include <sstmac/skeleton/actormatmul/actormatmul-computor.h>
#include <math.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;
  
  const std::string actormatmul::label_("actormatmul");

  template <typename T> T squared(T v) { return v*v; }
  template <typename T> T cubed(T v) { return v*v*v; }

  //
  // Called when somebody requests data.
  //
  void actormatmul::get_data(actorid requestor, int iteration) {
    // Data is always available (we're doing direct fetch).
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ":  get_data(" << requestor
                << ", " << iteration << ")\n";
    shared_ptr<putdata> data = putdata::construct(iteration);
    // Here we send the actual data content (it's a network send).
    //std::cerr << "DEBUG: actormatmul[" << mbox()->mypid() << ":  Send data "
    //          << " to " << requestor << " at time " << mbox()->now() << "\n";
    mbox()->send(requestor, data, squared(submatrix_dimension_));
  }

  //
  // Called when somebody delivers data.
  //
  void actormatmul::put_data(actorid provider, int iteration) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": put_data(" << provider << ", "
                << iteration << "): " << "pending_arrivals=="
                << remaining_.at(iteration).pending_arrivals << "\n";
    if((iteration < 0) || (iteration >= int(remaining_.size())))
      throw illformederror("actormatmul::put_data: Invalid iteration number");
    if(! remaining_.at(iteration).active)
      throw illformederror("actormatmul::put_data:  Fetching for an "
                           "inactive iteration.");
    if(remaining_.at(iteration).remaining >= 0)
      throw illformederror("actormatmul::put_data: Attempt to re-fetch "
                           "data for an iteration.");
    --remaining_.at(iteration).pending_arrivals;
    if(remaining_.at(iteration).pending_arrivals <= 0) {
      int64_t iters = cubed(workmatrix_splits_);
      remaining_.at(iteration).remaining = iters;
      remaining_stores_ += iters;
      alert_computors();
    }
  }

  //
  // Test whether we've hit the low water mark where we will no longer
  // accept worker computors from neighbors.
  //
  void actormatmul::test_watermark() {
    if((current_iteration_ == (sqrtnodes_-1)) &&
       remaining_.at(current_iteration_).remaining == computors_)
      {
        // Watermark.
        actorid up = mbox()->get_tag(flat_upper_);
        actorid down = mbox()->get_tag(flat_lower_);
        int sendlower = (flat_lower_ != flat_rank_ ? flat_lower_ : flat_upper_);
        int sendupper = (flat_upper_ != flat_rank_ ? flat_upper_ : flat_lower_);
        mbox()->send(up, lowernodeupdate::construct(sendlower), 8);
        mbox()->send(down, uppernodeupdate::construct(sendupper), 8);
      }
  }

  //
  // Called when a local computor completes a unit of work.
  //
  void actormatmul::local_computor_done(actorid thread, int iteration) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": local_computor_done(" << thread 
                << ", " << iteration << "):  remaining_stores=="
                << remaining_stores_ << "; current_iteration_=="
                << current_iteration_ << "\n";
    idle_local_computors_.push_back(thread);
    --remaining_stores_;
    test_watermark();
    if((current_iteration_ >= (sqrtnodes_-1)) && (remaining_stores_ == 0)) {
      // All stores are done.  Time to bail.
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << "  -- Transfers done -- reduce global counter\n";
      mbox()->move_global_counter(label_, -1);
    }
    bool out_of_blocks = ((current_iteration_ == (sqrtnodes_-1)) &&
                          remaining_.at(current_iteration_).active &&
                          (remaining_.at(current_iteration_).remaining == 0));
    if(out_of_blocks || (current_iteration_ >= sqrtnodes_)) {
      // We may be able to "donate" this computor to a neighboring node.
      // Only do load balancing if prefetching is also turned on.
      if(use_prefetch_) {
        int live_computors = mbox()->read_global_counter(label_);
        if((live_computors > 0) && (flat_upper_ != flat_rank_)) {
          if(flat_upper_ != flat_rank_) {
            actorid dest = mbox()->get_tag(flat_upper_);
            if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
              std::cerr << mbox()->mypid() << ":  Donating local computor.  "
                        << "current_iteration_==" << current_iteration_
                        << ", remaining_stores_==" << remaining_stores_
                        << ", remaining-comps=="
                        << remaining_.at(current_iteration_).remaining << "\n";
            idle_local_computors_.pop_back();
            shared_ptr<donatecomputor> msg = donatecomputor::construct(thread);
            mbox()->send(dest, msg, 8);
          }
        }
      }
    }
    else {
      alert_computors();
    }
    if(! use_prefetch_)
      request_data();
  }

  //
  // Called when a local computor completes a unit of work.
  //
  void actormatmul::remote_computor_done(actorid thread, int iteration) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": remote_computor_done(" << thread 
                << ", " << iteration << "):  remaining_stores=="
                << remaining_stores_ << "; current_iteration_=="
                << current_iteration_ << "\n";
    idle_remote_computors_.push_back(thread);
    --remaining_stores_;
    test_watermark();
    if(current_iteration_ >= (sqrtnodes_-1)) {
      if(remaining_stores_ == 0) {
        // All stores are done.  Time to bail.
        if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
          std::cerr << "  -- Transfers done -- reduce global counter\n";
        mbox()->move_global_counter(label_, -1);
      }
    }
    bool out_of_blocks = 
      ((current_iteration_ == (sqrtnodes_-1)) &&
       remaining_.at(current_iteration_).active &&
       (remaining_.at(current_iteration_).remaining <= 2*computors_));
    if(out_of_blocks || (current_iteration_ >= sqrtnodes_)) {
      // We may be able to "donate" this computor to a neighboring node.
      // Only do the donation if prefetch is turned on.
      if(use_prefetch_) {
        int live_computors = mbox()->read_global_counter(label_);
        if((live_computors > 0) && (flat_upper_ != flat_rank_)) {
          idle_remote_computors_.pop_back();
          if(flat_upper_ != flat_rank_) {
            actorid dest = mbox()->get_tag(flat_upper_);
            shared_ptr<donatecomputor> msg = donatecomputor::construct(thread);
            mbox()->send(dest, msg, 8);
          }
        }
      }
    }
    else {
      alert_computors();
    }
    if(! use_prefetch_)
      request_data();
  }

  //
  // We are being given an opportunity to get a computor from somewhere else.
  //
  void actormatmul::remote_computor_available(actorid thread) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": remote_computor_available("
                << thread << ")\n";
    if(local_computors_.find(thread) != local_computors_.end())
      return; // one of our own -- we might as well stop.
    bool out_of_blocks =
      ((current_iteration_ == (sqrtnodes_-1)) &&
       remaining_.at(current_iteration_).active &&
       (remaining_.at(current_iteration_).remaining <= (2*computors_)));
    if(out_of_blocks || (current_iteration_ >= sqrtnodes_)) {
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << mbox()->mypid() << ":  Not accepting donated computor\n";
      // We don't want this computor -- donate it on if appropriate
      int live_computors = mbox()->read_global_counter(label_);
      if((live_computors > 0) && (flat_upper_ != flat_rank_)) {
        if(flat_upper_ != flat_rank_) {
          if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
            std::cerr << mbox()->mypid() << ":  Bouncing donation upward\n";
          actorid dest = mbox()->get_tag(flat_upper_);
          shared_ptr<donatecomputor> msg = donatecomputor::construct(thread);
          mbox()->send(dest, msg, 8);
        }
      }
    }
    else {
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << mbox()->mypid() << ":  Accepting donated computor\n";
      idle_remote_computors_.push_back(thread);
      alert_computors();
    }
  }

  //
  // Update info on our lower neighbor.
  //
  void actormatmul::update_lower_neighbor(int new_flat_lower) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": update_lower_neighbor("
                << new_flat_lower << ")\n";
    if(flat_lower_ != flat_rank_) {
      // Compute the upward-distance to the flat_lower_ and new_flat_lower;
      // we will only accept this distance getting smaller (no hopping around)
      int olddist = (nodes_ - flat_rank_ + flat_lower_) % nodes_;
      int newdist = (nodes_ - flat_rank_ + new_flat_lower) % nodes_;
      if(olddist > newdist)
        flat_lower_ = new_flat_lower;
      else
        flat_lower_ = flat_rank_;
    }
  }

  //
  // Update info on our upper neighbor.
  //
  void actormatmul::update_upper_neighbor(int new_flat_upper) {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": update_upper_neighbor("
                << new_flat_upper << ")\n";
    if(flat_upper_ != flat_rank_) {
      // Compute the upward-distance to the flat_lower_ and new_flat_lower;
      // we will only accept this distance getting larger (no hopping around)
      int olddist = (nodes_ - flat_rank_ + flat_upper_) % nodes_;
      int newdist = (nodes_ - flat_rank_ + new_flat_upper) % nodes_;
      if(olddist < newdist)
        flat_upper_ = new_flat_upper;
      else
        flat_upper_ = flat_rank_;
    }
  }

  //
  // Our work here is done.
  //
  void actormatmul::all_done(actorid) {
    done_ = true;
  }

  //
  // Alert computors that work has arrived.
  //
  void actormatmul::alert_computors() {
    bool chat = (chatty_ && (chatty_ < 0 || flat_rank_ == 0));
    if(chat)
      std::cerr << mbox()->mypid() << ": alert_computors():  idle_local_computors=="
                << idle_local_computors_.size() << ", idle_remote_computors_=="
                << idle_remote_computors_.size() << "\n";
    while((current_iteration_ < sqrtnodes_) &&
          (! (idle_local_computors_.empty() && idle_remote_computors_.empty()))) {
      if(chat)
        std::cerr << "  -- current_iteration_==" << current_iteration_
                  << "; remaining_.at(current_iteration_) "
                  << (current_iteration_ < sqrtnodes_ ?
                      remaining_.at(current_iteration_).remaining : -1)
                  << "\n";
      if(remaining_.at(current_iteration_).remaining == 0) {
        // Iteration is done.
        ++current_iteration_;
        if(current_iteration_ >= sqrtnodes_) {
          if(remaining_stores_ == 0) {
            // All stores are done.  Time to bail.
            if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
              std::cerr << "  -- Transfers done -- reduce global counter\n";
            mbox()->move_global_counter(label_, -1);
          }
          break;
        }
      }
      if((current_iteration_ < sqrtnodes_) && 
         (remaining_.at(current_iteration_).remaining <= 0)) 
        {
          // Not ready for consumption
          break;
        }
      else {
        // We can consume this one.
        --remaining_.at(current_iteration_).remaining;
        actorid worker;
        if(! idle_remote_computors_.empty()) {
          worker = idle_remote_computors_.back();
          idle_remote_computors_.pop_back();
        }
        else if(! idle_local_computors_.empty()) {
          worker = idle_local_computors_.back();
          idle_local_computors_.pop_back();
        }
        else {
          throw illformederror("alert_computors:  All out of computors.");
        }
        shared_ptr<work> msg =
          work::construct(workmatrix_dimension_, current_iteration_,
                          mbox()->mypid());
        // Here we only send 8 bytes -- we are basically sharing a pointer
        mbox()->send(worker, msg, 8);
        // It may be time to prefetch more info.
        request_data();
      }
    }
  }

  //
  // Request next block of data.
  //
  void actormatmul::request_data() {
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << mbox()->mypid() << ": request_data():  "
                << "current_iteration_==" << current_iteration_ 
                << "; prefetch_for_iteration_==" << prefetch_for_iteration_
                << "\n";
    // We will not fetch for iterations out of range.
    bool valid_iter = ((current_iteration_ < sqrtnodes_) &&
                            (prefetch_for_iteration_ < sqrtnodes_));
    
    // We will fetch for the first iteration:
    // if we have not started fetching for it.
    bool first_iter = false;
    if(valid_iter)
      first_iter = ((current_iteration_ == 0) &&
                    (prefetch_for_iteration_ == 0) &&
                    (! remaining_.at(0).active));
    // And we will fetch for any iteration if we have fetched the preceding
    // iteration, but we will only prefetch one block at a time.
    bool roll_iter = false;
    if(valid_iter) {
      if(use_prefetch_) {
        roll_iter = ((current_iteration_ < sqrtnodes_) &&
                     (prefetch_for_iteration_ < sqrtnodes_) &&
                     (remaining_.at(current_iteration_).remaining >= 0) &&
                     (! remaining_.at(prefetch_for_iteration_).active) &&
                     (prefetch_for_iteration_ <= (current_iteration_ + 1)));
      }
      else {
        roll_iter = ((current_iteration_ < sqrtnodes_) &&
                     (remaining_stores_ == 0) &&
                     (! remaining_.at(prefetch_for_iteration_).active));
      }
    }
    // OK, now we can go.
    if(first_iter || roll_iter) {
      // Compute where our next data comes from
      // At some point, one or both blocks of data will come from "us"
      remaining_.at(prefetch_for_iteration_).active = true;
      remaining_.at(prefetch_for_iteration_).pending_arrivals = 2;
      // Compute the grid indices of me and my two neighbors.
      // This is the (fixed) cartesian index of this node
      int me[] = { flat_rank_ / sqrtnodes_,
                   flat_rank_ % sqrtnodes_ };
      // To start, A is shifted left by row number and B up by column number
      // At each iteration, A is shifted up and B left by iteration number
      int ar[] = {(me[0] + prefetch_for_iteration_) % sqrtnodes_,
                  (me[0] + me[1] + prefetch_for_iteration_) % sqrtnodes_};
      int br[] = {(me[0] + me[1] + prefetch_for_iteration_) % sqrtnodes_,
                  (me[1] + prefetch_for_iteration_) % sqrtnodes_};
      // Convert the neighbor grid ranks into flat ranks.
      int flata = ar[0]*sqrtnodes_ + ar[1];
      if(flata == flat_rank_) {
        --remaining_.at(prefetch_for_iteration_).pending_arrivals;
      }
      else {
        actorid nb = mbox()->get_tag(flata);
        shared_ptr<getdata> msg = getdata::construct(prefetch_for_iteration_);
        //std::cerr << "DEBUG:  Node " << mbox()->mypid() << " sending data "
        //          << "request to 'A' peer " << nb << " at time "
        //          << mbox()->now() << "\n";
        mbox()->send(nb, msg, 8);
      }
      int flatb = br[0]*sqrtnodes_ + br[1];
      if(flatb == flat_rank_) {
        --remaining_.at(prefetch_for_iteration_).pending_arrivals;
      }
      else {
        actorid nb = mbox()->get_tag(flatb);
        shared_ptr<getdata> msg = getdata::construct(prefetch_for_iteration_);
        //std::cerr << "DEBUG:  Node " << mbox()->mypid() << " sending data "
        //          << "request to 'B' peer " << nb << " at time "
        //          << mbox()->now() << "\n";
        mbox()->send(nb, msg, 8);
      }
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << "Node " << flat_rank_ << " at (" << me[0] << ", "
                  << me[1] << ") receiving A from (" << ar[0] << ", "
                  << ar[1] << ") [flat=" << flata << "] and B from ("
                  << br[0] << ", " << br[1] << ") [flat=" << flatb << "]\n";
      if(remaining_.at(prefetch_for_iteration_).pending_arrivals == 0) {
        int64_t iter = cubed(workmatrix_splits_);
        remaining_.at(prefetch_for_iteration_).remaining = iter;
        remaining_stores_ += iter;
        alert_computors();
      }
      ++prefetch_for_iteration_;
    }
  }

  //
  // Initialize data.
  //
  void actormatmul::init() {
    // First find out how many nodes we're running on.
    actorid myrank_ = mbox()->mypid();
    if(mbox()->tag_map().empty()) {
      const simplembox::world_t &world = mbox()->world();
      int id = 0;
      simplembox::world_t::const_iterator it;
      for(it = world.begin(); it != world.end(); ++it, ++id)
        mbox()->set_tag(id, *it);
    }
    nodes_ = check_cast<int>(mbox()->tag_map().size());
    sqrtnodes_ = check_cast<int>(sqrt(nodes_));
    // Next, make sure that we have a square number of cores
    // (it would be a modest change to remove this requirement).
    if(squared(sqrtnodes_) != nodes_) {
      throw valueerror("actormult:  Sorry, this program will currently only "
                       "run on a square number of nodes (1, 4, 9, ...).");
    }
    // If we have too many computors we can run into crazy mis-balance due to
    // over-eager load balancing.
    if(computors_ > (2*nodes_)) {
      int64_t oldcomp = computors_;
      int64_t oldthr = threads_per_computor_;
      int64_t threads_per_node = computors_ * threads_per_computor_;
      if((threads_per_node % (2*nodes_)) == 0) {
        // we can get away with a simple scaling-down
        computors_ = 2*nodes_;
        threads_per_computor_ = check_cast<int>(threads_per_node) / computors_;
      }
      else {
        // Do the best we can.
        while(computors_ > (2*nodes_)) {
          for(int divisor = 2; divisor <= computors_; ++divisor) {
            if((computors_ % divisor) == 0) {
              computors_ /= divisor;
              threads_per_computor_ *= divisor;
              break; // break back out to while(...) test.
            }
          }
        }
      }
      std::cerr << "Changed computor/thread ratio from " << oldcomp << "/" << oldthr
                << " to " << computors_ << "/" << threads_per_computor_ << "\n";
    }
    // Find our rank in the world
    flat_rank_ = -1;
    {
      const simplembox::tagmap_t &tags = mbox()->tag_map();
      simplembox::tagmap_t::const_iterator it;
      for(it = tags.begin(); it != tags.end(); ++it) {
        if(it->second == myrank_) {
          flat_rank_ = it->first;
          break;
        }
      }
      if(flat_rank_ < 0)
        throw illformederror("actormatmul: Didn't find "
                             "this task in the flat tag map.");
      flat_upper_ = (flat_rank_ + 1) % nodes_;
      flat_lower_ = (nodes_ + flat_rank_ - 1) % nodes_;
    }
    // Now we can set up the rest of the variables.
    if((fullmatrix_dimension_ % sqrtnodes_) != 0) {
      // pad the full matrix.
      fullmatrix_dimension_ +=
        (sqrtnodes_ - (fullmatrix_dimension_ % sqrtnodes_));
    }
    submatrix_dimension_ = fullmatrix_dimension_ / sqrtnodes_;
    // We can make the local submatrices any size we need.
    // Let's make it so we have at least the given factor as many tasks
    // as we have computors.
    workmatrix_splits_ = 1;
    if(granularity_factor_ < 2) granularity_factor_ = 2;
    while(cubed(workmatrix_splits_) < (granularity_factor_*computors_))
      ++workmatrix_splits_;
    workmatrix_dimension_ = 
      int64_t(ceil(double(submatrix_dimension_) / workmatrix_splits_));
    remaining_.clear();
    remaining_.resize(sqrtnodes_);
    current_iteration_ = 0;
    prefetch_for_iteration_ = 0;
    remaining_stores_ = 0;
    // Figure out what our runtime performance should be.
    timestamp fma = time_per_FMA_;
    if(mbox()->mypid().node().id == degraded_id_)
      fma *= 2;
    // Ready to go -- fire up the worker computors.
    shared_ptr<computor> scaffold =
      computor::construct(mbox()->mypid(), threads_per_computor_, fma);
    for(int i = 0; i < computors_; ++i) {
      actorid thrid = mbox()->spawn(mbox()->mypid().node(), scaffold, 4096);
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << mbox()->mypid() << ":  Spawning computor " << i
                  << " with pid " << thrid << "\n";
      local_computors_.insert(thrid);
      idle_local_computors_.push_back(thrid);
    }
    request_data();
  }

  //
  // Real constructor.
  //
  actormatmul::actormatmul(bool use_prefetch,
                           int64_t matrix, int64_t computors, int chatty,
                           const timestamp &time_per_FMA, int64_t degradedid,
                           double granularity_factor, int threadspercomp) :
    use_prefetch_(use_prefetch),
    fullmatrix_dimension_(matrix), computors_(check_cast<int>(computors)),
    chatty_(chatty),
    time_per_FMA_(time_per_FMA), degraded_id_(degradedid),
    granularity_factor_(granularity_factor),
    threads_per_computor_(threadspercomp), submatrix_dimension_(0),
    workmatrix_dimension_(0), workmatrix_splits_(0),
    flat_rank_(0), nodes_(0), sqrtnodes_(0)
  {
    if(fullmatrix_dimension_ < 1)
      throw valueerror("actormatmul:  Can't operate on a zero-size matrix.");
    if(computors_ < 1)
      throw valueerror("actormatmul:  Need at least one computor per node.");
    if(time_per_FMA_ <= timestamp(0))
      throw valueerror("actormatmul:  Time per FMA should be > 0.");
    if(granularity_factor_ <= 0)
      throw valueerror("actormatmul:  Granularity factor must be > 0.");
    if(threads_per_computor_ < 1)
      throw valueerror("actormatmul:  Threads per computer must be >= 1.");
  }

  //
  // Construct a new object.
  //
  shared_ptr<actormatmul>
  actormatmul::construct(bool use_prefetch,
                         int64_t matrix, int64_t computors, int chatty,
                         const timestamp &time_per_FMA, int64_t degraded_id,
                         double granularity_factor, int threadspercomp)
  {
    shared_ptr<actormatmul> retval
      (new actormatmul(use_prefetch, matrix, computors, chatty, time_per_FMA,
                       degraded_id, granularity_factor, threadspercomp));
    retval->self_ = retval;
    retval->set_self(retval);
    return retval;
  }

  //
  //
  actormatmul::~actormatmul() {
  }

  //
  // Get a fresh clone of this object.
  //
  shared_ptr<simplemboxapp> actormatmul::clone() const {
    return construct(use_prefetch_, fullmatrix_dimension_, computors_, chatty_, time_per_FMA_,
                     degraded_id_, granularity_factor_, threads_per_computor_);
  }

  //
  // Start the evaluation.
  //
  void actormatmul::run() {
    // Initialize data.
    this->init();
    mbox()->move_global_counter(label_, +1);
    done_ = false;
    // Debug.
    if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
      std::cerr << "fullmatrix_dimension_==" << fullmatrix_dimension_
                << "; computors_==" << computors_
                << "; submatrix_dimension_==" << submatrix_dimension_
                << "; workmatrix_dimension_==" << workmatrix_dimension_
                << "; workmatrix_splits_==" << workmatrix_splits_
                << "; flat_rank_==" << flat_rank_
                << "; nodes_==" << nodes_
                << "; sqrtnodes_==" << sqrtnodes_
                << "; current_iteration_==" << current_iteration_
                << "; prefetch_for_iteration_==" << prefetch_for_iteration_
                << "; idle_local_computors_.size()=="<<idle_local_computors_.size()
               << "; idle_remote_computors_.size()=="<<idle_remote_computors_.size()
                << "; remaining_stores_==" << remaining_stores_
                << "; done_==" << done_ << "\n";
    // Start run loop.
    alert_computors();
    while(! done_) {
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << mbox()->mypid() << ": "
                  << mbox()->read_global_counter(label_)
                  << " tasks live.  Iteration " << current_iteration_
                  << " of " << sqrtnodes_ << ":  Entering recv\n";
      simplembox::recvresult_t reply =
        mbox()->recv(actorid::any(), actorpattern::any());
      actorid id = reply.first;
      shared_ptr<base> msg = dynamic_pointer_cast<base>(reply.second);
      if(! msg)
        throw valueerror("actormult::run:  Received message not of type "
                         "actormult::base.");
      if(chatty_ && (chatty_ < 0 || flat_rank_ == 0))
        std::cerr << "Received type " << typeid(*msg).name() << "\n";
      msg->handle(id, self_.lock());
      // Alert any sleeping nodes that we are exiting.
      if(mbox()->read_global_counter(label_) <= 0) {
        if(! done_) {
          if(chatty_)
            std::cerr << mbox()->mypid() << ": Last actormatmul waking peers\n";
          shared_ptr<done> msg = done::construct();
          for(int i = 0; i < nodes_; ++i) {
            if(i != flat_rank_)
              mbox()->send(mbox()->get_tag(i), msg, 8);
          }
        }
        break;
      }
    }
    // Tell our computors to stop.
    if(chatty_)
      std::cerr << mbox()->mypid() << ":  Waking worker computors\n";
    shared_ptr<work> stopit = work::construct(-1, 0, mbox()->mypid());
    local_computors_t::iterator it, end = local_computors_.end();
    for(it = local_computors_.begin(); it != end; ++it)
      mbox()->send(*it, stopit, 8);
  }

} // end of namespace sstmac.
