/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/dummy/dummylauncher.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/networkinterface.h>
#include <sstmac/simulator/node.h>
#include <sstmac/skeleton/dummy/dummykernel.h>
#include <sstmac/skeleton/dummy/dummyserver.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/environment.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Handle completion.
  //
  class dummylauncher::completion : public eventhandler {
    // The owner that gets notified on completion
    shared_ptr<dummylauncher> owner_;
    // Construc a new object and assign it an owner.
    completion(const shared_ptr<dummylauncher>& owner) :
      owner_(owner)
    {}
  public:
    // Alternate construction method.
    static shared_ptr<completion> construct(const shared_ptr<dummylauncher>& owner) {
      return shared_ptr<completion>(new completion(owner));
    }
    // Goodbye.
    virtual ~completion() throw() {
    }
    // We're done here.
    void handle(const timestamp &now) {
      owner_->task_completed(now);
    }
  };

  //
  // Complete a task.
  //
  void dummylauncher::task_completed(const timestamp &now) {
    if(debugger::active()) debug() << "dummylauncher::task_completed(" << now << ")\n";
    shared_ptr<dummylauncher> stay_alive = this->get_self();
    --live_tasks_;
    if(live_tasks_ == 0)
      env_->exit();
    completed_ = true;
  }

  //
  // Hola.
  //
  dummylauncher::dummylauncher() :
    completed_(false), live_tasks_(0)
  {}

  //
  // Who am I?
  //
  void dummylauncher::set_self(const shared_ptr<dummylauncher>& self) {
    if(self.get() != this) {
      throw valueerror("dummylauncher::set_self:  'self' pointer "
                       "does not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<dummylauncher> dummylauncher::get_self() {
    return self_.lock();
  }
   
  //
  // Build a new object.
  //
  shared_ptr<dummylauncher> dummylauncher::construct() {
    shared_ptr<dummylauncher> retval(new dummylauncher);
    retval->set_self(retval);
    return retval;
  }

  //
  // Adios.
  //
  dummylauncher::~dummylauncher() throw() {
  }

  //
  // A string description for this type.
  //
  const std::string& dummylauncher::description() const {
    static const std::string desc("dummylauncher");
    return desc;
  }

  //
  // Run this process.
  //
  void dummylauncher::run(const timestamp &current_time,
                         const boost::shared_ptr<environment>& my_world)
  {
    if(debugger::active())
      debug() << "dummylauncher::run(" << current_time << ", "
              << my_world << ")\n";
    env_ = my_world;
    shared_ptr<tasklist> tlist = my_world->get_tasklist();
    int64_t size = tlist->size();
    for(int64_t i = 0; i < size; ++i) {
      taskid id(i);
      shared_ptr<nodeinterface> node = env_->get(id);
      shared_ptr<dummyserver> server;
      shared_ptr<networkinterface> nic = node->get_node()->get_networkinterface(0);
      if(nic->has_server(my_world->get_ppid())) {
        server = dynamic_pointer_cast<dummyserver>(nic->get_server(my_world->get_ppid()));
      }
      else {
        nodeid nid = tlist->node(id);
        server = dummyserver::construct(nid, tlist, my_world->get_ppid(),
                                        env_->get_node(id), nic);
        nic->register_server(my_world->get_ppid(), server);
      }
      shared_ptr<eventhandler> ev(completion::construct(self_.lock()));
      my_world->get(id)->add
        (current_time + timestamp(id.id),
         shared_ptr<kernel>(dummykernel::construct(tlist, server, ev)));
    }
    live_tasks_ = size;
    completed_ = false;
  }

  //
  // Test whether this process has completed.
  //
  bool dummylauncher::completed() const {
    return completed_;
  }


} // end of namespace sstmac
