/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_DUMMY_DUMMYLAUNCHER_H
#define SSTMAC_SKELETON_DUMMY_DUMMYLAUNCHER_H

#include <sstmac/simulator/launcher.h>
#include <sstmac/util/nocopy.h>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  /**
   * A basic test process.  Simply schedules a dummykernel on every
   * node assigned to the process.
   */
  class dummylauncher : public launcher, virtual public nocopy {
    /// Private event handler used to signal completion.
    class completion;
    
    bool completed_;
    boost::weak_ptr<dummylauncher> self_;

    /// The number of live tasks.
    int64_t live_tasks_;
    /// The environment interface during a run.
    boost::shared_ptr<environment> env_;

    /// Private method to count down the number of live tasks.
    void task_completed(const timestamp &now);
    
  protected:
    /// Hola.
    dummylauncher();

    /// Who am I?
    void set_self(const boost::shared_ptr<dummylauncher>& self);

    /// This is who I am.
    boost::shared_ptr<dummylauncher> get_self();
    
  public:
    /// Build a new object.
    static boost::shared_ptr<dummylauncher> construct();

    /// Adios.
    virtual ~dummylauncher() throw();

    /// A string description for this type.
    /// Used to output debugging and tracing information.
    virtual const std::string& description() const;

    /// Run this process.
    /// \param current_time The simulation time at which this job was started
    /// \param my_world Information regarding the current job, including
    ///        the parallel process id (of type ppid), the sub-processes
    ///        to be run on this node (allowing shared resources),
    ///        and the list of all other nodes involved in the run.
    ///        The lifetime of this object is managed outside the process,
    ///        creating persistent shared_ptr instances to my_world from within
    ///        the process may result in cyclic reference ownership.
    /// \return true if the process exited normally, false otherwise.
    virtual void run(const timestamp &current_time,
                     const boost::shared_ptr<environment>& my_world);

    /// Test whether this process has completed.
    virtual bool completed() const;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SKELETON_DUMMY_DUMMYLAUNCHER_H
