/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/dummy/dummyserver.h>
#include <sstmac/simulator/networkinterface.h>
#include <sstmac/simulator/node.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/networkdata.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/process/status.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  // Dummy network message type.
  class dummydata : public networkdata {
    shared_ptr<status> status_;
    int id_;
    taskid source_;
    taskid dest_;
    
  protected:
    dummydata(int id, const taskid &source, const taskid &dest) :
      status_(status::construct()), id_(id), source_(source), dest_(dest)
    {}

  public: 
    static shared_ptr<dummydata>
    construct(int id, const taskid &source, const taskid &dest) {
      return shared_ptr<dummydata>(new dummydata(id, source, dest));
    }
    virtual ~dummydata() throw() {}
    /// Get another object that corresponds to the receiving end.
    virtual boost::shared_ptr<networkdata> complement() const {
      return construct(id_, source_, dest_);
    }
    /// A status object that may get updated to hold transmission info
    virtual boost::shared_ptr<status> get_status() { return status_; }
    /// The size of the transfer in bytes, including any envelope information.
    virtual int64_t bytes() const { return 20; }
    /// Unique identifier for this message.
    int id() const { return id_; }
    /// The source task.
    virtual const taskid &source_task() const { return source_; }
    /// The destination task.
    virtual const taskid &dest_task() const { return dest_; }
  };

  //
  // Hi.
  //
  dummyserver::dummyserver(const nodeid &mynode, const shared_ptr<tasklist>& tlist,
                           const ppid &ppidarg,
                           const boost::shared_ptr<node> &nodearg,
                           const boost::shared_ptr<networkinterface> &nic) :
    nodeid_(mynode), tasklist_(tlist), next_ping_(1), ppid_(ppidarg),
    node_(nodearg), nic_(nic)
  {
    if(tasklist_ == 0) {
      throw nullerror("dummyserver(...):  Null tasklist pointer.");
    }
  }

  //
  // Hello.
  //
  shared_ptr<dummyserver>
  dummyserver::construct(const nodeid &mynode,
                         const shared_ptr<tasklist>& tlist,
                         const ppid &ppidarg,
                         const boost::shared_ptr<node> &nodearg,
                         const boost::shared_ptr<networkinterface> &nic) {
    return shared_ptr<dummyserver>(new dummyserver(mynode, tlist, ppidarg, nodearg, nic));
  }

  //
  // Goodbye.
  dummyserver::~dummyserver() throw() {
  }

  //
  // Ping-pong.
  //
  void dummyserver::ping(const taskid &me, const taskid &dest,
                         const shared_ptr<eventhandler>& notify)
  {
    if(nic_ == 0) {
      throw nullerror("dummyserver::ping:  Network not initialized.");
    }
    nodeid destnode = tasklist_->node(dest);
    if(destnode == nodeid_) {
      // Local ping -- service almost immediatly.
      node_->get_eventmanager()->schedule(node_->now() + timestamp::exact_nsec(1), notify);
    }
    else {
      // Network message.
      shared_ptr<dummydata> dd = dummydata::construct(next_ping_, me, dest);
      pending_pong_[next_ping_] = notify;
      ++next_ping_;
      nic_->send(node_->now(), ppid_, destnode, dd, eventhandler::ignore);
    }
  }

  //
  // Complete a receive.
  //
  void dummyserver::incoming_message(const timestamp &start_time,
                                     const nodeid &sender,
                                     const shared_ptr<networkdata>& load)
  {
    // This should be a pong.
    shared_ptr<dummydata> dd = dynamic_pointer_cast<dummydata>(load);
    if(dd == 0) {
      throw valueerror("dummyserver::incoming_message: Bad message type.");
    }
    int id = dd->id();
    if(id > 0) {
      // This is a ping
      // Let's send them a pong.
      shared_ptr<dummydata> response =
        dummydata::construct(-id, dd->dest_task(), dd->source_task());
      nic_->send(start_time, ppid_, sender, response, eventhandler::ignore);
    }
    else {
      // This is a pong.
      int aid = abs(id);
      pending_t::iterator it = pending_pong_.find(aid);
      if(it == pending_pong_.end()) {
        throw valueerror("dummyserver::incoming_message: If a pong falls in "
                         "the woods but nobody's listening, "
                         "does it make a sound?");
      }
      shared_ptr<eventhandler> ev = it->second;
      pending_pong_.erase(it);
      ev->handle(start_time);
    }
  }

} // end of namespace sstmac

