/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/minimd/minimd.h>
#include <sstmac/skeleton/minimd/minimd-in.h>
#include <sstmac/skeleton/minimd/minimd-atom.h>
#include <sstmac/skeleton/minimd/minimd-force.h>
#include <sstmac/skeleton/minimd/minimd-neighbor.h>
#include <sstmac/skeleton/minimd/minimd-integrate.h>
#include <sstmac/skeleton/minimd/minimd-thermo.h>
#include <sstmac/skeleton/minimd/minimd-comm.h>
#include <sstmac/skeleton/minimd/minimd-timer.h>
#include <sstmac/process/runinterpolator.h>
#include <sstmac/process/mpiapi.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/BasicStringTokenizer.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <math.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  // Utility function.
  inline void gettok(std::ifstream &infile, std::string &line,
                     std::deque<std::string> &tok, size_t min_tokens,
                     const char *message)
  {
    getline(infile, line);
    if(! infile.good())
      throw valueerror(std::string(message) + ":  Insufficient input lines.");
    tok.clear();
    pst::BasicStringTokenizer::tokenize(line, tok);
    if(tok.size() < min_tokens)
      throw valueerror(std::string(message) + "  Insufficient input fields.");
  }

  // It takes really long to type lexical_cast<> -- blc is boost::lexical_cast
  template <typename T>
  T blc(const std::string &value) {
    return lexical_cast<T>(value);
  }

  //
  // Initialize.
  //
  void minimd::input() {
    std::ifstream in(infile_.c_str());
    if(! in) {
      throw valueerror("minimd::input:  Failed to open infile.");
    }
    // Fixed format file.  First line is title, second line blank.
    std::deque<std::string> tok;
    std::string line;
    for(int i = 0; i < 2; ++i) std::getline(in, line);
    // Third line is size of problem (boxx boxy boxz)
    gettok(in, line, tok, 3, "minimd::input:  Problem size");
    in_->nx = blc<int>(tok[0]);
    in_->ny = blc<int>(tok[1]);
    in_->nz = blc<int>(tok[2]);
    // Fourth line is number of timesteps.
    gettok(in, line, tok, 1, "minimd::input:  Number of timesteps");
    integrate_->ntimes = blc<int>(tok[0]);
    // Fifth line is neighbor bins in each direction.
    gettok(in, line, tok, 3, "minimd::input:  Neighbor bins");
    neighbor_->set_bins(blc<int>(tok[0]), blc<int>(tok[1]), blc<int>(tok[2]));
    // Sixth line is timestep size in reduced units.
    gettok(in, line, tok, 1, "minimd::input:  Timestep size");
    integrate_->dt = blc<double>(tok[0]);
    // Seventh line is initial temperature in reduced units.
    gettok(in, line, tok, 1, "minimd::input:  Initial temperature");
    in_->t_request = blc<double>(tok[0]);
    // Eight line is density (again in reduced units).
    gettok(in, line, tok, 1, "minimd::input:  Density");
    in_->rho = blc<double>(tok[0]);
    // Ninth line is steps between reneighboring.
    gettok(in, line, tok, 1, "minimd::input:  Steps between reneighboring");
    neighbor_->every = blc<int>(tok[0]);
    // Tenth line is inner and outer cutoff distances.
    gettok(in, line, tok, 2, "minimd::input:  Cutoff distances");
    force_->cutforce = blc<double>(tok[0]);
    neighbor_->cutneigh = blc<double>(tok[1]);
    // Eleventh (and final) line is steps between thermo calculations.
    gettok(in, line, tok, 1, "minimd::input:  Steps between thermo");
    thermo_->nstat = blc<int>(tok[0]);
    //
    // We can now initialize the interpolator (pre-compute interpolations).
    execution_->initialize();
  }

  //
  // Set up the box.
  //
  void minimd::create_box() {
    double latt = pow((4/in_->rho), (1.0/3.0));
    atom_->box.set_prd(in_->nx*latt, in_->ny*latt, in_->nz*latt);
  }

  //
  // Pretend to compute all atom positions etc.
  //
  void minimd::create_atoms() {
    mpicomm world = mpi()->comm_world();
    int atoms = 4 * in_->nx * in_->ny * in_->nz;
    int size = world.size().id;
    int rank = world.rank().id;
    atom_->natoms = atoms;
    // We're faking the nlocal to be evenly divided (best-case scenario).
    int eachatoms = atoms / size;
    int sum = eachatoms * size;
    atom_->neighatoms.clear();
    atom_->neighatoms.resize(size, eachatoms);
    for(int i = 0; sum+i < atoms; ++i)
      ++atom_->neighatoms.at(i);
    atom_->nlocal = atom_->neighatoms.at(rank);
    // Now pretend to create the actual atoms like miniMD does.
    mpi()->compute(execution_->get("create_atoms", 0));
  }

  //
  // Pretend to assign velocities to all atoms.
  //
  void minimd::create_velocity() {
    mpicomm world = mpi()->comm_world();
    mpi()->compute(execution_->get("create_velocity", 0));
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
    mpi()->compute(execution_->get("create_velocity", 1));
  }

  //
  // We're sort of ignoring output for now.
  //
  void minimd::output() {
    // get the sum of all profiled timings for this method.
    mpi()->compute(execution_->get("output"));
  }

  //
  // Hi.
  //
  minimd::minimd(const std::string &infile, const shared_ptr<runinterpolator>& exe) :
    infile_(infile), execution_(exe), in_(new in), atom_(new atom),
    force_(new force), neighbor_(new neighbor), integrate_(new integrate),
    thermo_(new thermo), comm_(new comm), timer_(new timer)
  {}

  //
  // Build a new minimd object.
  //
  shared_ptr<minimd> minimd::construct(const std::string &infile,
                                       const shared_ptr<runinterpolator>& exe)
  {
    shared_ptr<minimd> retval(new minimd(infile, exe));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  minimd::~minimd() throw() {
  }

  //
  // Get a copy of this object.
  //
  shared_ptr<mpiapp> minimd::clone() {
    return construct(infile_, execution_->clone());
  }
  
  // 
  // This object gets access to the mpi api.
  //
  void minimd::init_mpi(const boost::shared_ptr<mpiapi>& mpi) {
    mpiapp::init_mpi(mpi);
    in_->init(execution_, mpi);
    atom_->init(execution_, mpi);
    force_->init(execution_, mpi);
    neighbor_->init(execution_, mpi);
    integrate_->init(execution_, mpi);
    thermo_->init(execution_, mpi);
    comm_->init(execution_, mpi);
    timer_->init(execution_, mpi);
  }

  //
  // Go.
  //
  void minimd::run() {
//#define SSTMAC_DB(command) std::cerr << rank << " " << #command << "\n"; command
#define SSTMAC_DB(command) command
    timestamp startat = mpi()->init();
    mpiid rank = mpi()->comm_world().rank();
    unsigned int seedp = rank.id * 12 + 123;
    SSTMAC_DB( this->input() );
    SSTMAC_DB( this->create_box() );

    // This is, conveniently, pretty much a fixed overhead.
    double randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.05;
    mpi()->compute(timestamp(0.0003481) * (1+randnoise));
    //SSTMAC_DB( mpi()->compute(execution_->get("main", 0)) );

    SSTMAC_DB( comm_->setup(neighbor_->cutneigh, atom_) );
    SSTMAC_DB( this->create_atoms() );
    SSTMAC_DB( this->create_velocity() );

    SSTMAC_DB( comm_->exchange(atom_) );
    SSTMAC_DB( comm_->borders(atom_) );
    SSTMAC_DB( neighbor_->build(atom_) );
    SSTMAC_DB( thermo_->compute(0, atom_, neighbor_, force_) );

    // Starting dynamics.
    SSTMAC_DB( timer_->barrier_start(timer::TIME_TOTAL) );
    SSTMAC_DB( integrate_->run(atom_, force_, neighbor_, comm_, thermo_, timer_) );
    SSTMAC_DB( timer_->barrier_stop(timer::TIME_TOTAL) );

    SSTMAC_DB( thermo_->compute(-1, atom_, neighbor_, force_) );
    SSTMAC_DB( this->output() );

    // This compute time is also not really dependent on problem size,
    // although it is a little noisier than the earlier value.
    randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.10;
    mpi()->compute(timestamp(0.001495) * (1+randnoise));
    //SSTMAC_DB( mpi()->compute(execution_->get("main", 1)) );

    timestamp endat = SSTMAC_DB( mpi()->finalize() );
    if(rank.id == 0)
      std::cout << "Total runtime: " << (endat-startat) << "\n";
#undef SSTMAC_DB
  }


} // end of namespace sstmac
