/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/fixedindexing.h>
#include <sstmac/skeleton/mpiparsers/parsedumpi.h>
#include <sstmac/skeleton/mpiparsers/parsedumpi-callbacks.h>
#include <sstmac/skeleton/mpiparsers/dumpimeta.h>
#include <sstmac/util/errors.h>
#include <boost/lexical_cast.hpp>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fstream>
#include <algorithm>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Hi there.
  //
  parsedumpi::parsedumpi(const std::string &fileroot, double timescaling) :
    fileroot_(fileroot), timescaling_(timescaling)
  {}
  
  //
  // Hi there.
  //
  parsedumpi::parsedumpi(const boost::shared_ptr<dumpimeta>& dmeta, double timescaling) :
    timescaling_(timescaling), dmeta_(dmeta)
  {
  	fileroot_ = dmeta_->dirplusfileprefix_;
  }

  //
  // Good news everyone!  We are getting a new parsedumpi.
  //
  shared_ptr<parsedumpi>
  parsedumpi::construct(const std::string &fileroot, double timescaling) {
    shared_ptr<parsedumpi> retval(new parsedumpi(fileroot, timescaling));
    retval->set_self(retval);
    return retval;
  }
  
  //
  // Good news everyone!  We are getting a new parsedumpi.
  //
  shared_ptr<parsedumpi>
  parsedumpi::construct(const shared_ptr<dumpimeta>& dmeta,
                        double timescaling)
  {
    shared_ptr<parsedumpi> retval(new parsedumpi(dmeta, timescaling));
    retval->set_self(retval);
    return retval;
  }

  //
  // Wait!  That's not good news at all!
  //
  parsedumpi::~parsedumpi() throw() {
  }

  //
  // IT'S AN ADDAMS!
  //
  shared_ptr<mpiapp> parsedumpi::clone() {
    // put if to check fileroot
    return construct(fileroot_, timescaling_);
  }

  std::string parsedumpi::dumpi_file_name(int rank) {
    static std::string here("parsedumpi::dumpi_file_name: ");
    // Figure out our input file name.
    // Try fileroot-"%01d", fileroot-"%02d", ... up to some reasonable length
    // (in the short term that we will probably not have over 10 billion peers)
    FILE *fp = NULL;
    char fname[256]; fname[255] = '\0';
    const char *fnamefmt = "%%s-%%0%dd.bin";
    char fmt[30]; fmt[29] = '\0';
    for(int i = 1; i < 10; ++i) {
      snprintf(fmt, 29, fnamefmt, i);
      snprintf(fname, 255, fmt, fileroot_.c_str(), rank);
      //std::cerr << here << "looking for file " << fname << std::endl;
      fp = fopen(fname, "r");
      if(fp) {
        fclose(fp);
        return std::string(fname);
      }
    }
    // If we get here, then no file was found.
    throw ioerror(here + "failed to find a file starting from the fileroot \""
                  + fileroot_ + "\": " + strerror(errno));
  }
  
  int parsedumpi::getnumprocs() {
    static std::string here("parsedumpi::getnumprocs: ");
    if (dmeta_) {
      return dmeta_->getnumprocs();
    } 
    int nrank = 0;
    try {
      while (true) {
        std::string fname = dumpi_file_name(nrank);
        nrank++;
      }
    } catch (ioerror&) {
      // ioerror is thrown when no more trace files can be found
    }
    std::cerr << here << "nrank = " << nrank << "\n";
    return nrank;
  }
  
  static void read_map_file(const std::string &here,
                            const std::string &mapfile,
                            std::map<std::string,std::vector<int> >&hostmap)
  {
    hostmap.clear();
    std::ifstream in(mapfile.c_str());
    if (in.bad()) {
      throw valueerror(here+"could not find node map file");
    }
    int nnode = -1;
    in >> nnode;
    if (nnode <= 0) throw valueerror(here+"bad nnode in node map file");
    int ncoor = -1;
    in >> ncoor;
    if (ncoor <= 0) throw valueerror(here+"bad ncoor in node map file");
    for (int i=0; i<nnode; i++) {

      std::string hostname;
      in >> hostname;

      if (hostname.size()==0) throw valueerror(here+"bad hostname in map file");
      std::vector<int> coor(ncoor);
      for (int j=0; j<ncoor; j++) {
        coor[j] = -1;
        in >> coor[j];
        if (coor[j] < 0) throw valueerror(here+"bad coordinate in map file");
      }

      std::cerr << "inserting " << hostname << " into map at coor";
      for (int j=0; j<ncoor; j++)
        std::cerr << " " << coor[j];
      std::cerr << std::endl;

      hostmap[hostname] = coor;
    }
  }
  
  void
  parsedumpi::layout(const std::string &mapfile,
                     const boost::shared_ptr<network> &net,
                     boost::shared_ptr<allocationstrategy> &alloc,
                     boost::shared_ptr<indexstrategy> &index)
  {
    static std::string here("parsedumpi::layout: ");

    std::map<std::string, std::vector<int> > hostmap;
    read_map_file(here, mapfile, hostmap);

    std::set<nodeid> nodeidset;
    int nrank = getnumprocs();
    std::vector<nodeid> process_to_nodeid(nrank);
    for (int i=0; i<nrank; i++) {
      std::string fname = dumpi_file_name(i);
      dumpi_profile *profile = undumpi_open(fname.c_str());
      if (profile == NULL) {
        throw ioerror(here + "unable to open \"" + fname + "\"");
      }
      dumpi_header *header = dumpi_build_header();
      if (header == NULL) {
        throw ioerror(here + " header is null");
      }
      dumpi_read_header(profile, header);
      if (header->hostname == NULL || *header->hostname == 0) {
        throw ioerror(here + " hostname is null or empty");
      }
      std::string hostname(header->hostname);
      dumpi_free_header(header);
      std::cerr << here << " rank " << i << " is on " << hostname
                << " with coordinate ";
      std::map<std::string, std::vector<int> >::iterator maploc
        = hostmap.find(hostname);
      if (maploc==hostmap.end()) throw valueerror(here+hostname+" not in map");
      std::vector<int> coor = maploc->second;
      std::copy(coor.begin(),coor.end(),
                std::ostream_iterator<int>(std::cerr," "));

      // find node index for this vertex
      int64_t nodeindex;
      vertex hostvert(coor);
      try {
        nodeindex = net->node_to_index(hostvert);
      }
      catch(invalidkeyerror e) { 
        std::cerr << e.what() << "\n" << here 
                  << "possible mismatch between network geometry and hostmap\n";
        throw;
      } 
      catch(std::exception const& e) {
        std::cerr << "can't find index for vertex: " << hostvert << "\n";
        throw;
      }
      std::cout << " and is on simulator node " << nodeindex;
      std::cout << "\n";
      
      process_to_nodeid[i] = nodeid(nodeindex);
      nodeidset.insert(nodeid(nodeindex));
    }
    std::vector<nodeid> nodeidvec;
    std::copy(nodeidset.begin(), nodeidset.end(), std::back_inserter(nodeidvec));
    alloc = fixedallocation::construct(nodeidvec);
    index = fixedindexing::construct(process_to_nodeid);
  }
  
  //
  // Parse the file please (this needs an update, there is no reason
  // for it to return a TRACEREC (which should not be uppercase).
  // Need to figure out how to give the caller access to header/footer recs.
  //
  void parsedumpi::run() {
    if((mpi() == 0) || (world().rank().id < 0))
      throw illformederror("parsedumpi::run:  Dumpi parser not initialized.");
    callbacks cbacks(this);
    std::string fname = dumpi_file_name(world().rank().id);
    // Ready to go.
    cbacks.parse_stream(fname.c_str());
  }

} // end of namespace sstmac

