/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/mpipingpong/mpipingpong.h>
#include <sstmac/process/mpiapi.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <string.h>
#include <stdio.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  //
  // Hi.
  //
  mpipingpong::mpipingpong(int iterations, int count) :
    iterations_(iterations), count_(count)
  {}

  //
  // Hello.
  //
  shared_ptr<mpipingpong> mpipingpong::construct(int iterations, int count) {
    shared_ptr<mpipingpong> obj(new mpipingpong(iterations, count));
    obj->set_self(obj);
    return obj;
  }

  //
  // Goodbye.
  //
  mpipingpong::~mpipingpong() throw() {
  }

  //
  // Get a copy.
  //
  shared_ptr<mpiapp> mpipingpong::clone() {
    return mpipingpong::construct(iterations_, count_);
  }

  //
  // Go.
  //
  void mpipingpong::run() {
    if(mpi() == 0) {
      throw nullerror("mpipingpong::run:  mpiapi pointer is null.");
    }
    mpicomm world = mpi()->comm_world();
    timestamp start = mpi()->init();
    mpiid rank = world.rank();
    mpiid size = world.size();
    //std::cerr << "mpipingpong[" << rank << "]::run(): mpi()->init() at " << start << "\n";
    mpitag tag(0);
    bool participant = true;
    if((size.id % 2) && (rank.id+1 >= size.id)) {
      // This is the odd-node-out -- communicating with no-one.
      participant = false;
    }
    if(participant) {
      mpiid buddy(rank.id ^ 1); // 0<=>1, 2<=>3, etc.
      boost::shared_ptr<const mpistatus> stat = mpistatus::construct();
      for(int half_cycle = 0; half_cycle < 2*iterations_; ++half_cycle) {
        if((half_cycle + rank.id) & 1) { // even values of half-cycle plus rank.
          if(debugger::active()) debug() << "mpipingpong[" << this << "]::main:  Peer " << rank
                  << " sending data at half-cycle " << half_cycle 
                  << " (out of " << 2*iterations_ << ")\n";
          timestamp st = mpi()->send(count_, mpitype::mpi_double, buddy, tag, world);
          //std::cerr << "mpipingpong[" << rank << "]::run(): mpi()->send() at " << st << "\n";
        }
        else {
          if(debugger::active()) debug() << "mpipingpong[" << this << "]::main:  Peer " << rank
                  << " receiving data at half-cycle " << half_cycle 
                  << " (out of " << 2*iterations_ << ")\n";
          timestamp st = mpi()->recv(count_, mpitype::mpi_double, buddy, tag, world, stat);
          //std::cerr << "mpipingpong[" << rank << "]::run(): mpi()->recv() at " << st << "\n";
        }
      }
    }
    timestamp done = mpi()->finalize();

    //std::cerr << "mpipingpong[" << rank << "]::run():  mpi()->finalize() at " << done << "\n";

    /*
    if(rank.id == 0) {
      double runtime = (done-start).sec();
      const size_t bsize = 2000;
      char *buffer = new char[bsize+1];
      buffer[bsize] = '\0';
      snprintf(buffer, bsize, "Peer %d started at %.12f and finished at %.12f "
               "with a total runtime of %g seconds\n", rank.id, start.sec(),
               done.sec(), runtime);
      if(!fwrite(buffer, sizeof(char), strlen(buffer), stderr)) {
        // Failed to write, but we can't do much about that now.
      }
      //std::cerr << "Peer " << rank << " started at " << start
      //          << " and finished at " << done << ", with a total runtime of "
      //          << runtime << " seconds\n";
    }
    */
  }

} // end of namespace sstmac.
