/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_UTIL_NORMALDIST_H
#define SSTMAC_UTIL_NORMALDIST_H

#include <sstmac/util/rng.h>
#include <sstmac/util/errors.h>
#include <math.h>

namespace sstmac {

  namespace RNG {

    /**
     * Base type for classes that provide normally distributed random numbers.
     * Utility functions provide alternate mean and variance.
     */
    class normaldist {
    public:
      virtual ~normaldist() {}

      /// Get a normally distributed random value.
      virtual double normal() = 0;

      /// Get a half-normally distributed value.
      virtual double halfnormal() {
        return fabs(this->normal());
      }

      /// Given an input mean, generate an output value with the desired
      /// variance.
      template <typename T>
      T normal(const T &mean, const T &variance) {
        double rv = normal();
        return (mean + variance*rv);
      }

      template <typename T>
      T halfnormal(const T &mean, const T &variance) {
        double rv = halfnormal();
        return (mean + variance*rv);
      }
    };
    
    /**
     * Use the Box-Muller transform to generate standard normally-distributed
     * random numbers.  
     */
    class boxmuller : public normaldist {
      /// Use a uniformly distributed integer RNG as a source.
      boost::shared_ptr<UniformInteger> gen_;
      /// We generate values two at a time but only hand them out one at a time.
      bool stashed_;
      double stash_;

      boxmuller(boost::shared_ptr<UniformInteger> gen) :
        gen_(gen), stashed_(false)
      {
        if(! gen) throw nullerror("boxmuller:  NULL input RNG source.");
      }

    public:
      static boost::shared_ptr<boxmuller> construct() {
        return boost::shared_ptr<boxmuller>
          (new boxmuller(SimpleCombo::construct()));
      }

      static boost::shared_ptr<boxmuller>
      construct(boost::shared_ptr<UniformInteger> gen) {
        return boost::shared_ptr<boxmuller>(new boxmuller(gen));
      }

      virtual ~boxmuller() {}

      /// Get a normally distributed random value.
      virtual double normal() {
        double rv;
        if(! stashed_) {
          // u1 and u2 are random real values in the interval (0,1]
          double u1 = gen_->realvalue(false, true);
          double u2 = gen_->realvalue(false, true);
          double sqt = sqrt(-2 * log(u1));
          double tht = M_PI * 2 * u2;
          rv = sqt * cos(tht);
          stash_ = sqt * sin(tht);
          stashed_ = true;
        }
        else {
          rv = stash_;
          stashed_ = false;
        }
        return rv;
      }
    };

  } // end of namespace RNG 

} // end of namespace sstmac

#endif // ! SSTMAC_UTIL_NORMALDIST_H
