/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// gcomp.cc: Driver program to analyze properties of various networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <cstdlib>
#include <fstream>
#include <algorithm>
#include <stdexcept>
#include <memory>
#include <iomanip>
#include <sstmac/network/gamma.h>
#include <sstmac/network/hypercube.h>
#include <sstmac/network/ring.h>
#include <sstmac/network/torus.h>
#include <sstmac/network/fattree.h>
#include <sstmac/network/clos.h>
#include <sstmac/network/tbird.h>

extern "C" {
    extern double sqrt(double);
};

using std::setw;
using std::binary_function;
using std::auto_ptr;
using std::ofstream;
using std::set;
using std::atoi;
using std::cout;
using std::endl;
using std::vector;
using std::min_element;
using std::max_element;
using std::logic_error;
using std::multiset;

static double
average(const vector<int> &d)
{
  double a = 0.0;
  for (int i=0; i<d.size(); i++) a+=d[i];
  return a/d.size();
}

static double
stddev(const vector<int> &d)
{
  double a = average(d);
  double s = 0.0;
  for (int i=0; i<d.size(); i++) s+=(a-d[i])*(a-d[i]);
  return sqrt(s/(d.size()-1));
}

static void
write_header()
{
  cout << setw(20) << ""
       << setw(6) << ""
       << " "
       << setw(6) << ""
       << setw(8) << ""
       << setw(6) << ""
       << setw(4) << ""
       << setw(4) << ""
       << setw(8) << "Hops "
       << setw(20) << "Max Link Load    "
       << endl;
  cout << setw(20) << "Graph  "
       << setw(6) << "Ncomm"
       << " "
       << setw(6) << "Nnode"
       << setw(8) << "Nswitch"
       << setw(6) << "Nlink"
       << setw(4) << "Deg"
       << setw(4) << "Dia"
       << setw(8) << "Ave "
       << setw(4) << "Max"
       << setw(8) << "Ave "
       << setw(8) << "StdDev"
       << endl;
}

static void
analyze_one_line(sstmac::network &n, int maxb, int maxp, bool tight_packing)
{
  vector<int> sc(maxb);
  n.stochastic_max_link_load(sc, maxp, tight_packing);
  int minsc = *min_element(sc.begin(), sc.end());
  int maxsc = *max_element(sc.begin(), sc.end());
  double avesc = average(sc);
  double devsc = stddev(sc);

//   int nmaxsc = 0;
//   for (int i=0; i<maxb; i++) { if (sc[i] == maxsc) { nmaxsc++; } }
//   double percent_max = 100*double(nmaxsc)/maxb;

  int maxp_p = (maxp==0?n.n():maxp);
  if (maxp_p > n.n()) maxp_p = n.n();

  cout.setf(std::ios_base::fixed, std::ios_base::floatfield);
  cout << setw(20) << n.name()
       << setw(6) << maxp_p
       << setw(1) << (tight_packing?'p':' ')
       << setw(6) << n.n()
       << setw(8) << n.nswitch()
       << setw(6) << n.nlink()
       << setw(4) << n.d()
       << setw(4) << n.diameter()
       << setw(8) << std::setprecision(4) << n.a()
       << setw(4) << maxsc
//        << setw(4) << int(percent_max+0.5)
       << setw(8) << std::setprecision(4) << avesc
       << setw(8) << std::setprecision(4) << devsc
       << endl;
}

template <class T>
struct order_less: public binary_function<T, T, bool>
{
    bool operator() (const T &a, const T &b) {
      if (a->n() < b->n()) return true;
      if (a->n() > b->n()) return false;
      return a->d() < b->d();
    }
};

template <class T>
struct degree_less: public binary_function<T, T, bool>
{
    bool operator() (const T &a, const T &b) {
      if (a->d() < b->d()) return true;
      if (a->d() > b->d()) return false;
      return a->n() < b->n();
    }
};

class network_sorter: public binary_function<boost::shared_ptr<sstmac::network>,
                                             boost::shared_ptr<sstmac::network>,
                                             bool>
{
    bool node_most_significant_;
    degree_less<boost::shared_ptr<sstmac::network> > degree_;
    order_less<boost::shared_ptr<sstmac::network> > order_;
  public:
    network_sorter(bool a): node_most_significant_(a) {}
    bool operator() (boost::shared_ptr<sstmac::network> const &a,
                     boost::shared_ptr<sstmac::network> const &b) {
      if (node_most_significant_) return order_(a,b);
      return degree_(a,b);
    }
};

typedef multiset<boost::shared_ptr<sstmac::network>, network_sorter> n_multiset;

//typedef multiset<sstmac::network*, degree_less<sstmac::network*> > n_multiset;

void
insert_gammas(int maxn, n_multiset &netset)
{
  cout << "forming gammas" << endl;
  for (int d=2; d<10; d++) {
      for (int k=1; k<=d; k++) {
          boost::shared_ptr<sstmac::network> n(sstmac::gamma::construct(d,k));
          if (n->n() > maxn) break;
          netset.insert(n);
        }
    }
}

void
insert_fattrees(int maxn, n_multiset &netset)
{
  cout << "forming fat trees" << std::endl;
  for (int l=2; l<12; l++) {
      for (int k=2; k<13; k+=2) {
          if (sstmac::fattree::size(l,k) > maxn) break;
          boost::shared_ptr<sstmac::network> n(sstmac::fattree::construct(l,k));
          netset.insert(n);
        }
    }
}

void
insert_clos(int maxn, n_multiset &netset)
{
  cout << "forming clos networks" << std::endl;
  for (int n=2; n<12; n++) {
      if (sstmac::clos::size(n) > maxn) break;
      boost::shared_ptr<sstmac::network> net(sstmac::clos::construct(n));
      netset.insert(net);
    }
}

void
insert_hypercubes(int maxn, n_multiset &netset)
{
  cout << "forming hypercubes" << endl;
  for (int d=2; d<11; d++) {
      boost::shared_ptr<sstmac::network> n(sstmac::hypercube::construct(d));
      if (n->n() > maxn) break;
      netset.insert(n);
    }
}

void
insert_rings(int maxn, n_multiset &netset)
{
   cout << "forming rings" << endl;
   for (int d=2; d<=maxn; d++) {
       boost::shared_ptr<sstmac::network> n(sstmac::ring::construct(d));
       if (n->n() > maxn) break;
       netset.insert(n);
     }
}


void
insert_2Dtori(int maxn, n_multiset &netset)
{
  cout << "forming 2D tori" << endl;
  for (int x=2; x<=maxn/2; x++) {
//      for (int y=2; y<=x; y++) {
          boost::shared_ptr<sstmac::network> n(sstmac::torus::construct_2D(x,x,0));
          if (n->n() > maxn) break;
          netset.insert(n);
          n = sstmac::torus::construct_2D(x,x,1);
          if (n->n() <= maxn) netset.insert(n);
          n = sstmac::torus::construct_2D(x,x,2);
          if (n->n() <= maxn) netset.insert(n);
//        }
    }
}


void
insert_3Dtori(int maxn, n_multiset &netset)
{
  cout << "forming 3D tori" << endl;
  for (int x=2; x<=maxn/4; x++) {
//      for (int y=2; y<=x; y++) {
//          for (int z=2; z<=y; z++) {
              boost::shared_ptr<sstmac::network> n(sstmac::torus::construct_3D(x,x,x,0));
              if (n->n() > maxn) break;
              netset.insert(n);
              n = sstmac::torus::construct_3D(x,x,x,1);
              if (n->n() <= maxn) netset.insert(n);
              n = sstmac::torus::construct_3D(x,x,x,2);
              if (n->n() <= maxn) netset.insert(n);
//            }
//        }
    }
}

void
insert_tbird(int maxn, n_multiset &netset)
{
  cout << "forming tbird" << endl;
  if (maxn >= 4608) {
      boost::shared_ptr<sstmac::network> n(sstmac::tbird::construct());
      netset.insert(n);
    }
}

void
analyze_networks(n_multiset &netset, int maxb, int maxp, bool tight_packing)
{
  write_header();
  for (n_multiset::iterator i = netset.begin();
       i != netset.end();
       i++) {
      analyze_one_line(**i, maxb, maxp, tight_packing);
    }
}

void
help(const char *argv0)
{
  std::cout << "Usage: " << argv0
            << " [-n maxn] [-b maxb] [-p np] [-l] [-t] [-h] [-g] [-f] [-y] [-r] [-2] [-3]"
            << std::endl;
  std::cout
      << "The options are:" << std::endl
      << "  -h      Print help information and exit." << std::endl
      << "  -d      Sort results by degree, rather than n_node." << std::endl
      << "  -n maxn Specifies the maximum number of nodes." << std::endl
      << "  -b maxb Gives the number of bisections to sample." << std::endl
      << "  -p np   Gives the number nodes to use in the bisections." << std::endl
      << "  -l      Use the lowest numbered nodes when not all nodes are used." << std::endl
      << "  -t      Include tbird." << std::endl
      << "  -c      Include CLOS networks." << std::endl
      << "  -g      Include gamma graphs." << std::endl
      << "  -f      Include fat trees." << std::endl
      << "  -y      Include hypercubes." << std::endl
      << "  -r      Include rings." << std::endl
      << "  -2      Include 2D tori." << std::endl
      << "  -3      Include 3D tori." << std::endl
      << "If no networks are specified, then all except rings will be used."
      << std::endl;
  exit(0);
}

int
main(int argc, char **argv)
{
  bool show_help = false;

  bool analyze_all = false;
  bool analyze_gamma = false;
  bool analyze_fattree = false;
  bool analyze_clos = false;
  bool analyze_hypercube = false;
  bool analyze_ring = false;
  bool analyze_2Dtori = false;
  bool analyze_3Dtori = false;
  bool analyze_tbird = false;

  bool degree_order = false;
  bool tight_packing = false;

  int maxn = 840;
  int maxb = 20;
  int maxp = 0;

  while(1) {
      int c = getopt(argc, argv, "lthdn:b:p:gfyr23c");
      if (c < 0) break;
      switch (c) {
      case 'h':
          show_help = true;
          break;
      case 'd':
          degree_order = true;
          break;
      case 'n':
          maxn = atoi(optarg);
          break;
      case 'p':
          maxp = atoi(optarg);
          break;
      case 'b':
          maxb = atoi(optarg);
          break;
      case 'g':
          analyze_gamma = true;
          break;
      case 't':
          analyze_tbird = true;
          break;
      case 'l':
          tight_packing = true;
          break;
      case 'f':
          analyze_fattree = true;
          break;
      case 'c':
          analyze_clos = true;
          break;
      case 'y':
          analyze_hypercube = true;
          break;
      case 'r':
          analyze_ring = true;
          break;
      case '2':
          analyze_2Dtori = true;
          break;
      case '3':
          analyze_3Dtori = true;
          break;
      default:
          show_help = true;
        }
    }

  if (show_help) help(argv[0]);

  if (!(analyze_gamma
        ||analyze_fattree
        ||analyze_clos
        ||analyze_tbird
        ||analyze_hypercube
        ||analyze_ring
        ||analyze_2Dtori
        ||analyze_3Dtori)) {
      analyze_all = true;
    }

  network_sorter sorter(!degree_order);
  n_multiset netset(sorter);

  if (analyze_all||analyze_gamma) insert_gammas(maxn, netset);
  if (analyze_all||analyze_fattree) insert_fattrees(maxn, netset);
  if (analyze_all||analyze_clos) insert_clos(maxn, netset);
  if (analyze_all||analyze_hypercube) insert_hypercubes(maxn, netset);
  if (analyze_ring) insert_rings(maxn, netset);
  if (analyze_all||analyze_2Dtori) insert_2Dtori(maxn, netset);
  if (analyze_all||analyze_3Dtori) insert_3Dtori(maxn, netset);
  if (analyze_all||analyze_tbird) insert_tbird(maxn, netset);

  cout << endl;
  analyze_networks(netset,maxb,maxp,tight_packing);

  return 0;
}
