/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


#include <sstmac/sstmac.h>
#include <string>
#include <iomanip>

using boost::shared_ptr;

double
run_pingpong(int nodes, int nps, int width, int ppn,
             const shared_ptr<sstmac::indexstrategy> &is,
             const std::string &isname)
{
  const size_t stack = 1<<15;
  int arrlen = 100000;
  int niter = 10;

  sstmac::timestamp lat(1.e-6);
  double bw(1.e9);

  shared_ptr<sstmac::manager> mgr = sstmac::manager::construct();
  shared_ptr<sstmac::network> net = sstmac::ring::construct(nodes,nps,width);
  shared_ptr<sstmac::interconnect> fab
    = sstmac::circuit::construct(lat, bw, net);

  mgr->set_interconnect(fab);
  mgr->set_eventmanager(sstmac::eventmap::construct());
  mgr->add_nodes(sstmac::simplenode::construct(int(1e10)), nodes);
  shared_ptr<sstmac::mpilaunch> job =
    sstmac::mpilaunch::construct(sstmac::mpipingpong::construct(niter, arrlen));
  job->set_scheduler(sstmac::scheduler::construct(stack,
                                                  stack*nodes*ppn,
                                                  false));
  mgr->add_launcher(sstmac::timestamp(0), job, nodes, ppn, is);
  sstmac::timestamp t = mgr->run();
  return t.sec();
}

int
main(int argc, char **argv)
{
  bool helpflag = false;
  int opt;
  int nodes[] = { 10, 20 };
  int nps[] = { 0, 1, 2, 4 };
  int width[] = { 1, 2, 4 };
  shared_ptr<sstmac::indexstrategy> rr(sstmac::roundrobinindexing::construct());
  shared_ptr<sstmac::indexstrategy> block(sstmac::blockindexing::construct());
  for (int inodes = 0; inodes < 2; inodes++) {
    for (int inps = 0; inps < 4; inps++) {
      for (int iwidth = 0; iwidth < 3; iwidth++) {
        for (int ppn = 1; ppn <= 2; ppn++) {
          /// Construct this iteration so we use the same seeds in the RNG
          /// even if the loop structure is changed.
          shared_ptr<sstmac::indexstrategy> rng(sstmac::randomindexing::construct());
          double tbl = run_pingpong(nodes[inodes], nps[inps], width[iwidth],
                                    ppn, block, "block");
          double trr = run_pingpong(nodes[inodes], nps[inps], width[iwidth],
                                    ppn, rr, "round-robin");
          double tra = run_pingpong(nodes[inodes], nps[inps], width[iwidth],
                                    ppn, rng, "random");

          std::cout << " " << std::setw(4) << nodes[inodes]
                    << " " << std::setw(2) << nps[inps]
                    << " " << std::setw(2) << width[iwidth]
                    << " " << std::setw(2) << ppn
                    << " " << std::setw(8)
                    << std::fixed << std::setprecision(6) << tbl
                    << " " << std::setw(8)
                    << std::fixed << std::setprecision(6) << trr
                    << " " << std::setw(8)
                    << std::fixed << std::setprecision(6) << tra
                    << std::endl;
        }
      }
    }
  }
  return 0;
}
