/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <iostream>
#include <sstream>
#include <sys/time.h>
#include <time.h>

namespace sstmac {

  static timeval start, now;
  inline void walldelta() {
    gettimeofday(&now, NULL);
    double t = (now.tv_sec - start.tv_sec) + (now.tv_usec - start.tv_usec)*1e-6;
    std::cerr << " -- elapsed walltime " << t << " seconds, cputime "
              << double(clock())/CLOCKS_PER_SEC << " seconds\n";
  }

  class testcoll : public mpiapp {
    /// The MPI interface.
    boost::shared_ptr<mpiapi> mpi_;

    /// Ensure a context switch.
    inline timestamp context_switch() {
      return mpi_->compute(timestamp::exact_sec(1));
    }

    // Barrier.
    inline void barrier(const mpiid&, const mpicomm &comm,
                        timestamp &earlier, timestamp &now,
                        bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering barrier at " << now << ".\n";
        walldelta();
        earlier = now;
      }
      now = mpi_->barrier(comm);
      if(report) {
        std::cerr << comm.rank() << " free from barrier at " << now 
                  << " (runtime " << (now-earlier) << ") \n";
        walldelta();
      }
      now = context_switch();
    }

    // Bcast.
    inline void bcast(const mpiid &root, const mpicomm &comm,
                      timestamp &earlier, timestamp &now,
                      bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering bcast at " << now << ".\n";
        walldelta();
        earlier = now;
      }
      now = mpi_->bcast(1<<16, mpitype::mpi_double, root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from bcast at " << now
                  << " (runtime " << (now - earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }
    
    // Scan.
    inline void scan(const mpiid&, const mpicomm &comm,
                     timestamp &earlier, timestamp &now,
                     bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering scan at " << now << ".\n";
        walldelta();
        earlier = now;
      }
      now = mpi_->scan(1<<16, mpitype::mpi_double, mpiop::sum, comm);
      if(report) {
        std::cerr << comm.rank() << " free from scan at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Scatter.
    inline void scatter(const mpiid &root, const mpicomm &comm,
                        timestamp &earlier, timestamp &now,
                        bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering scatter at " << now << "\n";
        walldelta();
        earlier = now;
      }
      now = mpi_->scatter(1<<10, mpitype::mpi_double,
                          1<<10, mpitype::mpi_double,
                          root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from scatter at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Gather.
    inline void gather(const mpiid &root, const mpicomm &comm,
                       timestamp &earlier, timestamp &now,
                       bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering gather at " << now << "\n";
        walldelta();
        earlier = now;
      }
      now = mpi_->gather(1<<10, mpitype::mpi_double,
                         1<<10, mpitype::mpi_double,
                         root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from gather at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Scatterv.
    inline void scatterv(const mpiid &root, const mpicomm &comm,
                         timestamp &earlier, timestamp &now,
                         bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering scatterv at " << now << "\n";
        walldelta();
        earlier = now;
      }
      std::vector<int> sendlen(comm.size().id, 1<<10);
      now = mpi_->scatterv(sendlen, mpitype::mpi_double,
                           1<<10, mpitype::mpi_double,
                           root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from scatterv at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Gatherv.
    inline void gatherv(const mpiid &root, const mpicomm &comm,
                        timestamp &earlier, timestamp &now,
                        bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering gatherv at " << now << "\n";
        walldelta();
        earlier = now;
      }
      std::vector<int> recvlen(comm.size().id, 1<<10);
      now = mpi_->gatherv(1<<10, mpitype::mpi_double,
                          recvlen, mpitype::mpi_double,
                          root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from gatherv at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Alltoallv.
    inline void alltoallv(const mpiid&, const mpicomm &comm,
                         timestamp &earlier, timestamp &now,
                         bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering alltoallv at " << now << "\n";
        walldelta();
        earlier = now;
      }
      std::vector<int> sendcnts(comm.size().id, 1<<10);
      std::vector<int> recvcnts(comm.size().id, 1<<10);
      now = mpi_->alltoallv(sendcnts, mpitype::mpi_double,
                           recvcnts, mpitype::mpi_double,
                           comm);
      if(report) {
        std::cerr << comm.rank() << " free from alltoallv at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Reduce_scatter.
    inline void reduce_scatter(const mpiid&, const mpicomm &comm,
                               timestamp &earlier, timestamp &now,
                               bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering reduce_scatter at "
                  << now << "\n";
        walldelta();
        earlier = now;
      }
      std::vector<int> recvcnts(comm.size().id, 1<<10);
      now = mpi_->reduce_scatter(recvcnts, mpitype::mpi_double,
                                 mpiop::sum, comm);
      if(report) {
        std::cerr << comm.rank() << " free from reduce_scatter at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Allgather
    inline void allgather(const mpiid&, const mpicomm &comm,
                          timestamp &earlier, timestamp &now,
                          bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering allgather at "
                  << now << "\n";
        walldelta();
        earlier = now;
      }
      int sendcnts = 1<<10, recvcnts = sendcnts;
      now = mpi_->allgather(sendcnts, mpitype::mpi_double,
                            recvcnts, mpitype::mpi_double,
                            comm);
      if(report) {
        std::cerr << comm.rank() << " free from allgather at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Allgather
    inline void allgatherv(const mpiid&, const mpicomm &comm,
                           timestamp &earlier, timestamp &now,
                           bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering allgatherv at "
                  << now << "\n";
        walldelta();
        earlier = now;
      }
      int sendcnts = 1<<10;
      const std::vector<int> recvcnts(comm.size().id, 1<<10);
      now = mpi_->allgatherv(sendcnts, mpitype::mpi_double,
                             recvcnts, mpitype::mpi_double,
                             comm);
      if(report) {
        std::cerr << comm.rank() << " free from allgatherv at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Reduce
    inline void reduce(const mpiid &root, const mpicomm &comm,
                       timestamp &earlier, timestamp &now,
                       bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering reduce at "
                  << now << "\n";
        walldelta();
        earlier = now;
      }
      int count = comm.size().id*(1<<10);
      now = mpi_->reduce(count, mpitype::mpi_double,
                         mpiop::sum, root, comm);
      if(report) {
        std::cerr << comm.rank() << " free from reduce at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

    // Reduce
    inline void allreduce(const mpiid&, const mpicomm &comm,
                       timestamp &earlier, timestamp &now,
                       bool report)
    {
      if(report) {
        std::cerr << comm.rank() << " entering allreduce at "
                  << now << "\n";
        walldelta();
        earlier = now;
      }
      int count = comm.size().id*(1<<10);
      now = mpi_->allreduce(count, mpitype::mpi_double,
                            mpiop::sum, comm);
      if(report) {
        std::cerr << comm.rank() << " free from allreduce at " << now
                  << " (runtime " << (now-earlier) << ")\n";
        walldelta();
      }
      now = context_switch();
    }

  protected:
    testcoll() {
    }

  public:
    /// Hi.
    static boost::shared_ptr<testcoll> construct() {
      boost::shared_ptr<testcoll> retval(new testcoll);
      retval->set_self(retval);
      return retval;
    }

    /// Goodbye.
    virtual ~testcoll() throw() {
    }

    /// Get a copy.
    boost::shared_ptr<mpiapp> clone() {
      return construct();
    }

    /// This is your access to the MPI API.
    virtual void init_mpi(const boost::shared_ptr<mpiapi>& mpi) {
      //std::cerr << "testmpicollectives::init_mpi\n";
      mpi_ = mpi;
    }

    /// Go.
    virtual void run() {
      mpi_->init();
      mpicomm world = mpi_->comm_world();
      mpiid rank = world.rank();
      mpiid size = world.size();

      bool first = (rank.id == 0);
      bool last = (rank.id == (size.id - 1));
      const bool report = (first || last);
      gettimeofday(&start, NULL);

      if(first) {
        std::cerr << "Parallel run on " << size.id << " nodes\n";
      }

      timestamp now = mpi_->wtime();
      timestamp earlier(0);
      if(report) {
        std::cerr << rank << " starting run at time " << now << "\n";
        walldelta();
      }

      for(int iteration = 0; iteration < 1; ++iteration) {
        mpiid root((iteration+1) % size.id);
        barrier(root, world, earlier, now, report);
        bcast(root, world, earlier, now, report);
        scan(root, world, earlier, now, report);
        scatter(root, world, earlier, now, report);
        gather(root, world, earlier, now, report);
        scatterv(root, world, earlier, now, report);
        gatherv(root, world, earlier, now, report);
        alltoallv(root, world, earlier, now, report);
        reduce_scatter(root, world, earlier, now, report);
        allgather(root, world, earlier, now, report);
        allgatherv(root, world, earlier, now, report);
        reduce(root, world, earlier, now, report);
        allreduce(root, world, earlier, now, report);
      } // end of iterations.

      // Ensure a context switch.
      context_switch();
      now = mpi_->finalize();

      if(first || last) {
        std::cerr << rank << " finalized at " << now << "\n";
        walldelta();
      }
    }

  };

} // end of namespace sstmac

using boost::shared_ptr;
using namespace sstmac;

struct args {
  bool exit;
  bool verbose;
  int nodes;
  int ppn;
  int stack;
};

template <typename argtype>
argtype getarg(const std::string arg, const std::string val) {
  std::string valstr;
  argtype retval;
  std::stringstream st;
  if(!(st << val || !(st >> retval) || !(st >> std::ws).eof())) {
    std::cerr << "Invalid integer argument \"" << val << "\""
              << " for " << arg << "\n";
    throw "a fit";
  }
  st >> retval;
  return retval;
}

template <typename argtype>
bool testmatch(int argc,
               char **argv,
               int &currarg,
               const std::string &shopt,
               const std::string &lopt,
               argtype &value,
               argtype minval)
{
  bool match = false;
  const std::string strval(argv[currarg]);
  bool shortmatch = (strval == shopt);
  bool longmatch = (strval == lopt);
  bool eqmatch = ((strval.size() >= lopt.size()) &&
                  (strval.substr(0, lopt.size()+1) == (lopt+"=")));
  argtype scratchvalue;
  if(shortmatch || longmatch) {
    if(currarg+1 >= argc) {
      std::cerr << "Missing integer argument for " << argv[currarg] << "\n";
      throw "a fit";
    }
    match = true;
    scratchvalue = getarg<argtype>(argv[currarg], argv[currarg+1]);
    currarg += 1;
  }
  else if(eqmatch) {
    match = true;
    scratchvalue = getarg<argtype>(lopt, strval.substr(lopt.size()+1));
  }
  if(match && value < minval) {
    std::cerr << "Value " << value << " is below the minumum (" << minval
              << ") for " << argv[currarg] << "\n";
    match = false;
  }
  else if(match) {
    value = scratchvalue;
  }
  return match;
}

bool matches(const std::string &v, const std::string &s, const std::string l) {
  return (v == s || v == l);
}

args parse_args(int argc, char **argv) {
  args a;
  a.exit = false;
  a.verbose = false;
  a.nodes = 4;
  a.ppn = 1;
  a.stack = 1<<16;

  for(int i = 1; i < argc; ++i) {
    bool good = false;
    if(matches(argv[i], "-h", "--help")) {
      std::cerr << "Usage: " << argv[0] << " [--help] [--verbose] "
                << "[--nodes=int(=4)] [--ppn=int(=2)] [--stack=int(=10240)]\n";
      a.exit = true;
      break;
    }
    if(matches(argv[i], "-v", "--verbose")) {
      good = true;
      a.verbose = true;
    }
    good = good || testmatch(argc, argv, i, "-n", "--nodes", a.nodes, 1);
    good = good || testmatch(argc, argv, i, "-p", "--ppn", a.ppn, 1);
    good = good || testmatch(argc, argv, i, "-s", "--stack", a.stack, 16384);
    if(! good) {
      std::cerr << "Invalid argument " << argv[i] << "\n";
      a.exit = true;
      break;
    }
  }
  return a;
}

int main(int argc, char **argv) {
  args a = parse_args(argc, argv);
  if(a.exit)
    return 1;

  const int totnodes = a.nodes * a.ppn;
  const bool mpr = (totnodes < 20000 ? true : false);
  
  if(a.verbose)
    debug().to_stderr();
  shared_ptr<manager> mgr = manager::construct();
  mgr->set_interconnect(fullnetwork::construct(timestamp(1e-9), 125000000));
  mgr->set_eventmanager(eventmap::construct());
  mgr->add_nodes(simplenode::construct(1e10), a.nodes);
  shared_ptr<mpilaunch> job =
    mpilaunch::construct(testcoll::construct());
  job->set_scheduler(scheduler::construct(a.stack, a.stack*totnodes, mpr));
  mgr->add_launcher(timestamp(1), job, a.nodes, a.ppn);
  timestamp t = mgr->run();
  std::cout << "end time is " << t << std::endl;
}
