/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <math.h> // pow
#include <iomanip>
#include <sys/time.h>
#include <sys/resource.h>

using boost::shared_ptr;

double get_wall_time()
{
  struct timeval tv;
  gettimeofday(&tv,0);
  return tv.tv_sec + 0.000001 * tv.tv_usec;
}

sstmac::timestamp
run_parsedumpi(const std::string &metafname,
               const std::string &mapfile,
               int nodes, int ppn,
               const std::string &indexing,
               const shared_ptr<sstmac::interconnect> &intercon,
               const shared_ptr<sstmac::network> &net,
               size_t stack)
{
  shared_ptr<sstmac::dumpimeta> meta = sstmac::dumpimeta::construct(metafname);
  shared_ptr<sstmac::manager> mgr = sstmac::manager::construct();
  mgr->set_interconnect(intercon);
  mgr->set_eventmanager(sstmac::eventmap::construct());
  mgr->add_nodes(sstmac::simplenode::construct(int(1e10)), nodes);
  shared_ptr<sstmac::parsedumpi> app = sstmac::parsedumpi::construct(meta, 1.0);

  shared_ptr<sstmac::indexstrategy> is;
  shared_ptr<sstmac::allocationstrategy> as;
  if (indexing == "random")          is = sstmac::randomindexing::construct();
  else if (indexing == "block")      is = sstmac::blockindexing::construct();
  else if (indexing == "roundrobin") is = sstmac::roundrobinindexing::construct();
  else if (indexing == "dumpi")      app->layout(mapfile, net, as, is);
  else throw sstmac::valueerror(indexing+": invalid indexing strategy");

  int nproc = app->getnumprocs();
  int nnode_needed = nproc/ppn;
  if (nproc%ppn) {
    std::string errmsg = "testparsedumpi: trace nproc must be divisible by ppn";
    std::cout << errmsg << std::endl;
    throw sstmac::valueerror(errmsg);
  }

  shared_ptr<sstmac::mpilaunch> job
    = sstmac::mpilaunch::construct(app);
  job->set_scheduler(sstmac::scheduler::construct(stack, stack*nodes*ppn, false));
  mgr->add_launcher(sstmac::timestamp(0), job, nnode_needed, ppn, is, as);
  sstmac::timestamp t = mgr->run();
  return t;
}

shared_ptr<sstmac::flower>
get_flower(const shared_ptr<sstmac::network> &net,
           double node_router_bw, double router_router_bw)
{
  shared_ptr<sstmac::ordered_flower>
    fl = sstmac::ordered_flower::construct(net, node_router_bw, router_router_bw);

  return fl;
}

void
run_parsedumpis(const std::string &tfname,
                const std::string &mapfile,
                int nodes, int ppn,
                const std::string &indexing, int repeats, size_t stack,
                bool debug)
{

  double node_router_bw = 1.0e9;
  double router_router_bw = 1.5e9;
  double latency = 1e-6;

  int nxyz = 1;
  while (nxyz*nxyz*nxyz < nodes) nxyz++;

  shared_ptr<sstmac::network> net = sstmac::hdtorus::construct_3D(nxyz, nxyz, nxyz, 1, 1);

  for (int j=0; j<repeats; j++) {
    shared_ptr<sstmac::flower> flower = get_flower(net,node_router_bw, router_router_bw);
    flower->debug = debug;
    shared_ptr<sstmac::interconnect> intercon
      = sstmac::sharedcircuit::construct(sstmac::timestamp(latency), 0.0, net, flower);
    sstmac::timestamp t = run_parsedumpi(tfname, mapfile,
                                         nodes, ppn, indexing, intercon, net, stack);
    std::cout << " " << std::fixed << std::setprecision(6) << 1e6*t.sec() << std::endl;
  }
}

void
print_help(const char *name)
{
  std::cout << name << " [-h] [-n nnode] [-p ppn]"
            << " [-I indexing] [-s stacksize] [-r repeats]"
            << " [-t meta_file_name] [-m map_file_name]"
            << "\n";
}

int
main(int argc, char **argv)
{
  bool helpflag = false;
  int opt;
  int nodes = 200, ppn = 4;
  std::string indexing("dumpi");
  std::string metafile("trace.meta");
  std::string mapfile("trace.map");
  int repeats = 1;
  int stacksize = 1<<16;
  bool debug = false;
  while(!helpflag && (opt = getopt(argc, argv, "hdn:p:i:r:s:I:t:m:")) != -1) {
    switch(opt) {
    case 'h':
      helpflag = true;
      break;
    case 'd':
      debug = true;
      sstmac::debug().to_stderr();
      break;
    case 'n':
      nodes = atoi(optarg);
      break;
    case 'p':
      ppn = atoi(optarg);
      break;
    case 'I':
      indexing = optarg;
      break;
    case 't':
      metafile = optarg;
      break;
    case 'm':
      mapfile = optarg;
      break;
    case 's':
      stacksize = atoi(optarg);
      break;
    case 'r':
      repeats = atoi(optarg);
      break;
    default:
      std::cerr << "Unknown argument -" << opt << "\n";
      helpflag = true;
    }
  }
  if(! helpflag) {
    bool errflag = false;
    if(nodes <= 0) {
      std::cerr << "Invalid node count \"" << nodes << "\"\n";
      errflag = true;
    }
    if(ppn <= 0) {
      std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
      errflag = true;
    }
    if (errflag) {
      return 1;
    }
  }
  else {
    print_help(argv[0]);
    return 0;
  }

//  try {
    run_parsedumpis(metafile, mapfile, nodes, ppn, indexing, repeats, stacksize, debug);
//  } catch (std::exception &e) {
//    std::cerr << "Caught exception: " << e.what() << std::endl;
//  }

  return 0;
}
