/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <math.h> // pow

namespace sstmac {

  using boost::shared_ptr;
  using sstmac::composite;
  using sstmac::line;
  using sstmac::ring;
  using sstmac::network;

  inline void fatnet(double latency, long int bandwidth,
		     int levels, int radix,
		     shared_ptr<interconnect> &fab, shared_ptr<network> &net)
  {
    net = sstmac::fattree::construct(levels, radix/2);
    fab = circuit::construct(timestamp(latency), bandwidth, net);
  }

  void pingpong(int nodes, int ppn, int iter, int arrlen, int idxrng,
		int allrng, int radix, int levels, size_t stack = 1<<15)
  {
    std::cerr << "pingpong(" << nodes << ", " << ppn << ", " << iter
              << ", " << arrlen << ", " << idxrng << ", " << allrng
	      << ", " << radix << ", " << levels << ", " << stack << ")\n";
    shared_ptr<manager> mgr = manager::construct();
    shared_ptr<interconnect> fab;
    shared_ptr<network> net;
    
    int halfwidth = radix/2;
    // A fat tree cannot have less than 2 levels.
    int maxslots = 0;
    if(levels < 2) {
      levels = 2;
      maxslots = halfwidth * halfwidth;
      while(maxslots<nodes) {
	maxslots *= halfwidth;
	levels++;
      }
    }
    else {
      maxslots = pow((double)halfwidth, (double)levels);
    }


    std::cerr << "Fat tree with radix " << radix << " and " << levels << " levels "
              << "giving " << maxslots << " max slots\n";
    fatnet(1e-9, 125000000, levels, radix, fab, net);

    mgr->set_interconnect(fab);
    std::cerr << "manager::open_slots() = " << mgr->open_slots() << "\n";
    mgr->set_eventmanager(eventmap::construct());
    mgr->add_nodes(simplenode::construct(1e10), maxslots);
    shared_ptr<mpilaunch> job =
      mpilaunch::construct(mpipingpong::construct(iter, arrlen));
    std::cerr << "mmap-ing " << stack*nodes*ppn << " bytes\n";
    job->set_scheduler(scheduler::construct(stack, stack*nodes*ppn, false));
    shared_ptr<indexstrategy> idx;
    shared_ptr<allocationstrategy> all;
    if(idxrng >= 0) {
      std::cerr << "Randomizing MPI rank indexing\n";
      idx = randomindexing::construct(idxrng);
    }
    if(allrng >= 0) {
      std::cerr << "Randomizing node allocation\n";
      all = randomallocation::construct(allrng);
    }
    mgr->add_launcher(timestamp(1), job, nodes, ppn, idx, all);
    timestamp t = mgr->run();
    std::cout << "end time is " << t << std::endl;
  }
}

inline void print_help(const char *name) {
  std::cerr << name << " [-h] [-n nodes] [-p ppn] [-i iterations] "
            << "[-a array_length] [-r indexing-rng] [-s allocation-rng] "
	    << "[-x radix] [-l levels]\n";
}

int main(int argc, char **argv) {
  bool helpflag = false;
  int opt;
  int verbose = 0, nodes = 1000, ppn = 1, iter = 10, arrlen = 4, idxrng = -1;
  int allrng = -1, radix = 32, levels=-1;
  while(!helpflag && (opt = getopt(argc, argv, "hvn:p:i:a:r:s:x:l:")) != -1) {
    switch(opt) {
    case 'h':
      helpflag = true;
      break;
    case 'v':
      verbose = 1;
      break;
    case 'n':
      nodes = atoi(optarg);
      break;
    case 'p':
      ppn = atoi(optarg);
      break;
    case 'i':
      iter = atoi(optarg);
      break;
    case 'a':
      arrlen = atoi(optarg);
      break;
    case 'r':
      idxrng = atoi(optarg);
      break;
    case 's':
      allrng = atoi(optarg);
      break;
    case 'x':
      radix = atoi(optarg);
      break;
    case 'l':
      levels = atoi(optarg);
      break;
    default:
      std::cerr << "Unknown argument -" << opt << "\n";
      helpflag = true;
    }
  }
  if(! helpflag) {
    bool errflag = false;
    if(verbose)
      sstmac::debug().to_stderr();
    if(nodes <= 0) {
      std::cerr << "Invalid node count \"" << nodes << "\"\n";
      errflag = true;
    }
    if(ppn <= 0) {
      std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
      errflag = true;
    }
    if(iter < 0) {
      std::cerr << "Invalid iteration count \"" << iter << "\"\n";
      errflag = true;
    }
    if(arrlen <= 0) {
      std::cerr << "Invalid array length \"" << arrlen << "\"\n";
      errflag = true;
    }
    if(radix <= 2) {
      std::cerr << "Invalid radix \"" << radix << "\"\n";
      errflag = true;
    }
    if(! errflag) {
      sstmac::pingpong(nodes, ppn, iter, arrlen, idxrng, allrng, radix, levels);
    }
  }
  else {
    print_help(argv[0]);
  }
}
