/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <math.h> // pow

#include <sstmac_testutil.h>

namespace sstmac {

  using boost::shared_ptr;
  using sstmac::composite;
  using sstmac::line;
  using sstmac::ring;
  using sstmac::network;

  void
  meshnet(double latency, long int bandwidth, int n1, int n2 ,int  n3,
	  shared_ptr<interconnect> &fab, shared_ptr<network> &net){
    net = composite::construct(
			       line::construct(n1),
			       composite::construct(
						    line::construct(n2),
						    ring::construct(n3)));
    fab = circuit::construct(timestamp(latency), bandwidth, net);
  }
  void
  fullnet(double latency, long int bandwidth, shared_ptr<interconnect> &fab,
	  shared_ptr<network> &net) {
    fab = fullnetwork::construct(timestamp(latency),bandwidth);
  }

  timestamp pingpong(int nodes, int ppn, int iter, int arrlen,
		int randomize = -1, size_t stack = 1<<16) {
    shared_ptr<manager> mgr = manager::construct();
    shared_ptr<interconnect> fab;
    shared_ptr<network> net;
#if 1
    fullnet(1e-9,125000000, fab, net);
#else
    int nx = int(pow(nodes,1.0/3.0)+0.5);
    int ny = nx;
    int nz = nx;
    meshnet(1e-9,125000000,nx,ny,nx, fab, net);
#endif
    mgr->set_interconnect(fab);
    mgr->set_eventmanager(eventmap::construct());
    mgr->add_nodes(simplenode::construct(int(1e10)), nodes);
    shared_ptr<mpilaunch> job =
      mpilaunch::construct(mpipingpong::construct(iter, arrlen));
    std::cerr << "mmap-ing " << stack*nodes*ppn << " bytes\n";
    job->set_scheduler(scheduler::construct(stack, stack*nodes*ppn, false));
    if(randomize < 0) {
      std::cerr << "Not randomizing\n";
      mgr->add_launcher(timestamp(1), job, nodes, ppn);
    }
    else {
      std::cerr << "Randomizing\n";
      mgr->add_launcher(timestamp(1), job, nodes, ppn,
			randomindexing::construct(randomize));
    }
    timestamp t = mgr->run();
    std::cout << "end time is " << t << std::endl;
    return t;
  }

}

void print_help(const char *name) {
  std::cerr << name << " [-h] [-n nodes] [-p ppn] [-i iterations] "
            << "[-a array_length] [-t]\n";
}

int main(int argc, char **argv) {
  bool helpflag = false, testflag=false;
  int opt;
  int nodes = 1000, ppn = 1, iter = 10, arrlen = 4, randomize = -1;
  int stacksize = 1<<16;
  int argc0 = argc;
  ::testing::InitGoogleTest(&argc, argv);
  // If any googletest flags were found, then run google test.
  if (argc != argc0) {
    testflag = true;
  }
  while(!helpflag && (opt = getopt(argc, argv, "hdn:p:i:a:r:s:t")) != -1) {
    switch(opt) {
    case 'h':
      helpflag = true;
      break;
    case 't':
      testflag = true;
      break;
    case 'd':
      sstmac::debug().to_stderr();
      break;
    case 'n':
      nodes = atoi(optarg);
      break;
    case 'p':
      ppn = atoi(optarg);
      break;
    case 'i':
      iter = atoi(optarg);
      break;
    case 'a':
      arrlen = atoi(optarg);
      break;
    case 'r':
      randomize = atoi(optarg);
      break;
    case 's':
      stacksize = atoi(optarg);
      break;
    default:
      std::cerr << "Unknown argument -" << opt << "\n";
      helpflag = true;
    }
  }
  if (helpflag) {
    print_help(argv[0]);
    return 0;
  }
  if (testflag) {
    return RUN_ALL_TESTS();
  }
  bool errflag = false;
  if(nodes <= 0) {
    std::cerr << "Invalid node count \"" << nodes << "\"\n";
    errflag = true;
  }
  if(ppn <= 0) {
    std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
    errflag = true;
  }
  if(iter < 0) {
    std::cerr << "Invalid iteration count \"" << iter << "\"\n";
    errflag = true;
  }
  if(arrlen < 0) {
    std::cerr << "Invalid array length \"" << arrlen << "\"\n";
    errflag = true;
  }
  if (errflag) {
    return 1;
  }
  sstmac::pingpong(nodes, ppn, iter, arrlen, randomize, stacksize);
  return 0;
}

TEST(testpingpong, n10)
{
  int nodes = 10, ppn = 1, iter = 10, arrlen = 4, randomize = -1;
  int stacksize = 1<<16;
  sstmac::timestamp t = sstmac::pingpong(nodes, ppn, iter, arrlen, randomize, stacksize);
  std::cout << "testpingpong-n10: " << t << std::endl;
  EXPECT_TRUE(timestamp_near(t, 1.000009501540));
}
TEST(testpingpong, n100)
{
  int nodes = 100, ppn = 1, iter = 10, arrlen = 4, randomize = -1;
  int stacksize = 1<<16;
  sstmac::timestamp t = sstmac::pingpong(nodes, ppn, iter, arrlen, randomize, stacksize);
  std::cout << "testpingpong-n100: " << t << std::endl;
  EXPECT_TRUE(timestamp_near(t, 1.000009501540));
}
TEST(testpingpong, n1000)
{
  int nodes = 1000, ppn = 1, iter = 10, arrlen = 4, randomize = -1;
  int stacksize = 1<<16;
  sstmac::timestamp t = sstmac::pingpong(nodes, ppn, iter, arrlen, randomize, stacksize);
  std::cout << "testpingpong-n1000: " << t << std::endl;
  EXPECT_TRUE(timestamp_near(t, 1.000009501540));
}
