/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <math.h> // pow
#include <iomanip>
#include <sys/time.h>
#include <sys/resource.h>

using boost::shared_ptr;

double get_wall_time()
{
  struct timeval tv;
  gettimeofday(&tv,0);
  return tv.tv_sec + 0.000001 * tv.tv_usec;
}

sstmac::timestamp
pingpong(int nodes, int ppn, int iter, int arrlen,
         const shared_ptr<sstmac::indexstrategy> &indexing,
         const shared_ptr<sstmac::interconnect> &intercon,
         size_t stack)
{
  shared_ptr<sstmac::manager> mgr = sstmac::manager::construct();
  mgr->set_interconnect(intercon);
  mgr->set_eventmanager(sstmac::eventmap::construct());
  mgr->add_nodes(sstmac::simplenode::construct(int(1e10)), nodes);

  shared_ptr<sstmac::mpilaunch> job
    = sstmac::mpilaunch::construct(sstmac::mpipingpong::construct(iter, arrlen));
  job->set_scheduler(sstmac::scheduler::construct(stack, stack*nodes*ppn, false));
  mgr->add_launcher(sstmac::timestamp(0), job, nodes, ppn, indexing);
  sstmac::timestamp t = mgr->run();
  return t;
}

shared_ptr<sstmac::flower>
get_flower(const std::string &name, const shared_ptr<sstmac::network> &net,
           double node_router_bw, double router_router_bw)
{
  shared_ptr<sstmac::flower> ret;
  if (name == "overcommit") {
    ret = sstmac::original_flower::construct(net, node_router_bw, router_router_bw,
                                              sstmac::OVERCOMMIT);
  }
  else if (name == "overcommit_flower") {
    ret = sstmac::overcommit_flower::construct(net, node_router_bw, router_router_bw);
  }
  else if (name == "ordered(1)") {
    shared_ptr<sstmac::original_flower>
      fl = sstmac::original_flower::construct(net, node_router_bw, router_router_bw,
                                    sstmac::ORDERED);
    fl->n_rebalance_max = 1;
    ret = fl;
  }
  else if (name == "ordered_flower(1)") {
    shared_ptr<sstmac::ordered_flower>
      fl = sstmac::ordered_flower::construct(net, node_router_bw, router_router_bw);
    fl->n_rebalance_max = 1;
    ret = fl;
  }
  else if (name == "ordered(2)") {
    shared_ptr<sstmac::original_flower>
      fl = sstmac::original_flower::construct(net, node_router_bw, router_router_bw,
                                    sstmac::ORDERED);
    fl->n_rebalance_max = 2;
    ret = fl;
  }
  else if (name == "ordered_flower(2)") {
    shared_ptr<sstmac::ordered_flower>
      fl = sstmac::ordered_flower::construct(net, node_router_bw, router_router_bw);
    fl->n_rebalance_max = 2;
    ret = fl;
  }
  else if (name == "ordered(-1)") {
    shared_ptr<sstmac::original_flower>
      fl = sstmac::original_flower::construct(net, node_router_bw, router_router_bw,
                                    sstmac::ORDERED);
    fl->n_rebalance_max = -1;
    ret = fl;
  }
  else if (name == "ordered_flower(-1)") {
    shared_ptr<sstmac::ordered_flower>
      fl = sstmac::ordered_flower::construct(net, node_router_bw, router_router_bw);
    fl->n_rebalance_max = -1;
    ret = fl;
  }
  else if (name == "throttle_overcommits") {
    ret = sstmac::original_flower::construct(net, node_router_bw, router_router_bw,
                                    sstmac::THROTTLE_OVERCOMMITS);
  }
  else if (name == "throttle_flower") {
    ret = sstmac::throttle_flower::construct(net, node_router_bw, router_router_bw);
  }
  else {
    std::cerr << "get_flower: invalid flower name: " << name << std::endl;
    abort();
  }
  return ret;
}

void
run_pingpongs(const std::string &flower_name,
              int nodes, int ppn, int iter, int arrlen,
              const std::string &indexing, int repeats, size_t stack,
              bool debug)
{
  std::vector<std::string> flower_names;
  if (flower_name == "*") {
    flower_names.push_back("overcommit");
    flower_names.push_back("overcommit_flower");
    flower_names.push_back("ordered(1)");
    flower_names.push_back("ordered_flower(1)");
    flower_names.push_back("ordered(2)");
    flower_names.push_back("ordered_flower(2)");
    flower_names.push_back("ordered(-1)");
    flower_names.push_back("ordered_flower(-1)");
    flower_names.push_back("throttle_overcommits");
    flower_names.push_back("throttle_flower");
  }
  else {
    flower_names.push_back(flower_name);
  }

  double node_router_bw = 1.0e9;
  double router_router_bw = 1.5e9;
  double latency = 1e-6;

  int nxyz = 1;
  while (nxyz*nxyz*nxyz < nodes) nxyz++;

  shared_ptr<sstmac::network> net = sstmac::hdtorus::construct_3D(nxyz, nxyz, nxyz, 1, 1);

  for (int i=0; i<flower_names.size(); i++) {
    shared_ptr<sstmac::indexstrategy> is;
    if (indexing == "random")     is = sstmac::randomindexing::construct();
    else if (indexing == "block") is = sstmac::blockindexing::construct();
    else                          is = sstmac::roundrobinindexing::construct();
    std::cout << std::setw(20) << flower_names[i];
    double t0 = get_wall_time();
    for (int j=0; j<repeats; j++) {
      shared_ptr<sstmac::flower> flower = get_flower(flower_names[i],net,node_router_bw, router_router_bw);
      flower->debug = debug;
      shared_ptr<sstmac::interconnect> intercon
        = sstmac::sharedcircuit::construct(sstmac::timestamp(latency), 0.0, net, flower);
      sstmac::timestamp t = pingpong(nodes, ppn, iter, arrlen, is, intercon, stack);
      std::cout << " " << std::fixed << std::setprecision(6) << 1e6*t.sec();
      std::cout.flush();
    }
    double t1 = get_wall_time();
    std::cout << std::endl;
    std::cerr << std::setw(20) << flower_names[i]
              << " " << std::fixed << std::setprecision(6) << t1-t0
              << " seconds" << std::endl;
  }
}

void
print_help(const char *name)
{
  std::cout << name << " [-h] [-n nnode] [-p ppn] [-i iterations]"
            << " [-a array_length] [-I indexing] [-s stacksize] [-r repeats]"
            << " [-f flower_name]"
            << "\n";
}

int
main(int argc, char **argv)
{
  bool helpflag = false;
  int opt;
  int nodes = 200, ppn = 1, iter = 2, arrlen = 4;
  std::string indexing("random");
  std::string flower_name("*");
  int repeats = 5;
  int stacksize = 1<<16;
  bool debug = false;
  while(!helpflag && (opt = getopt(argc, argv, "hdn:p:i:a:r:s:I:f:")) != -1) {
    switch(opt) {
    case 'h':
      helpflag = true;
      break;
    case 'd':
      debug = true;
      sstmac::debug().to_stderr();
      break;
    case 'n':
      nodes = atoi(optarg);
      break;
    case 'p':
      ppn = atoi(optarg);
      break;
    case 'i':
      iter = atoi(optarg);
      break;
    case 'a':
      arrlen = atoi(optarg);
      break;
    case 'I':
      indexing = optarg;
      break;
    case 'f':
      flower_name = optarg;
      break;
    case 's':
      stacksize = atoi(optarg);
      break;
    case 'r':
      repeats = atoi(optarg);
      break;
    default:
      std::cerr << "Unknown argument -" << opt << "\n";
      helpflag = true;
    }
  }
  if(! helpflag) {
    bool errflag = false;
    if(nodes <= 0) {
      std::cerr << "Invalid node count \"" << nodes << "\"\n";
      errflag = true;
    }
    if(ppn <= 0) {
      std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
      errflag = true;
    }
    if(iter < 0) {
      std::cerr << "Invalid iteration count \"" << iter << "\"\n";
      errflag = true;
    }
    if(arrlen < 0) {
      std::cerr << "Invalid array length \"" << arrlen << "\"\n";
      errflag = true;
    }
    if(indexing != "random" && indexing != "roundrobin" && indexing != "block") {
      std::cerr << "Invalid indexing scheme \"" << indexing << "\"\n";
      errflag = true;
    }
    if (errflag) {
      return 1;
    }
  }
  else {
    print_help(argv[0]);
    return 0;
  }

  run_pingpongs(flower_name, nodes, ppn, iter, arrlen, indexing, repeats, stacksize, debug);

  return 0;
}
