/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <iostream>
#include <sstream>
#include <sys/time.h>
#include <time.h>

namespace sstmac {

  class testsplitcoll : public mpiapp {
    // Debug
    void dump_comm(const mpicomm &comm) {
      const std::vector<taskid> &peers = comm.peers();
      std::cerr << "World rank " << mpi()->comm_world().rank().id
                << ".  Local rank: " << comm.rank().id
                << ".  Peers: " << peers.size()
                << " [";
      for(size_t i = 0; i < peers.size(); ++i)
        std::cerr << peers[i] << ((i < peers.size()-1) ? ", " : "");
      std::cerr << "]\n";
    }

    /// Test splitting the world and doing a set of reduce/allreduce calls.
    void split_n_go(int blocksize, const mpicomm &comm) {
      //std::cerr << "split_n_go(" << blocksize << ", " << comm
      //          << " {.size()=" << comm.size() << "})\n";
      mpicomm outcomm;
      int color = comm.rank().id / blocksize;
      int key = comm.rank().id % blocksize;
      mpi()->comm_split(comm, color, key, outcomm);
      //mpi()->reduce(1, mpitype::mpi_int, mpiop::sum, mpiid(0), outcomm);
      //mpi()->reduce(1, mpitype::mpi_int, mpiop::sum, mpiid(0), outcomm);
      for(int i = 2; i < outcomm.size().id; ++i)
        split_n_go(i, outcomm);
      //std::cerr << "  mpi()->allreduce(1, int, sum, " << outcomm
      //          << " {.size()=" << outcomm.size() << "}) starting\n";
      //dump_comm(outcomm);
      mpi()->allreduce(5, mpitype::mpi_int, mpiop::sum, outcomm);
      //std::cerr << "  mpi()->allreduce(1, int, sum, " << outcomm
      //          << " {.size()=" << outcomm.size() << "}) done\n";
    }
    
  protected:
    testsplitcoll() {
    }

  public:
    /// Hi.
    static boost::shared_ptr<testsplitcoll> construct() {
      boost::shared_ptr<testsplitcoll> retval(new testsplitcoll);
      retval->set_self(retval);
      return retval;
    }

    /// Goodbye.
    virtual ~testsplitcoll() throw() {
    }

    /// Get a copy.
    boost::shared_ptr<mpiapp> clone() {
      return construct();
    }

    /// Go.
    virtual void run() {
      mpi()->init();
      mpicomm world = mpi()->comm_world();
      mpiid rank = world.rank();
      mpiid size = world.size();
      for(int i = 2; i < size.id; ++i) {
        split_n_go(i, world);
      }
      mpi()->finalize();
    }

  };

} // end of namespace sstmac

using boost::shared_ptr;
using namespace sstmac;

struct args {
  bool exit;
  bool verbose;
  int nodes;
  int ppn;
  int stack;
};

template <typename argtype>
argtype getarg(const std::string arg, const std::string val) {
  std::string valstr;
  argtype retval;
  std::stringstream st;
  if(!(st << val || !(st >> retval) || !(st >> std::ws).eof())) {
    std::cerr << "Invalid integer argument \"" << val << "\""
              << " for " << arg << "\n";
    throw "a fit";
  }
  st >> retval;
  return retval;
}

template <typename argtype>
bool testmatch(int argc,
               char **argv,
               int &currarg,
               const std::string &shopt,
               const std::string &lopt,
               argtype &value,
               argtype minval)
{
  bool match = false;
  const std::string strval(argv[currarg]);
  bool shortmatch = (strval == shopt);
  bool longmatch = (strval == lopt);
  bool eqmatch = ((strval.size() >= lopt.size()) &&
                  (strval.substr(0, lopt.size()+1) == (lopt+"=")));
  argtype scratchvalue;
  if(shortmatch || longmatch) {
    if(currarg+1 >= argc) {
      std::cerr << "Missing integer argument for " << argv[currarg] << "\n";
      throw "a fit";
    }
    match = true;
    scratchvalue = getarg<argtype>(argv[currarg], argv[currarg+1]);
    currarg += 1;
  }
  else if(eqmatch) {
    match = true;
    scratchvalue = getarg<argtype>(lopt, strval.substr(lopt.size()+1));
  }
  if(match && value < minval) {
    std::cerr << "Value " << value << " is below the minumum (" << minval
              << ") for " << argv[currarg] << "\n";
    match = false;
  }
  else if(match) {
    value = scratchvalue;
  }
  return match;
}

bool matches(const std::string &v, const std::string &s, const std::string l) {
  return (v == s || v == l);
}

args parse_args(int argc, char **argv) {
  args a;
  a.exit = false;
  a.verbose = false;
  a.nodes = 4;
  a.ppn = 1;
  a.stack = 1<<16;

  for(int i = 1; i < argc; ++i) {
    bool good = false;
    if(matches(argv[i], "-h", "--help")) {
      std::cerr << "Usage: " << argv[0] << " [--help] [--verbose] "
                << "[--nodes=int(=4)] [--ppn=int(=2)] [--stack=int(=10240)]\n";
      a.exit = true;
      break;
    }
    if(matches(argv[i], "-v", "--verbose")) {
      good = true;
      a.verbose = true;
    }
    good = good || testmatch(argc, argv, i, "-n", "--nodes", a.nodes, 1);
    good = good || testmatch(argc, argv, i, "-p", "--ppn", a.ppn, 1);
    good = good || testmatch(argc, argv, i, "-s", "--stack", a.stack, 16384);
    if(! good) {
      std::cerr << "Invalid argument " << argv[i] << "\n";
      a.exit = true;
      break;
    }
  }
  return a;
}

int main(int argc, char **argv) {
  args a = parse_args(argc, argv);
  if(a.exit)
    return 1;

  const int totnodes = a.nodes * a.ppn;
  const bool mpr = (totnodes < 20000 ? true : false);
  
  if(a.verbose)
    debug().to_stderr();
  shared_ptr<manager> mgr = manager::construct();
  mgr->set_interconnect(fullnetwork::construct(timestamp(1e-9), 125000000));
  mgr->set_eventmanager(eventmap::construct());
  mgr->add_nodes(simplenode::construct(1e10), a.nodes);
  shared_ptr<mpilaunch> job =
    mpilaunch::construct(testsplitcoll::construct());
  job->set_scheduler(scheduler::construct(a.stack, a.stack*totnodes, mpr));
  mgr->add_launcher(timestamp(0), job, a.nodes, a.ppn);
  timestamp t = mgr->run();
  std::cout << "end time is " << t << std::endl;
}
